"""
This script sets up the serial postprocessing of all BayesWaveBurst run results in the directory in which it is
launched. It assumes that the individual run directories are labelled job_*. Once the script is done, the
postprocessing may be completed by submitting the dagfile it produces to condor.

This script takes the optional boolean flag radec_flag. This enables the search for the injected right ascension
and declination values to be plotted in the skymap. These are obtained from an xml injection file, assuming the
naming convention m1_m2.xml. radec_flag=False by default 

This script was tested on ldas-pcdev1.ligo.caltech.edu only.
"""

__author__ = "Francesco Pannarale, Ronaldas Macas"
__email__ = "francesco.pannarale@ligo.org, ronaldas.macas@gmail.com"
__version__ = "1.2"
__date__ = "27.08.2014"

######################################################################################################################
# Import modules
######################################################################################################################
import argparse
import math
import os
import pwd
import subprocess
import sys

######################################################################################################################
# Parse the arguments
######################################################################################################################
parser = argparse.ArgumentParser(description='Set up a dag file to launch multiple BayesWaveBurst postprocessing\n runs on condor.')
parser.add_argument('--radec_flag', action='store_true', default=False, dest="radec", help="enable plot of injection locations in skymaps")
#parser.add_argument('--match', action='store_true', default=False, dest="match", help="Add matches and injected waveforms to the webpage")
parser.add_argument('--match', default=False, help="Add matches and injected waveforms to the webpage")
args = parser.parse_args()

######################################################################################################################
# Function that copies many_plots_1job.sh in all run directories to analyze
######################################################################################################################
def ObtainScript(jobDirName, PWD, many_plots_script):
    destination_path=PWD+'/'+jobDirName+'/'
    os.system("cp "+many_plots_script+" "+destination_path)
    
######################################################################################################################
# Function that writes the condor submission script
######################################################################################################################
def WriteSub(jobDirName, PWD, ra, dec, match):
    path=PWD+'/'+jobDirName+'/'
    SubFile = open(path+'submitWebpage.sub', 'w')
    SubFile.write('executable='+path+'many_plots_1job.sh\n')
    if str(ra)+str(dec)=="":
        if str(match)=="False":
            SubFile.write('arguments='+path+'\n')
        else:
            SubFile.write('arguments='+path+' '+str(match)+'\n')
    else:
        if str(match)=="False":
            SubFile.write('arguments='+path+' '+str(ra)+' '+str(dec)+'\n')
        else:
            SubFile.write('arguments='+path+' '+str(ra)+' '+str(dec)+' '+str(match)+'\n')
    SubFile.write('universe=vanilla\n')
    SubFile.write('getenv=True\n')
    SubFile.write('output='+path+'condorManyPlotsOut.txt\n')
    SubFile.write('error='+path+'condorManyPlotsError.txt\n')
    SubFile.write('log='+path+'condorManyPlotsLog.txt\n')
    SubFile.write('notification=never\n')
    SubFile.write('should_transfer_files=YES\n')
    SubFile.write('when_to_transfer_output=ON_EXIT\n')
    SubFile.write('stream_error=True\n')
    SubFile.write('stream_output=True\n')
    SubFile.write('queue 1\n')
    SubFile.close()

######################################################################################################################
# Function that generates a single dag file for multiple postprocessing runs with many_plots_1job.sh 
######################################################################################################################
def WriteDagFile(jobDirName, PWD):
    # Might need to make sure the file exists instead of appending (but the + should take care of that)
    DagFile = open('submitWebpages.dag', 'a')
    DagFile.write('JOB '+jobDirName+' '+PWD+'/'+jobDirName+'/submitWebpage.sub\n')
    DagFile.write('RETRY '+jobDirName+' 1\n\n')
    DagFile.close()

######################################################################################################################
# Function that reads the sky location of an injected event from the XML injection file and expresses it in degrees
######################################################################################################################
def GetRADec(jobDirName, PWD):
    # Find XML file name assuming the format "m1_m2.xml"
    m1 = subprocess.Popen(["echo "+jobDirName+" | cut -d _ -f 2"], stdout=subprocess.PIPE, shell=True).communicate()[0].rstrip()
    m2 = subprocess.Popen(["echo "+jobDirName+" | cut -d _ -f 3"], stdout=subprocess.PIPE, shell=True).communicate()[0].rstrip()
    job_index = subprocess.Popen(["echo "+jobDirName+" | cut -d _ -f 4"], stdout=subprocess.PIPE, shell=True).communicate()[0].rstrip()

    job_index = int(job_index) + 1 # adding +1 because command 'tail' (used below) requires non-zero index 

    longitude = subprocess.Popen(["lwtprint "+str(m1)+"_"+str(m2)+".xml -t sim_inspiralgroup:sim_inspiral -c longitude | head -n"+str(job_index)+" | tail -n1"], stdout=subprocess.PIPE, shell=True).communicate()[0].rstrip()
    latitude = subprocess.Popen(["lwtprint "+str(m1)+"_"+str(m2)+".xml -t sim_inspiralgroup:sim_inspiral -c latitude | head -n"+str(job_index)+"| tail -n1"], stdout=subprocess.PIPE, shell=True).communicate()[0].rstrip()

    #Converting from radians to degrees
    longitude_deg = float(longitude)*180./math.pi
    if longitude_deg > 180:
        ra = longitude_deg - 360.
    else:
        ra = longitude_deg
    dec = float(latitude)*180./math.pi

    return (ra, dec)

######################################################################################################################
# Main
######################################################################################################################

# Find out the path to the BayesWaveBurst executable
p = subprocess.Popen(["which","BayesWaveBurst"],stdout=subprocess.PIPE,stderr=subprocess.STDOUT)
for line in iter(p.stdout.readline, b''):
    BayesWaveBurst = line.rstrip()
if not os.path.isfile(BayesWaveBurst):
    print "****************************************************************************************************************"
    sys.exit("\nMake sure the BayesWaveBurst command is in your path before running this script!\n\n****************************************************************************************************************")
# Store the absolute path to the post-processing script many_plots_1job.sh
many_plots_script = BayesWaveBurst.split("src")[0]+"postprocess/many_plots_1job.sh"

# Determine what directory this script was called from and drop the program name from the list of arguments
sys.argv.pop(0)
username = pwd.getpwuid(os.getuid())[0]
pathname = os.getcwd().split(username)[1]
pathname = '/home/'+username+pathname
PWD = os.path.abspath(pathname)

# Get a list of all the content of the PWD
PWD_ls = subprocess.Popen(["ls",PWD],stdout=subprocess.PIPE,stderr=subprocess.STDOUT)
job_list = []
jobs_number = 0

#TODO: CHANGE THIS!!
# get match argument which should be passed to many_plots_1job.sh
match = args.match
#if match :
#    match = 500
# For each element in the PWD...
for line in iter(PWD_ls.stdout.readline, b''):
    jobDirName = line.rstrip()
    # ...check if it is a BayesWaveBurst job directory. If it is:
    if "job_" in jobDirName:
        # 1) store its name (may be useful in the future)
        job_list.append(jobDirName)
        # 2) copy the post-processing script to the job directory 
        ObtainScript(jobDirName, PWD, many_plots_script)
        # 3) if asked by the user get the right ascension and declination of the injected event for the skymap
	radec = args.radec
	if radec :
            ra, dec = GetRADec(jobDirName, PWD)
	else:
	    ra, dec = "",""
        # 4) write a submit script in it to call the post processing script
        WriteSub(jobDirName, PWD, ra, dec, match)
        # 5) tell the dag file to postprocess this job
        WriteDagFile(jobDirName, PWD)
        # 6) keep count of the jobs found
        jobs_number = jobs_number + 1

# All done! Report to the user
print "****************************************************************************************************************"
print "\n\nJobs to analyze created:", jobs_number
print "\n\n****************************************************************************************************************"
print "\n\nTo run the condor jobs, type\n condor_submit_dag submitWebpages.dag"
print "\nTo monitor a job\'s output type\n tail -f <job-directory>/condorManyPlotsOutput.txt"
print "\nTo monitor a job\'s errors type\n tail -f <job-directory>/condorManyPlotsError.txt"
print "\n To move the postprocessing results in order to make them visible. Run:"
print "\n rsync -av --progress "+PWD+" <path to the desired directory in the public_html>  --exclude '*/*snr' --exclude '*/waveforms' --exclude '*/chains' --include='*/' --include '*html' --include '*png' --include condorOut_BWB.txt --exclude='*'\n\n"
print "****************************************************************************************************************"

######################################################################################################################
######################################################################################################################

