"""
This script sets up a dag file to launch multiple BayesWaveBurst and BayesWavePost runs on condor.
It picks the BBH injection trigger times from a file provided by the user.

Running this script requires six line arguments:
 1) triggers filename (assuming it lives in the directory this script is called from)
 2) first trigger to be analyzed (positive integer)
 3) last trigger to be analyzed (positive integer)

Additional information:
 1) the lowest frequency is fixed to 32 Hz
 2) the sampling rate is fixed to 4096 Hz 
 3) the length of the analysis window is set to 4 s [starts at trigger time - 2s]

This script was tested on ldas-pcdev1.ligo.caltech.edu only.
"""

__author__ = "Francesco Pannarale"
__email__ = "francesco.pannarale@ligo.org"
__version__ = "1.0"
__date__ = "09.09.2014"

######################################################################################################################
# Import modules
######################################################################################################################
import argparse
import math
import numpy as np
import os
import pwd
import subprocess
import sys

######################################################################################################################
# Parse the arguments
######################################################################################################################
parser = argparse.ArgumentParser(description='Set up a dag file to launch multiple BayesWaveBurst and BayesWavePost runs on condor.')
parser.add_argument('triggersFile', type=str, help='Filename for the triggers list')
parser.add_argument('nFirst', type=int, default=1, help='Number of first trigger to run')
parser.add_argument('nLast', type=int, default=40, help='Number of last trigger to run')
args = parser.parse_args()

######################################################################################################################
# Function that writes the condor submission script
# Will work for BayesWave_type=BWB (BayesWaveBurst) and BayesWave_type=BWP (BayesWavePost)
######################################################################################################################
def WriteSub(jobDirName, PWD, TriggerTime, flow, srate, duration, BayesWave_type):
    path=PWD+'/'+jobDirName+'/'
    SubFile = open(path+"submit_"+BayesWave_type+".sub", "w")
    if BayesWave_type == "BWB":
    	SubFile.write("executable="+BayesWaveBurst+"\n")
        SubFile.write("arguments=--ifo H1 --H1-flow "+str(flow)+" --H1-channel H1:GAUSSIAN \\\n")
        SubFile.write("--ifo L1 --L1-flow "+str(flow)+" --L1-channel L1:GAUSSIAN \\\n")
        SubFile.write("--H1-cache /mnt/zfs2/reed.essick/LIB_SKY_LOC/2015_ortFIX/BBH-app/H1.lcf \\\n")
        SubFile.write("--L1-cache /mnt/zfs2/reed.essick/LIB_SKY_LOC/2015_ortFIX/BBH-app/L1.lcf \\\n")
        SubFile.write("--trigtime "+str(TriggerTime)+" --srate "+str(srate)+" --seglen "+str(duration)+" \\\n")
        SubFile.write("--bayesLine --PSDstart "+str(TriggerTime)+" --PSDlength "+str(duration)+" \\\n")
        SubFile.write("--Niter 2000000 \\\n")
        SubFile.write("--Qmax 40 --Dmax 20 --clusterProposal \\\n")
        SubFile.write("--MDC-channel [H1:Science,L1:Science] \\\n")
        SubFile.write("--outputDir "+path+" \\\n")
        SubFile.write("--dataseed 12345 \\\n")
        SubFile.write("--MDC-cache [/mnt/zfs2/reed.essick/LIB_SKY_LOC/2015_ortFIX/BBH-app/MDC.lcf,/mnt/zfs2/reed.essick/LIB_SKY_LOC/2015_ortFIX/BBH-app/MDC.lcf] \n\n")
    if BayesWave_type == "BWP":
    	SubFile.write("executable="+BayesWavePost+"\n")
        SubFile.write("arguments=--ifo H1 --H1-flow "+str(flow)+" --H1-channel LALAdLIGO \\\n")
        SubFile.write("--ifo L1 --L1-flow "+str(flow)+" --L1-channel LALAdLIGO \\\n")
        SubFile.write("--H1-cache LALAdLIGO \\\n")
        SubFile.write("--L1-cache LALAdLIGO \\\n")
        SubFile.write("--trigtime "+str(TriggerTime)+" --srate "+str(srate)+" --seglen "+str(duration)+" \\\n")
        SubFile.write("--bayesLine --PSDstart "+str(TriggerTime)+" --PSDlength "+str(duration)+" \\\n")
        SubFile.write("--MDC-channel [H1:Science,L1:Science] \\\n")
        SubFile.write("--outputDir "+path+" \\\n")
        SubFile.write("--dataseed 12345 \\\n")
        SubFile.write("--MDC-cache [/mnt/zfs2/reed.essick/LIB_SKY_LOC/2015_ortFIX/BBH-app/MDC.lcf,/mnt/zfs2/reed.essick/LIB_SKY_LOC/2015_ortFIX/BBH-app/MDC.lcf] \\\n")
    	SubFile.write("--0noise \n\n")
    SubFile.write("universe=standard\n")
    SubFile.write("output="+path+"condorOut_"+BayesWave_type+".txt\n")
    SubFile.write("error="+path+"condorError_"+BayesWave_type+".txt\n")
    SubFile.write("log="+path+"condorLog_"+BayesWave_type+".txt\n")
    SubFile.write("notification=never\n")
    SubFile.write("should_transfer_files=YES\n")
    SubFile.write("when_to_transfer_output=ON_EXIT\n")
    SubFile.write("stream_error=True\n")
    SubFile.write("stream_output=True\n")
    SubFile.write("queue 1\n")
    SubFile.close()

######################################################################################################################
# Function that generates a single dag file for multiple BayesWaveBurst and BayesWavePost runs
# Will work for BayesWave_type=BWB (BayesWaveBurst) and BayesWave_type=BWP (BayesWavePost)
######################################################################################################################
def WriteDagFile(jobDirName, PWD, BayesWave_type):
    DagFile = open("submitRuns.dag", "a")
    if BayesWave_type == "BWB":
    	DagFile.write("JOB "+jobDirName+" "+PWD+"/"+jobDirName+"/submit_BWB.sub\n")
    	DagFile.write("RETRY "+jobDirName+" 1\n\n")
    if BayesWave_type == "BWP":
    	DagFile.write("JOB "+jobDirName+"_post "+PWD+"/"+jobDirName+"/submit_BWP.sub\n")
    	DagFile.write("RETRY "+jobDirName+"_post 1\n\n")
    	DagFile.write("PARENT "+jobDirName+" CHILD "+jobDirName+"_post \n\n")
    DagFile.close()

######################################################################################################################
# Main
######################################################################################################################

# Find out the path to the BayesWaveBurst executable
p = subprocess.Popen(["which","BayesWaveBurst"],stdout=subprocess.PIPE,stderr=subprocess.STDOUT)
for line in iter(p.stdout.readline, b''):
    BayesWaveBurst = line.rstrip()
if not os.path.isfile(BayesWaveBurst):
    print "****************************************************************************************************************"
    sys.exit("\nMake sure the BayesWaveBurst command is in your path before running this script!\n\n****************************************************************************************************************")

# Find out the path to the BayesWavePost executable                                                                                                                    
p = subprocess.Popen(["which","BayesWavePost"],stdout=subprocess.PIPE,stderr=subprocess.STDOUT)
for line in iter(p.stdout.readline, b''):
    BayesWavePost = line.rstrip()
if not os.path.isfile(BayesWavePost):
    print "****************************************************************************************************************"
    sys.exit("\nMake sure the BayesWavePost command is in your path before running this script!\n\n**********************************************************************************************")

# Determine what directory this is and drop the program name from the list of arguments
sys.argv.pop(0)
username = pwd.getpwuid(os.getuid())[0]
pathname = os.getcwd().split(username)[1]
pathname = '/home/'+username+pathname
PWD = os.path.abspath(pathname)

# The triggers file name
trigfile = os.path.join(pathname, args.triggersFile)
trigList = np.loadtxt(trigfile)

# Prepare to elaborate only triggers specified by the user
indexTriggers = 0
nFirst = args.nFirst - 1
nLast = args.nLast + 1

# Keep track of the number of triggers ending up in the dagfile
countTriggers = 0

# Loop over triggers
for trigger in trigList:
    # Get the trigger and convert it to a string
    TriggerTime = str(trigger)
    # Update trigger index
    indexTriggers = indexTriggers + 1
    if indexTriggers > nFirst and indexTriggers < nLast:
        # Define directory and create folder if no previous data exists
        jobDirName = "job_"+TriggerTime
        if not os.path.exists(jobDirName):
            os.makedirs(jobDirName)
            # Update trigger counter
            countTriggers = countTriggers + 1
            # Write the condor submission scripts to the job directory
            WriteSub(jobDirName, PWD, TriggerTime, 32, 4096, 4, "BWB")
            WriteSub(jobDirName, PWD, TriggerTime, 32, 4096, 4, "BWP")
            # Write the dag file entries 
            WriteDagFile(jobDirName, PWD, "BWB")
            WriteDagFile(jobDirName, PWD, "BWP")

# All done! Remind the user of what to do...
print "****************************************************************************************************************"
print "\n\nJobs created: "+str(countTriggers)
print "\n\n****************************************************************************************************************"
print "\n\nTo run the condor jobs, type\n condor_submit_dag submitRuns.dag"
print "\nTo monitor a job\'s output type\n tail -f <job-directory>/condorOutput*.txt"
print "\nTo monitor a job\'s errors type\n tail -f <job-directory>/condorError*.txt\n"
print "****************************************************************************************************************"

######################################################################################################################
######################################################################################################################

