"""
This script sets up a dag file to launch multiple BayesWaveBurst and BayesWavePost runs on condor.
It generates the necessary BBH waveform injections in an XML file using lalapps_inspinj.

Running this script requires six line arguments:
 1) the fixed mass value for the primary component [solar masses]
 2) the fixed mass value for the primary component [solar masses]
 3) the lowest frequency used for generating a waveform and the detector noise [Hz]
 4) the highest frequency used for reconstructing the power signal spectrogram [Hz]
 5) the GPS time at which the set of injections is started [s]
 6) the GPS time at which the set of injections ends [s]
 7) the time separation between two succesive injections [s]
 8) the length of the analysis window [starts at trigger time - 2s] [s]
 9) the minumum SNR used to generate the uniform SNR distribution 
10) the maximum SNR used to generate the uniform SNR distribution 
11) the number of iterations
12) the value for the seed of the random number generator

As the injections are uniformly distributed in SNR, it may be useful to know the SNRs of the single injections
beforehand. To do, uncomment the following lines

char SnrName[200];
sprintf(SnrName,"SNRfile.txt");
FILE * snrout = fopen(SnrName,"a");
fprintf(snrout,"%lf\n",proposedSNR);
fclose(snrout);

in lalapps/src/inspiral/inspinj.c and recompile LALapps. This will produce a txt file with the network SNR for
each injection.

This script was tested on ldas-pcdev1.ligo.caltech.edu only.
"""

__author__ = "Francesco Pannarale, Ronaldas Macas"
__email__ = "francesco.pannarale@ligo.org, ronaldas.macas@gmail.com"
__version__ = "1.4"
__date__ = "10.09.2014"

######################################################################################################################
# Import modules
######################################################################################################################
import argparse
import math
import os
import pwd
import subprocess
import sys

######################################################################################################################
# Parse the arguments
######################################################################################################################
parser = argparse.ArgumentParser(description='Set up a dag file to launch multiple BayesWaveBurst and BayesWavePost runs on condor.')
parser.add_argument('m1', type=int, default=5, help='First mass in solar masses')
parser.add_argument('m2', type=int, default=5, help='Second mass in solar masses')
parser.add_argument('flow', type=int, default=40, help='Lowest frequency for a waveform and detector noise in Hz')
parser.add_argument('fhigh', type=int, default= 512, help='Cut-off frequency for a waveform and detector noise in Hz')
parser.add_argument('GPSstart', type=int, default=900000000, help='GPS start time in seconds')
parser.add_argument('GPSend', type=int, default=900002000, help='GPS end time in seconds')
parser.add_argument('time_step', type=int, default=100, help='Time separation between two successive injections in seconds')
parser.add_argument('duration', type=int, default= 4, help='Duration of analysis in seconds [starts at TrigTime-2]')
parser.add_argument('minSNR', type=int, default=10, help='Minimum SNR value for the SNR uniform distribution')
parser.add_argument('maxSNR', type=int, default=30, help='Maximum SNR value for the SNR uniform distribution')
parser.add_argument('GWmodel', type=str, default='IMRPhenomBthreePointFivePN', help='Waveform model (e.g., TaylorT4threePointFivePN, IMRPhenomBthreePointFivePN)')
parser.add_argument('Niter', type=int, default=100000, help='Number of iterations used for reconstructing the signal')
parser.add_argument('seed', type=int, default=1234, help='Seed for the random number generator')
args = parser.parse_args()

######################################################################################################################
# Function that writes the condor submission script
# Will work for BayesWave_type=BWB (BayesWaveBurst) and BayesWave_type=BWP (BayesWavePost)
######################################################################################################################
def WriteSubFile(jobDirName, PWD, TriggerTime, flow, fhigh, duration, xml_injection_file, event_number, BayesWave_type):
    path=PWD+'/'+jobDirName+'/'
    SubFile = open(path+"submit_"+BayesWave_type+".sub", "w")
    if BayesWave_type == "BWB":
        SubFile.write("executable="+BayesWaveBurst+"\n")
    if BayesWave_type == "BWP":
        SubFile.write("executable="+BayesWavePost+"\n")
    SubFile.write("universe=standard\n")
    SubFile.write("arguments=--ifo H1 --H1-flow "+str(flow)+" --H1-channel H1:LALAdLIGO \\\n")
    SubFile.write("--ifo L1 --L1-flow "+str(flow)+" --L1-channel L1:LALAdLIGO \\\n")
    SubFile.write("--ifo V1 --V1-flow "+str(flow)+" --V1-channel V1:LALAdLIGO \\\n")
    SubFile.write("--H1-cache LALAdLIGO \\\n")
    SubFile.write("--L1-cache LALAdLIGO \\\n")
    SubFile.write("--V1-cache LALAdLIGO \\\n")
    SubFile.write("--trigtime "+str(TriggerTime)+" --srate "+str(fhigh)+" --seglen "+str(duration)+" \\\n")
    SubFile.write("--bayesLine --PSDstart "+str(TriggerTime)+" --PSDlength "+str(duration)+" \\\n")
    SubFile.write("--outputDir "+path+" \\\n")
    if BayesWave_type == "BWB":
    	SubFile.write("--Niter "+str(Niter)+" \\\n")
    SubFile.write("--dataseed 12345 \\\n")
    if BayesWave_type == "BWP":
    	SubFile.write("--0noise \\\n")
    	SubFile.write("--inj "+str(PWD)+"/"+str(xml_injection_file)+" --event "+str(event_number)+" \n\n")
    if BayesWave_type == "BWB":
    	SubFile.write("--inj "+str(PWD)+"/"+str(xml_injection_file)+" --event "+str(event_number)+" \\\n")
        SubFile.write("--NCmin 25 --NCmax 25 --noAdaptTemperature --tempSpacing 1.5 \\\n")
        if flow % 2 == 0:
            SubFile.write("--clusterProposal \\\n")
    	SubFile.write("--gnuplot\n\n")
    SubFile.write("output="+path+"condorOut_"+BayesWave_type+".txt\n")
    SubFile.write("error="+path+"condorError_"+BayesWave_type+".txt\n")
    SubFile.write("log="+path+"condorLog_"+BayesWave_type+".txt\n")
    SubFile.write("notification=never\n")
    SubFile.write("should_transfer_files=YES\n")
    SubFile.write("when_to_transfer_output=ON_EXIT\n")
    SubFile.write("stream_error=True\n")
    SubFile.write("stream_output=True\n")
    SubFile.write("queue 1\n")
    SubFile.close()

######################################################################################################################
# Function that generates a single dag file for multiple BayesWaveBurst and BayesWavePost runs
# Will work for BayesWave_type=BWB (BayesWaveBurst) and BayesWave_type=BWP (BayesWavePost)
######################################################################################################################
def WriteDagFile(jobDirName, PWD, BayesWave_type):
    DagFile = open("submitRuns.dag", "a")
    if BayesWave_type == "BWB":
    	DagFile.write("JOB "+jobDirName+" "+PWD+"/"+jobDirName+"/submit_BWB.sub\n")
    	DagFile.write("RETRY "+jobDirName+" 1\n\n")
    if BayesWave_type == "BWP":
    	DagFile.write("JOB "+jobDirName+"_post "+PWD+"/"+jobDirName+"/submit_BWP.sub\n")
    	DagFile.write("RETRY "+jobDirName+"_post 1\n\n")
    	DagFile.write("PARENT "+jobDirName+" CHILD "+jobDirName+"_post \n\n")
    DagFile.close()

######################################################################################################################
# Main
######################################################################################################################

# Find out the path to the BayesWaveBurst executable
p = subprocess.Popen(["which","BayesWaveBurst"],stdout=subprocess.PIPE,stderr=subprocess.STDOUT)
for line in iter(p.stdout.readline, b''):
    BayesWaveBurst = line.rstrip()
if not os.path.isfile(BayesWaveBurst):
    print "****************************************************************************************************************"
    sys.exit("\nMake sure the BayesWaveBurst command is in your path before running this script!\n\n****************************************************************************************************************")

# Find out the path to the BayesWavePost executable                                                                                                                    
p = subprocess.Popen(["which","BayesWavePost"],stdout=subprocess.PIPE,stderr=subprocess.STDOUT)
for line in iter(p.stdout.readline, b''):
    BayesWavePost = line.rstrip()
if not os.path.isfile(BayesWavePost):
    print "****************************************************************************************************************"
    sys.exit("\nMake sure the BayesWavePost command is in your path before running this script!\n\n**********************************************************************************************")

# Determine what directory this is and drop the program name from the list of arguments
sys.argv.pop(0)
username = pwd.getpwuid(os.getuid())[0]
pathname = os.getcwd().split(username)[1]
pathname = '/home/'+username+pathname
PWD = os.path.abspath(pathname)

# Determine the masses for the injections
m1 = args.m1
m2 = args.m2
# Determine the frequency interval
flow = args.flow
fhigh = args.fhigh
# Determine the GPS time interval for the injections and their time separation
GPSstart = args.GPSstart
GPSend = args.GPSend
time_step = args.time_step
# Determine the duration of the analysis
duration = args.duration
# Determine the minimum and maximum for the uniform SNR distribution used to generate the injections
minSNR = args.minSNR
maxSNR = args.maxSNR
# Determine the GWform model
GWmodel = args.GWmodel
# Determine the number of iterations
Niter =args.Niter
# Determine the seed for the random number generator
seed = args.seed
# The xml injection file name
xml_injection_file = ""+str(m1)+"_"+str(m2)+".xml"

# Create an XML file to generate the waveforms
inspinj_command = "lalapps_inspinj --gps-start-time "+str(GPSstart)+" --gps-end-time "+str(GPSend)+" --waveform "+str(GWmodel)+" --l-distr random --i-distr uniform --m-distr fixMasses --f-lower "+str(flow)+" --fixed-mass1 "+str(m1)+" --fixed-mass2 "+str(m2)+" --disable-spin --t-distr fixed --time-step "+str(time_step)+" --snr-distr uniform --min-snr "+str(minSNR)+" --max-snr "+str(maxSNR)+" --ifos H1,L1,V1 --virgo-fake-psd LALAdVirgo --ligo-fake-psd LALAdLIGO --ligo-start-freq "+str(flow)+" --virgo-start-freq "+str(flow)+" --output "+str(xml_injection_file)+" --seed "+str(seed)
os.system(inspinj_command)

TriggerTime = GPSstart
event_number = 0

# BayesWaveBurst analysis window ends 2 seconds after the trigger time
while TriggerTime+2.0 < GPSend:
    # Define directory and create folder
    if event_number > 9:
        event_numberStr = str(event_number)
    else:
        event_numberStr = "0"+str(event_number)
    jobDirName = "job_"+str(m1)+"_"+str(m2)+"_"+event_numberStr
    os.system("mkdir "+jobDirName)

    # Write the condor submission scripts to the job directory
    WriteSubFile(jobDirName, PWD,TriggerTime, flow, fhigh, duration, xml_injection_file, event_number, "BWB")
    WriteSubFile(jobDirName, PWD,TriggerTime, flow, fhigh, duration, xml_injection_file, event_number, "BWP")

    # Write the dag file entries
    WriteDagFile(jobDirName, PWD, "BWB")
    WriteDagFile(jobDirName, PWD, "BWP")

    TriggerTime = TriggerTime + time_step
    event_number = event_number + 1

os.system('mv SNRfile.txt '+str(m1)+"_"+str(m2)+'_SNR.txt')

# All done! Remind the user of what to do...
print "****************************************************************************************************************"
print "\n\nJobs created:", event_number
print "\n\n****************************************************************************************************************"
print "\n\nTo run the condor jobs, type\n condor_submit_dag submitRuns.dag"
print "\nTo monitor a job\'s output type\n tail -f <job-directory>/condorOutput*.txt"
print "\nTo monitor a job\'s errors type\n tail -f <job-directory>/condorError*.txt\n"
print "****************************************************************************************************************"

######################################################################################################################
######################################################################################################################

