import matplotlib
matplotlib.use('Agg')
import numpy as np
import matplotlib.pyplot as plt
import glob
import os
import sys
from pylal.xlal.datatypes.ligotimegps import LIGOTimeGPS
from pylal.xlal.date import XLALGreenwichSiderealTime
import find_trig as ft
import healpy as hp
from subprocess import call
# from scipy.optimize import curve_fit
import acor
from pylal import bayespputils
import sky_area.sky_area_clustering as sky
import getopt

# -------------------------------------------
# Module to make skymaps, skyview webpage, etc.
# Works with single output directory from 
# BayesWaveBurst
# -------------------------------------------
def make_skyview(directory='.', mdc=None, NSIDE=512, results=None, sim=True):

    # -- Close any open figures
    plt.close('all')

    # ----------------
    # Read S6 MDC log file
    # -----------------
    if mdc is None and sim:
        print "WARNING: No MDC object provided, using default (see find_trig.py)"
        print "Reading MDC log (this may take a few minutes) ..."
        mdc = ft.Mdc()
        print "I read the mdc log!"

    # -- Save PWD for future reference
    topdir = os.getcwd()

    # -- Enter requested directory
    os.chdir(directory)
    print "Entered", os.getcwd()
    jobname = glob.glob('*bayeswave.run')[0]
    
    # ---------------------------------------
    # Extract information from bayeswave.run
    # Note: We may only need the trigger time from this
    # ---------------------------------------
    bayeswave = open(jobname, 'r')
    ifoNames = []
    for line in bayeswave:
        spl = line.split()
        if len(spl) < 1: continue
        
        # Determine names of IFOs and MDC scale factor
        if spl[0] == 'Command' and spl[1] == 'line:':
            for index, splElement in enumerate(spl):
                if splElement == '--ifo':
                    ifoNames.append(spl[index+1]) 
                    
        del spl[-1]
        # Determine number of IFOs
        if ' '.join(spl) == 'number of detectors':
            spl = line.split()
            ifoNum = spl[3]
            continue         
        # Determine gps trigger time
        if ' '.join(spl) == 'trigger time (s)':
            spl = line.split()
            gps = spl[3]
            break
    bayeswave.close()

    # --------------------------------
    # Create skymap summary statistics
    # --------------------------------
    # -- Input skymap data
    # num, L, ralist, sin_dec, psi, e, dA, dphi, dt = np.loadtxt(filename, unpack=True)
    print "Extracting RA/DEC samples"
    filename = './chains/signal_1_extchain.dat.0'
    data = np.loadtxt(filename, unpack=True)
    ralist = data[2]
    sin_dec = data[3]
    declist = np.arcsin(sin_dec)
    thetalist = np.pi/2.0 - declist

    # -- Determin auto-correlation length
    tau, mean, sigma = acor.acor(ralist)
    acl = int(tau)
    
    # -- Thin the chains
    ralist = ralist[::acl]
    declist = declist[::acl]
    thetalist = thetalist[::acl]
    radec = np.column_stack((ralist, declist))

    # -- Make the KDE map
    redo = True
    count = 0
    while redo:
        try:
            kde = sky.ClusteredKDEPosterior(radec)
            redo = False
        except:
            count += 1
            print "KDE failure number {0}".format(count)
            print "Will attempt to try again"

    # -- Get the sidereal time
    print "Finding sidereal time"
    trigtime = float(gps)
    lalgps = LIGOTimeGPS(trigtime)
    sidtime = XLALGreenwichSiderealTime(lalgps, 0)
    sidtime = sidtime % (np.pi*2)

    # -- Get the injection location
    injtheta, injphi = mdc.get_theta_phi(trigtime)
    injra = injphi + sidtime
    injdec = np.pi/2 - injtheta
    print injtheta, injra

    # -- Plot the skymap and injection location
    skymap = kde.as_healpix(NSIDE, nest=True)
    hp.mollview(skymap, title='Skymap from KDE clusters', nest=True)
    hp.projscatter(injtheta, injra, s=200, c='yellow', marker=(5,1), alpha=0.5)
    plt.savefig('skymap.png')
    plt.close()
    hp.write_map("skymap_{0}.fits".format(gps), skymap, nest=True)

    # -- Get the found contour and searched areas
    sq_deg = (180.0/np.pi)**2
    print "Calculating p value of injection"
    injcontour = kde.p_values( np.array([[injra, injdec]]) )[0]
    print "Got contour of {0}".format(injcontour)
    
    print "Calculating sky area ..."
    (area50, area90) = kde.sky_area( [0.5, 0.9] )*sq_deg
    print "Got area50 of {0}".format(area50)
    
    print "Calculating searched area..."
    searcharea = (kde.searched_area( np.array([[injra, injdec]]) )*sq_deg)[0]
    print "Got searched area of {0}".format(searcharea)
    
    pixarea = hp.nside2pixarea(NSIDE, degrees=True)

    if results is not None:
        for name in ['injcontour', 'area50', 'area90', 'searcharea']:
            if name not in results: results[name] = []
            results[name].append(eval(name)) 

    # -- Compute autocorrelation functions
    acf = bayespputils.autocorrelation(ralist)
    plt.figure()
    plt.plot(acf,'o', markersize=2)
    plt.title('RA autocorrelation', fontsize=14)
    plt.savefig('ra_acf.png')
    plt.close()

    # -- Try scatter plot of ra
    plt.figure()
    plt.plot(ralist, 'o', markersize=2)
    plt.title('RA chain: {0} samples'.format(ralist.size), fontsize=14)
    plt.savefig('ra_chain.png')

    # -- Make an output file with key statistics
    outfile = open('skystats.txt', 'w')
    outfile.write("# area50  area90  searcharea  injcontour \n")
    outfile.write("{0} {1} {2} {3}".format(area50, area90, searcharea, injcontour))
    outfile.close()

    # -- Since we've done all the sky stuff, let's try making the skyview web page right here
    skyview = open('skyview.html','w')
    skyview.write('<html><head><title>Skyview</title></head><body>')
    skyview.write('<h1>Sky View for trigger at {0}</h1>'.format(trigtime))
    skyview.write('<h2>Links</h2>')
    skyview.write('<a href="./index.html">Event Overview</a> for this trigger<br/>')
    skyview.write('<a href="../../top.html">Top level page</a> for this injection set<br/>')
    skyview.write('<h2>Statistics</h2>')
    skyview.write('The area per pixel is {0:.2f} sq. degrees.<br/>'.format(pixarea))
    skyview.write('The area of the 50% region is {0:.2f} sq. degrees.<br/>'.format(area50))
    skyview.write('The area of the 90% region is {0:.2f} sq. degrees.<br/>'.format(area90))
    skyview.write('The searched area is {0:.2f} sq. degrees.<br/>'.format(searcharea))
    skyview.write('The injection was found at the {0:.2f} percent area contour.<br/>'.format(injcontour * 100.0))
    skyview.write('<h2>Plots</h2>')
    skyview.write('<img src="skymap.png" width=600>')
    skyview.write('<img src="ra_chain.png" width=600>')
    skyview.write('<img src="ra_acf.png" width=600>')   
    skyview.write('</body></html>')
    skyview.close()
        
    os.chdir(topdir)

# -- Write main script for command line running
if __name__ == "__main__":

    opts, args = getopt.getopt(sys.argv[1:], "", ['directory=', 'mdc=', 'NSIDE=', 'sim='])
    
    # -- Set default argument values
    directory='.'
    mdc=None
    NSIDE=512
    results=None
    sim=True
    
    print "Got these arguments"
    print opts

    print "Missed these:"
    print args

    for opt, arg in opts:
        print opt
        if opt=='--mdc':
            # -- mdc argument should be the name of the MDC log
            print "Reading MDC log {0}".format(arg)
            mdc = ft.Mdc(arg)
        if opt == '--directory':
            directory = arg
        if opt == '--NSIDE':
            NSIDE = int(arg)
        if opt == '--sim':
            sim = 'True' == arg

    make_skyview(directory, mdc, NSIDE, results, sim)

