/*
 Compile: See Makefile in bayeswave directory
 */

/***************************  REQUIRED LIBRARIES  ***************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <math.h>
#include <sys/stat.h>

#include "Constants.h"
#include "Declarations.h"
#include "numrec.h"
#include "BayesLine.h"

/*************  PROTOTYPE DECLARATIONS FOR INTERNAL FUNCTIONS  **************/



void BurstRJMCMC(struct Data *data, struct Model **model, struct Chain *chain, struct Prior *prior, double *logEvidence, double *varLogEvidence);

void print_help_message(void);
void print_run_flags(FILE *fptr, struct Data *data, struct Prior *prior);
void print_run_stats(FILE *fptr, struct Data *data, struct Chain *chain);

void write_gnuplot_script_header(FILE *script, char modelname[], double Tobs);
void write_gnuplot_script_frame (FILE *script, char modelname[], char runname[], double Tobs, int phase, int signal, int glitch, int cycle, int NI);
void write_bayesline_gnuplot_script_frame(FILE *script, char modelname[], double Tobs, int cycle);
void write_bayeswave_gnuplot_script_frame(FILE *script, char modelname[], char runname[], double Tobs, double fmin, double fmax, double SAmin, double SAmax, int phase, int signal, int glitch, int cycle, int NI);
void write_evidence_gnuplot_script(FILE *script, char modelname[]);

void print_chain_status(int mc, struct Data *data, struct Chain *chain, struct Model **model, int searchFlag, int NC);
void print_time_domain_waveforms(char filename[], double *h, int N, double *Snf, double eta, double Tobs, double fix, int imin, int imax, double tmin, double tmax);
void print_bayesline_spectrum(char filename[], double *f, double *power, double *Sbase, double *Sline, int N);








/* ============================  MAIN PROGRAM  ============================ */

int main(int argc, char *argv[])
{

  /*   Variable declaration   */
  int i, ic, ifo, imin, imax;

  char filename[100];

  /* LAL data structure that will contain all of the data from the frame files */
  LALInferenceRunState *runState = XLALCalloc(1,sizeof(LALInferenceRunState));

  runState->commandLine = LALInferenceParseCommandLine(argc,argv);

  /******************************************************************************/
  /*                                                                            */
  /*  Usage statment if --help                                                  */
  /*                                                                            */
  /******************************************************************************/

  if(LALInferenceGetProcParamVal(runState->commandLine, "--help") || argc==1)
  {
    print_help_message();
    return 0;
  }

  /******************************************************************************/
  /*                                                                            */
  /*  Read/store  data                                                          */
  /*                                                                            */
  /******************************************************************************/
  fprintf(stdout, "\n");
  fprintf(stdout, " ======= LALInferenceReadData() ======\n");
  runState->data = LALInferenceReadData(runState->commandLine);
  fprintf(stdout, "\n");


  /******************************************************************************/
  /*                                                                            */
  /*  read/store data                                                           */
  /*                                                                            */
  /******************************************************************************/
  LALInferenceIFOData *dataPtr = NULL;

  dataPtr = runState->data;

  double *SNRinj = NULL;

  int NI;
  int N = dataPtr->timeData->data->length;
  double Tobs = (double)N*dataPtr->timeData->deltaT;

  ifo=0;
  while(dataPtr!=NULL)
  {
    dataPtr = dataPtr->next;
    ifo++;
  }
  NI=ifo;

  double **freqData = dmatrix(0,NI-1,0,N-1);
  double **psd      = dmatrix(0,NI-1,0,N/2-1);

  struct Data *data   = malloc(sizeof(struct Data));
  struct Chain *chain = malloc(sizeof(struct Chain));
  struct Prior *prior = malloc(sizeof(struct Prior));

  parse_command_line(data, chain, prior, runState->commandLine);

  //Move to the directory specified by the user, or stay in ./
  chdir(data->outputDirectory);
  fprintf(stdout,"\nOutputting all data to: %s \n\n", data->outputDirectory);

  mode_t process_mask = umask(0);
  mkdir("chains",   S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
  mkdir("waveforms",S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
  sprintf(filename,"%ssnr",data->runName);
  mkdir(filename,   S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
  umask(process_mask);

  // burstMDC injection
  if(LALInferenceGetProcParamVal(runState->commandLine, "--MDC-cache"))
  {
    SNRinj = malloc((NI+1)*sizeof(double));
    fprintf(stdout, " ==== LALInferenceInjectFromMDC(): started. ====\n");
    InjectFromMDC(runState->commandLine, runState->data, SNRinj);
    fprintf(stdout, " ==== LALInferenceInjectFromMDC(): finished. ====\n\n");
    free(SNRinj);
  }

  // CBC injection
  if(LALInferenceGetProcParamVal(runState->commandLine, "--inj"))
  {
    fprintf(stdout, " ==== LALInferenceInjectInspiralSignal(): started. ====\n");
    LALInferenceInjectInspiralSignal(runState->data,runState->commandLine);
    fprintf(stdout, " ==== LALInferenceInjectInspiralSignal(): finished. ====\n\n");
  }


  //Output Fourier domain data for BayesLine
  dataPtr = runState->data;

  //TODO: We have to pick frequency resolution for TF proposal.  What's optimal?
  //TODO: A continuous density would be better and do-able -- store (fraction of) burnin wavelets, ~KDE (has Q information)

  double fmin   = dataPtr->fLow;
  double fmax   = ceil(dataPtr->fHigh);
  double deltaT = dataPtr->timeData->deltaT;

  //set frequency resolution to be 32 Hz because why not?
  double df,dt;

  df = 32.0;
  dt = 0.5/df;

  int tsize = (int)(Tobs/dt);

  //Copy Fourier domain data & PSD from LAL data types into simple arrays
  ifo=0;
  imin = (int)(fmin*Tobs);
  while(dataPtr!=NULL)
  {
    for(i=0; i<N/2; i++)
    {
      if(i<imin)
      {
        freqData[ifo][2*i]   = 0.0;
        freqData[ifo][2*i+1] = 0.0;
        psd[ifo][i] = 1.0;
      }
      else
      {
        freqData[ifo][2*i]   = creal(dataPtr->freqData->data->data[i]);
        freqData[ifo][2*i+1] = cimag(dataPtr->freqData->data->data[i]);

        /*
         Innerproducts are expecting <n_i^2> instead of Sn(f):
         So psd arrays == <n_i^2> = T/2 * Sn(f)
         */
        psd[ifo][i] = dataPtr->oneSidedNoisePowerSpectrum->data->data[i]*Tobs/2.0;
      }
    }
    dataPtr = dataPtr->next;
    ifo++;
  }

  /******************************************************************************/
  /*                                                                            */
  /*  Setup DATA, CHAIN, PRIOR, MODEL structures                                */
  /*                                                                            */
  /******************************************************************************/

  /*
   DATA
   */

  initialize_data(data,freqData,N,tsize,psd,Tobs,NI,fmin,fmax);


  // Print Fourier domain data to file for post-processing
  FILE *FDdata = NULL;
  int re,im;
  for(ifo=0; ifo<data->NI; ifo++)
  {
    sprintf(filename,"chains/%sfourier_domain_data_ifo%i.dat",data->runName,ifo);

    FDdata = fopen(filename,"w");
    for(i=0; i<data->N/2; i++)
    {
      re = 2*i;
      im = 2*i+1;

      if(i<data->imin || i>data->imax)
        fprintf(FDdata,"%e %e\n", 0.0 , 0.0);
      else
        fprintf(FDdata,"%e %e\n", data->s[ifo][re] , data->s[ifo][im]);
    }
    fclose(FDdata);
  }

  //likelhood is a pointer to a function which returns logL
  if(data->constantLogLFlag)
  {
    // logL=const to test detailed balance
    data->intrinsic_likelihood = EvaluateConstantLogLikelihood;
    data->extrinsic_likelihood = EvaluateExtrinsicConstantLogLikelihood;
  }
  else
  {
    // Start with search-phase log likelihood
    data->intrinsic_likelihood = EvaluateMarkovianLogLikelihood;//EvaluateSearchLogLikelihood;

    // Jump right into characterization likelihood if noise-only run
    if(!data->glitchFlag && !data->signalFlag) data->intrinsic_likelihood = EvaluateMarkovianLogLikelihood;

    //data->extrinsic_likelihood = EvaluateExtrinsicSearchLogLikelihood;
    data->extrinsic_likelihood = EvaluateExtrinsicMarkovianLogLikelihood;
  }

  data->detector  = malloc(NI*sizeof(LALDetector*));
  data->epoch     = runState->data->epoch;

  ifo=0;
  dataPtr = runState->data;
  while(dataPtr!=NULL)
  {
    data->detector[ifo] = dataPtr->detector;
    ifo++;
    dataPtr = dataPtr->next;
  }

  /*
   CHAIN
   */
  chain->NC = chain->NCmax;
  allocate_chain(chain,NI);
  for(ic=0; ic<chain->NC; ic++) chain->index[ic]=ic;

  /*
   PRIOR
   */
  int Dmax = data->Dmax;
  data->Dmax = 200;

  if(data->noiseOnlyFlag) Dmax=data->Dmax=1;
  int omax = 10;
  if(data->Dmax < omax) omax = data->Dmax;

  // maximum number of sineGaussians
  prior->gmin = ivector(0,data->NI-1);
  prior->gmax = ivector(0,data->NI-1);
  /* Wavelet parameters */
  prior->range = dmatrix(0,4,0,1);


  // sanity check use of clustering prior
  if(data->clusterPriorFlag)
  {
    //sprintf(filename,"%s/cluster_norm_flow%i_seglen%i_srate%i_Qmax%i.dat",prior->path,(int)data->fmin,(int)data->Tobs,(int)data->fmax*2,(int)data->Qmax);
    sprintf(filename,"%s/cluster_norm_flow%i_twin%i_srate%i_Qmax%i.dat",prior->path,(int)data->fmin,(int)data->Twin,(int)data->fmax*2,(int)data->Qmax);
    printf("opening file %s\n",filename);
    if(!checkfile(filename))
    {
      fprintf(stdout,"\n");
      fprintf(stdout,"ERROR: Missing file to correct bias in clustering prior:\n");
      fprintf(stdout,"  path:   %s\n",filename);
      fprintf(stdout,"  flow:   %i\n",(int)data->fmin);
      fprintf(stdout,"  seglen: %i\n",(int)data->Tobs);
      fprintf(stdout,"  srate:  %i\n",(int)data->fmax*2);
      fprintf(stdout,"  Qmax:   %i\n",(int)data->Qmax);
      fprintf(stdout,"TRY:\n");
      fprintf(stdout,"  Creating file to correct for bias\n");
      fprintf(stdout,"  Checking your --clusterPath\n");
      fprintf(stdout,"  Removing --clusterPrior\n\n");
      return 1;
    }
  }
  initialize_priors(data, prior, omax);

  if(data->amplitudePriorFlag)
  {
    data->glitch_amplitude_proposal = draw_glitch_amplitude;
    data->glitch_amplitude_prior    = glitch_amplitude_prior;

    if(data->signalAmplitudePriorFlag)
    {
      data->signal_amplitude_proposal = draw_signal_amplitude;
      data->signal_amplitude_prior    = signal_amplitude_prior;
    }
    else
    {
      data->signal_amplitude_proposal = draw_glitch_amplitude;
      data->signal_amplitude_prior    = glitch_amplitude_prior;
    }
  }

  /*
   MODEL
   */
  struct Model **model = malloc(chain->NC*sizeof(struct Model*));
  for(ic=0; ic<chain->NC; ic++)
  {
    model[ic] = malloc(sizeof(struct Model));
    initialize_model(model[ic], data, prior, &chain->seed);
  }

  //print run summary file
  sprintf(filename,"%sbayeswave.run",data->runName);
  chain->runFile = fopen(filename,"w");
  fprintf(chain->runFile,"  %s\n\n",LALInferencePrintCommandLine(data->commandLine) );
  print_run_stats(chain->runFile, data, chain);
  fclose(chain->runFile);

  /******************************************************************************/
  /*                                                                            */
  /*  PSD search & glitch cleaning stages are only done once                    */
  /*  Results are dumped to file so that we can rerun individual models         */
  /*  without redoing the entire cleaning phase                                 */
  /*                                                                            */
  /******************************************************************************/

  FILE *psdFile = NULL;
  FILE *asdFile = NULL;
  FILE *resFile = NULL;

  struct BayesLineParams **bayesline = malloc(NI*sizeof(struct BayesLineParams *));
  data->bayesline = bayesline;

  if(data->bayesLineFlag)
  {
    fprintf(stdout,"\n ============ BayesLine ==============\n");
    fflush(stdout);

    for(ifo=0; ifo<NI; ifo++)
    {
      bayesline[ifo] = malloc(sizeof(struct BayesLineParams));
      bayesline[ifo]->priors = malloc(sizeof(BayesLinePriors));

      sprintf(filename,"chains/%ssearch_splinechain_ifo%i.dat",data->runName,ifo);
      data->bayesline[ifo]->splineChainFile = fopen(filename,"w");
      sprintf(filename,"chains/%ssearch_lorentzchain_ifo%i.dat",data->runName,ifo);
      data->bayesline[ifo]->lineChainFile = fopen(filename,"w");

      //Set BayesLine priors based on the data channel being used

      //S6 strain
      if(!strcmp(data->channels[ifo],"LDAS-STRAIN"))
      {
        bayesline[ifo]->priors->SAmin = 1.0e-48;
        bayesline[ifo]->priors->SAmax = 1.0e-25;
        bayesline[ifo]->priors->LQmin = 1.0e2;
        bayesline[ifo]->priors->LQmax = 1.0e8;
        bayesline[ifo]->priors->LAmin = 1.0e-44;
        bayesline[ifo]->priors->LAmax = 1.0e-20;
      }

      //VSR2 strain
      else if (!strcmp(data->channels[ifo],"h_16384Hz"))
      {
        bayesline[ifo]->priors->SAmin = 1.0e-48;
        bayesline[ifo]->priors->SAmax = 1.0e-30;
        bayesline[ifo]->priors->LQmin = 1.0e2;
        bayesline[ifo]->priors->LQmax = 1.0e8;
        bayesline[ifo]->priors->LAmin = 1.0e-44;
        bayesline[ifo]->priors->LAmax = 1.0e-30;
      }

      //S6 LIGO DARM Error
      else if (!strcmp(data->channels[ifo],"LSC-DARM_ERR"))
      {
        bayesline[ifo]->priors->SAmin = 1.0e-18;
        bayesline[ifo]->priors->SAmax = 1.0e0;
        bayesline[ifo]->priors->LQmin = 1.0e2;
        bayesline[ifo]->priors->LQmax = 1.0e8;
        bayesline[ifo]->priors->LAmin = 1.0e-14;
        bayesline[ifo]->priors->LAmax = 1.0e0;
      }

      else if (!strcmp(data->channels[ifo],"LALAdLIGO"))
      {
        bayesline[ifo]->priors->SAmin = 1.0e-50;
        bayesline[ifo]->priors->SAmax = 1.0e-43;
        bayesline[ifo]->priors->LQmin = 1.0e2;
        bayesline[ifo]->priors->LQmax = 1.0e8;
        bayesline[ifo]->priors->LAmin = 1.0e-47;
        bayesline[ifo]->priors->LAmax = 1.0e-43;
      }

      //Use S6 strain as default
      else
      {
        bayesline[ifo]->priors->SAmin = 1.0e-48;
        bayesline[ifo]->priors->SAmax = 1.0e-25;
        bayesline[ifo]->priors->LQmin = 1.0e2;
        bayesline[ifo]->priors->LQmax = 1.0e8;
        bayesline[ifo]->priors->LAmin = 1.0e-44;
        bayesline[ifo]->priors->LAmax = 1.0e-20;
      }

      fprintf(stdout,"Running search phase for IFO %i\n", ifo);
      BayesLineSearch(bayesline[ifo], freqData[ifo], fmin, fmax, deltaT, Tobs);
    }
    data->fmax = bayesline[0]->data->fhigh;
  }

  if(!LALInferenceGetProcParamVal(runState->commandLine, "--restart"))
  {
    /******************************************************************************/
    /*                                                                            */
    /*  BayesLine Characterization Phase                                          */
    /*                                                                            */
    /******************************************************************************/

    if(data->bayesLineFlag)
    {

      for(ifo=0; ifo<NI; ifo++)
      {
        fprintf(stdout,"Running characterization phase for IFO %i\n", ifo);
        BayesLineRJMCMC(bayesline[ifo], freqData[ifo], data->Snf[ifo], data->SnS[ifo], N, chain->cycle);

        fclose(data->bayesline[ifo]->splineChainFile);
        fclose(data->bayesline[ifo]->lineChainFile);
      }
      fprintf(stdout,"\n");
      fflush(stdout);
    }

    /******************************************************************************/
    /*                                                                            */
    /*  The Full-time MCMC Run (Cleaning phase)                                   */
    /*                                                                            */
    /******************************************************************************/


    if(!data->skipCleanFlag)
    {

      //lie to the code about the number of chains & iterations
      data->Dmax = 200;

      int M    = chain->count;
      int NC   = chain->NC;
      int CP   = data->clusterPriorFlag;
      data->clusterPriorFlag = 0;
      chain->count = M/5;
      chain->NC    = NC/5;
      if(chain->NC < 1)chain->NC=1;

      //Initialize proposal ratios for glitch cleaning phase
      chain->modelRate   = 0.0; //even chance for each channel
      chain->rjmcmcRate  = 0.01;                   //model fixed to one wavelet

      data->glitchFlag = 1;
      data->signalFlag = 0;

      data->runPhase = 0;
      if(data->bayesLineFlag) data->psdFitFlag = 0;

      initialize_chain(chain,1);

      reset_likelihood(data);
      reset_priors(data, prior);
      reset_model(data, chain, prior, model);

      double cleanLogZ, cleanVarZ;
      printf("characterizing PSD+glitch model...\n\n");

      BurstRJMCMC(data, model, chain, prior, &cleanLogZ, &cleanVarZ);

      /******************************************************************************/
      /*                                                                            */
      /*  Remove excess power outside the trigger time, reset dimension priors      */
      /*                                                                            */
      /******************************************************************************/

      data->runPhase = 1;

      chain->count = M;
      chain->NC    = NC;
      data->Dmax   = Dmax;
      data->clusterPriorFlag = CP;

      reset_priors(data, prior);

      //Permanently regress any glitches outside of the restricted prior range
      fprintf(stdout, "\n");
      fprintf(stdout, " Removing excess power outside of t=[%g,%g]\n",prior->range[0][0],prior->range[0][1]);
      int j;
      double t0;

      //SNR of signal remaining in data
      double rSNR=0.0;

      struct Wavelet *glitch = NULL;
      for(ifo=0; ifo<NI; ifo++)
      {
        glitch = model[chain->index[0]]->glitch[ifo];
        for(j=1; j<=glitch->size; j++)
        {
          i  = glitch->index[j];
          t0 = glitch->intParams[i][0];
          if(t0<prior->range[0][0] || t0>prior->range[0][1])
          {
            //regress from the data
            SineGaussian(data->s[ifo], glitch->intParams[i], N, -1, data->Tobs);

            //regress from glitch model se we can compute window SNR
            SineGaussian(glitch->templates, glitch->intParams[i], N, -1, data->Tobs);

          }
        }
        rSNR += fourier_nwip(data->imin,data->imax,glitch->templates,glitch->templates,data->Snf[ifo]);
      }
      fprintf(stdout, " Residual SNR = %g\n",sqrt(rSNR));

      //free model structure
      for(ic=0; ic<chain->NC; ic++)
      {
        free_model(model[ic], data, prior);
        free(model[ic]);
      }
      free(model);
      free_chain(chain, NI);

      //use glitch SNR to determine total number of chains
      chain->NC = 0.0;
      if(rSNR>1) chain->NC += (int)ceil(log(rSNR)/log(chain->tempStep));
      if(chain->NC<chain->NCmin)      chain->NC = chain->NCmin;
      else if(chain->NC>chain->NCmax) chain->NC = chain->NCmax;
      fprintf(stdout, " Using %i chains with initial spacing of %g\n\n",chain->NC,chain->tempStep);

      //chain structure
      allocate_chain(chain,NI);

      model = malloc(chain->NC*sizeof(struct Model*));
      for(ic=0; ic<chain->NC; ic++)
      {
        model[ic] = malloc(sizeof(struct Model));
        initialize_model(model[ic], data, prior, &chain->seed);
      }

      //Reset likelihood normalization (mostly aesthetic)
      data->logLc = 0.0;
      if(data->noiseSimFlag)for(ifo=0; ifo<NI; ifo++) data->logLc += loglike(data->imin, data->imax, data->s[ifo], 1.0, data->Snf[ifo]);


      //Export BayesLine PSD and cleaned data so we can re-start runs
      if(data->bayesLineFlag)
      {
        for(ifo=0; ifo<NI; ifo++)
        {
          sprintf(filename,"%sIFO%i_res.dat",data->runName,ifo);
          resFile = fopen(filename,"w");
          sprintf(filename,"%sIFO%i_psd.dat",data->runName,ifo);
          psdFile = fopen(filename,"w");
          sprintf(filename,"%sIFO%i_asd.dat",data->runName,ifo);
          asdFile = fopen(filename,"w");

          imax = (int)(floor(data->bayesline[0]->data->fhigh * data->bayesline[0]->data->Tobs));
          imin = (int)(floor(data->bayesline[0]->data->flow  * data->bayesline[0]->data->Tobs));

          //output PSD
          for(i=imin; i<imax; i++)
          {
            fprintf(psdFile, "%.12g %.12g\n",(double)i/data->Tobs, data->Snf[ifo][i]/(Tobs/2.0));
            fprintf(asdFile, "%.12g %.12g\n",(double)i/data->Tobs, sqrt(data->Snf[ifo][i]/(Tobs/2.0)));
            fprintf(resFile, "%.12g %.12g %.12g\n",(double)i/data->Tobs,data->s[ifo][2*i],data->s[ifo][2*i+1]);
          }

          fclose(psdFile);
          fclose(asdFile);
          fclose(resFile);
        }
      }

      //Shut off BayesLine in favor of PSD fitting
      data->bayesLineFlag = 0;
      data->psdFitFlag    = 1;

    }
    if(data->cleanOnlyFlag)
    {
      /******************************************************************************/
      /*                                                                            */
      /*  Free memory, exit cleanly                                                 */
      /*                                                                            */
      /******************************************************************************/

      for(ic=0; ic<chain->NC; ic++)
      {
        free_model(model[ic], data, prior);
        free(model[ic]);
      }
      free(model);
      free_chain(chain,NI);

      return 0;
    }
  }else //skip cleaning phase and read PSD & residual from file
  {
    fprintf(stdout,"\n ==== Restart BayesWave from file ====\n");
    fflush(stdout);

    data->runPhase = 1;
    data->Dmax = Dmax;
    reset_priors(data, prior);

    //free model structure
    for(ic=0; ic<chain->NC; ic++)
    {
      free_model(model[ic], data, prior);
      free(model[ic]);
    }
    free(model);
    free_chain(chain,NI);

    //get BayesLine PSD and cleaned data so we can re-start runs
    for(ifo=0; ifo<NI; ifo++)
    {
      sprintf(filename,"%sIFO%i_res.dat",data->runName,ifo);
      resFile = fopen(filename,"r");
      sprintf(filename,"%sIFO%i_psd.dat",data->runName,ifo);
      psdFile = fopen(filename,"r");
      sprintf(filename,"%sIFO%i_bl.dat",data->runName,ifo);

      //read in PSD
      double junk;
      for(i=0; i<N/2; i++)
      {
        fscanf(psdFile, "%lg %lg",&junk, &data->Snf[ifo][i]);
        fscanf(resFile, "%lg %lg %lg",&junk,&data->s[ifo][2*i],&data->s[ifo][2*i+1]);
      }

      fclose(psdFile);
      fclose(resFile);
    }

    //chain structure
    allocate_chain(chain,NI);

    model = malloc(chain->NC*sizeof(struct Model*));
    for(ic=0; ic<chain->NC; ic++)
    {
      model[ic] = malloc(sizeof(struct Model));
      initialize_model(model[ic], data, prior, &chain->seed);
    }

    //Reset likelihood normalization (mostly aesthetic)
    data->logLc = 0.0;
    if(data->noiseSimFlag)for(ifo=0; ifo<NI; ifo++) data->logLc += loglike(data->imin, data->imax, data->s[ifo], 1.0, data->Snf[ifo]);

    data->bayesLineFlag = 0;
    data->psdFitFlag    = 1;

  }


  /******************************************************************************/
  /*                                                                            */
  /* Here we redo the injections into the original data just to get an updated  */
  /* estimate of the SNR using the clean PSD.  This does not affect the data    */
  /* used by BayesWave (which lives in the Data structure) because the          */
  /* injection codes only look at the LALInferenceRunState structure            */
  /*                                                                            */
  /******************************************************************************/

  if(LALInferenceGetProcParamVal(runState->commandLine, "--inj"))
  {

    ifo=0;
    dataPtr = runState->data;
    while(dataPtr!=NULL)
    {
      for(i=0; i<N/2; i++) dataPtr->oneSidedNoisePowerSpectrum->data->data[i] = psd[ifo][i]*2.0/Tobs;
      dataPtr = dataPtr->next;
      ifo++;
    }

    fprintf(stdout, " ==== LALInferenceInjectInspiralSignal(): started. ====\n");
    LALInferenceInjectInspiralSignal(runState->data,runState->commandLine);
    fprintf(stdout, " ==== LALInferenceInjectInspiralSignal(): finished. ====\n\n");
  }

  if(LALInferenceGetProcParamVal(runState->commandLine, "--MDC-cache"))
  {
    fprintf(stdout, " ====== Recompute burstMDC SNR =======\n");
    SNRinj = malloc((NI+1)*sizeof(double));

    ifo=0;
    dataPtr = runState->data;
    while(dataPtr!=NULL)
    {
      // Fill runState PSD with results from BayesLine
      for(i=0; i<N/2; i++) dataPtr->oneSidedNoisePowerSpectrum->data->data[i] = data->Snf[ifo][i]*2.0/Tobs;
      dataPtr = dataPtr->next;
      ifo++;
    }

    //Redo the injection to get injected SNR
    InjectFromMDC(runState->commandLine, runState->data, SNRinj);


    double SNRmin = SNRinj[NI];
    for(i=0; i<NI; i++) if(SNRinj[i]<SNRmin) SNRmin=SNRinj[i];

    FILE *snrFile = NULL;
    sprintf(filename,"%ssnr/snr.dat",data->runName);
    snrFile = fopen(filename,"w");
    fprintf(snrFile,"%.12g %.12g\n",SNRmin,SNRinj[NI]);
    fclose(snrFile);

    free(SNRinj);

  }

  /******************************************************************************/
  /*                                                                            */
  /*  Full PT & RJMCMC run for each model under consideration                   */
  /*  Evidence automatically calculated using thermodynamic integration         */
  /*                                                                            */
  /******************************************************************************/

  /*
   Initialize logZ to logL of NULL model, i.e. -(d|d)/2
   That way if models are skipped the Bayes factors tell
   us about the improvement over the naive Gaussian noise
   model
   */
  double logZglitch = data->logLc;
  double logZnoise  = data->logLc;
  double logZsignal = data->logLc;
  double logZfull   = data->logLc;

  double varZglitch = 0.0;
  double varZnoise  = 0.0;
  double varZsignal = 0.0;
  double varZfull   = 0.0;

  data->Dmax = Dmax;

  FILE *evidence;

  //output file format for evidence histogram
  /*
   if(data->noiseOnlyFlag)
   sprintf(data->runName,"%snoiseOnly_",data->runName);
   else if(data->glitchOnlyFlag)
   sprintf(data->runName,"%sglitchOnly_",data->runName);
   else if(data->signalOnlyFlag)
   sprintf(data->runName,"%ssignalOnly_",data->runName);
   else if(data->fullOnlyFlag)
   sprintf(data->runName,"%sfullOnly_",data->runName);
   */

  sprintf(filename,"%sevidence.dat",data->runName);
  evidence = fopen(filename,"w");


  if(data->noiseOnlyFlag || (!data->signalOnlyFlag && !data->glitchOnlyFlag && !data->fullOnlyFlag) )
  {
    /******************************************************************************/
    /*                                                                            */
    /*  Gaussian Noise Characterization phase                                     */
    /*                                                                            */
    /******************************************************************************/

    data->runPhase = 1;
    reset_priors(data, prior);

    //Initialize proposal ratios for GW search phase
    chain->rjmcmcRate  = 0.0;

    //set flags for no wavelets in model
    data->glitchFlag = 0;
    data->signalFlag = 0;

    initialize_chain(chain,1);

    reset_likelihood(data);
    reset_priors(data, prior);
    reset_model(data, chain, prior, model);

    //use fewer iterations and chains and coarser ladder for noise-only model
    //int NC = chain->NC;
    //double dT = chain->tempStep;

    //chain->NC = chain->NCmin;
    chain->count /= 10;
    //chain->tempStep = 1.5;
    BurstRJMCMC(data, model, chain, prior, &logZnoise, &varZnoise);
    //chain->tempStep = dT;
    chain->count *= 10;
    //chain->NC = NC;
  }
  fprintf(evidence,"noise %lg %lg\n",logZnoise,varZnoise);
  fflush(evidence);

  if(data->glitchOnlyFlag || (!data->noiseOnlyFlag && !data->signalOnlyFlag && !data->fullOnlyFlag) )
  {
    /******************************************************************************/
    /*                                                                            */
    /*  Glitch Characterization phase                                             */
    /*                                                                            */
    /******************************************************************************/

    data->runPhase = 1;
    reset_priors(data, prior);

    //Initialize proposal ratios for GW search phase
    chain->modelRate   = 1;//1./(double)(data->NI); //even chance for each channel
    chain->rjmcmcRate  = 0;//0.5;                   //model fixed to one wavelet

    data->glitchFlag = 1;
    data->signalFlag = 0;

    initialize_chain(chain,1);

    reset_likelihood(data);
    reset_priors(data, prior);
    reset_model(data, chain, prior, model);

    BurstRJMCMC(data, model, chain, prior, &logZglitch, &varZglitch);

    /*if(data->glitchOnlyFlag==1 && strcmp(data->channels[0],"AdLIGO"))
     outputFrameFile(runState->commandLine, data, model[chain->index[0]]);*/
  }
  fprintf(evidence,"glitch %lg %lg\n",logZglitch,varZglitch);
  fflush(evidence);

  if(data->signalOnlyFlag || (!data->noiseOnlyFlag && !data->glitchOnlyFlag && !data->fullOnlyFlag) )
  {
    /******************************************************************************/
    /*                                                                            */
    /*  Signal Characterization phase                                             */
    /*                                                                            */
    /******************************************************************************/

    data->runPhase = 1;
    reset_priors(data, prior);

    //Initialize proposal ratios for GW search phase
    chain->modelRate   = 1;   //always do signal model updates
    chain->rjmcmcRate  = 0;   //model fixed to one wavelet for extrinsic search

    data->glitchFlag = 0;
    data->signalFlag = 1;

    initialize_chain(chain,1);

    reset_likelihood(data);
    reset_priors(data, prior);
    reset_model(data, chain, prior, model);

    BurstRJMCMC(data, model, chain, prior, &logZsignal, &varZsignal);
  }
  fprintf(evidence,"signal %lg %lg\n",logZsignal,varZsignal);
  fflush(evidence);

  if(data->fullOnlyFlag)
  {
    /******************************************************************************/
    /*                                                                            */
    /*  The full model RJMCMC Run (Signal+Glitch) Characterization phase          */
    /*                                                                            */
    /******************************************************************************/

    data->runPhase = 1;
    reset_priors(data, prior);

    //Initialize proposal ratios for GW search phase
    chain->modelRate   = 1;   //always do signal model updates during search
    chain->rjmcmcRate  = 0;   //model fixed to one wavelet for extrinsic search

    data->glitchFlag = 1;
    data->signalFlag = 1;

    if(data->bayesLineFlag) data->psdFitFlag = 0;

    initialize_chain(chain,1);

    reset_likelihood(data);
    reset_priors(data, prior);
    reset_model(data, chain, prior, model);

    BurstRJMCMC(data, model, chain, prior, &logZfull, &varZfull);

    fprintf(evidence,"full %lg %lg\n",logZsignal,varZsignal);
    fflush(evidence);
  }

  /******************************************************************************/
  /*                                                                            */
  /*  Output log- and normalized evidence                                       */
  /*                                                                            */
  /******************************************************************************/

  fclose(evidence);

  //output file format for stacked histogram
  sprintf(filename,"%sevidence_stacked.dat",data->runName);
  evidence = fopen(filename,"w");

  fprintf(evidence,"%lg %lg %lg %lg %lg %lg\n",logZnoise,logZglitch,logZsignal,varZnoise,varZglitch,varZsignal);

  fclose(evidence);

  //write gnuplot script to make single-run evidence histograms
  sprintf(filename,"%sevidence.gpi",data->runName);
  evidence = fopen(filename,"w");
  write_evidence_gnuplot_script(evidence, data->runName);


  /******************************************************************************/
  /*                                                                            */
  /*  Free memory, exit cleanly                                                 */
  /*                                                                            */
  /******************************************************************************/

  for(ic=0; ic<chain->NC; ic++)
  {
    free_model(model[ic], data, prior);
    free(model[ic]);
  }
  free(model);
  free_chain(chain,NI);

  return 0;

}

void BurstRJMCMC(struct Data *data, struct Model **model, struct Chain *chain, struct Prior *prior, double *logEvidence, double *varLogEvidence)
{

  print_run_stats(stdout, data, chain);
  print_run_flags(stdout, data, prior);

  int N  = data->N;
  int NI = data->NI;
  int M  = chain->count;
  int NC = chain->NC;
  double Tobs     = data->Tobs;

  int i, j, ic;
  int ix,jx;
  int ifo;
  double fix;
  double x;

  struct Wavelet *wave    = NULL;
  struct Wavelet *signal  = NULL;
  struct Wavelet **glitch = NULL;

  char filename[100];
  char modelname[100];

  // parameters to control how many frames are output to model animations
  int frame=0;
  int frameCount=200;

  // gnuplot animations cadence can be linear or logarithmic in time
  //TODO: linear/log frame cadence should be a command line switch
  //double frameInterval = log((double)M)/(double)frameCount;
  double frameInterval = (double)M/(double)frameCount;

  // I don't understand the factor of 8 here in the scaling of the time domain reconstructions....
  fix = (sqrt(data->Tobs))*sqrt(0.5*(double)N)/8.0;

  // parameters for autocorrelation length
  int count=0;
  int nPoints = M/2/chain->cycle;//M/4/chain->cycle;
  double **logLchain=dmatrix(0,chain->NC-1,0,nPoints-1);


  /******************************************************************************/
  /*                                                                            */
  /*  Set up data structures                                                    */
  /*                                                                            */
  /******************************************************************************/

  /*
   OUTPUT DATA
   */
  int *index = malloc(NC*sizeof(int));

  //give output files common name
  if(data->runPhase==0)sprintf(modelname,"%sclean",data->runName);
  else
  {
    if(!data->glitchFlag && !data->signalFlag)
      sprintf(modelname,"%snoise",data->runName);

    else if(data->glitchFlag && !data->signalFlag)
      sprintf(modelname,"%sglitch",data->runName);

    else if(!data->glitchFlag && data->signalFlag)
      sprintf(modelname,"%ssignal",data->runName);

    else
      sprintf(modelname,"%sfull",data->runName);
  }

  /*
   PTMCMC TEMPERATURE LADDER
   */
  for(ic=0; ic<chain->NC; ic++)
  {
    chain->dT[ic]=chain->tempStep;
    chain->index[ic]=ic;
    if(ic==0)chain->temperature[0] = chain->tempMin;
    else chain->temperature[ic] = chain->temperature[ic-1]*chain->dT[ic-1];
    chain->ptprop[ic]=1;
    chain->ptacc[ic]=0;
    chain->avgLogLikelihood[ic] = 0.0;
    chain->varLogLikelihood[ic] = 0.0;
  }

  /******************************************************************************/
  /*                                                                            */
  /*  Print run stats to file in current directory                              */
  /*  so that the run is replicable                                             */
  /*                                                                            */
  /******************************************************************************/

  sprintf(filename,"%s.run",modelname);
  chain->runFile = fopen(filename,"w");

  fprintf(chain->runFile,"\n =========== BayesWaveBurst ===========\n");
  fprintf(chain->runFile,"  %s\n\n",LALInferencePrintCommandLine(data->commandLine) );
  print_run_flags(chain->runFile, data, prior);


  /******************************************************************************/
  /*                                                                            */
  /*  Set up each models' parameters and output files                           */
  /*                                                                            */
  /******************************************************************************/
  for(ic=0; ic<NC; ic++)
  {
    index[ic]=ic;
    model[ic]->size = 0;

    signal = model[ic]->signal;
    glitch = model[ic]->glitch;

    // clear out whatever is in the signal/glitch model
    for(i=0; i<data->N; i++)
    {
      signal->templates[i] = 0.0;
      for(ifo=0; ifo<NI; ifo++)
      {
        model[ic]->response[ifo][i] = 0.0;
        glitch[ifo]->templates[i] = 0.0;
      }
    }

    // start model parameters anywhere
    if(data->glitchFlag)
    {
      for(ifo=0; ifo<NI; ifo++)
      {
        glitch[ifo]->size = data->Dmin;//1;
        model[ic]->size++;
        for(j=1; j<=glitch[ifo]->size; j++)
        {
          draw_wavelet_params(ifo, glitch[ifo]->intParams[j], data, prior->range, &chain->seed);
          if(data->amplitudePriorFlag)  data->glitch_amplitude_proposal(glitch[ifo]->intParams[j], data->Snf[ifo], 1.0, &chain->seed, data->Tobs, prior->SNRpeak);

          SineGaussian(glitch[ifo]->templates, glitch[ifo]->intParams[j], data->N, 1, Tobs);
        }
      }
    }

    if(data->signalFlag)
    {
      signal->size = 1;
      model[ic]->size++;
      //extrinsic_uniform_proposal(&chain->seed,model[ic]->extParams);

      /*
       WaveformProject() was split into two functions:
       -computeProjectionCoeffs() computes the time delays, F+ and Fx, and
       exp(i Phi(f)), which only change with new extrinsic parameters.
       -waveformProject uses the stored coeffs and does the convolution between the
       current geocenter waveform and the detector response function.
       */
      computeProjectionCoeffs(data, model[ic]->projection, model[ic]->extParams);
      Shf_Geocenter(data, model[ic]->projection, data->SnGeo, model[ic]->extParams);

      for(j=1; j<=signal->size; j++)
      {
        draw_wavelet_params(ifo, signal->intParams[j], data, prior->range, &chain->seed);
        if(data->amplitudePriorFlag)  data->signal_amplitude_proposal(signal->intParams[j], data->SnGeo, 1.0, &chain->seed, data->Tobs, prior->SNRpeak);
        SineGaussian(signal->templates, signal->intParams[j], data->N, 1, Tobs);
      }
      waveformProject(data, model[ic]->projection, model[ic]->extParams, model[ic]->response, signal->templates, data->fmin, data->fmax);
    }

    // initialize the noise parameters at some random value drawn from the (uniform) prior range
    for(ifo=0; ifo<NI; ifo++) model[ic]->Sn[ifo] = 1.0;//prior->Snmin + (prior->Snmax-prior->Snmin)*ran2(&chain->seed);

    // form up residual and compute initial likelihood
    model[ic]->logL = 0.0;

    for(ifo=0; ifo<NI; ifo++)
    {
      model[ic]->detLogL[ifo] = 0.0;

      for(i=0; i<data->N; i++)
      {
        data->r[ifo][i] = data->s[ifo][i];
        if(data->signalFlag) data->r[ifo][i] -= model[ic]->response[ifo][i];
        if(data->glitchFlag) data->r[ifo][i] -= glitch[ifo]->templates[i];
      }
      if(!data->constantLogLFlag)
      {
        model[ic]->detLogL[ifo] = loglike(data->imin, data->imax, data->r[ifo], model[ic]->Sn[ifo], data->Snf[ifo]);
        model[ic]->logL += model[ic]->detLogL[ifo];
      }
    }

    // start TFdensity proposal as uniform in TF volume
    initialize_TFproposal(data, prior, signal);
    for(ifo=0; ifo<NI; ifo++) initialize_TFproposal(data, prior, glitch[ifo]);

    //Chain files
    if(ic==0 || data->verboseFlag)
    {
      //Basic chain file with likelihood and model dimensions
      sprintf(filename,"chains/%s_intchain.dat.%i",modelname,ic);
      chain->intChainFile[ic] = fopen(filename,"w");

      //Parameters for reproducing glitch model
      if(data->glitchFlag)
      {
        for(ifo=0; ifo<NI; ifo++)
        {
          sprintf(filename,"chains/%s_glitchchain_ifo%i.dat.%i",modelname,ifo,ic);
          chain->intGlitchChainFile[ifo][ic] = fopen(filename,"w");
        }
      }

      //Parameters for reproducing signal model
      if(data->signalFlag)
      {
        //extrinsic
        sprintf(filename,"chains/%s_extchain.dat.%i",modelname,ic);
        chain->extChainFile[ic] = fopen(filename,"w");

        //intrinsic
        sprintf(filename,"chains/%s_wavechain.dat.%i",modelname,ic);
        chain->intWaveChainFile[ic] = fopen(filename,"w");
      }

      //Parameters for reproducing PSD model
      if(data->bayesLineFlag && ic==0)
      {
        for(ifo=0; ifo<NI; ifo++)
        {
          sprintf(filename,"chains/%s_splinechain_ifo%i.dat.0",modelname,ifo);
          data->bayesline[ifo]->splineChainFile = fopen(filename,"w");

          sprintf(filename,"chains/%s_lorentzchain_ifo%i.dat.0",modelname,ifo);
          data->bayesline[ifo]->lineChainFile = fopen(filename,"w");
        }
      }

    }//end chain files initialization
  }

  FILE *script,*psdscript,*bwscript,*logLscript;
  if(data->gnuplotFlag)
  {
    /******************************************************************************/
    /*                                                                            */
    /*  Set up gnuplot script for waveform movies                                 */
    /*                                                                            */
    /******************************************************************************/
    //write_gnuplot_script_header(script, modelname,data->Tobs);
    sprintf(filename,"waveforms/%s_waveforms_animate.gpi",modelname);
    script=fopen(filename,"w");
    fprintf(script,"set terminal gif small size 800,600 animate optimize delay 0.1 enhanced font\n");
    fprintf(script,"set output \"%s_waveforms.gif\"                     \n",modelname);
    fprintf(script,"                                   \n");
    //fprintf(script,"set xrange[%g:%g]                  \n", data->Tobs-2.0-0.06, data->Tobs-2.0+0.06);
    fprintf(script,"set xrange[%g:%g]                  \n", -data->Twin/2, +data->Twin/2);
    //fprintf(script,"set yrange[-40:40]                 \n");
    //fprintf(script,"set xtics 0.02                     \n");
    fprintf(script,"                                   \n");
    fprintf(script,"set xlabel \"time (s)\"            \n");
    fprintf(script,"set ylabel \"whitened strain\"     \n");
    fprintf(script,"                                   \n");
    fflush(script);

    /******************************************************************************/
    /*                                                                            */
    /*  Set up gnuplot script for PSD movies                                      */
    /*                                                                            */
    /******************************************************************************/
    if(data->bayesLineFlag)
    {
      sprintf(filename,"waveforms/%s_psd_animate.gpi",modelname);
      psdscript=fopen(filename,"w");
      fprintf(psdscript,"set terminal gif small size 800,600 animate optimize delay 0.1 enhanced font\n");
      fprintf(psdscript,"set output \"%s_psd.gif\"          \n",modelname);
      fprintf(psdscript,"                                   \n");
      fprintf(psdscript,"set xrange[%g:%g]                  \n", data->fmin, data->fmax);
      fprintf(psdscript,"set yrange[1e-48:1e-37]            \n");
      fprintf(psdscript,"set logscale y                     \n");
      fprintf(psdscript,"                                   \n");
      fprintf(psdscript,"set xlabel \"frequency (Hz)\"      \n");
      fprintf(psdscript,"set ylabel \"PSD\"                 \n");
      fprintf(psdscript,"                                   \n");
      fflush(psdscript);
    }

    /******************************************************************************/
    /*                                                                            */
    /*  Set up gnuplot script for BayesWave movies                                */
    /*                                                                            */
    /******************************************************************************/
    if(data->bayesLineFlag)
    {
      sprintf(filename,"waveforms/%s_bayeswave_animate.gpi",modelname);
      bwscript=fopen(filename,"w");
      fprintf(bwscript,"set terminal gif small size 800,600 animate optimize delay 0.1 enhanced font\n");
      fprintf(bwscript,"set output \"%s_bayeswave.gif\"          \n",modelname);
      fprintf(bwscript,"                                         \n");
      fflush(bwscript);
    }

    /******************************************************************************/
    /*                                                                            */
    /*  Set up gnuplot script for <logL> movies                                   */
    /*                                                                            */
    /******************************************************************************/
    sprintf(filename,"waveforms/%s_avgLogL_animate.gpi",modelname);
    logLscript=fopen(filename,"w");
    fprintf(logLscript,"set terminal gif small size 800,600 animate optimize delay 0.5 enhanced font\n");
    fprintf(logLscript,"set output \"%s_avgLogL.gif\"          \n",modelname);
    fprintf(logLscript,"                                       \n");
    fprintf(logLscript,"set logscale x                         \n");
    fprintf(logLscript,"set xrange [1e-8:1]                    \n");
    fprintf(logLscript,"set xlabel \"1/T\"                     \n");
    fprintf(logLscript,"set ylabel \"<logL>\"                  \n");
    fflush(logLscript);

  }

  /******************************************************************************/
  /*                                                                            */
  /*  The BurstRJMCMC loop                                                      */
  /*                                                                            */
  /******************************************************************************/

  int mc=0;

  /*
   FILE *TF;
   */

  //work-space for chain diagnostics (periodic time-domain waveforms)
  double *ty=dvector(0,data->N-1);

  chain->intPropRate = 0.9; //split intrinsic proposals between U[prior] and Fisher

  int burnFlag   = 0;
  int searchFlag = 0;

  print_run_stats(chain->runFile, data, chain);

  while(mc<M)
  {
    //Reset proposal ratios after GW search phase
    if((searchFlag==0 && mc>chain->burn/10) || (data->glitchFlag && !data->signalFlag) )
    {
      //heavily favor RJ moves (they get accepted at ~10%)
      if(data->glitchFlag || data->signalFlag)  chain->rjmcmcRate  = 0.5;

      searchFlag=1;
    }

    //Slowly add RJ option to GW model-only
    if(searchFlag==1 && mc>chain->burn/2)
    {
      chain->modelRate   = 1./(double)(NI+1); //even chance for each channel

      //heavily favor RJ moves (they get accepted at ~10%)
      if(data->glitchFlag || data->signalFlag)  chain->rjmcmcRate  = 0.5;

      //heavily favor Fisher moves (the all-sky ones are worthless now)
      chain->intPropRate = 0.9;

      searchFlag=2;
    }

    //Reset counters after initial burn-in & switch to Markovian likelihood function
    if(burnFlag==0 && mc>chain->burn)
    {
      initialize_chain(chain,0);
      if(!data->constantLogLFlag)
      {
        data->intrinsic_likelihood=EvaluateMarkovianLogLikelihood;
        data->extrinsic_likelihood=EvaluateExtrinsicMarkovianLogLikelihood;


        //clone chains
        for(ic=1; ic<NC; ic++)
        {
          copy_int_model(model[chain->index[0]], model[chain->index[ic]], data->N, data->NI, -1);
          for(ifo=0; ifo<data->NI; ifo++)copy_int_model(model[chain->index[0]], model[chain->index[ic]], data->N, data->NI, ifo);
          copy_ext_model(model[chain->index[0]], model[chain->index[ic]], data->N, data->NI);
        }



        //recompute likelihood
        for(ic=0; ic<NC; ic++)
        {
          for(ifo=0; ifo<NI; ifo++)
          {
            for(i=0; i<data->N; i++)
            {
              data->r[ifo][i] = data->s[ifo][i];
              if(data->glitchFlag) data->r[ifo][i] -= model[ic]->glitch[ifo]->templates[i];
              if(data->signalFlag) data->r[ifo][i] -= model[ic]->response[ifo][i];
            }
          }

          model[ic]->logL = 0.0;
          for(ifo=0; ifo<NI; ifo++)
          {
            model[ic]->detLogL[ifo] = loglike(data->imin, data->imax, data->r[ifo], model[ic]->Sn[ifo], data->Snf[ifo]);
            model[ic]->logL += model[ic]->detLogL[ifo];
          }
        }
      }
      burnFlag=1;
    }

    //print status update to screen
    if(mc%10000==0) print_chain_status(mc, data, chain, model, searchFlag, NC);

    //periodically update extrinsic parameter fisher matrix (numerical derivatives of response to geocenter waveform)
    if(mc%1000==0 && data->signalFlag)
    {
      for(ic=0; ic<NC; ic++)extrinsic_fisher_update(data, model[chain->index[ic]]);
    }

    //use history of where wavelets have gone as a differential evolution-like proposal distribution
    if(mc%1000==0 && mc<M/2 && data->clusterProposalFlag)
    {
      for(ic=0; ic<NC; ic++)
      {
        //turn TFhistory into useful proposal
        for(ifo=0; ifo<=NI; ifo++)
        {
          /*
           if (chain->index[ic]==0)
           {
           if(ifo==1)TF=fopen("TFdensity.dat.0","w");
           if(ifo==2)TF=fopen("TFdensity.dat.1","w");
           }
           */
          //use x to normalize distribution
          x = 0.0;

          /* Geocenter signal model */
          if(ifo==0) wave = model[chain->index[ic]]->signal;
          else       wave = model[chain->index[ic]]->glitch[ifo-1];

          for(i=0; i<data->fsize; i++)
          {
            for(j=0; j<data->tsize; j++)
            {
              //printf("ic%i:ifo%i,i=%i/%i,j=%i/%i\n",ic,ifo,i,data->fsize,j,data->tsize);
              wave->TFdensity[i][j] = wave->TFhistory[i][j];

              /*
               if(ifo==1) if(chain->index[ic]==0)
               {
               if(wave->TFhistory[i][j]==0)fprintf(TF,"%g %g %g\n", (double)i*data->fmax/(double)data->fsize, (double)j*data->Tobs/(double)data->tsize,1.e-10);
               else fprintf(TF,"%g %g %g\n", (double)i*data->fmax/(double)data->fsize, (double)j*data->Tobs/(double)data->tsize,wave->TFhistory[i][j]);
               }
               */

              x += wave->TFhistory[i][j];
            }
          }
          x*=0.5; // each pixel has area 1/2.
          wave->TFdensityMax = 0.0;
          for(i=0; i<data->fsize; i++)
          {
            for(j=0; j<data->tsize; j++)
            {
              wave->TFdensity[i][j] /= x;
              if(wave->TFdensity[i][j]>wave->TFdensityMax)wave->TFdensityMax = wave->TFdensity[i][j];
            }
          }

          /*
           if (chain->index[ic]==0 && (ifo==1||ifo==2)) fclose(TF);
           */

        }//end loop over channels
      }
    }

    //Loop over chains doing model updates
    for(ic=0; ic<NC; ic++)
    {
      chain->beta = 1./chain->temperature[ic];

      //This function executes [cycle] intrinsic parameter updates for the geocenter signal(RJMCMC or MCMC)
      EvolveIntrinsicParameters(data, prior, model, chain, &chain->seed, ic);

      //This function executes [cycle] extrinsic parameter updates, common to all geocenter wavelets.
      if(model[chain->index[ic]]->signal->size>0)EvolveExtrinsicParameters(data, prior, model, chain, &chain->seed, ic);

    }

    //update PSD with BayesLine
    if(data->bayesLineFlag) EvolveBayesLineParameters(data, model, chain, prior);

    //Parallel tempering
    if(NC>1)
    {
      PTMCMC(model, chain, &chain->seed, NC);

      //adapt PTMCMC spacing and reset various counters
      if(mc<M/2 && data->adaptTemperatureFlag && burnFlag)
      {

        //if hottest chain is absurdly so remove it from the run
        if(!burnFlag)adapt_temperature_ladder(data, model, chain, NC);

        for(ic=0; ic<NC-1; ic++)
          chain->temperature[ic+1] = chain->temperature[ic]*chain->dT[ic];

      }//end mc=M/2 condition

      //if hottest chain is absurdly so remove it from the run
      if(mc>M/2 && chain->temperature[NC-1]>1.0e4 && NC>chain->NCmin) NC--;

    }//end NC>1 condition for parallel tempering

    //add current wavelets to proposal density
    if(mc<M/2)
    {
      for(ic=0; ic<NC; ic++)
      {
        for(ifo=0; ifo<=NI; ifo++)
        {
          /* Geocenter signal model */
          if(ifo==0) wave = model[chain->index[ic]]->signal;
          else       wave = model[chain->index[ic]]->glitch[ifo-1];

          for(j=1; j<=wave->size; j++)
          {
            i  = wave->index[j];
            ix = (int)(wave->intParams[i][1]/data->df);
            jx = (int)(wave->intParams[i][0]/data->dt);
            wave->TFhistory[ix][jx] += 1.0;
          }
        }
      }//end loop over channels
    }//end mc<chain->burn

    // take an occasional look at the whitened reconstructed waveforms in the time domain (print ~100 total)

    //if(mc>=exp(frame*frameInterval) && data->gnuplotFlag)
    if(mc>=frame*frameInterval && data->gnuplotFlag)
    {
      //print waveforms to file for diagnostics
      for(ifo=0; ifo<NI; ifo++)
      {
        sprintf(filename,"waveforms/%s_data_%d.dat.%i", modelname, frame,ifo);
        for(i=0; i< N; i++) ty[i] = data->s[ifo][i];
        print_time_domain_waveforms(filename, ty, N, data->Snf[ifo],model[chain->index[0]]->Sn[ifo],data->Tobs,fix, data->imin, data->imax, prior->range[0][0], prior->range[0][1]);

        if(data->glitchFlag)
        {
          sprintf(filename,"waveforms/%s_glitch_%d.dat.%i", modelname, frame,ifo);
          for(i=0; i< N; i++) ty[i] = model[chain->index[0]]->glitch[ifo]->templates[i];
          print_time_domain_waveforms(filename, ty, N, data->Snf[ifo],model[chain->index[0]]->Sn[ifo],data->Tobs,fix, data->imin, data->imax, prior->range[0][0], prior->range[0][1]);
        }

        if(data->signalFlag)
        {
          sprintf(filename,"waveforms/%s_wave_%d.dat.%i", modelname, frame, ifo);
          for(i=0; i<N; i++) ty[i] = model[chain->index[0]]->response[ifo][i];
          print_time_domain_waveforms(filename, ty, N, data->Snf[ifo],model[chain->index[0]]->Sn[ifo],data->Tobs,fix, data->imin, data->imax, prior->range[0][0], prior->range[0][1]);
        }

        if(data->bayesLineFlag)
        {
          sprintf(filename,"waveforms/%s_psd_%d.dat.%i", modelname, frame,ifo);
          print_bayesline_spectrum(filename, data->bayesline[ifo]->freq, data->bayesline[ifo]->spow, data->bayesline[ifo]->Sbase, data->bayesline[ifo]->Sline, data->bayesline[ifo]->data->ncut);
        }
      }
      write_gnuplot_script_frame(script, modelname, data->runName, data->Tobs, data->runPhase, data->signalFlag, data->glitchFlag, frame, data->NI);
      if(data->bayesLineFlag)
      {
        write_bayesline_gnuplot_script_frame(psdscript, modelname, data->Tobs, frame);
        write_bayeswave_gnuplot_script_frame(bwscript, modelname, data->runName, data->Tobs, data->fmin, data->fmax, data->bayesline[0]->priors->SAmin, data->bayesline[0]->priors->SAmax, data->runPhase, data->signalFlag, data->glitchFlag, frame, data->NI);
      }
      frame++;
    }

    // Compute Evidence/Bayes Factor using last 1/2 of samples
    if(mc >= M/2)
    {
      for(ic=0; ic<NC; ic++) logLchain[ic][count] = model[chain->index[ic]]->logL;//  - data->logLc;
      count++;
    }

    mc+=chain->cycle;
  }


  /******************************************************************************/
  /*                                                                            */
  /*  Here we are experimenting with different ways to                          */
  /*  estimate the mean and variance of each chain                              */
  /*                                                                            */
  /******************************************************************************/

  /*
   Directly downsample chain to Neff = 100
   Compute <logL> from downsampled chain and var(<logL>) = var(logL)/Neff
   If ACL ~ Nsamples then use var(<logL>) = var(logL)
   */
  average_log_likelihood_via_direct_downsampling(M, logLchain, chain);

  //print logZ integrand
  sprintf(filename,"%s_evidence.dat",modelname);
  FILE *zFile = fopen(filename,"w");
  for(ic=0; ic<NC; ic++) fprintf(zFile,"%.12g %.12g %.12g\n", 1./chain->temperature[ic], chain->avgLogLikelihood[ic], sqrt(chain->varLogLikelihood[ic]));
  fclose(zFile);

  /*
   Downsample chain by ACL computed directly
   Compute <logL> from downsampled chain and var(<logL>) = var(logL)/Neff
   If ACL ~ Nsamples then use var(<logL>) = var(logL)
   */

  /*
   average_log_likelihood_via_ACL_downsampling(M, logLchain, chain);

   //print logZ integrand
   sprintf(filename,"%s_evidence_DSTyson.dat",modelname);
   zFile = fopen(filename,"w");
   for(ic=0; ic<NC; ic++) fprintf(zFile,"%.12g %.12g %.12g\n", 1./chain->temperature[ic], chain->avgLogLikelihood[ic], sqrt(chain->varLogLikelihood[ic]));
   fclose(zFile);
   */

  /*
   Downsample chain by ACL estimated using recursion relation.
   Compute <logL> from downsampled chain and var(<logL>) = var(logL)/Neff
   */

  /*
   average_log_likelihood_via_recursion_downsampling(M, logLchain, chain);

   //print logZ integrand
   sprintf(filename,"%s_evidence_DSWill.dat",modelname);
   zFile = fopen(filename,"w");
   for(ic=0; ic<NC; ic++) fprintf(zFile,"%.12g %.12g %.12g\n", 1./chain->temperature[ic], chain->avgLogLikelihood[ic], sqrt(chain->varLogLikelihood[ic]));
   fclose(zFile);
   */

  /*
   Compute var(<logL>) using the autocorrelation function a la
   http://en.wikipedia.org/wiki/Unbiased_estimation_of_standard_deviation
   */

  /*
   average_log_likelihood_via_ACF(M, logLchain, chain);

   //print logZ integrand
   sprintf(filename,"%s_evidence_ACL.dat",modelname);
   zFile = fopen(filename,"w");
   for(ic=0; ic<NC; ic++) fprintf(zFile,"%.12g %.12g %.12g\n", 1./chain->temperature[ic], chain->avgLogLikelihood[ic], sqrt(chain->varLogLikelihood[ic]));
   fclose(zFile);
   */



  /******************************************************************************/
  /*                                                                            */
  /*  Trapezoid integration of <logL> dBeta                                     */
  /*  Obsolete thanks to Neil's Monte Carlo integration of the output files     */
  /*  which will eventually get ported into BayesWave as a function             */
  /*                                                                            */
  /******************************************************************************/

  double beta[NC];
  for(ic=0; ic<NC; ic++)
    beta[ic] = 1./chain->temperature[ic];

  double dBeta= 0.0;
  double logZ = 0.0;
  double varZ = 0.0;

  for(ic=0; ic<NC-1; ic++)
  {
    dBeta = beta[ic]-beta[ic+1];
    logZ += (chain->avgLogLikelihood[ic]+chain->avgLogLikelihood[ic+1])*dBeta/2.0;
    if(ic==0)
      varZ += chain->varLogLikelihood[ic]*dBeta*dBeta/4.0;
    else
      varZ += chain->varLogLikelihood[ic]*pow(beta[ic+1]-beta[ic-1],2)/4.0;
  }
  varZ += chain->varLogLikelihood[NC-1]*pow(beta[NC-2]-beta[NC-1],2)/4.0;


  printf("Trapezoid Rule logZ = %g +/- %g\n",logZ,sqrt(varZ));

  /******************************************************************************/
  /*                                                                            */
  /*  Neil's Monte Carlo integration of <logL> v beta                           */
  /*                                                                            */
  /******************************************************************************/

  ThermodynamicIntegration(chain->temperature, chain->avgLogLikelihood, chain->varLogLikelihood, NC, &logZ, &varZ);

  printf("Monte Carlo logZ = %g +/- %g\n",logZ,sqrt(varZ));

  /******************************************************************************/
  /*                                                                            */
  /*  Neil's Monte Carlo integration of <logL> v beta                           */
  /*  including chain-to-chain correlations                                     */
  /*                                                                            */
  /******************************************************************************/
  if(!data->constantLogLFlag && data->bootstrapEvidenceFlag)
  {
    BootstrapThermodynamicIntegration(chain->temperature, logLchain, nPoints, NC, modelname, &logZ, &varZ);

    printf("Bootstrap Monte Carlo logZ = %g +/- %g\n",logZ,sqrt(varZ));
  }
  free_dmatrix(logLchain,0,chain->NC-1,0,nPoints-1);

  /******************************************************************************/
  /*                                                                            */
  /*  Return logZ, and exit cleanly so we can test other models                 */
  /*                                                                            */
  /******************************************************************************/

  free_dvector(ty,0,N-1);
  free(index);
  fclose(chain->runFile);

  if(data->gnuplotFlag)
  {
    fclose(script);
    if(data->bayesLineFlag)
    {
      fclose(bwscript);
      fclose(psdscript);
    }
  }

  for(ic=0; ic<NC; ic++)
  {
    if(ic==0 || data->verboseFlag)
    {
      if(data->glitchFlag || data->signalFlag)
      {
        fclose(chain->intChainFile[ic]);
        if(data->signalFlag)
        {
          fclose(chain->intWaveChainFile[ic]);
          fclose(chain->extChainFile[ic]);
        }
        if(data->glitchFlag) for(ifo=0; ifo<NI; ifo++)
          fclose(chain->intGlitchChainFile[ifo][ic]);
      }

      if(data->bayesLineFlag && ic==0)
      {
        for(ifo=0; ifo<NI; ifo++)
        {
          fclose(data->bayesline[ifo]->splineChainFile);
          fclose(data->bayesline[ifo]->lineChainFile);
        }
      }

    }
  }

  *logEvidence    = logZ;
  *varLogEvidence = varZ;

}


void write_gnuplot_script_header(FILE *script, char modelname[], double Tobs)
{
  char filename[100];
  sprintf(filename,"waveforms/%s_waveforms_animate.gpi",modelname);
  script=fopen(filename,"w");
  fprintf(script,"set terminal gif giant enhanced size 800,600 animate delay 1 \n");
  fprintf(script,"set output \"%s_waveforms.gif\"                     \n",modelname);
  fprintf(script,"                                   \n");
  //fprintf(script,"set xrange[%g:%g]                  \n", -0.02, +0.02);
  //fprintf(script,"set yrange[-20:20]                 \n");
  //fprintf(script,"set xtics 0.02                     \n");
  fprintf(script,"                                   \n");
  fprintf(script,"set xlabel \"time (s)\"            \n");
  fprintf(script,"set ylabel \"whitened strain\"     \n");
  fprintf(script,"                                   \n");
}

void write_gnuplot_script_frame (FILE *script, char modelname[], char runname[], double Tobs, int phase, int signal, int glitch, int cycle, int NI)
{
  int i;

  char datafile[100];
  sprintf(datafile,"%s_data_%i.dat.%i", modelname,cycle,0);


  fprintf(script,"set multiplot  \n");
  fprintf(script,"set size 0.5,%f \n",1./(double)NI);

//  fprintf(script,"stats \"%s\" u 2 nooutput\n", datafile);
//  fprintf(script,"min0=STATS_min\n");
//  fprintf(script,"max0=STATS_max\n");
//  fprintf(script,"if(abs(min0)>max0)max0=abs(min0)\n");
//
//  for(i=1; i<NI; i++)
//  {
//    sprintf(datafile,"%s_data_%i.dat.%i", modelname,cycle,i);
//
//    fprintf(script,"stats \"%s\" u 2 nooutput\n", datafile);
//    fprintf(script,"min1=STATS_min\n");
//    fprintf(script,"max1=STATS_max\n");
//    fprintf(script,"if(abs(min1)>max1)max1=abs(min1)\n");
//    fprintf(script,"if(abs(max1)>max0)max0=max1\n");
//  }
//  fprintf(script,"set yrange[-max0:max0]\n");

  for(i=0; i<NI; i++)
  {
    sprintf(datafile,"%s_data_%i.dat.%i", modelname,cycle,i);

    fprintf(script,"set origin 0,%f \n",(double)i/(double)NI);
    fprintf(script,"set title \"IFO %i\" \n", i);
    fprintf(script,"p ");
    fprintf(script,"\"%s\" u 1:2 title \"data\" w l ls 1 lw 2, ",datafile);
    if(signal)fprintf(script,"\"%s_wave_%i.dat.%i\" u 1:2 title \"signal model\" w l ls 2 lw 2, " ,modelname,cycle,i);
    if(glitch)fprintf(script,"\"%s_glitch_%i.dat.%i\" u 1:2 title \"glitch model\" w l ls 3 lw 2 ",modelname,cycle,i);
    fprintf(script,"\n");
  }
  /*
   fprintf(script,"set origin 0,0\n");
   fprintf(script,"set title \"L1\" \n");
   fprintf(script,"p ");
   fprintf(script,"\"%s_data_%i.dat.%i\" u 1:2 title \"data\" w l ls 1 lw 2, "         ,modelname,cycle,1);
   fprintf(script,"\"%s_wave_%i.dat.%i\" u 1:2 title \"signal model\" w l ls 2 lw 2, " ,modelname,cycle,1);
   fprintf(script,"\"%s_glitch_%i.dat.%i\" u 1:2 title \"glitch model\" w l ls 3 lw 2 ",modelname,cycle,1);
   fprintf(script,"\n");
   */
  fprintf(script,"unset multiplot \n");
  fflush(script);
}

void write_bayesline_gnuplot_script_frame(FILE *script, char modelname[], double Tobs, int cycle)
{
  fprintf(script,"set multiplot  \n");
  fprintf(script,"set size 1,0.5 \n");
  fprintf(script,"set origin 0,0.5 \n");
  fprintf(script,"set title \"H1\" \n");
  fprintf(script,"p ");
  fprintf(script,"\"%s_psd_%i.dat.%i\" u 1:2 title \"residual\" w l ls 1 lw 2, "     ,modelname,cycle,0);
  fprintf(script,"\"%s_psd_%i.dat.%i\" u 1:3 title \"spline model\" w l ls 2 lw 2, " ,modelname,cycle,0);
  fprintf(script,"\"%s_psd_%i.dat.%i\" u 1:4 title \"line model\" w l ls 3 lw 2 "    ,modelname,cycle,0);
  fprintf(script,"\n");
  fprintf(script,"set origin 0,0\n");
  fprintf(script,"set title \"L1\" \n");
  fprintf(script,"p ");
  fprintf(script,"\"%s_psd_%i.dat.%i\" u 1:2 title \"residual\" w l ls 1 lw 2, "     ,modelname,cycle,1);
  fprintf(script,"\"%s_psd_%i.dat.%i\" u 1:3 title \"spline model\" w l ls 2 lw 2, " ,modelname,cycle,1);
  fprintf(script,"\"%s_psd_%i.dat.%i\" u 1:4 title \"line model\" w l ls 3 lw 2 "    ,modelname,cycle,1);
  fprintf(script,"\n");
  fprintf(script,"unset multiplot \n");
  fflush(script);
}

void write_bayeswave_gnuplot_script_frame(FILE *script, char modelname[], char runname[], double Tobs, double fmin, double fmax, double SAmin, double SAmax, int phase, int signal, int glitch, int cycle, int NI)
{
  int i;

  char datafile[100];
  sprintf(datafile,"%s_data_%i.dat.%i", modelname,cycle,0);

  fprintf(script,"set multiplot  \n");
  fprintf(script,"set size 0.5,0.5 \n");

  fprintf(script,"set xrange[%g:%g]                  \n", fmin, fmax);
  fprintf(script,"set yrange[%g:%g]                  \n", 1.0e-48, 1.0e-38);
  fprintf(script,"set xtics 100                       \n");
  fprintf(script,"set logscale y                     \n");
  fprintf(script,"                                   \n");
  fprintf(script,"set xlabel \"frequency (Hz)\"      \n");
  fprintf(script,"set ylabel \"PSD\"                 \n");

  for(i=0; i<NI; i++)
  {
    fprintf(script,"set origin 0,%f \n",0.5*(double)(1-i));
    fprintf(script,"set title \"IFO %i\" \n", i);
    fprintf(script,"p ");
    fprintf(script,"\"%s_psd_%i.dat.%i\" u 1:2 title \"residual\" w l ls 2, "     ,modelname,cycle,i);
    fprintf(script,"\"%s_psd_%i.dat.%i\" u 1:3 title \"spline model\" w l ls 1, " ,modelname,cycle,i);
    fprintf(script,"\"%s_psd_%i.dat.%i\" u 1:4 title \"line model\" w l ls 3 "    ,modelname,cycle,i);
    fprintf(script,"\n");
  }


  /*
   fprintf(script,"set origin 0,0\n");
   fprintf(script,"set title \"L1\" \n");
   fprintf(script,"p ");
   fprintf(script,"\"%s_psd_%i.dat.%i\" u 1:2 title \"residual\" w l ls 1, "     ,modelname,cycle,1);
   fprintf(script,"\"%s_psd_%i.dat.%i\" u 1:3 title \"spline model\" w l ls 2, " ,modelname,cycle,1);
   fprintf(script,"\"%s_psd_%i.dat.%i\" u 1:4 title \"line model\" w l ls 3 "    ,modelname,cycle,1);
   fprintf(script,"\n");
   */

  fprintf(script,"unset logscale y \n");
  fprintf(script,"\n");


  if(phase==0)
  {
    fprintf(script,"set xrange[%g:%g]\n",-Tobs+3.,1.);
    fprintf(script,"set xtics 1\n");
    //fprintf(script,"set yrange[-40:40]\n");
  }
  if(phase==1)
  {
    fprintf(script,"set xrange[%g:%g]\n", -1., 1.);
    fprintf(script,"set xtics 0.1\n");
    //fprintf(script,"set yrange[-20:20]\n");
  }

  fprintf(script,"set xrange[*:*]\n");
  fprintf(script,"set yrange[*:*]\n");
  //  fprintf(script,"stats \"%s\" u 2 nooutput\n", datafile);
  //
  //  fprintf(script,"min0=STATS_min\n");
  //  fprintf(script,"max0=STATS_max\n");
  //
  //  for(i=1; i<NI; i++)
  //  {
  //    fprintf(script,"stats \"%s\" u 2 nooutput\n", datafile);
  //
  //    fprintf(script,"min1=STATS_min\n");
  //    fprintf(script,"max1=STATS_max\n");
  //    fprintf(script,"if(abs(min0)>max0)max0=abs(min0)\n");
  //    fprintf(script,"if(abs(min1)>max1)max1=abs(min1)\n");
  //    fprintf(script,"if(abs(max1)>max0)max0=max1\n");
  //  }
  //  fprintf(script,"set yrange[-max0:max0]\n");

  fprintf(script,"                                   \n");
  fprintf(script,"set xlabel \"time (s)\"            \n");
  fprintf(script,"set ylabel \"whitened strain\"     \n");


  for(i=0; i<NI; i++)
  {

    sprintf(datafile,"%s_data_%i.dat.%i", modelname,cycle,i);

    fprintf(script,"set origin 0.5,%f \n",0.5*(double)(1-i));
    fprintf(script,"set title \"IFO %i\" \n",i);
    fprintf(script,"p ");
    fprintf(script,"\"%s\" u 1:2 title \"data\" w l ls 2, " ,datafile);

    //if(signal) fprintf(script,"\"%s_wave_%i.dat.%i\" u 1:2 title \"signal model\" w l ls 3, " ,modelname,cycle,i);
    if(signal) fprintf(script,"\"<paste %s_data_%i.dat.%i %s_wave_%i.dat.%i\" u 1:($2-$4) title \"residual\" w l ls 1 ", modelname,cycle,i,modelname,cycle,i);
    if(glitch) fprintf(script,"\"<paste %s_data_%i.dat.%i %s_glitch_%i.dat.%i\" u 1:($2-$4) title \"residual\" w l ls 1 ", modelname,cycle,i,modelname,cycle,i);
    //if(glitch && phase==0) fprintf(script,"\"<paste %s_data_%i.dat.%i %s_glitch_%i.dat.%i\" u 1:($2-$4) title \"residual\" w l ls 1 ", modelname,cycle,i,modelname,cycle,i);
    //if(glitch && phase!=0) fprintf(script,"\"%s_glitch_%i.dat.%i\" u 1:2 title \"glitch model\" w l ls 1 ",modelname,cycle,i);
    fprintf(script,"\n");
  }

  /*
   fprintf(script,"set origin 0.5,0\n");
   fprintf(script,"set title \"L1\" \n");
   fprintf(script,"p ");
   fprintf(script,"\"%s_data_%i.dat.%i\" u 1:2 title \"data\" w l ls 1, "         ,modelname,cycle,1);
   fprintf(script,"\"%s_wave_%i.dat.%i\" u 1:2 title \"signal model\" w l ls 2, " ,modelname,cycle,1);
   fprintf(script,"\"%s_glitch_%i.dat.%i\" u 1:2 title \"glitch model\" w l ls 3 ",modelname,cycle,1);
   fprintf(script,"\n");
   */

  fprintf(script,"unset multiplot \n");
  fflush(script);
}

void write_evidence_gnuplot_script(FILE *script, char modelname[])
{
  fprintf(script,"reset\n");
  fprintf(script,"set terminal postscript color solid enhanced linewidth 2.0 \"Helvetica\" 24\n");
  fprintf(script,"\n");
  fprintf(script,"unset key\n");
  fprintf(script,"set border 3 linewidth 0.5\n");
  fprintf(script,"set nox2tics\n");
  fprintf(script,"set noy2tics\n");
  fprintf(script,"set tics scale 2\n");
  fprintf(script,"set style fill solid 1.0 border lt -1\n");
  fprintf(script,"set xtics nomirror axis offset 0,0.3 scale 0.5\n");
  fprintf(script,"set ytics nomirror\n");
  fprintf(script,"\n");
  fprintf(script,"set mytics\n");
  fprintf(script,"\n");
  fprintf(script,"set xlabel \"Model\"\n");
  fprintf(script,"set ylabel \"p( Model | d )\"\n");
  fprintf(script,"set logscale y\n");
  fprintf(script,"set format y \"10^{%%T}\"\n");
  fprintf(script,"\n");
  fprintf(script,"set xtics font \"Helvetica,20\"\n");
  fprintf(script,"set ytics font \"Helvetica,20\"\n");
  fprintf(script,"\n");
  fprintf(script,"set style data histograms\n");
  fprintf(script,"set style histogram cluster gap 1\n");
  fprintf(script,"\n");
  fprintf(script,"\n");
  fprintf(script,"set output \"evidence.eps\"\n");
  fprintf(script,"p \"%sevidence.dat\" using 3:xtic(1)\n",modelname);
}

void print_bayesline_spectrum(char filename[], double *f, double *power, double *Sbase, double *Sline, int N)
{
  int i;
  FILE *waveout = fopen(filename,"w");

  for(i=0; i<N; i++) fprintf(waveout,"%e %e %e %e\n", f[i], power[i], Sbase[i], Sline[i]);

  fclose(waveout);
}


void print_help_message(void)
{
  fprintf(stdout,"\n ==== BayesWaveBurst() USAGE: ====\n\n");
  fprintf(stdout,"REQUIRED:\n");
  fprintf(stdout,"  --ifo IFO              interferometer (H1,L1,V1)\n");
  fprintf(stdout,"  --IFO-flow             minimum frequency (Hz)\n");
  fprintf(stdout,"  --IFO-cache            full path to cache file\n");
  fprintf(stdout,"  --IFO-channel          channel name (IFO:LSC-STRAIN)\n");
  fprintf(stdout,"  --trigtime             GPS trigger time\n");
  fprintf(stdout,"  --srate                sampling rate (Hz)\n");
  fprintf(stdout,"  --seglen               duration of data (s)\n");
  fprintf(stdout,"  --PSDstart             GPS start time for PSD estimation\n");
  fprintf(stdout,"  --PSDlength            duration of PSD estimation length\n");
  fprintf(stdout,"  --dataseed             Required if using simulated noise e.g. --H1-cache LALAdLIGO\n");
  fprintf(stdout,"\n");
  fprintf(stdout,"OPTIONAL:\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --- Run parameters   -------------------------------------------------------------\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --Niter                number of iterations (2000000)\n");
  fprintf(stdout,"  --NCmin                minimum number of parallel chains (25)\n");
  fprintf(stdout,"  --NCmax                maximum number of parallel chains (25)\n");
  fprintf(stdout,"  --Ncycle               number of model updates per BurstRJMCMC iteration (100)\n");
  fprintf(stdout,"  --Nburnin              number of burn-in iterations (100000)\n");
  fprintf(stdout,"  --chainSeed            random number seed for Markov chain (1234)\n");
  fprintf(stdout,"  --runName              run name for output files\n");
  fprintf(stdout,"  --outputDir            absolute path of where the output should be written\n");
  fprintf(stdout,"  --0noise               no noise realization\n");
  fprintf(stdout,"  --zeroLogL             logL = constant test\n");
  fprintf(stdout,"  --restart              restart run using psd and residual from file\n");
  fprintf(stdout,"  --gnuplot              output files for gnuplot animations\n");
  fprintf(stdout,"  --verbose              output hot chains\n");
  fprintf(stdout,"  --window               duration of time window for model characterization runs\n");
  fprintf(stdout,"\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --- Model parameters   -----------------------------------------------------------\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --fullOnly             require signal && glitch model\n");
  fprintf(stdout,"  --noClean              skip cleaning phase and go right to reduced window\n");
  fprintf(stdout,"  --cleanOnly            run bayesline & glitch cleaning phase only\n");
  fprintf(stdout,"  --noiseOnly            use noise model only (no signal or glitches)\n");
  fprintf(stdout,"  --signalOnly           use signal model only (no glitches)\n");
  fprintf(stdout,"  --glitchOnly           use glitch model only (no signal)\n");
  fprintf(stdout,"  --noPSDfit             keep PSD parameters fixed\n");
  fprintf(stdout,"  --bayesLine            use BayesLine for PSD model\n");
  fprintf(stdout,"\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --- Priors & Proposals -----------------------------------------------------------\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --Dmin                 minimum number of wavelets total (1)\n");
  fprintf(stdout,"  --Dmax                 maximum number of wavelets per channel (20)\n");
  fprintf(stdout,"  --Qmax                 maximum quality factor for wavelets (40)\n");
  fprintf(stdout,"  --clusterPrior         use metric-based clustering prior\n");
  fprintf(stdout,"  --clusterPath          full path to cluster prior normalization files\n");
  fprintf(stdout,"  --clusterAlpha         distance between wavelets to be considered a cluster (2)\n");
  fprintf(stdout,"  --clusterBeta          cluster prior exp(-beta K) (4)\n");
  fprintf(stdout,"  --clusterGamma         occam penalty exp(gamma*J) (0)\n");
  fprintf(stdout,"  --noOrientationProposal disable simplex algorithm to propose orientation params\n");
  fprintf(stdout,"  --noAmplitudePrior     don't use SNR-dependent amplitude prior\n");
  fprintf(stdout,"  --sameAmplitudePrior   use same SNR-dependent prior for glitch and signal model\n");
  fprintf(stdout,"  --noClusterProposal    don't use clustering & TF density proposal\n");
  fprintf(stdout,"  --ampPriorPeak         SNR where amplitude prior peaks (5)\n");
  fprintf(stdout,"  --dimensionDecayRate   e-folding time for number of wavelets (1000)\n");
  fprintf(stdout,"\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --- Parallel Tempering parameters   ----------------------------------------------\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --tempMin              minimum temperature chain (1)\n");
  fprintf(stdout,"  --adaptTemperature     adjust PT ladder to maintain acc. rate from 40%% to 60%%\n");
  fprintf(stdout,"  --tempSpacing          set temperature spacing for geometric ladder (1.5)\n");
  fprintf(stdout,"  --bootstrapEvidence    use bootstrap thermodynamic integration\n");
  fprintf(stdout,"\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --- LALInference injection  options  ---------------------------------------------\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --inj injfile.xml      Injection XML file to use\n");
  fprintf(stdout,"  --event N              Event number from Injection XML file to use\n");
  fprintf(stdout,"\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --- Burst MDC injection  options  ------------------------------------------------\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --MDC-channel          IFO1-chan, IFO2-chan, etc\n");
  fprintf(stdout,"  --MDC-cache            IFO1-mdcframe, IFO2-mdcframe, etc\n");
  fprintf(stdout,"  --MDC-prefactor        Rescale injection amplitude (1.0)\n");
  fprintf(stdout,"\n");
  fprintf(stdout,"EXAMPLE:\n");
  fprintf(stdout,"  BayesWaveBurst --ifo H1 --H1-flow 32 --H1-cache LALSimAdLIGO --H1-channel LALSimAdLIGO --trigtime 900000000.00 --srate 512 --seglen 4 --PSDstart 900000000 --PSDlength 1024 --NCmin 2 --NCmax 2 --dataseed 1234\n");
  fprintf(stdout,"\n");
}

void print_run_stats(FILE *fptr, struct Data *data, struct Chain *chain)
{
  fprintf(fptr, " ============ Run Settings ===========\n");
  fprintf(fptr, "  number of detectors          %i\n",   data->NI);
  fprintf(fptr, "  trigger time (s)             %s\n",   LALInferenceGetProcParamVal(data->commandLine,"--trigtime")->value);
  fprintf(fptr, "  duration of data (s)         %.0f\n", data->Tobs);
  fprintf(fptr, "  duration of window (s)       %.1f\n", data->Twin);
  fprintf(fptr, "  sampling rate (Hz)           %i\n",   data->N/(int)data->Tobs);
  fprintf(fptr, "  number of samples            %i\n",   data->N);
  fprintf(fptr, "  number of iterations         %i\n",   chain->count);
  fprintf(fptr, "  number of burn-in iterations %i\n",   chain->burn);
  fprintf(fptr, "  RNG seed for chain           %li\n",  chain->seed);
  fprintf(fptr, "  max # of wavelets/channel    %i\n",   data->Dmax);
  fprintf(fptr, "  max Q for wavelets           %.0f\n", data->Qmax);
  fprintf(fptr, "  number of chains             %i\n",   chain->NC);
  fprintf(fptr, "  coldest chain temperature    %.0f\n", chain->tempMin);
  fprintf(fptr, "  initial temperature spacing  %.1f\n", chain->tempStep);
  fprintf(fptr, "  number of independent steps  %i\n",   chain->cycle);

  fprintf(fptr,"\n");
  fflush(fptr);
}

void print_run_flags(FILE *fptr, struct Data *data, struct Prior *prior)
{
  fprintf(fptr, " ============= Run Flags =============\n");

  fprintf(fptr, "  noise realization is ..... ");
  if(data->noiseSimFlag) fprintf(fptr, "INCLUDED");
  else fprintf(fptr, "NOT INCLUDED");
  fprintf(fptr, "\n");

  fprintf(fptr, "  PSD fitting is ........... ");
  if(data->psdFitFlag) fprintf(fptr, "ENABLED");
  else fprintf(fptr, "DISABLED");
  fprintf(fptr, "\n");

  fprintf(fptr, "  BayesLine PSD model is ... ");
  if(data->bayesLineFlag) fprintf(fptr, "ENABLED");
  else fprintf(fptr, "DISABLED");
  fprintf(fptr, "\n");

  fprintf(fptr, "  cleaning phase is ........ ");
  if(!data->skipCleanFlag)   fprintf(fptr, "ENABLED");
  else                       fprintf(fptr, "DISABLED");
  fprintf(fptr, "\n");

  fprintf(fptr, "  glitch model is .......... ");
  if(!data->glitchFlag)      fprintf(fptr, "DISABLED");
  else if(!data->signalFlag) fprintf(fptr, "REQUIRED");
  else                       fprintf(fptr, "ENABLED");
  fprintf(fptr, "\n");

  fprintf(fptr, "  signal model is .......... ");
  if(!data->signalFlag)      fprintf(fptr, "DISABLED");
  else if(!data->glitchFlag) fprintf(fptr, "REQUIRED");
  else                       fprintf(fptr, "ENABLED");
  fprintf(fptr, "\n");

  fprintf(fptr, "  clustering prior is ...... ");
  if(data->clusterPriorFlag) fprintf(fptr, "(%i,%i,%i)",(int)prior->alpha, (int)prior->beta, (int)prior->gamma);
  else fprintf(fptr, "DISABLED");
  fprintf(fptr, "\n");

  fprintf(fptr, "  amplitude prior is ....... ");
  if(data->amplitudePriorFlag) fprintf(fptr, "ENABLED");
  else fprintf(fptr, "DISABLED");
  fprintf(fptr, "\n");

  fprintf(fptr, "  signal amp prior is ...... ");
  if(data->amplitudePriorFlag && data->signalAmplitudePriorFlag) fprintf(fptr, "ENABLED");
  else fprintf(fptr, "DISABLED");
  fprintf(fptr, "\n");

  fprintf(fptr, "  density proposal is ...... ");
  if(data->clusterProposalFlag) fprintf(fptr, "ENABLED");
  else fprintf(fptr, "DISABLED");
  fprintf(fptr, "\n");

  fprintf(fptr, "  logL=const test is ....... ");
  if(data->constantLogLFlag) fprintf(fptr, "ENABLED");
  else fprintf(fptr, "DISABLED");
  fprintf(fptr, "\n");

  fprintf(fptr, "  PTMCMC adaptation is ..... ");
  if(data->adaptTemperatureFlag) fprintf(fptr,"ENABLED");
  else fprintf(fptr, "DISABLED");
  fprintf(fptr, "\n");

  fprintf(fptr, "  gnuplot output is ........ ");
  if(data->gnuplotFlag) fprintf(fptr,"ENABLED");
  else fprintf(fptr, "DISABLED");
  fprintf(fptr, "\n");

  fprintf(fptr, "\n");
  fflush(fptr);
}

void print_time_domain_waveforms(char filename[], double *h, int N, double *Snf, double eta, double Tobs, double fix, int imin, int imax, double tmin, double tmax)
{
  /*
   TODO: invFFT comes out time-reversed.
   LAL uses exp(-i2pift) for their Fourier transforms
   Why doesn't drealft(ty-1,N,1) work for inverse FFT?
   In the meantime, hacked invFFT by changing sign of imaginary part
   */
  int i;
  double x,t;
  FILE *waveout = fopen(filename,"w");

  h[0] = 0.0;
  h[1] = 0.0;
  for(i=1; i< N/2; i++)
  {
    if(i>imin && i<imax)
    {
      x = sqrt(Snf[i]*eta);
      h[2*i] /= x;
      h[2*i+1] /= -x;
    }
    else
    {
      h[2*i]=h[2*i+1]=0.0;
    }
  }

  drealft(h-1,N,-1);

  double t0 = Tobs-2.0;

  for(i=0; i<N; i++)
  {
    t = (double)(i)/(double)(N)*Tobs;
    if(t>=tmin && t<tmax)fprintf(waveout,"%e %e\n", t-t0, h[i]/fix);
  }

  fclose(waveout);
}

void print_chain_status(int mc, struct Data *data, struct Chain *chain, struct Model **model, int searchFlag, int NC)
{
  int ifo;
  int NI = data->NI;

  double SNR;

  //print overal status
  fprintf(stdout, "RJMCMC: %i/%i (%i,%i)\n",mc,chain->count,chain->burnFlag,searchFlag);

  //  if(data->bayesLineFlag)
  //  {
  //    //print chain temperatures
  //    fprintf(stdout, "PTMCMC:\n");
  //
  //    fprintf(stdout, "  Temps: ");
  //    for(ic=0; ic<NC; ic++) fprintf(stdout, "%.2f (%.2g,%.2g) ",chain->temperature[ic],chain->dT[ic],(double)chain->ptacc[ic]/(double)chain->ptprop[ic]);
  //    fprintf(stdout, "\n");
  //  }
  SNR = 0.0;
  SNR = network_snr((int)floor(data->fmin*data->Tobs), (int)floor(data->fmax*data->Tobs), model[chain->index[0]]->response, data->Snf, model[chain->index[0]]->Sn, data->NI);
  fprintf(stdout, "  logL=%f hSNR=%.0f ",  model[chain->index[0]]->logL - data->logLc, SNR);
  for(ifo=0; ifo<NI; ifo++)
  {
    SNR = detector_snr((int)floor(data->fmin*data->Tobs), (int)floor(data->fmax*data->Tobs), model[chain->index[0]]->glitch[ifo]->templates, data->Snf[ifo], model[chain->index[0]]->Sn[ifo]);
    fprintf(stdout, "  g%iSNR=%.0f ",  ifo, SNR);
  }
  fprintf(stdout, "\n");
  
  
  //print intrinsic parameter stats
  SNR = 0.0;
  if(!data->constantLogLFlag) SNR = network_snr((int)floor(data->fmin*data->Tobs), (int)floor(data->fmax*data->Tobs), model[chain->index[0]]->response, data->Snf, model[chain->index[0]]->Sn, data->NI);
  fprintf(stdout, "  INT: ");
  fprintf(stdout, "DGW=%d ",  model[chain->index[0]]->signal->size);
  for(ifo=0; ifo<NI; ifo++) fprintf(stdout, "D%i=%d ", ifo, model[chain->index[0]]->glitch[ifo]->size);
  fprintf(stdout, "m0=%i m1=%i m2=%i m3=%i rjacc=%.1e macc=%.1e",chain->mod0, chain->mod1, chain->mod2, chain->mod3, (double)chain->sacc/(double)(chain->scount), (double)chain->macc/(double)(chain->mcount));
  fprintf(stdout, "\n");
  
  
  //print extrinsic parameter stats
  if(data->signalFlag)
  {
    fprintf(stdout, "  EXT: xacc=%.1e ",(double)chain->xacc/(double)chain->xcount);
    fprintf(stdout, "\n");
  }
  
  
  fflush(stdout);
}
