/*
 Compile: See Makefile in bayeswave directory
 */

/***************************  REQUIRED LIBRARIES  ***************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <math.h>
#include <sys/stat.h>

#include "Constants.h"
#include "Declarations.h"
#include "numrec.h"
#include "BayesLine.h"
#include <lal/LALInferenceTemplate.h> 
#include <lal/LALSimIMR.h> 
#include <lal/LALSimulation.h> 

/*************  PROTOTYPE DECLARATIONS FOR INTERNAL FUNCTIONS  **************/

struct Moments
{
  double *overlap;
  double *energy;
  double *frequency;
  double *time;
  double *bandwidth;
  double *duration;
  double *snr;
  double networkOverlap;
  double *hmax;
  double *t_at_hmax;
};

void print_help_message(void);
void print_time_domain_waveforms(char filename[], double *h, int N, double *Snf, double eta, double Tobs, double fix, int imin, int imax, double tmin, double tmax);
void ComputeWaveformOverlap(struct Data *data, double **hinj, double **hrec, double **psd, struct Moments *m);
void ComputeWaveformMoments(struct Data *data, double deltaF, double deltaT, double **h, double **psd, struct Moments *m);
void initialize_moments(struct Moments *m, int NI);
void setup_output_files(int NI, int injectionFlag, char *outdir, char *model, char *weighting, FILE **momentsOut);
void parse_glitch_parameters(struct Data *data, struct Model *model, FILE **paramsfile, double **grec);
void parse_signal_parameters(struct Data *data, struct Model *model, FILE *paramsfile, double **hrec, double **hrecPlus, double **hrecCross);
void discard_burnin_samples(FILE *params, int *COUNT, int *BURNIN);
void print_moments(int NI, int injectionFlag, FILE **outfile, struct Moments *m);
void print_waveforms(struct Data *data, double **h, double **psd, FILE **ofile);
void print_stats(struct Data *data, LIGOTimeGPS GPS, int injectionFlag, char *outdir, char *model, int **dimension);

double get_energy(int imin, int imax, double *a, double deltaF);
double get_rhof(int imin, int imax, double *a, double deltaF, double *arg);
double get_rhof_net(int imin, int imax, double **a, double deltaF, double *arg, int NI);
double rhonorm(int imax, double *a, double deltaF);
double get_f0(int imax, double *a, double deltaF);
double get_band(int imax, double *a, double deltaF, double f0);
void get_hmax_and_t(int imax, double *a, double deltaT, double *hmax, double *t_at_hmax);
void get_time_domain_waveforms(double *hoft, double *h, int N, int imin, int imax);
double get_rhot(int N, double *a, double deltaT, double *arg);
double get_rhot_net(int N, double **a, double deltaT, double *arg, int NI);
void whiten_data(int imin, int imax, double *a, double *Snf);

void write_time_freq_samples(struct Data *data);
void compute_injected_plus_and_cross(LALInferenceRunState *runState, double deltaT, double deltaF, int N, int imin, char *outdir, double **timeDataPlus, double **timeDataCross, double **freqDataPlus, double **freqDataCross);

void Lorentzian(double *Slines, double Tobs, lorentzianParams *lines, int ii, int N);


void print_single_waveform(struct Data *data, double **h, double **psd, FILE *ofile);

/* ============================  MAIN PROGRAM  ============================ */

int main(int argc, char *argv[])
{

  /*   Variable declaration   */
  int i, n, ifo, imin;//, imax;

  char filename[100];
  char burnrows[100000];

  /* LAL data structure that will contain all of the data from the frame files */
  LALInferenceRunState *runState = XLALCalloc(1,sizeof(LALInferenceRunState));

  runState->commandLine = LALInferenceParseCommandLine(argc,argv);

  /******************************************************************************/
  /*                                                                            */
  /*  Usage statment if --help                                                  */
  /*                                                                            */
  /******************************************************************************/

  if(LALInferenceGetProcParamVal(runState->commandLine, "--help") || argc==1)
  {
    print_help_message();
    return 0;
  }

  /******************************************************************************/
  /*                                                                            */
  /*  Read/store  data                                                          */
  /*                                                                            */
  /******************************************************************************/

  /*
   Unlike in BayesWaveBurst, we DO NOT want to read and store the IFO data,
   but we have to use LALInferenceReadData to properly setup the runState
   structure and use the injection software.

   This is a hack to fill the runState detector with zeros so that all we store
   in runState is the injected waveform.
   */

  if( LALInferenceGetProcParamVal(runState->commandLine, "--0noise") )
  {
    fprintf(stdout,"Found --0noise in the command line\n");
  } else {
    fprintf(stdout," \n");
    fprintf(stdout,"Error:  You must include --0noise in the command line\n");
    return 0;
  }

  fprintf(stdout, "\n");
  fprintf(stdout, " ======= LALInferenceReadData() ======\n");
  runState->data = LALInferenceReadData(runState->commandLine);
  fprintf(stdout, "\n");


  /******************************************************************************/
  /*                                                                            */
  /*  read/store data                                                           */
  /*                                                                            */
  /******************************************************************************/
  LALInferenceIFOData *dataPtr = NULL;

  dataPtr = runState->data;

  double *SNRinj = NULL;

  int NI;
  int N = dataPtr->timeData->data->length;
  double Tobs = (double)N*dataPtr->timeData->deltaT;

  ifo=0;
  while(dataPtr!=NULL)
  {
    dataPtr = dataPtr->next;
    ifo++;
  }
  NI=ifo;

  double **freqData = dmatrix(0,NI-1,0,N-1);
  double **psd      = dmatrix(0,NI-1,0,N/2-1);
  double **one      = dmatrix(0,NI-1,0,N/2-1);

  struct Data *data = malloc(sizeof(struct Data));

  /*
   Unused, but needed for command line and initialization functions
   */
  struct Prior *prior = malloc(sizeof(struct Prior));
  struct Chain *chain = malloc(sizeof(struct Chain));

  parse_command_line(data, chain, prior, runState->commandLine);

  /*
   Move to the directory specified by the user, or stay in ./
   */
  chdir(data->outputDirectory);
  fprintf(stdout,"\nOutputting all data to: %s \n\n", data->outputDirectory);

  /*
   OUTPUT DIRECTORY 
   */
  char outdir[100];
  sprintf(outdir,"%spost",data->runName);
  mkdir(outdir,S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);

  /*
   Get trigger time in GPS seconds
  */
  LALStatus status;
  char *chartmp=NULL;
  LIGOTimeGPS GPStrig;
  ProcessParamsTable *ppt = NULL;

  memset(&status,0,sizeof(status));
  ppt = LALInferenceGetProcParamVal(runState->commandLine,"--trigtime");
  LALStringToGPS(&status,&GPStrig,ppt->value,&chartmp);


  // burstMDC injection
  int injectionFlag = 0;
  int xmlInjectionFlag = 0;
  if(LALInferenceGetProcParamVal(runState->commandLine, "--MDC-cache"))
  {
    SNRinj = malloc((NI+1)*sizeof(double));
    fprintf(stdout, " ==== LALInferenceInjectFromMDC(): started. ====\n");
    InjectFromMDC(runState->commandLine, runState->data, SNRinj);
    fprintf(stdout, " ==== LALInferenceInjectFromMDC(): finished. ====\n\n");
    free(SNRinj);

    injectionFlag = 1;
  }

  //CBC injection
  if(LALInferenceGetProcParamVal(runState->commandLine, "--inj"))
  {
    fprintf(stdout, " ==== LALInferenceInjectInspiralSignal(): started. ====\n");
    LALInferenceInjectInspiralSignal(runState->data,runState->commandLine);
    fprintf(stdout, " ==== LALInferenceInjectInspiralSignal(): finished. ====\n\n");

    injectionFlag = 1;
    xmlInjectionFlag = 1;
  }

  //Output Fourier domain data for BayesLine
  dataPtr = runState->data;

  double fmin   = dataPtr->fLow;
  double fmax   = dataPtr->fHigh;
  double deltaT = dataPtr->timeData->deltaT;
  double deltaF = dataPtr->freqData->deltaF;

  int tsize = dataPtr->timeData->data->length/4;

  //Copy Fourier domain data & PSD from LAL data types into simple arrays
  ifo=0;
  imin = (int)(fmin*Tobs);
  while(dataPtr!=NULL)
  {
    for(i=0; i<N/2; i++)
    {
      if(i<imin)
      {
        freqData[ifo][2*i]   = 0.0;
        freqData[ifo][2*i+1] = 0.0;
        psd[ifo][i] = 1.0;
      }
      else
      {
        freqData[ifo][2*i]   = creal(dataPtr->freqData->data->data[i]);
        freqData[ifo][2*i+1] = cimag(dataPtr->freqData->data->data[i]);

        /*
         Innerproducts are expecting <n_i^2> instead of Sn(f):
         So psd arrays == <n_i^2> = T/2 * Sn(f)
         */
        psd[ifo][i] = dataPtr->oneSidedNoisePowerSpectrum->data->data[i]*Tobs/2.0;
        one[ifo][i] = 1.0;
      }
    }
    dataPtr = dataPtr->next;
    ifo++;
  }

  //Storage for h+(t) and hx(t) of all IFOs
  double **timeDataPlus = dmatrix(0,NI-1,0,N-1);
  double **timeDataCross = dmatrix(0,NI-1,0,N-1);
  //Storage for h+(f) and hx(f) of all IFOs
  double **freqDataPlus = dmatrix(0,NI-1,0,N-1);
  double **freqDataCross = dmatrix(0,NI-1,0,N-1);
  if(xmlInjectionFlag)
  {
    //Compute the + and x polarizations of the injected waveform, then write them
    //to file following the conventions used by LALInferenceReadData to produce
    //the *_freqInjection.dat and *_timeInjection.dat files.
    compute_injected_plus_and_cross(runState, deltaT, deltaF, N, imin, outdir, timeDataPlus, timeDataCross, freqDataPlus, freqDataCross);
  }

  /******************************************************************************/
  /*                                                                            */
  /*  Setup DATA, CHAIN, PRIOR, MODEL structures                                */
  /*                                                                            */
  /******************************************************************************/

  /*
   DATA
   */
  initialize_data(data,freqData,N,tsize,psd,Tobs,NI,fmin,fmax);

  //Write out time and freq. samples
  write_time_freq_samples(data);

  data->detector  = malloc(NI*sizeof(LALDetector*));
  data->epoch     = runState->data->epoch;

  ifo=0;
  dataPtr = runState->data;
  while(dataPtr!=NULL)
  {
    data->detector[ifo] = dataPtr->detector;
    ifo++;
    dataPtr = dataPtr->next;
  }

  /*
   CHAIN
   */
  chain->NC = chain->NCmin;
  allocate_chain(chain,NI);

  /*
   PRIOR
   */
  int Dmax = data->Dmax;
  data->Dmax = 200;

  if(data->noiseOnlyFlag) Dmax=data->Dmax=1;
  int omax = 10;
  if(data->Dmax < omax) omax = data->Dmax;

  //Maximum number of sineGaussians
  prior->gmin = ivector(0,data->NI-1);
  prior->gmax = ivector(0,data->NI-1);
  //Wavelet parameters
  prior->range = dmatrix(0,4,0,1);

  initialize_priors(data, prior, omax);

  /*
   MODEL
   */
  struct Model *model = malloc(sizeof(struct Model));
  initialize_model(model, data, prior, &chain->seed);

  //Use Welch averaged PSD from LALInferenceReadData
  for(ifo=0; ifo<data->NI; ifo++)
  {
    for(i=0; i<N/2; i++)
    {
      data->SnS[ifo][i] = psd[ifo][i];
    }
  }

  /*
   BAYESLINE PSD
   */
  int PSDimin = 0;
  int PSDimax = data->N/2;

  if(data->bayesLineFlag)
  {

    struct BayesLineParams **bayesline = malloc(NI*sizeof(struct BayesLineParams *));
    data->bayesline = bayesline;


    FILE *splinechain;
    FILE *linechain;

    int count;
    int Nspline;
    int Nline;

    double *f =malloc(N/2*sizeof(double));
    double *Sn=malloc(N/2*sizeof(double));

    splineParams *spline    = NULL;
    lorentzianParams *lines = NULL;

    for(i=0; i<N/2; i++) f[i] = (double)(i)/data->Tobs;

    for(ifo=0; ifo<NI; ifo++)
    {

      //Get BayesLine parameter files
      sprintf(filename,"chains/%sclean_lorentzchain_ifo%i.dat.0",data->runName,ifo);
      linechain = fopen(filename,"r");
      if(linechain==NULL)
      {
        fprintf(stdout,"\nERROR:  Can not find %s \nTry running without --bayesLine argument\n\n",filename);
        return 1;
      }

      sprintf(filename,"chains/%sclean_splinechain_ifo%i.dat.0",data->runName,ifo);
      splinechain = fopen(filename,"r");
      if(splinechain==NULL)
      {
        fprintf(stdout,"\nERROR: Can not find %s \nTry running without --bayesLine argument\n\n",filename);
        return 1;
      }


      /*
       THE BROADBAND PSD MODEL
       */

      //Count lines in output files
      count=0;
      while(!feof(splinechain))
      {
        fgets(burnrows,100000,splinechain);
        count++;
      }
      count--;
      rewind(splinechain);

      //Now burn off all but the last row
      for(i=0; i<count-1; i++)fgets(burnrows,100000,splinechain);

      //Get dimension of spline model
      fscanf(splinechain,"%i",&Nspline);


      spline = malloc(sizeof(splineParams));
      create_splineParams(spline,Nspline);

      //Store spline points
      for(i=0; i<Nspline; i++) fscanf(splinechain,"%lg %lg",&spline->points[i],&spline->data[i]);
      PSDimin = (int)(spline->points[0]*data->Tobs);
      PSDimax = (int)(spline->points[Nspline-1]*data->Tobs);

      //Interpolate control points
      /*
       This is tricky:
       - The GSL cubic spline routine throws errors if interpolation
       is outside of control points.
       - Bayesline only fits to frequencies above fmin
       - Some index gymnastics are needed
       */
      for(i=0; i<N/2; i++) data->SnS[ifo][i] = 1.0;

      CubicSplineGSL(Nspline,spline->points,spline->data,PSDimax-PSDimin,f+PSDimin,Sn+PSDimin);

      //Map interpolated points to PSD used later on
      for(i=PSDimin; i<PSDimax; i++) data->SnS[ifo][i]=exp(Sn[i]);

      destroy_splineParams(spline);

      fclose(splinechain);


      /*
       THE SPECTRAL LINES MODEL
       */

       //Now burn off all but the last row
       for(i=0; i<count-1; i++)fgets(burnrows,1000000,linechain);

       //Get dimension of line model
       fscanf(linechain,"%i",&Nline);

       lines = malloc( sizeof(lorentzianParams) );
       create_lorentzianParams(lines, Nline);

       //Store line parameters
      for(i=0; i<Nline; i++)
      {
        fscanf(linechain,"%lg %lg %lg",&lines->f[i],&lines->A[i],&lines->Q[i]);

        Lorentzian(data->SnS[ifo], data->Tobs, lines, i, data->N);
      }

       destroy_lorentzianParams(lines);


      fclose(linechain);
    }

    free(f);
    free(Sn);
  }




  /******************************************************************************/
  /*                                                                            */
  /*  Declare the list of metrics to compute as part                            */
  /*  of the post-processing step.                                              */
  /*                                                                            */
  /******************************************************************************/

  fprintf(stdout, "\n");
  fprintf(stdout, " ===== BayesWave Post Processing =====\n");

  //User-friendly arrays for PSD, signal, and injected waveform
  double **hrec      = malloc(NI*sizeof(double*));
  double **hrecPlus  = malloc(NI*sizeof(double*));
  double **hrecCross = malloc(NI*sizeof(double*));
  double **grec      = malloc(NI*sizeof(double*));
  double **hinj      = malloc(NI*sizeof(double*));
  double **hinjPlus  = malloc(NI*sizeof(double*));
  double **hinjCross = malloc(NI*sizeof(double*));
  double **hoft      = malloc(NI*sizeof(double*));

  for(ifo=0; ifo<NI; ifo++)
  {
    hrec[ifo]      = malloc(N*sizeof(double));
    hrecPlus[ifo]  = malloc(N*sizeof(double));
    hrecCross[ifo] = malloc(N*sizeof(double));
    grec[ifo]      = malloc(N*sizeof(double));
    hinj[ifo]      = malloc(N*sizeof(double));
    hinjPlus[ifo]  = malloc(N*sizeof(double));
    hinjCross[ifo] = malloc(N*sizeof(double)); 
    hoft[ifo]      = malloc(N*sizeof(double));
  }

  //Fill hinj, hinjPlus, and hCrossPlus with the injected waveform (held in data because did a 0-noise simulation) and its + and x polarizations
  for(ifo=0; ifo<NI; ifo++) for(i=0; i<N; i++) hinj[ifo][i] = data->s[ifo][i];
  if(xmlInjectionFlag)
  {
    for(ifo=0; ifo<NI; ifo++) for(i=0; i<N; i++) hinjPlus[ifo][i] = freqDataPlus[ifo][i];
    for(ifo=0; ifo<NI; ifo++) for(i=0; i<N; i++) hinjCross[ifo][i] = freqDataCross[ifo][i];
  }

  // Different structures for whitened and "colored" moments
  struct Moments *whitenedInjectionMoments      = malloc(sizeof(struct Moments));
  struct Moments *whitenedPlusInjectionMoments  = malloc(sizeof(struct Moments));
  struct Moments *whitenedCrossInjectionMoments = malloc(sizeof(struct Moments));
  struct Moments *whitenedSignalMoments         = malloc(sizeof(struct Moments));
  struct Moments *whitenedPlusSignalMoments     = malloc(sizeof(struct Moments));
  struct Moments *whitenedCrossSignalMoments    = malloc(sizeof(struct Moments));
  struct Moments *whitenedGlitchMoments         = malloc(sizeof(struct Moments));
  struct Moments *whitenedCleanMoments          = malloc(sizeof(struct Moments));

  initialize_moments(whitenedInjectionMoments, NI);
  if(xmlInjectionFlag)
  {
    initialize_moments(whitenedPlusInjectionMoments, NI);
    initialize_moments(whitenedCrossInjectionMoments, NI);
  }
  initialize_moments(whitenedSignalMoments, NI);
  initialize_moments(whitenedPlusSignalMoments, NI);
  initialize_moments(whitenedCrossSignalMoments, NI);
  initialize_moments(whitenedGlitchMoments, NI);
  initialize_moments(whitenedCleanMoments, NI);

  struct Moments *coloredInjectionMoments      = malloc(sizeof(struct Moments));
  struct Moments *coloredPlusInjectionMoments  = malloc(sizeof(struct Moments));
  struct Moments *coloredCrossInjectionMoments = malloc(sizeof(struct Moments));
  struct Moments *coloredSignalMoments         = malloc(sizeof(struct Moments));
  struct Moments *coloredPlusSignalMoments     = malloc(sizeof(struct Moments));
  struct Moments *coloredCrossSignalMoments    = malloc(sizeof(struct Moments));
  struct Moments *coloredGlitchMoments         = malloc(sizeof(struct Moments));
  struct Moments *coloredCleanMoments          = malloc(sizeof(struct Moments));

  initialize_moments(coloredInjectionMoments, NI);
  if(xmlInjectionFlag)
  {
    initialize_moments(coloredPlusInjectionMoments, NI);
    initialize_moments(coloredCrossInjectionMoments, NI);
  }
  initialize_moments(coloredSignalMoments, NI);
  initialize_moments(coloredPlusSignalMoments, NI);
  initialize_moments(coloredCrossSignalMoments, NI);
  initialize_moments(coloredGlitchMoments, NI);
  initialize_moments(coloredCleanMoments, NI);


  // Keep track of dimension distribution for output stats
  int **dimension = malloc(NI*sizeof(int *));
  for(ifo=0; ifo<NI; ifo++) dimension[ifo] = malloc(data->Dmax*sizeof(int));

  // Shortcut pointers to different members of the Model structre
  struct Wavelet **glitch = NULL;
  struct Wavelet *signal  = NULL;


  int COUNT; //total number of chain iterations
  int BURNIN;//burn-in samples (COUNT/2)


  // Different files for whitened and "colored" output
  FILE **whitenedFile      = malloc((NI+1)*sizeof(FILE*));
  FILE **coloredFile       = malloc((NI+1)*sizeof(FILE*));
  FILE **whitenedPlusFile  = malloc((NI+1)*sizeof(FILE*));
  FILE **coloredPlusFile   = malloc((NI+1)*sizeof(FILE*));
  FILE **whitenedCrossFile = malloc((NI+1)*sizeof(FILE*));
  FILE **coloredCrossFile  = malloc((NI+1)*sizeof(FILE*));

  // Output injected moments separately
  FILE **whitenedInjFile      = malloc((NI+1)*sizeof(FILE*));
  FILE **coloredInjFile       = malloc((NI+1)*sizeof(FILE*));
  FILE **whitenedPlusInjFile  = malloc((NI+1)*sizeof(FILE*));
  FILE **coloredPlusInjFile   = malloc((NI+1)*sizeof(FILE*));
  FILE **whitenedCrossInjFile = malloc((NI+1)*sizeof(FILE*));
  FILE **coloredCrossInjFile  = malloc((NI+1)*sizeof(FILE*));

  // Output a subsample of all whitened waveforms to a single (gigantic) file
  FILE **recWhitenedWaveformFile      = malloc(NI*sizeof(FILE*));
  FILE **recPlusWhitenedWaveformFile  = malloc(NI*sizeof(FILE*));
  FILE **recCrossWhitenedWaveformFile = malloc(NI*sizeof(FILE*));
  FILE **recColoredWaveformFile       = malloc(NI*sizeof(FILE*));
  FILE **recPlusColoredWaveformFile   = malloc(NI*sizeof(FILE*));
  FILE **recCrossColoredWaveformFile  = malloc(NI*sizeof(FILE*));

  FILE **injWhitenedWaveformFile      = malloc(NI*sizeof(FILE*));
  FILE **injPlusWhitenedWaveformFile  = malloc(NI*sizeof(FILE*));
  FILE **injCrossWhitenedWaveformFile = malloc(NI*sizeof(FILE*));
  FILE **injColoredWaveformFile       = malloc(NI*sizeof(FILE*));
  FILE **injPlusColoredWaveformFile   = malloc(NI*sizeof(FILE*));
  FILE **injCrossColoredWaveformFile  = malloc(NI*sizeof(FILE*));



  /********************************/
  /*                              */
  /*  Post process data           */
  /*                              */
  /********************************/

  // read data from file made by BayesWaveBurst
  FILE *FDdata = NULL;

  // Output a subsample of all whitened waveforms to a single (gigantic) file
  FILE **whitenedDataFile = malloc(NI*sizeof(FILE*));
  FILE **coloredDataFile  = malloc(NI*sizeof(FILE*));

  for(ifo=0; ifo<NI; ifo++)
  {
    sprintf(filename,"chains/%sfourier_domain_data_ifo%i.dat",data->runName,ifo);

    if( (FDdata = fopen(filename,"r")) == NULL)
    {
      fprintf(stdout,"Could not find %s.  Filling data array with 0's\n",filename);
      for(i=0; i<data->N; i++) hoft[ifo][i] = 0.0;
    }
    else
    {
      for(i=0; i<data->N/2; i++) fscanf(FDdata,"%lg %lg",&hoft[ifo][2*i],&hoft[ifo][2*i+1]);
      fprintf(stdout, "Getting data into same format as waveforms\n");
      for(i=0; i<N/2; i++) psd[ifo][i] = data->SnS[ifo][i]; //data->SnS holds the "smooth" part of the PSD
    }

    fclose(FDdata);

  }

  //print fourier-domain data and model PSD
  FILE *PSDfile=NULL;
  for(ifo=0; ifo<NI; ifo++)
  {
    sprintf(filename,"%s/gaussian_noise_model_ifo%i.dat",outdir,ifo);
    PSDfile = fopen(filename,"w");
    for(i=PSDimin; i<PSDimax; i++) fprintf(PSDfile,"%g %g %g\n",(double)i/data->Tobs,hoft[ifo][2*i]*hoft[ifo][2*i]+hoft[ifo][2*i+1]*hoft[ifo][2*i+1],psd[ifo][i]);
    fclose(PSDfile);

    sprintf(filename,"%s/colored_data.dat.%i",outdir,ifo);
    coloredDataFile[ifo] = fopen(filename,"w");

    sprintf(filename,"%s/whitened_data.dat.%i",outdir,ifo);
    whitenedDataFile[ifo] = fopen(filename,"w");

  }

  //print time-domain data 
  print_waveforms(data, hoft, psd, whitenedDataFile);
  print_waveforms(data, hoft, one, coloredDataFile);


  for(ifo=0; ifo<NI; ifo++)
  {
      fclose(whitenedDataFile[ifo]);
      fclose(coloredDataFile[ifo]);
  }

  free(whitenedDataFile);
  free(coloredDataFile);


  /********************************/
  /*                              */
  /*  Post process injection      */
  /*                              */
  /********************************/
  if(injectionFlag)
  {

    fprintf(stdout, "Computing moments for injected waveform\n");

    for(ifo=0; ifo<NI; ifo++)
    {
      sprintf(filename,"%s/injected_whitened_waveform.dat.%i",outdir,ifo);
      injWhitenedWaveformFile[ifo] = fopen(filename,"w");
      if(xmlInjectionFlag)
      {
        sprintf(filename,"%s/injected_plus_whitened_waveform.dat.%i",outdir,ifo);
        injPlusWhitenedWaveformFile[ifo] = fopen(filename,"w");
        sprintf(filename,"%s/injected_cross_whitened_waveform.dat.%i",outdir,ifo);
        injCrossWhitenedWaveformFile[ifo] = fopen(filename,"w");
      }

      sprintf(filename,"%s/injected_colored_waveform.dat.%i",outdir,ifo);
      injColoredWaveformFile[ifo] = fopen(filename,"w");
      if(xmlInjectionFlag)
      {
        sprintf(filename,"%s/injected_plus_colored_waveform.dat.%i",outdir,ifo);
        injPlusColoredWaveformFile[ifo] = fopen(filename,"w");
        sprintf(filename,"%s/injected_cross_colored_waveform.dat.%i",outdir,ifo);
        injCrossColoredWaveformFile[ifo] = fopen(filename,"w");
      }
    }

    setup_output_files(NI, injectionFlag, outdir, "injection", "whitened", whitenedInjFile);
    setup_output_files(NI, injectionFlag, outdir, "injection", "colored", coloredInjFile);
    if(xmlInjectionFlag)
    {
      setup_output_files(NI, injectionFlag, outdir, "injectionPlus", "whitened", whitenedPlusInjFile);
      setup_output_files(NI, injectionFlag, outdir, "injectionPlus", "colored", coloredPlusInjFile);
      setup_output_files(NI, injectionFlag, outdir, "injectionCross", "whitened", whitenedCrossInjFile);
      setup_output_files(NI, injectionFlag, outdir, "injectionCross", "colored", coloredCrossInjFile);
    }

    for(ifo=0; ifo<NI; ifo++) for(i=0; i<N/2; i++) psd[ifo][i] = data->SnS[ifo][i]; //data->SnS holds the "smooth" part of the PSD

    for(ifo=0; ifo<NI; ifo++)
    {
      whitenedInjectionMoments->overlap[ifo] = 1.0;
      coloredInjectionMoments->overlap[ifo]  = 1.0;
      if(xmlInjectionFlag)
      {
        whitenedPlusInjectionMoments->overlap[ifo]  = 1.0;
        coloredPlusInjectionMoments->overlap[ifo]   = 1.0;
        whitenedCrossInjectionMoments->overlap[ifo] = 1.0;
        coloredCrossInjectionMoments->overlap[ifo]  = 1.0;
      }
    }
    whitenedInjectionMoments->networkOverlap = 1.0;
    coloredInjectionMoments->networkOverlap  = 1.0;

    ComputeWaveformMoments(data, deltaF, deltaT, hinj, psd, whitenedInjectionMoments);
    ComputeWaveformMoments(data, deltaF, deltaT, hinj, one, coloredInjectionMoments);

    if(xmlInjectionFlag)
    {
      ComputeWaveformMoments(data, deltaF, deltaT, hinjPlus, psd, whitenedPlusInjectionMoments);
      ComputeWaveformMoments(data, deltaF, deltaT, hinjPlus, one, coloredPlusInjectionMoments);
      ComputeWaveformMoments(data, deltaF, deltaT, hinjCross, psd, whitenedCrossInjectionMoments);
      ComputeWaveformMoments(data, deltaF, deltaT, hinjCross, one, coloredCrossInjectionMoments);
    }

    /* 
     Whitened hrss is meaningless and colored SNR are meaningless
     Swap their values in the structures so the moments files are self-contained
    
    for(ifo=0; ifo<NI; ifo++)
    {
      whitenedInjectionMoments->energy[ifo] = coloredInjectionMoments->energy[ifo];
      coloredInjectionMoments->snr[ifo]     = whitenedInjectionMoments->snr[ifo];
    }
    */

    //Detector-by-detector moments
    for(ifo=0; ifo<NI; ifo++)
    {
      print_moments(ifo, injectionFlag, whitenedInjFile, whitenedInjectionMoments);
      print_moments(ifo, injectionFlag, coloredInjFile, coloredInjectionMoments);
      if(xmlInjectionFlag)
      {
        print_moments(ifo, injectionFlag, whitenedPlusInjFile, whitenedPlusInjectionMoments);
        print_moments(ifo, injectionFlag, coloredPlusInjFile, coloredPlusInjectionMoments);
        print_moments(ifo, injectionFlag, whitenedCrossInjFile, whitenedCrossInjectionMoments);
        print_moments(ifo, injectionFlag, coloredCrossInjFile, coloredCrossInjectionMoments);
      }
    }
    //Network moments
    /*
    print_moments(NI, injectionFlag, whitenedInjFile, whitenedInjectionMoments);
    print_moments(NI, injectionFlag, coloredInjFile, coloredInjectionMoments);
    if(xmlInjectionFlag)
    {
      print_moments(NI, injectionFlag, whitenedPlusInjFile, whitenedPlusInjectionMoments);
      print_moments(NI, injectionFlag, coloredPlusInjFile, coloredPlusInjectionMoments);
      print_moments(NI, injectionFlag, whitenedCrossInjFile, whitenedCrossInjectionMoments);
      print_moments(NI, injectionFlag, coloredCrossInjFile, coloredCrossInjectionMoments);
    }
     */

    //Print injected time-domain waveforms and their + and x polarizations
    print_waveforms(data, hinj, psd, injWhitenedWaveformFile);
    print_waveforms(data, hinj, one, injColoredWaveformFile);
    if(xmlInjectionFlag)
    {
      print_waveforms(data, hinjPlus, psd, injPlusWhitenedWaveformFile);
      print_waveforms(data, hinjCross, psd, injCrossWhitenedWaveformFile);
      print_waveforms(data, hinjPlus, one, injPlusColoredWaveformFile);
      print_waveforms(data, hinjCross, one, injCrossColoredWaveformFile);
    }

    for(ifo=0; ifo<NI; ifo++)
    {
      fclose(injWhitenedWaveformFile[ifo]);
      fclose(injColoredWaveformFile[ifo]);
      fclose(whitenedInjFile[ifo]);
      fclose(coloredInjFile[ifo]);
      if(xmlInjectionFlag)
      {
        fclose(injPlusWhitenedWaveformFile[ifo]);
        fclose(injCrossWhitenedWaveformFile[ifo]);
        fclose(injPlusColoredWaveformFile[ifo]);
        fclose(injCrossColoredWaveformFile[ifo]);
        fclose(whitenedPlusInjFile[ifo]);
        fclose(whitenedCrossInjFile[ifo]);
        fclose(coloredPlusInjFile[ifo]);
        fclose(coloredCrossInjFile[ifo]);
      }
    }

    free(injWhitenedWaveformFile);
    free(injColoredWaveformFile);
    free(whitenedInjFile);
    free(coloredInjFile);
    if(xmlInjectionFlag)
    {
      free(injPlusWhitenedWaveformFile);
      free(injCrossWhitenedWaveformFile);
      free(injPlusColoredWaveformFile);
      free(injCrossColoredWaveformFile);
      free(whitenedPlusInjFile);
      free(whitenedCrossInjFile);
      free(coloredPlusInjFile);
      free(coloredCrossInjFile);
    }

  }



  /********************************/
  /*                              */
  /*  Post process signal model   */
  /*                              */
  /********************************/

  //File containing signal-model waveform parameters
  sprintf(filename,"chains/%ssignal_wavechain.dat.0",data->runName);
  FILE *signalParams;
//  FILE *temp;
  if( (signalParams = fopen(filename,"r")) == NULL)
  {
    fprintf(stdout,"Could not find %s.  Skipping to glitch model\n",filename);
  }
  else
  {
    fprintf(stdout, "Starting loop over %s  ...\n", filename);

    signal = model->signal;

    discard_burnin_samples(signalParams, &COUNT, &BURNIN);

    for(ifo=0; ifo<NI; ifo++)
    {
      sprintf(filename,"%s/signal_recovered_whitened_waveform.dat.%i",outdir,ifo);
      recWhitenedWaveformFile[ifo] = fopen(filename,"w");
      sprintf(filename,"%s/signal_recovered_plus_whitened_waveform.dat.%i",outdir,ifo);
      recPlusWhitenedWaveformFile[ifo] = fopen(filename,"w");
      sprintf(filename,"%s/signal_recovered_cross_whitened_waveform.dat.%i",outdir,ifo);
      recCrossWhitenedWaveformFile[ifo] = fopen(filename,"w");

      sprintf(filename,"%s/signal_recovered_colored_waveform.dat.%i",outdir,ifo);
      recColoredWaveformFile[ifo] = fopen(filename,"w");
      sprintf(filename,"%s/signal_recovered_plus_colored_waveform.dat.%i",outdir,ifo);
      recPlusColoredWaveformFile[ifo] = fopen(filename,"w");
      sprintf(filename,"%s/signal_recovered_cross_colored_waveform.dat.%i",outdir,ifo);
      recCrossColoredWaveformFile[ifo] = fopen(filename,"w");
    }

    setup_output_files(NI, injectionFlag, outdir, "signal", "whitened", whitenedFile);
    setup_output_files(NI, injectionFlag, outdir, "signal", "colored", coloredFile);
    setup_output_files(NI, injectionFlag, outdir, "plusSignal", "whitened", whitenedPlusFile);
    setup_output_files(NI, injectionFlag, outdir, "plusSignal", "colored", coloredPlusFile);
    setup_output_files(NI, injectionFlag, outdir, "crossSignal", "whitened", whitenedCrossFile);
    setup_output_files(NI, injectionFlag, outdir, "crossSignal", "colored", coloredCrossFile);

    for(ifo=0; ifo<NI; ifo++) for(n=0; n<data->Dmax; n++) dimension[ifo][n] = 0;

    for(n=0; n<COUNT-BURNIN; n++)
    {

      /******************************************************************************/
      /*                                                                            */
      /*  Get current state of MCMC output into format for BayesWave code library   */
      /*                                                                            */
      /******************************************************************************/

      /*
       SIGNAL MODEL
       */
      parse_signal_parameters(data, model, signalParams, hrec, hrecPlus, hrecCross);
      dimension[0][signal->size]++;

      /*
       PSD MODEL
       */
      //Fill psd with current model PSD
      for(ifo=0; ifo<NI; ifo++) for(i=0; i<N/2; i++) psd[ifo][i] = data->SnS[ifo][i]; //data->SnS holds the "smooth" part of the PSD

      /******************************************************************************/
      /*                                                                            */
      /*  Compute various moments/overlaps to quantify waveform reconstruction      */
      /*                                                                            */
      /******************************************************************************/

      /*
       MATCH BETWEEN INJECTED AND RECOVERED WAVEFORM
       */

      if(injectionFlag)
      {
        ComputeWaveformOverlap(data, hinj, hrec, psd, whitenedSignalMoments);
        ComputeWaveformOverlap(data, hinj, hrec, one, coloredSignalMoments);
        if(xmlInjectionFlag)
        {
          ComputeWaveformOverlap(data, hinjPlus, hrecPlus, psd, whitenedPlusSignalMoments);
          ComputeWaveformOverlap(data, hinjPlus, hrecPlus, one, coloredPlusSignalMoments);
          ComputeWaveformOverlap(data, hinjCross, hrecCross, psd, whitenedCrossSignalMoments);
          ComputeWaveformOverlap(data, hinjCross, hrecCross, one, coloredCrossSignalMoments);
        }
      }

      ComputeWaveformMoments(data, deltaF, deltaT, hrec, psd, whitenedSignalMoments);
      ComputeWaveformMoments(data, deltaF, deltaT, hrec, one, coloredSignalMoments);
      ComputeWaveformMoments(data, deltaF, deltaT, hrecPlus, psd, whitenedPlusSignalMoments);
      ComputeWaveformMoments(data, deltaF, deltaT, hrecPlus, one, coloredPlusSignalMoments);
      ComputeWaveformMoments(data, deltaF, deltaT, hrecCross, psd, whitenedCrossSignalMoments);
      ComputeWaveformMoments(data, deltaF, deltaT, hrecCross, one, coloredCrossSignalMoments);

      /*
       Whitened hrss is meaningless and colored SNR are meaningless
       Swap their values in the structures so the moments files are self-contained

      for(ifo=0; ifo<NI; ifo++)
      {
        whitenedSignalMoments->energy[ifo] = coloredSignalMoments->energy[ifo];
        coloredSignalMoments->snr[ifo]     = whitenedSignalMoments->snr[ifo];

        whitenedPlusSignalMoments->energy[ifo] = coloredPlusSignalMoments->energy[ifo];
        coloredPlusSignalMoments->snr[ifo]     = whitenedPlusSignalMoments->snr[ifo];

        whitenedCrossSignalMoments->energy[ifo] = coloredCrossSignalMoments->energy[ifo];
        coloredCrossSignalMoments->snr[ifo]     = whitenedCrossSignalMoments->snr[ifo];
      }*/

      for(ifo=0; ifo<NI; ifo++)
      {
        if(signal->size>0)
        {
          print_moments(ifo, injectionFlag, whitenedFile, whitenedSignalMoments);
          print_moments(ifo, injectionFlag, coloredFile, coloredSignalMoments);
          print_moments(ifo, injectionFlag, whitenedPlusFile, whitenedPlusSignalMoments);
          print_moments(ifo, injectionFlag, coloredPlusFile, coloredPlusSignalMoments);
          print_moments(ifo, injectionFlag, whitenedCrossFile, whitenedCrossSignalMoments);
          print_moments(ifo, injectionFlag, coloredCrossFile, coloredCrossSignalMoments);
        }
      }

      //Print reconstructed time-domain waveforms and their + and x polarizations
      if(n%1000 == 0)
      {
        /*
        sprintf(filename,"%s/temp.dat.%i.%i",outdir,ifo,n);
        temp = fopen(filename,"w");
        print_single_waveform(data, hrec, psd, temp);
        fclose(temp);
         */
        print_waveforms(data, hrec, psd, recWhitenedWaveformFile);
        print_waveforms(data, hrec, one, recColoredWaveformFile);
        print_waveforms(data, hrecPlus, psd, recPlusWhitenedWaveformFile);
        print_waveforms(data, hrecPlus, one, recPlusColoredWaveformFile);
        print_waveforms(data, hrecCross, psd, recCrossWhitenedWaveformFile);
        print_waveforms(data, hrecCross, one, recCrossColoredWaveformFile);
      }

    }

    fclose(signalParams);

    for(ifo=0; ifo<NI; ifo++)
    {
      fclose(recWhitenedWaveformFile[ifo]);
      fclose(recColoredWaveformFile[ifo]);
      fclose(whitenedFile[ifo]);
      fclose(coloredFile[ifo]);
      fclose(recPlusWhitenedWaveformFile[ifo]);
      fclose(recCrossWhitenedWaveformFile[ifo]);
      fclose(recPlusColoredWaveformFile[ifo]);
      fclose(recCrossColoredWaveformFile[ifo]);
      fclose(whitenedPlusFile[ifo]);
      fclose(whitenedCrossFile[ifo]);
      fclose(coloredPlusFile[ifo]);
      fclose(coloredCrossFile[ifo]);
    }

    print_stats(data, GPStrig, injectionFlag, outdir, "signal", dimension);


  } //End post processing of signal model





  /********************************/
  /*                              */
  /*  Post process glitch model   */
  /*                              */
  /********************************/

  FILE **glitchParams = malloc(NI*sizeof(FILE *));

  sprintf(filename,"chains/%sglitch_glitchchain_ifo0.dat.0",data->runName);

  if( (glitchParams[0] = fopen(filename,"r")) == NULL)
  {
    fprintf(stdout,"Could not find %s.  Skipping to cleaning model\n",filename);
  }
  else
  {
    fprintf(stdout, "Starting loop over %s  ...\n", filename);

    glitch = model->glitch;

    glitch[0] = model->glitch[0];

    for(ifo=1; ifo<NI; ifo++)
    {
      sprintf(filename,"chains/%sglitch_glitchchain_ifo%i.dat.0",data->runName,ifo);
      glitchParams[ifo] = fopen(filename,"r");

      glitch[ifo] = model->glitch[ifo];
    }

    for(ifo=0; ifo<NI; ifo++)
      discard_burnin_samples(glitchParams[ifo], &COUNT, &BURNIN);

    for(ifo=0; ifo<NI; ifo++)
    {
      sprintf(filename,"%s/glitch_recovered_whitened_waveform.dat.%i",outdir,ifo);
      recWhitenedWaveformFile[ifo] = fopen(filename,"w");

      sprintf(filename,"%s/glitch_recovered_colored_waveform.dat.%i",outdir,ifo);
      recColoredWaveformFile[ifo] = fopen(filename,"w");
    }

    setup_output_files(NI, injectionFlag, outdir, "glitch", "whitened", whitenedFile);
    setup_output_files(NI, injectionFlag, outdir, "glitch", "colored", coloredFile);

    for(ifo=0; ifo<NI; ifo++) for(n=0; n<data->Dmax; n++) dimension[ifo][n] = 0;

    for(n=0; n<COUNT-BURNIN; n++)
    {

      /******************************************************************************/
      /*                                                                            */
      /*  Get current state of MCMC output into format for BayesWave code library   */
      /*                                                                            */
      /******************************************************************************/

      /*
       GLITCH MODEL
       */
      parse_glitch_parameters(data, model, glitchParams, grec);
      for(ifo=0; ifo<NI; ifo++) dimension[ifo][glitch[ifo]->size]++;

      /*
       PSD MODEL
       */
      //Fill psd with current model PSD
      for(ifo=0; ifo<NI; ifo++) for(i=0; i<N/2; i++) psd[ifo][i] = data->SnS[ifo][i]; //data->SnS holds the "smooth" part of the PSD

      /******************************************************************************/
      /*                                                                            */
      /*  Compute various moments/overlaps to quantify waveform reconstruction      */
      /*                                                                            */
      /******************************************************************************/

      /*
       MATCH BETWEEN INJECTED AND RECOVERED WAVEFORM
       */

      if(injectionFlag)
      {
        ComputeWaveformOverlap(data, hinj, grec, psd, whitenedGlitchMoments);
        ComputeWaveformOverlap(data, hinj, grec, one, coloredGlitchMoments);
      }
      ComputeWaveformMoments(data, deltaF, deltaT, grec, psd, whitenedGlitchMoments);
      ComputeWaveformMoments(data, deltaF, deltaT, grec, one, coloredGlitchMoments);

      for(ifo=0; ifo<NI; ifo++)
      {
        if(glitch[ifo]->size>0)
        {
          print_moments(ifo, injectionFlag, whitenedFile, whitenedGlitchMoments);
          print_moments(ifo, injectionFlag, coloredFile, coloredGlitchMoments);
        }
      }

      //print time-domain waveforms
      if(n%1000 == 0)
      {
        print_waveforms(data, grec, psd, recWhitenedWaveformFile);
        print_waveforms(data, grec, one, recColoredWaveformFile);
      }

    }

    for(ifo=0; ifo<NI; ifo++)
    {
      fclose(recWhitenedWaveformFile[ifo]);
      fclose(recColoredWaveformFile[ifo]);
      fclose(glitchParams[ifo]);
      fclose(whitenedFile[ifo]);
      fclose(coloredFile[ifo]);
    }

    print_stats(data, GPStrig, injectionFlag, outdir, "glitch", dimension);

  } //End post processing of glitch model





  /**********************************/
  /*                                */
  /*  Post process cleaning phase   */
  /*                                */
  /**********************************/

  FILE **cleanParams = malloc(NI*sizeof(FILE *));

  sprintf(filename,"chains/%sclean_glitchchain_ifo0.dat.0",data->runName);
  if( (cleanParams[0] = fopen(filename,"r")) == NULL)
  {
    fprintf(stdout,"Could not find %s.  Finishing...\n",filename);
  }
  else
  {

    fprintf(stdout, "Starting loop over %s  ...\n", filename);

    glitch = model->glitch;

    glitch[0] = model->glitch[0];

    for(ifo=1; ifo<NI; ifo++)
    {
      sprintf(filename,"chains/%sclean_glitchchain_ifo%i.dat.0",data->runName,ifo);
      cleanParams[ifo] = fopen(filename,"r");

      glitch[ifo] = model->glitch[ifo];
    }

    for(ifo=0; ifo<NI; ifo++)
      discard_burnin_samples(cleanParams[ifo], &COUNT, &BURNIN);

    for(ifo=0; ifo<NI; ifo++)
    {
      sprintf(filename,"%s/clean_recovered_whitened_waveform.dat.%i",outdir,ifo);
      recWhitenedWaveformFile[ifo] = fopen(filename,"w");

      sprintf(filename,"%s/clean_recovered_colored_waveform.dat.%i",outdir,ifo);
      recColoredWaveformFile[ifo] = fopen(filename,"w");
    }

    setup_output_files(NI, injectionFlag, outdir, "clean", "whitened", whitenedFile);
    setup_output_files(NI, injectionFlag, outdir, "clean", "colored", coloredFile);

    for(ifo=0; ifo<NI; ifo++) for(n=0; n<data->Dmax; n++) dimension[ifo][n] = 0;

    for(n=0; n<COUNT-BURNIN; n++)
    {

      /******************************************************************************/
      /*                                                                            */
      /*  Get current state of MCMC output into format for BayesWave code library   */
      /*                                                                            */
      /******************************************************************************/

      /*
       GLITCH MODEL
       */
      parse_glitch_parameters(data, model, cleanParams, grec);
      for(ifo=0; ifo<NI; ifo++) dimension[ifo][glitch[ifo]->size]++;

      /*
       PSD MODEL
       */
      //Fill psd with current model PSD
      for(ifo=0; ifo<NI; ifo++) for(i=0; i<N/2; i++) psd[ifo][i] = data->SnS[ifo][i]; //data->SnS holds the "smooth" part of the PSD

      /******************************************************************************/
      /*                                                                            */
      /*  Compute various moments/overlaps to quantify waveform reconstruction      */
      /*                                                                            */
      /******************************************************************************/

      /*
       MATCH BETWEEN INJECTED AND RECOVERED WAVEFORM
       */

      if(injectionFlag)
      {
        ComputeWaveformOverlap(data, hinj, grec, psd, whitenedCleanMoments);
        ComputeWaveformOverlap(data, hinj, grec, one, coloredCleanMoments);
      }
      ComputeWaveformMoments(data, deltaF, deltaT, grec, psd, whitenedCleanMoments);
      ComputeWaveformMoments(data, deltaF, deltaT, grec, one, coloredCleanMoments);

      for(ifo=0; ifo<NI; ifo++)
      {
        if(glitch[ifo]->size>0)
        {
          print_moments(ifo, injectionFlag, whitenedFile, whitenedCleanMoments);
          print_moments(ifo, injectionFlag, coloredFile, coloredCleanMoments);
        }
      }

      //print time-domain waveforms
      if(n%1000 == 0)
      {
        print_waveforms(data, grec, psd, recWhitenedWaveformFile);
        print_waveforms(data, grec, one, recColoredWaveformFile);
      }
    }

    for(ifo=0; ifo<NI; ifo++)
    {
      fclose(recWhitenedWaveformFile[ifo]);
      fclose(recColoredWaveformFile[ifo]);
      fclose(cleanParams[ifo]);
      fclose(whitenedFile[ifo]);
      fclose(coloredFile[ifo]);
    }

    print_stats(data, GPStrig, injectionFlag, outdir, "clean", dimension);

  } //End post processing of cleaning phase

  /******************************************************************************/
  /*                                                                            */
  /*  Free memory, exit cleanly                                                 */
  /*                                                                            */
  /******************************************************************************/

  free(recWhitenedWaveformFile);
  free(recColoredWaveformFile);
  free(whitenedFile);
  free(coloredFile);

  fprintf(stdout, " ======= DONE! ======\n");

  return 0;

}


void print_help_message(void)
{
  fprintf(stdout,"\n ==== BayesWavePost() USAGE: ====\n\n");
  fprintf(stdout,"REQUIRED:\n");
  fprintf(stdout,"  --ifo IFO              interferometer (H1,L1,V1)\n");
  fprintf(stdout,"  --IFO-flow             minimum frequency (Hz)\n");
  fprintf(stdout,"  --IFO-cache LALAdLIGO  lie about the PSD\n");
  fprintf(stdout,"  --trigtime             GPS trigger time\n");
  fprintf(stdout,"  --srate                sampling rate (Hz)\n");
  fprintf(stdout,"  --seglen               duration of data (s)\n");
  fprintf(stdout,"  --PSDstart             GPS start time for PSD estimation\n");
  fprintf(stdout,"  --PSDlength            duration of PSD estimation length\n");
  fprintf(stdout,"  --dataseed 1234        Lie about random seed for data\n");
  fprintf(stdout,"  --0noise               no noise realization\n");
  fprintf(stdout,"\n");
  fprintf(stdout,"OPTIONAL:\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --- Run parameters   -------------------------------------------------------------\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --runName              run name for output files\n");
  fprintf(stdout,"\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --- Model parameters   -----------------------------------------------------------\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --bayesLine            use BayesLine for PSD model\n");
  fprintf(stdout,"\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --- LALInference injection  options  ---------------------------------------------\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --inj injfile.xml      Injection XML file to use\n");
  fprintf(stdout,"  --event N              Event number from Injection XML file to use\n");
  fprintf(stdout,"\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --- Burst MDC injection  options  ------------------------------------------------\n");
  fprintf(stdout,"  ----------------------------------------------------------------------------------\n");
  fprintf(stdout,"  --MDC-channel          IFO1-chan, IFO2-chan, etc\n");
  fprintf(stdout,"  --MDC-cache            IFO1-mdcframe, IFO2-mdcframe, etc\n");
  fprintf(stdout,"  --MDC-prefactor        Rescale injection amplitude (1.0)\n");
  fprintf(stdout,"\n");
  fprintf(stdout,"EXAMPLE:\n");
  fprintf(stdout,"./BayesWavePost --ifo H1 --H1-flow 32 --H1-cache LALSimAdLIGO --H1-channel LALSimAdLIGO --trigtime 900000000.00 --srate 512 --seglen 4 --PSDstart 900000000 --PSDlength 1024 --dataseed 1234 --0noise\n");
  fprintf(stdout,"\n");
}


void print_time_domain_waveforms(char filename[], double *h, int N, double *Snf, double eta, double Tobs, double fix, int imin, int imax, double tmin, double tmax)
{
  /*
   TODO: invFFT comes out time-reversed.
   LAL uses exp(-i2pift) for their Fourier transforms
   Why doesn't drealft(ty-1,N,1) work for inverse FFT?
   In the meantime, hacked invFFT by changing sign of imaginary part
   */
  int i;
  double x,t;
  FILE *waveout = fopen(filename,"w");

  h[0] = 0.0;
  h[1] = 0.0;
  for(i=1; i< N/2; i++)
  {
    if(i>imin && i<imax)
    {
      x = sqrt(Snf[i]*eta);
      h[2*i] /= x;
      h[2*i+1] /= -x;
    }
    else
    {
      h[2*i]=h[2*i+1]=0.0;
    }
  }

  drealft(h-1,N,-1);

  double t0 = Tobs-2.0;

  for(i=0; i<N; i++)
  {
    t = (double)(i)/(double)(N)*Tobs;
    if(t>=tmin && t<tmax) fprintf(waveout,"%e %e\n", t-t0, h[i]/fix);
  }

  fclose(waveout);
}


void ComputeWaveformOverlap(struct Data *data, double **hinj, double **hrec, double **psd, struct Moments *m)
{
  int N = data->N;
  int NI = data->NI;
  int imin = data->imin;

  int ifo;

  double hihi; //Detector (hinj|hinj)
  double hrhr; //Detector (hrec|hrec)
  double hrhi; //Detector (hrec|hinj)
  double ii=0; //Network (hinj|hinj)
  double rr=0; //Network (hrec|hrec)
  double ri=0; //Network (hrec|hinj)

  for(ifo=0; ifo<NI; ifo++)
  {
    hihi = fourier_nwip(imin, N/2, hinj[ifo], hinj[ifo], psd[ifo]);
    hrhr = fourier_nwip(imin, N/2, hrec[ifo], hrec[ifo], psd[ifo]);
    hrhi = fourier_nwip(imin, N/2, hrec[ifo], hinj[ifo], psd[ifo]);

    ii += hihi;
    rr += hrhr;
    ri += hrhi;

    //Store inj-rec overlap of each detector to file
    m->overlap[ifo] =  hrhi/sqrt(hihi*hrhr); // (hrec|hinj)/sqrt((hrec|hrec)(hinj|hinj))
  }

  //Store inj-rec networ overlap to file
  m->networkOverlap = ri/sqrt(ii*rr); // (hrec|hinj)/sqrt((hrec|hrec)(hinj|hinj)) 
}


void ComputeWaveformMoments(struct Data *data, double deltaF, double deltaT, double **h, double **psd, struct Moments *m)
{
  int N = data->N;
  int NI = data->NI;
  int imin = data->imin;
  int imax = data->imax;

  int ifo,n;
  double *rho = malloc(N*sizeof(double));
  double **wt = malloc(NI*sizeof(double *));
  double **wf = malloc(NI*sizeof(double *));

  for(ifo=0; ifo<NI; ifo++)
  {
    wt[ifo]=malloc(N*sizeof(double));
    wf[ifo]=malloc(N*sizeof(double));
  }

  //get input waveform into right format (either colored or whitened)
  for(ifo=0; ifo<NI; ifo++)
  {
    for(n=0; n<N/2; n++)
    {
      wf[ifo][2*n]   = h[ifo][2*n]/sqrt(psd[ifo][n]);
      wf[ifo][2*n+1] = h[ifo][2*n+1]/sqrt(psd[ifo][n]);
    }
    //transform waveform into time-domain
    get_time_domain_waveforms(wt[ifo], wf[ifo], N, imin, imax);
  }

  //SNR
  for(ifo=0; ifo<NI; ifo++) m->snr[ifo] = detector_snr(imin, imax, h[ifo], psd[ifo], 1.0);

  /******************************/
  /*                            */
  /*  Frequency domain moments  */
  /*                            */
  /******************************/

  for(ifo=0; ifo<=NI; ifo++)
  {
    /*
     RECOVERED SIGNAL ENERGY
     */
    if(ifo<NI)
      m->energy[ifo] = get_rhof(imin, imax, wf[ifo], deltaF, rho);
    else
      m->energy[ifo] = get_rhof_net(imin, imax, wf, deltaF, rho, NI);

    /*
     RECOVERED CENTRAL FREQUENCY
     */
    m->frequency[ifo] = get_f0(imax, rho, deltaF);


    /*
     RECOVERED BANDWIDTH
     */
    m->bandwidth[ifo] = get_band(imax, rho, deltaF, m->frequency[ifo]);
  }



  /******************************/
  /*                            */
  /*     Time domain moments    */
  /*                            */
  /******************************/

  for(ifo=0; ifo<=NI; ifo++)
  {
    /*
     RECOVERED SIGNAL ENERGY
     */
    if(ifo<NI)
      m->energy[ifo] = get_rhot(N, wt[ifo], deltaT, rho);
    else
      m->energy[ifo] = get_rhot_net(N, wt, deltaT, rho, NI);

    /*
     RECOVERED CENTRAL TIME
     */
    m->time[ifo] = get_f0(N, rho, deltaT);

    /*
     RECOVERED WAVEFORM DURATION
     */
    m->duration[ifo] = get_band(N, rho, deltaT, m->time[ifo]);

    /*
     WAVEFORM MAXIMUM AND TIME AT MAXIMUM
     */
    if(ifo<NI)
    {
      get_hmax_and_t(N, wt[ifo], deltaT, &m->hmax[ifo], &m->t_at_hmax[ifo]);
    }

  }


  for(ifo=0; ifo<NI; ifo++)
  {
    free(wf[ifo]);
    free(wt[ifo]);
  }
  free(rho);
  free(wt);
  free(wf);


}

void initialize_moments(struct Moments *m, int NI)
{
  m->overlap   = malloc((NI+1)*sizeof(double));
  m->energy    = malloc((NI+1)*sizeof(double));
  m->frequency = malloc((NI+1)*sizeof(double));
  m->time      = malloc((NI+1)*sizeof(double));
  m->bandwidth = malloc((NI+1)*sizeof(double));
  m->duration  = malloc((NI+1)*sizeof(double));
  m->snr       = malloc((NI+1)*sizeof(double));
  m->hmax      = malloc((NI+1)*sizeof(double));
  m->t_at_hmax = malloc((NI+1)*sizeof(double));
}


void setup_output_files(int NI, int injectionFlag, char *outdir, char *model, char *weighting, FILE **momentsOut)
{
  int ifo;
  char filename[100];

  //Open files
  for(ifo=0; ifo<NI; ifo++)
  {
    sprintf(filename,"%s/%s_%s_moments.dat.%i", outdir, model, weighting, ifo);
    momentsOut[ifo] = fopen(filename,"w");
  }
  sprintf(filename,"%s/%s_%s_moments.dat.net", outdir, model, weighting);
  momentsOut[NI] = fopen(filename,"w");

  //Write headers
  for(ifo=0; ifo<=NI; ifo++)
  {
    fprintf(momentsOut[ifo], "# ");
    fprintf(momentsOut[ifo], "snr ");
    fprintf(momentsOut[ifo], "t_energy_rec ");
    fprintf(momentsOut[ifo], "hrss ");
    fprintf(momentsOut[ifo], "t0_rec ");
    fprintf(momentsOut[ifo], "dur_rec ");
    fprintf(momentsOut[ifo], "f0_rec ");
    fprintf(momentsOut[ifo], "band_rec ");
    if(injectionFlag)
    {
      fprintf(momentsOut[ifo], "overlap ");
      fprintf(momentsOut[ifo], "network_overlap "); 
    }
    fprintf(momentsOut[ifo], "h_max ");
    fprintf(momentsOut[ifo], "t_at_h_max ");
    fprintf(momentsOut[ifo], "\n");
  }
}

void parse_glitch_parameters(struct Data *data, struct Model *model, FILE **paramsfile, double **grec)
{
  int ifo;
  int d;
  int i;

  struct Wavelet **glitch = model->glitch;

  for(ifo=0; ifo<data->NI; ifo++)
  {
    //Zero out signal->templates array (holds the linear combination of wavelets)
    for(i=0; i<data->N; i++) glitch[ifo]->templates[i] = 0.0;

    //Get the number of wavelet basis functions for the current sample
    fscanf(paramsfile[ifo], "%i", &glitch[ifo]->size);


    //Get each wavelets parameters and fill up signal structure
    for(d=1; d<=glitch[ifo]->size; d++)
    {
      glitch[ifo]->index[d]=d;
      for(i=0; i<NG; i++)
        fscanf(paramsfile[ifo],"%lg ", &glitch[ifo]->intParams[d][i]);

      //Append new wavelet to glitch[ifo]->templates array
      SineGaussian(glitch[ifo]->templates, glitch[ifo]->intParams[d], data->N, 1, data->Tobs);

    }

    //fill hrec with reconstructed waveform
    for(i=0; i<data->N; i++) grec[ifo][i] = glitch[ifo]->templates[i];
  }
}

void parse_signal_parameters(struct Data *data, struct Model *model, FILE *paramsfile, double **hrec, double **hrecPlus, double **hrecCross)
{
  int ifo;
  int d;
  int i;

  struct Wavelet *signal = model->signal;

  //Zero out signal->templates array (holds the linear combination of wavelets)
  for(i=0; i<data->N; i++) signal->templates[i] = 0.0;

  //Get the number of wavelet basis functions for the current sample
  fscanf(paramsfile, "%i", &signal->size);

  //First in the wavechain file are the extrinsic parameters
  for(i=0; i<NE; i++)
    fscanf(paramsfile,"%lg", &model->extParams[i]);

  //Compute network projection coefficients
  computeProjectionCoeffs(data, model->projection, model->extParams);

  //Get each wavelets parameters and fill up signal structure
  for(d=1; d<=signal->size; d++)
  {
    signal->index[d]=d;
    for(i=0; i<NG; i++)
      fscanf(paramsfile,"%lg ", &signal->intParams[d][i]);

    //Append new wavelet to signal->templates array
    SineGaussian(signal->templates, signal->intParams[d], data->N, 1, data->Tobs);

  }

  //Project geocenter waveform onto network
  waveformProject(data, model->projection, model->extParams, model->response, signal->templates, data->fmin, data->fmax);

  //Get + and x polarizations of hrec
  compute_reconstructed_plus_and_cross(data, model->projection, model->extParams, hrecPlus, hrecCross, signal->templates, data->fmin, data->fmax);

  //fill hrec with reconstructed waveform
  for(ifo=0; ifo<data->NI; ifo++) for(i=0; i<data->N; i++) hrec[ifo][i] = model->response[ifo][i];
}

void discard_burnin_samples(FILE *params, int *COUNT, int *BURNIN)
{
  int n;
  char burnrows[100000];
  (*COUNT) = 0;
  while(!feof(params))
  {
    fgets(burnrows,100000,params);
    (*COUNT)++;
  }
  (*COUNT)--;
  (*BURNIN) = (*COUNT)/2;
  rewind(params);

  for(n=0; n<(*BURNIN); n++) fgets(burnrows,100000,params);
}

void print_moments(int ifo, int injectionFlag, FILE **outfile, struct Moments *m)
{
  fprintf(outfile[ifo], "%e ",m->snr[ifo]);
  fprintf(outfile[ifo], "%e ",m->energy[ifo]);
  fprintf(outfile[ifo], "%e ",log10(sqrt(m->energy[ifo]/2.)));
  fprintf(outfile[ifo], "%e ",m->time[ifo]);
  fprintf(outfile[ifo], "%e ",m->duration[ifo]);
  fprintf(outfile[ifo], "%e ",m->frequency[ifo]);
  fprintf(outfile[ifo], "%e ",m->bandwidth[ifo]);
  if(injectionFlag)
  {
    fprintf(outfile[ifo], "%e ",m->overlap[ifo]);
    //if(injectionFlag)fprintf(outfile[ifo], "%e\n",m->overlap[ifo]);
    fprintf(outfile[ifo], "%e ",m->networkOverlap);
  }
  fprintf(outfile[ifo], "%e ",m->hmax[ifo]);
  fprintf(outfile[ifo], "%e ",m->t_at_hmax[ifo]);
  fprintf(outfile[ifo], "\n");
}

void print_waveforms(struct Data *data, double **h, double **psd, FILE **ofile)
{
  int i;
  int ifo;

  int N = data->N;
  int NI = data->NI;
  int imin = data->imin;
  int imax = data->imax;

  double *hoft = malloc(N*sizeof(double));
  double *w    = malloc(N*sizeof(double));

  for(ifo=0; ifo<NI; ifo++)
  {
    //whiten waveforms
    for(i=0; i<N/2; i++)
    {
      w[2*i]   =   h[ifo][2*i] / sqrt(psd[ifo][i]);
      w[2*i+1] = h[ifo][2*i+1] / sqrt(psd[ifo][i]);
    }

    get_time_domain_waveforms(hoft, w, N, imin, imax);

    for(i=0; i<N; i++)
      fprintf(ofile[ifo], "%e ", hoft[i]);
      //fprintf(ofile[ifo], "%e\n", hoft[i]);

    fprintf(ofile[ifo], "\n");
  }

  free(hoft);
  free(w);
}

void print_single_waveform(struct Data *data, double **h, double **psd, FILE *ofile)
{
  int i;
  int ifo;

  int N = data->N;
  int imin = data->imin;
  int imax = data->imax;

  double *hoft = malloc(N*sizeof(double));
  double *w    = malloc(N*sizeof(double));

  double maxh = 0.0;
  double maxt = 0.0;
  double samp = data->Tobs/(double)data->N;

  ifo=0;
  //whiten waveforms
  for(i=0; i<N/2; i++)
  {
    w[2*i]   =   h[ifo][2*i] / sqrt(psd[ifo][i]);
    w[2*i+1] = h[ifo][2*i+1] / sqrt(psd[ifo][i]);
  }

  //Obtain waveform in the time domain
  get_time_domain_waveforms(hoft, w, N, imin, imax);

  for(i=0; i<N; i++) fprintf(ofile, "%e\n", hoft[i]);

  for(i=0; i<N; i++)
  {
    if(sqrt(hoft[i]*hoft[i])>maxh )
    {
      maxh = sqrt(hoft[i]*hoft[i]);
      maxt = (double)i*samp;
    }
  }

  printf("max waveform %g at %g s \n",maxh,maxt);

  free(hoft);
  free(w);
}

void print_stats(struct Data *data, LIGOTimeGPS GPS, int injectionFlag, char *outdir, char *model, int **dimension)
{
  int i;
  int ifo;
  int NI = data->NI;

  int map,max;
  double bayesfactor;
  int Zs,Zn;

  FILE *momentsfile = NULL;
  FILE *statsfile   = NULL;

  char filename[100];
  char filerow[10000];
  int count;

  double *snr       = NULL;
  double *time      = NULL;
  double *duration  = NULL;
  double *frequency = NULL;
  double *bandwidth = NULL;
  double *hmax      = NULL;
  double *t_at_hmax = NULL;

  int seconds = GPS.gpsSeconds;
  int nanoseconds = GPS.gpsNanoSeconds;
  double GPStime = (double)seconds + 1.0e-9*nanoseconds;

  for(ifo=0; ifo<NI; ifo++)
  {

    //find model with most posterior support

    map=max=-1;
    bayesfactor=0.0;
    Zn = 0;
    Zs = 0;

    if(!strcmp(model,"signal"))
    {
      for(i=0; i<data->Dmax; i++)
      {
        //evidence for each model
        if(i==0) Zn += dimension[0][i];
        else     Zs += dimension[0][i];

        //map model
        if(dimension[0][i]>max)
        {
          max = dimension[0][i];
          map = i;
        }
      }
      if(ifo==0)printf("map dimension for channel \"%s\" = %i\n",model,map);
    }
    else
    {
      for(i=0; i<data->Dmax; i++)
      {
        //evidence for each model
        if(i==0) Zn += dimension[ifo][i];
        else     Zs += dimension[ifo][i];

        //map model
        if(dimension[ifo][i]>max)
        {
          max = dimension[ifo][i];
          map = i;
        }
      }
      printf("map dimension for channel \"%s[%i]\" = %i\n",model,ifo,map);
    }

    if(Zn==0) bayesfactor = 1000.0;
    else      bayesfactor = (double)Zs/(double)Zn;


    sprintf(filename,"%s/%s_whitened_moments.dat.%i",outdir,model,ifo);
    momentsfile = fopen(filename,"r");

    sprintf(filename,"%s/%s_stats.dat.%i",outdir,model,ifo);
    statsfile = fopen(filename,"w");

    // print header for stats file
    fprintf(statsfile, "# ");
    fprintf(statsfile, "map_D ");
    fprintf(statsfile, "bayesfactor ");
    fprintf(statsfile, "snr ");
    fprintf(statsfile, "time ");
    fprintf(statsfile, "duration ");
    fprintf(statsfile, "frequency ");
    fprintf(statsfile, "bandwidth ");
    fprintf(statsfile, "h_max ");
    fprintf(statsfile, "t_at_h_max ");
    fprintf(statsfile, "\n");

    count=0;

    //ignore header
    fgets(filerow,10000,momentsfile);

    //get size of file
    while(!feof(momentsfile))
    {
      fgets(filerow,10000,momentsfile);
      count++;
    }
    rewind(momentsfile);
    count--;

    //allocate memory, store burst table moments
    double junk;
    snr       = malloc(count*sizeof(double));
    time      = malloc(count*sizeof(double));
    duration  = malloc(count*sizeof(double));
    frequency = malloc(count*sizeof(double));
    bandwidth = malloc(count*sizeof(double));
    hmax = malloc(count*sizeof(double));
    t_at_hmax = malloc(count*sizeof(double));


    //ignore header again
    fgets(filerow,10000,momentsfile);

    for(i=0; i<count; i++)
    {
      fscanf(momentsfile, "%lg",&snr[i]);
      fscanf(momentsfile, "%lg",&junk);
      fscanf(momentsfile, "%lg",&junk); //FP: bug? There are two values between snr and t0_rec but only one was going to &junk
      fscanf(momentsfile, "%lg",&time[i]);
      fscanf(momentsfile, "%lg",&duration[i]);
      fscanf(momentsfile, "%lg",&frequency[i]);
      fscanf(momentsfile, "%lg",&bandwidth[i]);
      if(injectionFlag)
      {
        fscanf(momentsfile, "%lg",&junk);
        fscanf(momentsfile, "%lg",&junk);
      }
      fscanf(momentsfile, "%lg",&hmax[i]);
      fscanf(momentsfile, "%lg",&t_at_hmax[i]);
    }

    //sort each moment array
    gsl_sort(snr,1,count);
    gsl_sort(time,1,count);
    gsl_sort(duration,1,count);
    gsl_sort(frequency,1,count);
    gsl_sort(bandwidth,1,count);
    gsl_sort(hmax,1,count);
    gsl_sort(t_at_hmax,1,count);

    //get and print medians for stats file
    fprintf(statsfile, "%i ",    map );
    fprintf(statsfile, "%g ",    bayesfactor);
    fprintf(statsfile, "%g ",    gsl_stats_median_from_sorted_data (snr,1,count) );
    fprintf(statsfile, "%.16g ", gsl_stats_median_from_sorted_data (time,1,count)  + (GPStime + 2.0 - data->Tobs) );
    fprintf(statsfile, "%g ",    gsl_stats_median_from_sorted_data (duration,1,count) );
    fprintf(statsfile, "%g ",    gsl_stats_median_from_sorted_data (frequency,1,count) );
    fprintf(statsfile, "%g ",    gsl_stats_median_from_sorted_data (bandwidth,1,count) );
    fprintf(statsfile, "%g ",    gsl_stats_median_from_sorted_data (hmax,1,count));
    fprintf(statsfile, "%g ",    gsl_stats_median_from_sorted_data (t_at_hmax,1,count) + (GPStime + 2.0 - data->Tobs));
    fprintf(statsfile, "\n");

    fclose(momentsfile);
    fclose(statsfile);

    free(snr);
    free(time);
    free(duration);
    free(frequency);
    free(bandwidth);

  }
}

double get_energy(int imin, int imax, double *a, double deltaF)
{
  int i;
  double result;

  // Calculate total signal energy
  result = 0.0;
  for(i=imin; i<imax; i++){
    result += (a[i*2]*a[i*2] + a[i*2+1]*a[i*2+1]);
  }
  result *= 2*deltaF;
  return(result);
}

double get_rhof(int imin, int imax, double *a, double deltaF, double *arg)
{
  // Calculate rho_f (Normalized signal energy frequency distribution.  See moments.pdf eqn 1.5)

  int i;
  double norm;

  // We'll normalize by the total signal energy
  norm = get_energy(imin, imax, a, deltaF);
  for(i=0; i<imax; i++) {
    if (i<imin) {
      arg[i] = 0;
    } else {
      arg[i] = 2*(a[i*2]*a[i*2] + a[i*2+1]*a[i*2+1])/norm;
    }
  }
  // Return value is the total signal energy
  return(norm);
}

double get_rhof_net(int imin, int imax, double **a, double deltaF, double *arg, int NI)
{
  // Calculate rho_f (Normalized signal energy frequency distribution.  See moments.pdf eqn 1.5)

  int i,ifo;
  double norm;

  // We'll normalize by the total signal energy
  norm=0.0;
  for(ifo=0; ifo<NI; ifo++) norm += get_energy(imin, imax, a[ifo], deltaF);

  for(i=0; i<imax; i++)
  {
    arg[i]=0.0;
    if(i>=imin) for(ifo=0; ifo<NI; ifo++) arg[i] += 2*(a[ifo][i*2]*a[ifo][i*2] + a[ifo][i*2+1]*a[ifo][i*2+1])/norm;
  }

  // Return value is the total signal energy
  return(norm);
}

double get_rhot(int N, double *a, double deltaT, double *arg)
{
  // Calculate rho_t (Normalized signal energy frequency distribution.  See moments.pdf eqn 1.5)
  int i;
  double norm;

  // We'll normalize by the total signal energy
  norm = 0;
  for(i=0;i<N; i++) {
    norm += a[i]*a[i];
  }
  norm *= deltaT;

  for(i=0; i<N; i++) {
    arg[i] = a[i]*a[i]/norm;
  }
  // Return value is the total signal energy
  return(norm);
}

double get_rhot_net(int N, double **a, double deltaT, double *arg, int NI)
{
  // Calculate rho_t (Normalized signal energy frequency distribution.  See moments.pdf eqn 1.5)
  int i,ifo;
  double norm;

  // We'll normalize by the total signal energy
  norm = 0;
  for(ifo=0; ifo<NI; ifo++)
  {
    for(i=0;i<N; i++) {
      norm += a[ifo][i]*a[ifo][i];
    }
  }
  norm *= deltaT;

  for(i=0; i<N; i++) {
    arg[i]=0.0;
    for(ifo=0; ifo<NI; ifo++) arg[i] += a[ifo][i]*a[ifo][i]/norm;
  }
  // Return value is the total signal energy
  return(norm);
}


double rhonorm(int imax, double *a, double deltaF)
{
  // Calculate normalization.  This is a sanity check, since rho is supposed to have a normalization of 1.
  int i;
  double norm;

  norm = 0.0;
  for(i=0; i<imax; i++) {
    norm += (a[i]);
  }
  norm *= deltaF;
  return(norm);
}

double get_f0(int imax, double *a, double deltaF)
{
  // Calculate the central frequency (aka first moment).
  int i;
  double result;

  result = 0.0;
  for(i=0; i<imax; i++) {
    result += (a[i]*i);
  }
  result *= deltaF*deltaF;
  return(result);
}

double get_band(int imax, double *a, double deltaF, double f0)
{
  // Calculate the bandwidth (aka 2nd moment).
  int i;
  double result;
  
  result = 0.0;
  for(i=0; i<imax; i++) {
    result += ( a[i]*(i*deltaF - f0)*(i*deltaF - f0) );
  }
  result *= deltaF;
  result = sqrt(result);
  return(result);
}

void get_hmax_and_t(int imax, double *a, double deltaT, double *hmax, double *t_at_hmax)
{
  //Find the maximum value of the vector and the time there
  int i;
  double a_abs;

  *hmax = 0.0;
  *t_at_hmax = 0.0;
  
  for(i=0; i<imax; i++)
  {
    a_abs = sqrt(a[i]*a[i]); 
    if(a_abs>*hmax)
    {
      *hmax = a_abs;
      *t_at_hmax = (double)i*deltaT;
    }
  }
}

void whiten_data(int imin, int imax, double *a, double *Snf)
{
  int i;
  for(i=imin; i<imax; i++) {
    a[i*2] = a[i*2]/sqrt(Snf[i]);
    a[i*2+1] = a[i*2+1]/sqrt(Snf[i]);
  }
}

void get_time_domain_waveforms(double *hoft, double *h, int N, int imin, int imax)
{
  /*
   TODO: invFFT comes out time-reversed.
   LAL uses exp(-i2pift) for their Fourier transforms
   Why doesn't drealft(ty-1,N,1) work for inverse FFT?
   In the meantime, hacked invFFT by changing sign of imaginary part
   */
  
  int i;
  
  hoft[0] = 0.0;
  hoft[1] = 0.0;
  for(i=1; i< N/2; i++)
  {
    if(i>imin && i<imax)
    {
      hoft[2*i]   = h[2*i];
      hoft[2*i+1] = -h[2*i+1];
    }
    else
    {
      hoft[2*i]=hoft[2*i+1]=0.0;
    }
  }
  
  drealft(hoft-1,N,-1);
  
}

void write_time_freq_samples(struct Data *data)
{

  int N    = data->N;
  int imax = data->imax;

  double Tobs   = data->Tobs;
  double deltaF = 1./data->Tobs;

  char filename[100];
  sprintf(filename,"%spost/timesamp.dat",data->runName);
  FILE *timesamp = fopen(filename, "w");
  sprintf(filename,"%spost/freqsamp.dat",data->runName);
  FILE *freqsamp = fopen(filename, "w");
  double t0 = Tobs-2.0;
  double f;
  double t;
  int i;
  for(i=0; i<N; i++) {
    t = (i*Tobs)/N;
    fprintf(timesamp, "%e \n", t-t0);
  }
  fclose(timesamp);
  for (i=0;i<imax;i++) {
    f = i*deltaF;
    fprintf(freqsamp, "%e \n", f);
  }
  fclose(freqsamp);
}

void compute_injected_plus_and_cross(LALInferenceRunState *runState, double deltaT, double deltaF, int N, int imin, char *outdir, double **timeDataPlus, double **timeDataCross, double **freqDataPlus, double **freqDataCross) 
{
  //This function was inspired by a portion of
  //LALInferenceInjectInspiralSignal (in LALInferenceReadData.c)
  fprintf(stdout, " ======= Computing injected h+ and hx ======\n");
  ProcessParamsTable *procparam=NULL;
  SimInspiralTable *injTable=NULL;
  procparam=LALInferenceGetProcParamVal(runState->commandLine,"--inj");
  if(procparam) SimInspiralTableFromLIGOLw(&injTable,procparam->value,0,0);
  //Select the desired injection event from the injection table
  int event=atoi(LALInferenceGetProcParamVal(runState->commandLine,"--event")->value);
  int i=0;
  while(i<event){
    i++;
    injTable = injTable->next;
  }
  SimInspiralTable *injEvent=NULL;
  injEvent = injTable;
  injEvent->next = NULL;
  //Reference frequency
  REAL8 fref = 100.;
  if(LALInferenceGetProcParamVal(runState->commandLine,"--inj-fref"))
    fref = atoi(LALInferenceGetProcParamVal(runState->commandLine,"--inj-fref")->value);
  //Tidal deformabilities
  REAL8 lambda1 = 0.;
  if(LALInferenceGetProcParamVal(runState->commandLine,"--inj-lambda1"))
    lambda1=atof(LALInferenceGetProcParamVal(runState->commandLine,"--inj-lambda1")->value);
  REAL8 lambda2 = 0.;
  if(LALInferenceGetProcParamVal(runState->commandLine,"--inj-lambda2"))
    lambda2=atof(LALInferenceGetProcParamVal(runState->commandLine,"--inj-lambda2")->value);
  //PN orders of the spin terms and of the tidal effects
  LALSimInspiralWaveformFlags *waveFlags = XLALSimInspiralCreateWaveformFlags();
  LALSimInspiralSpinOrder spinO = -1;
  if(LALInferenceGetProcParamVal(runState->commandLine,"--inj-spinOrder")) {
    spinO = atoi(LALInferenceGetProcParamVal(runState->commandLine,"--inj-spinOrder")->value);
    XLALSimInspiralSetSpinOrder(waveFlags, spinO);
  }
  LALSimInspiralTidalOrder tideO = -1;
  if(LALInferenceGetProcParamVal(runState->commandLine,"--inj-tidalOrder")) {
    tideO = atoi(LALInferenceGetProcParamVal(runState->commandLine,"--inj-tidalOrder")->value);
    XLALSimInspiralSetTidalOrder(waveFlags, tideO);
  }
  LALSimInspiralTestGRParam *nonGRparams = NULL; 
  //PN order of the model
  LALPNOrder order = XLALGetOrderFromString(injEvent->waveform);
  //Amplitude PN order of the model
  INT4 amporder=0;
  amporder = injEvent->amp_order; 
  //Get approximant value
  Approximant approximant = XLALGetApproximantFromString(injEvent->waveform);
  //Starting frequency for waveform generation
  REAL8 f_min = fLow2fStart(injEvent->f_lower, amporder, approximant);
  //Generate the raw + and x-polarization waveforms in the time domain
  REAL8TimeSeries *hp=NULL;
  REAL8TimeSeries *hc=NULL;
  XLALSimInspiralChooseTDWaveform(&hp, &hc, injEvent->coa_phase, deltaT,
                                  injEvent->mass1*LAL_MSUN_SI, injEvent->mass2*LAL_MSUN_SI, injEvent->spin1x,
                                  injEvent->spin1y, injEvent->spin1z, injEvent->spin2x, injEvent->spin2y,
                                  injEvent->spin2z, f_min, fref, injEvent->distance*LAL_PC_SI*1.0e6,
                                  injEvent->inclination, lambda1, lambda2, waveFlags,
                                  //0.0, lambda1, lambda2, waveFlags, //Set inclination to 0?
                                  nonGRparams, amporder, order, approximant);
  //Sanity check
  if(!hp || !hc) {
    fprintf(stderr,"Error: XLALSimInspiralChooseWaveform() failed to produce waveform.\n");
    exit(-1);
  }
  //Free some memory
  XLALSimInspiralDestroyWaveformFlags(waveFlags);
  XLALSimInspiralDestroyTestGRParam(nonGRparams);
  //Resample h+ and hx and place them at the correct event GPS time
  XLALResampleREAL8TimeSeries(hp,deltaT);
  XLALResampleREAL8TimeSeries(hc,deltaT);
  XLALGPSAddGPS(&(hp->epoch), &(injEvent->geocent_end_time));
  XLALGPSAddGPS(&(hc->epoch), &(injEvent->geocent_end_time));
  //Loop over the IFOs to:
  // - place h+ and hx at the right IFO GPS time
  // - window h+ and hx
  // - generate the data in the frequency domain
  LALInferenceIFOData *dataPtr = runState->data;
  int ifo=0;
  while(dataPtr!=NULL)
  {
    //Workspace for the + and x polarizations of the actual injection
    REAL8TimeSeries *hplus=(REAL8TimeSeries *)XLALCreateREAL8TimeSeries("hplus",&dataPtr->timeData->epoch,0.0,deltaT,
                                                                         &lalStrainUnit,dataPtr->timeData->data->length);
    for(i=0;i<hplus->data->length;i++) hplus->data->data[i]=0.0;
    XLALSimAddInjectionREAL8TimeSeries(hplus, hp, NULL);
    REAL8TimeSeries *hcross=(REAL8TimeSeries *)XLALCreateREAL8TimeSeries("hcross",&dataPtr->timeData->epoch,0.0,deltaT,
                                                        &lalStrainUnit,dataPtr->timeData->data->length);
    for(i=0;i<hcross->data->length;i++) hcross->data->data[i]=0.0;
    XLALSimAddInjectionREAL8TimeSeries(hcross, hc, NULL);
    //The two polarizations are windowed (as for the injection), checked,
    //and copied from LAL data types into simple arrays
    ///WinNorm - Window normalisation is applied only in frequency domain)
    for(i=0;i<hplus->data->length;i++){
      hplus->data->data[i]*=dataPtr->window->data->data[i];
      if(isnan(hplus->data->data[i])) hplus->data->data[i]=0.0;
      timeDataPlus[ifo][i]=hplus->data->data[i];
    }
    for(i=0;i<hcross->data->length;i++){
      hcross->data->data[i]*=dataPtr->window->data->data[i];
      if(isnan(hcross->data->data[i])) hcross->data->data[i]=0.0;
      timeDataCross[ifo][i]=hcross->data->data[i];
    }
    //Obtain h+ and hx in the frequency domain
    COMPLEX16FrequencySeries *htildeplus=(COMPLEX16FrequencySeries *)XLALCreateCOMPLEX16FrequencySeries("htildeplus",&dataPtr->timeData->epoch,
                                                                                                        0.0,deltaF,
                                                                                                        &lalDimensionlessUnit,
                                                                                                        dataPtr->freqData->data->length);
    XLALREAL8TimeFreqFFT(htildeplus,hplus,dataPtr->timeToFreqFFTPlan);
    COMPLEX16FrequencySeries *htildecross=(COMPLEX16FrequencySeries *)XLALCreateCOMPLEX16FrequencySeries("htildecross",&dataPtr->timeData->epoch,
                                                                                                         0.0,deltaF,
                                                                                                         &lalDimensionlessUnit,
                                                                                                         dataPtr->freqData->data->length);
    XLALREAL8TimeFreqFFT(htildecross,hcross,dataPtr->timeToFreqFFTPlan);
    //Copy Fourier domain h+ and hx from LAL data types into simple arrays
    for(i=0; i<N/2; i++)
    {
      if(i<imin)
      {
        freqDataPlus[ifo][2*i]   = 0.0;
        freqDataPlus[ifo][2*i+1] = 0.0;
        freqDataCross[ifo][2*i]   = 0.0;
        freqDataCross[ifo][2*i+1] = 0.0;
      }
      else
      {
        freqDataPlus[ifo][2*i]   = 0.5*creal(htildeplus->data->data[i])/deltaT;
        freqDataPlus[ifo][2*i+1] = 0.5*cimag(htildeplus->data->data[i])/deltaT;
        freqDataCross[ifo][2*i]   = 0.5*creal(htildecross->data->data[i])/deltaT;
        freqDataCross[ifo][2*i+1] = 0.5*cimag(htildecross->data->data[i])/deltaT;
      }
    }
    //Destroy the workspaces for h+(t), hx(t), h+(f), and hx(f)
    XLALDestroyREAL8TimeSeries(hplus);
    XLALDestroyREAL8TimeSeries(hcross);
    XLALDestroyCOMPLEX16FrequencySeries(htildeplus);
    XLALDestroyCOMPLEX16FrequencySeries(htildecross);
    //Write the waveforms to file following the conventions of LALInferenceReadData
    //to produce the *_freqInjection.dat and *_timeInjection.dat files.
    FILE *filePlusCrossInj;
    char filename[100];
    //Time domain
    sprintf(filename, "%s/%s_timeInjectionPlusCross.dat", outdir, dataPtr->name);
    filePlusCrossInj=fopen(filename, "w");
    for(i=0;i<N;i++) fprintf(filePlusCrossInj, "%.6f\t%lg\t%lg\n", XLALGPSGetREAL8(&dataPtr->timeData->epoch)+dataPtr->timeData->deltaT*i,
                                               //(double)(i)/(double)(N)*data->Tobs-(data->Tobs-2.0),
                                               timeDataPlus[ifo][i], timeDataCross[ifo][i]);
    fclose(filePlusCrossInj);
    //Frequency domain
    sprintf(filename, "%s/%s_freqInjectionPlusCross.dat", outdir, dataPtr->name);
    filePlusCrossInj=fopen(filename, "w");
    for(i=0;i<N/2;i++) fprintf(filePlusCrossInj, "%lg\t%lg\t%lg\t%lg\t%lg\n", deltaF*i,
                                                 2.0*freqDataPlus[ifo][2*i]*deltaT, 2.0*freqDataPlus[ifo][2*i+1]*deltaT,
                                                 2.0*freqDataCross[ifo][2*i]*deltaT, 2.0*freqDataCross[ifo][2*i+1]*deltaT);
    fclose(filePlusCrossInj);
    //Move on to the next IFO
    dataPtr = dataPtr->next;
    ifo++;
  }
}


void Lorentzian(double *Slines, double Tobs, lorentzianParams *lines, int ii, int N)
{
  int i;
  double f2,f4;
  double deltf;
  double freq,fsq, x, z, deltafmax, spread;
  double amplitude;
  int istart, istop, imid, idelt;

  double A = lines->A[ii];
  double Q = lines->Q[ii];
  double f = lines->f[ii];


  // here we figure out how many frequency bins are needed for the line profile
  imid = (int)((f)*Tobs);
  spread = (1.0e-2*Q);

  if(spread < 50.0) spread = 50.0;  // maximum half-width is f_resonance/50
  deltafmax = f/spread;
  deltf = 4.0*deltafmax;
  idelt = (int)(deltf*Tobs)+1;


  istart = imid-idelt;
  istop = imid+idelt;
  if(istart < 0)  istart = 0;
  if(istop > N/2) istop = N/2;



  // add or remove the old line
  f2=f*f;
  f4=f2*f2;
  amplitude = A*f4/(f2*Q*Q);
  for(i=istart; i<istop; i++)
  {
    freq = (double)i/Tobs;
    fsq = freq*freq;
    x = fabs(f-freq);
    z = 1.0;
    if(x > deltafmax) z = exp(-(x-deltafmax)/deltafmax);

    Slines[i] += z*amplitude/(fsq*(fsq-f2)*(fsq-f2));
  }
  
}
















