/* ********************************************************************************** */
/*                                                                                    */
/*                                  LAL Dependencies                                  */
/*                                                                                    */
/* ********************************************************************************** */
#include <lal/LALInspiral.h>
#include <lal/LALCache.h>
#include <lal/LALFrStream.h>
#include <lal/TimeFreqFFT.h>
#include <lal/LALDetectors.h>
#include <lal/ResampleTimeSeries.h>
#include <lal/TimeSeries.h>
#include <lal/FrequencySeries.h>
#include <lal/Units.h>
#include <lal/Date.h>
#include <lal/StringInput.h>
#include <lal/VectorOps.h>
#include <lal/Random.h>
#include <lal/LALNoiseModels.h>
#include <lal/XLALError.h>
#include <lal/GenerateInspiral.h>
#include <lal/LIGOLwXMLRead.h>
#include <lal/LIGOLwXMLInspiralRead.h>

#include <lal/SeqFactories.h>
#include <lal/DetectorSite.h>
#include <lal/GenerateInspiral.h>
#include <lal/GeneratePPNInspiral.h>
#include <lal/SimulateCoherentGW.h>
#include <lal/Inject.h>
#include <lal/LIGOMetadataTables.h>
#include <lal/LIGOMetadataUtils.h>
#include <lal/LIGOMetadataInspiralUtils.h>
#include <lal/LIGOMetadataRingdownUtils.h>
#include <lal/LALInspiralBank.h>
#include <lal/FindChirp.h>
#include <lal/LALInspiralBank.h>
#include <lal/GenerateInspiral.h>
#include <lal/NRWaveInject.h>
#include <lal/GenerateInspRing.h>
#include <lal/LALErrno.h>

#include <lal/LALInference.h>
#include <lal/LALInferenceReadData.h>

/* ********************************************************************************** */
/*                                                                                    */
/*                                  Data structures                                   */
/*                                                                                    */
/* ********************************************************************************** */
struct Wavelet
{
  int size;
  int smax;
  int *index;

  //clustering prior
  int Ncluster;
  double cosy;
  double logp;

  double *templates;
  double **intParams;

  double **TFhistory;
  double **TFdensity;
  double TFdensityMax;

  char *name;
};

struct Model
{
  int size;

  double *Sn;
  double logL;
  double *detLogL;
  double *extParams;
  double **response;

  struct Wavelet *signal;
  struct Wavelet **glitch;

  struct Network *projection;

  struct FisherMatrix *fisher;

};

struct Network
{
  double **expPhase;
  double *deltaT;
  double *Fplus;
  double *Fcross;
};

struct Chain
{
  int NC;     // Number of parallel chains
  int NCmin;  // Minimum number of parallel chains
  int NCmax;  // Maximum number of parallel chains
  int burn;   // Number of burn in samples
  int mod0;   // Counter keeping track of noise model
  int mod1;   // Counter keeping track of glitch model
  int mod2;   // Counter keeping track of signal model
  int mod3;   // Counter keeping track of glitch+signal model

  int cycle;  // # of model updates per BurstRJMCMC iteration
  int count;  // # of BurstRJMCMC iterations
  int mcount; // # of MCMC attempts
  int scount; // # of RJMCMC attempts
  int xcount; // # of extrinsic attempts
  int macc;   // # of MCMC successess
  int sacc;   // # of RJMCMC successes
  int xacc;   // # of extrinsic successes

  int burnFlag;// flag telling proposals if we are in burn-in phase

  int *index; // keep track of which chain is at which temperature
  int *ptacc;
  int *ptprop;

  long seed;

  double modelRate;   //fraction of updates to signal model
  double rjmcmcRate;  //fraction of trans- vs. inter-dimensional moves
  double intPropRate; //fraction of prior vs. fisher intrinsic moves

  double *dT;
  double beta;
  double tempMin;
  double tempStep;
  double *temperature;
  double *avgLogLikelihood;
  double *varLogLikelihood;


  FILE *runFile;
  FILE **intChainFile;
  FILE **extChainFile;
  FILE **intWaveChainFile;
  FILE ***intGlitchChainFile;

};

struct Prior
{
  int smin;
  int smax;
  int *gmin;
  int *gmax;
  int omax;

  double Snmin;
  double Snmax;

  double TFV;
  double logTFV;

  double *bias;
  double **range;

  //cluster prior parameters
  double alpha;
  double beta;
  double gamma;

  //amplitude prior parameters
  double SNRpeak;

  //dimension prior parameters
  double Dtau;

  char path[100];
  
};

struct Data
{
  int N;
  int NI;
  int tsize;
  int fsize;
  int imin;
  int imax;
  int Dmax;
  int Dmin;

  int runPhase;
  int psdFitFlag;
  int noiseSimFlag;
  int bayesLineFlag;

  int fullOnlyFlag;
  int noiseOnlyFlag;
  int glitchOnlyFlag;
  int signalOnlyFlag;
  int cleanOnlyFlag;
  int skipCleanFlag;
  
  int glitchFlag;
  int signalFlag;
  int noiseFlag;

  int clusterPriorFlag;
  int amplitudePriorFlag;
  int signalAmplitudePriorFlag;
  int orientationProposalFlag;
  int clusterProposalFlag;
  int adaptTemperatureFlag;
  int bootstrapEvidenceFlag;
  int constantLogLFlag;
  int gnuplotFlag;
  int verboseFlag;

  double dt;
  double df;
  double fmin;
  double fmax;
  double Tobs;
  double Twin;
  double Qmax;
  double gmst;
  double logLc;
  double TwoDeltaToverN;

  double **r;
  double **s;
  double **rh;
  double *h;
  double **Snf;
  double **SnS;
  double *SnGeo;

  void (*signal_amplitude_proposal)(double *, double *, double, long *, double, double);
  void (*glitch_amplitude_proposal)(double *, double *, double, long *, double, double);

  double (*signal_amplitude_prior)   (double *, double *, double, double, double);
  double (*glitch_amplitude_prior)   (double *, double *, double, double, double);

  double (*intrinsic_likelihood)(int, int, int, struct Model*, struct Model*, struct Prior*, struct Chain*, struct Data*);
  double (*extrinsic_likelihood)(struct Network*, double *, double *, double *, double **, struct Data*, double, double);

  char runName[100];
  char outputDirectory[1000];
  char **channels;

  struct BayesLineParams **bayesline;

  ProcessParamsTable *commandLine;
  LALDetector **detector;
  LIGOTimeGPS epoch;
};


struct FisherMatrix
{
	double *sigma;
	double *evalue;
	double **evector;
	double **matrix;
	double *aamps;
	double **count;
};

/* ********************************************************************************** */
/*                                                                                    */
/*                                    MCMC tools                                      */
/*                                                                                    */
/* ********************************************************************************** */

void EvolveBayesLineParameters(struct Data *data, struct Model **model, struct Chain *chain, struct Prior *prior);
void EvolveIntrinsicParameters(struct Data *data, struct Prior *prior, struct Model **model, struct Chain *chain, long *seed, int ic);
void EvolveExtrinsicParameters(struct Data *data, struct Prior *prior, struct Model **model, struct Chain *chain, long *seed, int ic);

void extrinsic_fisher_update  (struct Data *data, struct Model *model);
void fisher_information_matrix(struct FisherMatrix *fisher, double *params, double *Sn, double *geo, double **glitch, struct Data *data);
void extrinsic_fisher_proposal(struct FisherMatrix *fisher, double *params, long *seed, int c, double *y);
void extrinsic_uniform_proposal(long *seed, double *y);
void sky_ring_proposal(double *x, long *seed, double *y, struct Data *data);
void uniform_orientation_proposal(double *x, long *seed, double *y, struct Data *data);
void burst2bayes_sky(double phi, double theta, struct Data *data, double *ra, double *dec);

void intrinsic_fisher_update  (double *params,  double *dparams, double *Snf, double Snx, double Tobs);
void intrinsic_fisher_proposal(double **range, double *paramsx, double *paramsy, double *Snf, double Snx, double *logpx, double *logpy, double scale, long *seed, double Tobs, int *test);

void adapt_temperature_ladder(struct Data *data, struct Model **model, struct Chain *chain, int NC);

void reset_params(struct Wavelet *wave_x, struct Wavelet *wave_y, double **range);

int checkrange(double *p, double **r);

int extrinsic_checkrange(double *p);


void RJMCMC_SG(int M, int N, int tsize, double **s, double ***TFdensity, double **Shf, long *seed, int NC, double Tobs, int NI, double fmin, double fmax);
void PTMCMC(struct Model **model, struct Chain *chain, long *seed, int NC);

void average_log_likelihood_via_direct_downsampling(int M, double **logLchain, struct Chain *chain);
void average_log_likelihood_via_recursion_downsampling(int M, double **logLchain, struct Chain *chain);
void average_log_likelihood_via_ACF(int M, double **logLchain, struct Chain *chain);

void ThermodynamicIntegration(double *temperature, double *avgLogLikelihood, double *varLogLikelihood, int NC, double *logEvidence, double *varLogEvidence);
void BootstrapThermodynamicIntegration(double *temperature, double **loglikelihood, int M, int NC, char modelname[], double *logEvidence, double *varLogEvidence);
/* ********************************************************************************** */
/*                                                                                    */
/*                            Wavelet basis functions                                 */
/*                                                                                    */
/* ********************************************************************************** */

void SineGaussian(double *hs, double *sigpar, int NMAX, int flag, double Tobs);
void SineGaussianBandwidth(double *sigpar, double *fmin, double *fmax);

double detector_snr(int imin, int imax, double *g, double *psd, double eta);
double network_snr(int imin, int imax, double **h, double **psd, double *eta, int NI);
double wavlet_proposal_density(int ifo, double f, double t, struct Data *data);
double wavelet_proximity_density(double f, double t, double **params, int *larray, int cnt, struct Prior *prior);

void wavelet_proximity(double *paramsy, double **params, int *larray, double **range, int cnt, long *seed);
void wavelet_sample(double *params, double **range, double **prop, int tsize, double Tobs, double pmax, long *seed);
double wavelet_proximity_prior(struct Wavelet *wave, struct Prior *prior);

void draw_wavelet_params(int ifo, double *params, struct Data *data, double **range, long *seed);
void draw_time_frequency(int ifo, int ii, struct Wavelet *wx, struct Wavelet *wy, struct Data *data, double **range, long *seed, double fr);

double signal_amplitude_prior (double *params, double *Snf, double Sa, double Tobs, double SNRpeak);
double glitch_amplitude_prior (double *params, double *Snf, double Sa, double Tobs, double SNRpeak);

void draw_signal_amplitude (double *params, double *Snf, double Sa, long *seed, double Tobs, double SNRpeak);
void draw_glitch_amplitude (double *params, double *Snf, double Sa, long *seed, double Tobs, double SNRpeak);

double TFprop(double f, double t, double **prop, int tsize, int fsize, double Tobs);

/* ********************************************************************************** */
/*                                                                                    */
/*                             Likelihood Functions                                   */
/*                                                                                    */
/* ********************************************************************************** */

void max_array_element(double *max, int *index, double *array, int n);
void phase_blind_time_shift(double *corr, double *corrf, double *data1, double *data2, double *psd, int n);

void computeProjectionCoeffs(struct Data *data, struct Network *projection, double *extParams);
void waveformProject(struct Data *data, struct Network *projection, double *extParams, double **response, double *geo, double fmin, double fmax);
void waveformTrivialProject(double **response, double *geo, int N, int NI);
void compute_reconstructed_plus_and_cross(struct Data *data, struct Network *projection, double *extParams, double **hrecPlus, double **hrecCross, double *geo, double fmin, double fmax);

double loglike(int imin, int imax, double *r, double Sn, double *Snf);
double loglike_maxNET(struct Data *data, struct Prior *prior, double **sNET, double **hNET, double *SnNET, double t0, double *dt, double *dphi, double *dA);

double Sum(int imin, int imax, double **a, double **b, double **Sn, int n, int NI);
double Sum_Extreme(double **a, double **b, double **Sn, int n, double *delt, double *pshift, double Tobs, int NI, int imin, int imax, double t0);
double Single_Extreme(double *a, double *b, double *Sn, int n, double *delt, double *pshift, double Tobs);

double EvaluateSearchLogLikelihood   (int typ, int ii, int det, struct Model *mx, struct Model *my, struct Prior *prior, struct Chain *chain, struct Data *data);
double EvaluateConstantLogLikelihood (int typ, int ii, int det, struct Model *mx, struct Model *my, struct Prior *prior, struct Chain *chain, struct Data *data);
double EvaluateMarkovianLogLikelihood(int typ, int ii, int det, struct Model *mx, struct Model *my, struct Prior *prior, struct Chain *chain, struct Data *data);

double EvaluateExtrinsicSearchLogLikelihood   (struct Network *projection, double *params, double *Sn, double *geo, double **g, struct Data *data, double fmin, double fmax);
double EvaluateExtrinsicConstantLogLikelihood (struct Network *projection, double *params, double *Sn, double *geo, double **g, struct Data *data, double fmin, double fmax);
double EvaluateExtrinsicMarkovianLogLikelihood(struct Network *projection, double *params, double *Sn, double *geo, double **g, struct Data *data, double fmin, double fmax);

/* ********************************************************************************** */
/*                                                                                    */
/*                                    Math tools                                      */
/*                                                                                    */
/* ********************************************************************************** */

double ran2(long *idum);
double gasdev2(long *idum);
double fourier_nwip(int imin, int imax, double *a, double *b, double *Sn);
double network_nwip(int imin, int imax, double **a, double **b, double **Sn, double *eta, int NI);

double gaussian_norm(double min, double max, double sigma);

void nrerror(char error_text[]);


/* ********************************************************************************** */
/*                                                                                    */
/*                             Instrument noise routines                              */
/*                                                                                    */
/* ********************************************************************************** */

void Shf_Geocenter(struct Data *data, struct Network *projection, double *SnGeo, double *params);

/* ********************************************************************************** */
/*																					                                          */
/*                                 Fourier Routines                                   */
/*																				                                         	  */
/* ********************************************************************************** */

void Fourier_Scaling(double *ft, int n, double Tobs);
void dfour1(double data[], unsigned long nn, int isign);
void drealft(double data[], unsigned long n, int isign);

/* ********************************************************************************** */
/*                                                                                    */
/*                        Data handling and injection routines                        */
/*                                                                                    */
/* ********************************************************************************** */

REAL8TimeSeries *readTseries(CHAR *cachefile, CHAR *channel, LIGOTimeGPS start, REAL8 length);

void InjectFromMDC(ProcessParamsTable *commandLine, LALInferenceIFOData *IFOdata, double *SNR);

void getChannels(ProcessParamsTable *commandLine, char **channel);

int outputFrameFile(ProcessParamsTable *commandLine, struct Data *data, struct Model *model);

void parse_command_line(struct Data *data, struct Chain *chain, struct Prior *prior, ProcessParamsTable *commandLine);

/* ********************************************************************************** */
/*                                                                                    */
/*                           Memory (de)allocation routines                           */
/*                                                                                    */
/* ********************************************************************************** */

void initialize_fisher(struct FisherMatrix *fisher);
void free_fisher(struct FisherMatrix *fisher);

void free_chain        (struct Chain *chain, int NI);
void reset_chain       (struct Chain *chain);
void allocate_chain    (struct Chain *chain, int NI);
void initialize_chain  (struct Chain *chain, int flag);

void reset_priors      (struct Data *data, struct Prior *prior);
void initialize_priors (struct Data *data, struct Prior *prior, int omax);

void initialize_data   (struct Data *data, double **s, int N, int tsize, double **psd, double Tobs, int NI, double fmin, double fmax);

void reset_model       (struct Data *data, struct Chain *chain, struct Prior *prior, struct Model **model);
void initialize_model  (struct Model *model, struct Data *data, struct Prior *prior, long *seed);
void free_model        (struct Model *model, struct Data *data, struct Prior *prior);
void copy_int_model    (struct Model *origin, struct Model *copy, int N, int NI, int det);
void copy_ext_model    (struct Model *origin, struct Model *copy, int N, int NI);

void reset_likelihood(struct Data *data);

void initialize_wavelet(struct Wavelet *wave, int N, int smax, int size, int fsize, int tsize);
void free_wavelet      (struct Wavelet *wave, int N, int smax, int fsize, int tsize);
void copy_wavelet      (struct Wavelet *origin, struct Wavelet *copy, int N);

void initialize_TFproposal(struct Data *data, struct Prior *prior, struct Wavelet *wave);

void initialize_network(struct Network *projection, int N, int NI);
void free_network(struct Network *projection, int N, int NI);

int checkfile(char *name);

int *ivector(long nl, long nh);
int **imatrix(long nrl, long nrh, long ncl, long nch);
int ***i3tensor(long nrl, long nrh, long ncl, long nch, long ndl, long ndh);

unsigned long *lvector(long nl, long nh);

double *dvector(long nl, long nh);
double **dmatrix(long nrl, long nrh, long ncl, long nch);
double ***d3tensor(long nrl, long nrh, long ncl, long nch, long ndl, long ndh);

void free_ivector(int *v, long nl, long nh);
void free_dvector(double *v, long nl, long nh);
void free_lvector(unsigned long *v, long nl, long nh);
void free_imatrix(int **m, long nrl, long nrh, long ncl, long nch);
void free_dmatrix(double **m, long nrl, long nrh, long ncl, long nch);
void free_i3tensor(int ***t, long nrl, long nrh, long ncl, long nch, long ndl, long ndh);
void free_d3tensor(double ***t, long nrl, long nrh, long ncl, long nch, long ndl, long ndh);

void KILL(char* Message);

/* ********************************************************************************** */
/*                                                                                    */
/*                The Sausage Factory:  Custom proposal distributions                 */
/*                                                                                    */
/* ********************************************************************************** */

void amoeba_orientation_proposal(double *x, long *seed, double *y, struct Data *data);
void amoeba(double **p, double *y, double *pass, int ndim, double ftol, double fmin, int *nfunk, struct Data *data);
double amotry(double **p, double *y, double *psum, double *pass, int ndim, int ihi, float fac, struct Data *data);
double mfunc(double *psum, double *pass, struct Data *data);
