#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <gsl/gsl_statistics.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_cblas.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_blas.h>


#include "Declarations.h"
#include "BayesLine.h"
#include "Constants.h"
#include "numrec.h"



/* ********************************************************************************** */
/*                                                                                    */
/*                                    MCMC tools                                      */
/*                                                                                    */
/* ********************************************************************************** */


void PTMCMC(struct Model **model, struct Chain *chain, long *seed, int NC)
{
  /*
   NC is passed to PTMCMC redundendtly because the main code
   can reduce the number of chains based on the temperature
   of the hottest chain
   */
	int a, b;
	int olda, oldb;

	double heat1, heat2;
	double logL1, logL2;
	double dlogL;
	double H;
	double alpha;
	double beta;

	//b = (int)(ran2(seed)*((double)(chain->NC-1)));
  for(b=NC-1; b>0; b--)
  {
    a = b - 1;

    olda = chain->index[a];
    oldb = chain->index[b];

    heat1 = chain->temperature[a];
    heat2 = chain->temperature[b];

    logL1 = model[olda]->logL;
    logL2 = model[oldb]->logL;

    //Hot chains jump more rarely
    if(ran2(seed)<1.0)///heat1)
    {
      dlogL = logL2 - logL1;
      H  = (heat2 - heat1)/(heat2*heat1);

      alpha = exp(dlogL*H);
      beta  = ran2(seed);

      chain->ptprop[a]++;

      if(alpha >= beta)
      {
        chain->ptacc[a]++;
        chain->index[a] = oldb;
        chain->index[b] = olda;
      }
    }
  }
}

void EvolveBayesLineParameters(struct Data *data, struct Model **model, struct Chain *chain, struct Prior *prior)
{
  int ic,ifo,i;

  int N  = data->N;
  int NI = data->NI;
  int NC = chain->NC;

  struct Wavelet *wave;

  //pick cold chain
  ic = chain->index[0];

  //update PSD for each interferometer
  for(ifo=0; ifo<NI; ifo++)
  {
    //copy over current multi-template & current residual
    for(i=0; i<N; i++)
    {
      data->r[ifo][i] = data->s[ifo][i];
      if(data->signalFlag) data->r[ifo][i] -= model[ic]->response[ifo][i];
      if(data->glitchFlag) data->r[ifo][i] -= model[ic]->glitch[ifo]->templates[i];
    }
    //re-run Markovian, full spectrum, full model part of BayesLine
    BayesLineRJMCMC(data->bayesline[ifo], data->r[ifo], data->Snf[ifo], data->SnS[ifo], N, chain->cycle);
  }

  Shf_Geocenter(data, model[ic]->projection, data->SnGeo, model[ic]->extParams);

  //recompute likelihoods & priors of each chain
  for(ic=0; ic<NC; ic++)
  {
    model[chain->index[ic]]->logL = 0.0;
    data->logLc = 0.0;

    for(ifo=0; ifo<NI; ifo++)
    {
      for(i=0; i< N; i++)
      {
        data->r[ifo][i] = data->s[ifo][i];
        if(data->signalFlag) data->r[ifo][i] -= model[chain->index[ic]]->response[ifo][i];
        if(data->glitchFlag) data->r[ifo][i] -= model[chain->index[ic]]->glitch[ifo]->templates[i];
      }

      model[chain->index[ic]]->detLogL[ifo] = loglike(data->imin, data->imax, data->r[ifo], model[chain->index[ic]]->Sn[ifo], data->Snf[ifo]);
      if(!data->constantLogLFlag)
        model[chain->index[ic]]->logL += model[chain->index[ic]]->detLogL[ifo];

      wave = model[chain->index[ic]]->glitch[ifo];
      wave->logp = 0.0;
      for(i=1; i<=wave->size; i++)
      {
        wave->logp += (data->glitch_amplitude_prior(wave->intParams[wave->index[i]],data->Snf[ifo], model[chain->index[ic]]->Sn[ifo], data->Tobs, prior->SNRpeak));
      }
    }

    wave = model[chain->index[ic]]->signal;
    wave->logp = 0.0;
    for(i=1; i<=wave->size; i++)
    {
      wave->logp += (data->glitch_amplitude_prior(wave->intParams[wave->index[i]],data->SnGeo, 1.0, data->Tobs, prior->SNRpeak));
    }

  }
}

void EvolveIntrinsicParameters(struct Data *data, struct Prior *prior, struct Model **model, struct Chain *chain, long *seed, int ic)
{
  int mc;
  int i,j,k;
  int ii=0,jj,kk;
  int ifo;
  int rj;
  int typ;
  int det;
  int test;
  int acc;
  double alpha;
  double logqx,logqy;
  double logH;
  double SNR;
  double Tobs = data->Tobs;

  //ratio of TF proposals to proximity proposals
  double TFtoProx = 0.5;

  double scale = 1./sqrt(6.);//??

  int burn = chain->burnFlag;

  //Unpack structures and use convenient (and familiar) names

  /* DATA */
  int N  = data->N;
  int NI = data->NI;

  /* PRIOR */
  int dmin = 0;
  int dmax = 1;

  double Snmin   = prior->Snmin;
  double Snmax   = prior->Snmax;
  double **range = prior->range;

  //FILE *chainFile = chain->intChainFile[ic];

  /* MODEL */
  struct Model *model_x = model[chain->index[ic]];

  model_x->size=model_x->signal->size;
  for(ifo=0; ifo<NI; ifo++) model_x->size+=model_x->glitch[ifo]->size;

  double *Snx = model_x->Sn;

  struct Wavelet *wave_x;
  int *larrayx;
  double **paramsx;

  /* CHAIN */
  int M = chain->cycle;//*(model_x->size+1);

  /* PROPOSED MODEL */
  struct Model *model_y = malloc(sizeof(struct Model));
  initialize_model(model_y, data, prior, seed);
  copy_ext_model(model_x, model_y, N, NI);
  copy_int_model(model_x, model_y, N, NI,-1);
  for(ifo=0; ifo<NI; ifo++) copy_int_model(model_x, model_y, N, NI, ifo);
  model_y->size=model_x->size;

  double *Sny = model_y->Sn;

  struct Wavelet *wave_y;
  int    *larrayy;
  double **paramsy;

  Shf_Geocenter(data, model_x->projection, data->SnGeo, model_x->extParams);

  acc=0;
  for(mc=1; mc<=M; mc++)
  {
    if(data->signalFlag) copy_int_model(model_x, model_y, N, NI,-1);
    if(data->glitchFlag) for(ifo=0; ifo<NI; ifo++) copy_int_model(model_x, model_y, N, NI, ifo);
    if(data->psdFitFlag) for(ifo=0; ifo<NI; ifo++) Sny[ifo] = Snx[ifo];
    else                 for(ifo=0; ifo<NI; ifo++) Sny[ifo] = 1.0;


    //decide if we're doing a noise or signal update
    if(!data->signalFlag)
      alpha = 1.0;
    else if(data->signalFlag && !data->glitchFlag )
      alpha = 0.0;
    else
      alpha = ran2(seed);

    //update geocenter model
    if(alpha<chain->modelRate)
    {
      det = -1;

      dmin = prior->smin;
      dmax = prior->smax;

      wave_x = model_x->signal;
      wave_y = model_y->signal;

    }
    //update noise model
    else
    {
      det = (int)floor(ran2(seed)*(float)NI);

      dmin = 0;//prior->gmin[det];
      dmax = prior->gmax[det];

      wave_x = model_x->glitch[det];
      wave_y = model_y->glitch[det];

    }

    larrayx = wave_x->index;
    larrayy = wave_y->index;

    paramsx = wave_x->intParams;
    paramsy = wave_y->intParams;

    acc=0; //acc gets set to 1 if model_y gets accepted (no need to copy at next iteration)

    //set prior and proposal terms to 0 (updated with += later on logp)
    wave_x->logp = 0.0;
    wave_y->logp = 0.0;
    logqx        = 0.0;
    logqy        = 0.0;

    rj   = 0;
    test = 0;


    /*******************************************************/
    /*                                                     */
    /*              TRANS DIMENSION RJMCMC MOVE            */
    /*                                                     */
    /*******************************************************/
    if(ran2(seed) < chain->rjmcmcRate)
    {
      rj = 1;

      if(ran2(seed) < 0.5)  // try and add a new wavelet
      {
        wave_y->size = wave_x->size+1;
        model_y->size = model_x->size+1;
        typ = 2;
      }
      else // try and remove wavelet
      {
        wave_y->size = wave_x->size-1;
        model_y->size = model_x->size-1;

        typ = 3;
      }

      if(wave_y->size < 0 || model_y->size < prior->smin || wave_y->size >= prior->smax) test = 1;

      if(test == 0)
      {
        // death move
        if(wave_y->size < wave_x->size)
        {
          i=(int)(ran2(seed)*(double)(wave_x->size))+1; // pick a term to try and kill
          k = 0;
          for(j=1; j<= wave_x->size; j++)
          {
            if(j != i)
            {
              k++;
              larrayy[k] = larrayx[j];
              jj = larrayx[j];
              for(kk=0; kk <= 4; kk++) paramsy[jj][kk] = paramsx[jj][kk];
            }
            if(j == i) ii = larrayx[j];  // take note of who's demise is being proposed
          }

          //reshift the indexing array for occupied wavelets
          if(wave_y->size == 0) for(j=1; j<=dmax; j++) larrayy[j] = j;

          // TF density for proposal
          if(data->clusterProposalFlag)
          {
            logqy +=  0.0;
            logqx += log( TFtoProx*TFprop(paramsx[ii][1], paramsx[ii][0], wave_x->TFdensity, data->tsize, data->fsize, data->Tobs) + (1.-TFtoProx)*wavelet_proximity_density(paramsx[ii][1], paramsx[ii][0], paramsy, larrayy, wave_y->size, prior) );

          }
          // Uniform in TF proposal
          else
          {
            logqx += -prior->logTFV;
            logqy += 0.0;
          }

          if(data->amplitudePriorFlag)
          {
            if(det==-1) logqx += ( data->signal_amplitude_prior(paramsx[ii], data->SnGeo, 1.0, data->Tobs, prior->SNRpeak) );
            else        logqx += ( data->glitch_amplitude_prior(paramsx[ii], data->Snf[det], Snx[det], data->Tobs, prior->SNRpeak) );
          }
        }//end death move

        // birth move
        if(wave_y->size > wave_x->size)
        {

          // find a label that isn't in use
          ii = 0;
          do
          {
            ii++;
            k = 0;
            for(j=1; j<= wave_x->size; j++)
            {
              if(ii==larrayx[j]) k = 1;
            }
          } while(k == 1);

          //reshift the indexing array for occupied wavelets
          larrayy[wave_y->size] = ii;

          // Draw all wavelet parameters from prior
          draw_wavelet_params(det, paramsy[ii], data, range, seed);

          // TF density for proposal
          if(data->clusterProposalFlag)
          {
            // Draw new time and frequency from TFV density proposal
            draw_time_frequency(det, ii, wave_x, wave_y, data, range, seed, TFtoProx);

            logqy += log( TFtoProx*TFprop(paramsy[ii][1], paramsy[ii][0], wave_x->TFdensity, data->tsize, data->fsize, data->Tobs) + (1.-TFtoProx)*wavelet_proximity_density(paramsy[ii][1], paramsy[ii][0], paramsx, larrayx, wave_x->size, prior) );
            logqx +=  0.0;
          }
          // Uniform in TF proposal
          else
          {
            logqy += -prior->logTFV;
            logqx += 0.0;
          }

          if(data->amplitudePriorFlag)
          {
            if(det==-1)
            {
              data->signal_amplitude_proposal(paramsy[ii], data->SnGeo, 1.0, seed, data->Tobs, prior->SNRpeak);
              logqy += ( data->signal_amplitude_prior(paramsy[ii], data->SnGeo, 1.0, data->Tobs, prior->SNRpeak) );
            }
            else
            {
              data->glitch_amplitude_proposal(paramsy[ii], data->Snf[det], Sny[det], seed, data->Tobs, prior->SNRpeak);
              logqy += (data->glitch_amplitude_prior(paramsy[ii], data->Snf[det], Sny[det], data->Tobs, prior->SNRpeak) );
            }
          }


        }//end birth move
      }//end test condition
    }//end trans-dimensional move

    /*******************************************************/
    /*                                                     */
    /*               FIXED DIMENSION MCMC MOVE             */
    /*                                                     */
    /*******************************************************/
    else
    {
      typ = 1;

      for(ifo=0; ifo<NI; ifo++)
      {
        if(data->psdFitFlag) Sny[ifo] = Snx[ifo]+gasdev2(seed)/sqrt(0.5*(double)N);
        if(Sny[ifo] > Snmax || Sny[ifo] < Snmin) test = 1;
      }

      //Now update wavelet models if they are enabled
      if( (data->signalFlag || data->glitchFlag) && wave_x->size>0 )
      {
        i  = (int)(ran2(seed)*(double)(wave_x->size))+1; // pick a term to update
        ii = larrayx[i];

        alpha = ran2(seed);

        // replace existing wavelet (birth/death move)
        if(alpha > chain->intPropRate/chain->temperature[ic])
        {

          // Draw all wavelet parameters from prior
          draw_wavelet_params(det, paramsy[ii], data, range, seed);

          // TF density for proposal
          if(data->clusterProposalFlag)
          {

            //TODO:  Why did clusterProposal break detailed balance for birth/death moves but not RJ moves?

            /* Draw new time and frequency from TFV density + proximity proposal */
            draw_time_frequency(det, ii, wave_x, wave_y, data, range, seed, TFtoProx);
            logqx += log( TFtoProx*TFprop(paramsx[ii][1], paramsx[ii][0], wave_x->TFdensity, data->tsize, data->fsize, data->Tobs) + (1.-TFtoProx)*wavelet_proximity_density(paramsx[ii][1], paramsx[ii][0], paramsy, larrayy, wave_y->size, prior) );
            logqy += log( TFtoProx*TFprop(paramsy[ii][1], paramsy[ii][0], wave_x->TFdensity, data->tsize, data->fsize, data->Tobs) + (1.-TFtoProx)*wavelet_proximity_density(paramsy[ii][1], paramsy[ii][0], paramsx, larrayx, wave_x->size, prior) );


            /* Draw new time and frequency from TFV proximity proposal
            draw_time_frequency(det, ii, wave_x, wave_y, data, range, seed, 0);
            logqx += log( wavelet_proximity_density(paramsx[ii][1], paramsx[ii][0], paramsy, larrayy, wave_y->size, prior->TFV) );
            logqy += log( wavelet_proximity_density(paramsy[ii][1], paramsy[ii][0], paramsx, larrayx, wave_x->size, prior->TFV) );
            */

            /* Draw new time and frequency from TFV density proposal
            draw_time_frequency(det, ii, wave_x, wave_y, data, range, seed, 1);
            logqx += log( TFprop(paramsx[ii][1], paramsx[ii][0], wave_x->TFdensity, data->tsize, data->fsize, data->Tobs) );
            logqy += log( TFprop(paramsy[ii][1], paramsy[ii][0], wave_x->TFdensity, data->tsize, data->fsize, data->Tobs) );
            */
            
            //logqy += -prior->logTFV;
            //logqx += -prior->logTFV;

          }

          // Uniform in TF proposal
          else
          {
            logqy += -prior->logTFV;
            logqx += -prior->logTFV;
          }

          if(data->amplitudePriorFlag)
          {
            if(det==-1)
            {
              data->signal_amplitude_proposal(paramsy[ii], data->SnGeo, 1.0, seed, data->Tobs, prior->SNRpeak);
              logqy += (data->signal_amplitude_prior(paramsy[ii], data->SnGeo, 1.0, data->Tobs, prior->SNRpeak));
              logqx += (data->signal_amplitude_prior(paramsx[ii], data->SnGeo, 1.0, data->Tobs, prior->SNRpeak));
            }
            else
            {
              data->glitch_amplitude_proposal(paramsy[ii], data->Snf[det], Sny[det], seed, data->Tobs, prior->SNRpeak);
              logqy += (data->glitch_amplitude_prior(paramsy[ii], data->Snf[det], Sny[det], data->Tobs, prior->SNRpeak));
              logqx += (data->glitch_amplitude_prior(paramsx[ii], data->Snf[det], Snx[det], data->Tobs, prior->SNRpeak));
            }
          }
        }

        // perturb existing wavelet & PSD model
        else
        {
          /*
           proposal distribution uses the Fisher at x to propose y,
           then computes the Fisher at y to compute the proposal
           densities to maintain detailed balance
           */
          if(det==-1) intrinsic_fisher_proposal(range, paramsx[ii], paramsy[ii], data->SnGeo,    1.0, &logqx, &logqy, scale, seed, Tobs, &test);
          else        intrinsic_fisher_proposal(range, paramsx[ii], paramsy[ii], data->Snf[det], 1.0, &logqx, &logqy, scale, seed, Tobs, &test);

        }//end perturb step
      }//end !noiseOnlyFlag check

    }//end model update


    // If new parameters are within prior compute priors & likelihoods
    if(test == 0)
    {
      // compute the likelihood using appropriate function
      model_y->logL = data->intrinsic_likelihood(typ, ii, det, model_x, model_y, prior, chain, data);

      if( (data->signalFlag || data->glitchFlag) && model_y->logL > -1.0e60)
      {
        // clustering prior
        if(data->clusterPriorFlag)
        {
          wave_y->logp += wavelet_proximity_prior(wave_y, prior);
          wave_x->logp += wavelet_proximity_prior(wave_x, prior);
        }

        // uniform prior
        else
        {
          wave_x->logp += -(double)(wave_x->size)*prior->logTFV;
          wave_y->logp += -(double)(wave_y->size)*prior->logTFV;
        }

        //amplitude prior
        if(data->amplitudePriorFlag)
        {
          for(i=1; i<=wave_y->size; i++)
          {
            if(det==-1)
              wave_y->logp += (data->signal_amplitude_prior(paramsy[larrayy[i]],data->SnGeo, 1.0, data->Tobs, prior->SNRpeak));
            else
              wave_y->logp += (data->glitch_amplitude_prior(paramsy[larrayy[i]],data->Snf[det], Sny[det], data->Tobs, prior->SNRpeak));
          }
          for(i=1; i<=wave_x->size; i++)
          {
            if(det==-1)
              wave_x->logp += (data->signal_amplitude_prior(paramsx[larrayx[i]],data->SnGeo, 1.0, data->Tobs, prior->SNRpeak));
            else
              wave_x->logp += (data->glitch_amplitude_prior(paramsx[larrayx[i]],data->Snf[det], Snx[det], data->Tobs, prior->SNRpeak));
          }
        }

        //dimension prior
        wave_x->logp -= wave_x->size/prior->Dtau;
        wave_y->logp -= wave_y->size/prior->Dtau;

      }

      // and finally, the Hastings ratio
      //if(data->psdFitFlag)
      //printf("logLy=%g logLx=%g py=%g px=%g qy=%g qx=%g\n",model_y->logL,model_x->logL,wave_y->logp,wave_x->logp,logqy,logqx);
      logH = (model_y->logL - model_x->logL)*chain->beta + wave_y->logp - wave_x->logp - logqy + logqx;

      /*
       if(model_x->size == model_y->size)
       {
        printf("%i: %i->%i logLy=%g logLx=%g py=%g px=%g qy=%g qx=%g, dx=%i,dy=%i,num=%g, den=%g, logH=%g\n",typ,model_x->size,model_y->size,model_y->logL,model_x->logL,wave_y->logp,wave_x->logp,logqy,logqx,wave_x->size,wave_y->size,wave_y->logp+ logqx,- wave_x->logp - logqy, logH);
       }
       */

      /*
       with the maximization on the sytsem effectively has fewer degrees of freedom per SG
       compensate with a prior penalty
       */
      if(burn)
      {
        if(det==-1) logH += 0.0*(double)(wave_x->size-wave_y->size);//*chain->beta;
        else        logH += 0.0*(double)(wave_x->size-wave_y->size);//*chain->beta;
      }

    }// end prior test condition
    else logH = -1.0e60; // Rejection sample at prior boundaries


    // Update proposal counters for acceptance ratios
    if(chain->index[ic]==0)
    {
      if(rj) chain->scount++;
      else
      {
        if(wave_x->size>0)chain->mcount++;
      }
    }

    /*******************************************************/
    /*                                                     */
    /*      Metropolis-Hastings acceptance/rejection       */
    /*                                                     */
    /*******************************************************/

    // Update current position (model_x) depending on acceptance/rejection

    alpha = log(ran2(seed));

    if(logH > alpha)
    {
      acc = 1;

      copy_int_model(model_y, model_x, N, data->NI, det);

      if(chain->index[ic]==0)
      {
        if(rj) chain->sacc++;
        else
        {
          if(wave_x->size>0)chain->macc++;
        }
      }
      wave_x->logp = wave_y->logp;
    }


    //print chain status to file
    if(mc%10 == 0 && (ic==0 || data->verboseFlag))// && (data->signalFlag || data->glitchFlag || data->noiseOnlyFlag))
    {
      //TODO: Output physical parameters (t,phase,amp have int. and ext. components)
      SNR = network_snr( (int)floor(data->fmin*data->Tobs), (int)floor(data->fmax*data->Tobs), model_x->response, data->Snf, model_x->Sn, data->NI );

      fprintf(chain->intChainFile[ic],"%d %f ", mc/10, model_x->logL-data->logLc);
      fprintf(chain->intChainFile[ic],"%d ", model_x->signal->size);
      for(ifo=0; ifo<NI; ifo++) fprintf(chain->intChainFile[ic],"%d ", model_x->glitch[ifo]->size);
      if(data->clusterPriorFlag)
      {
        fprintf(chain->intChainFile[ic],"%d ", model_x->signal->Ncluster);
        for(ifo=0; ifo<NI; ifo++) fprintf(chain->intChainFile[ic],"%d ", model_x->glitch[ifo]->Ncluster);
      }
      for(ifo=0; ifo<NI; ifo++) fprintf(chain->intChainFile[ic], "%f ",Snx[ifo]);
      fprintf(chain->intChainFile[ic],"%g ",wave_x->logp);


      //print intrinsic parameters of signal model
      if(data->signalFlag)
      {
        fprintf(chain->intWaveChainFile[ic],"%i ",model_x->signal->size);
        for(k=0; k<NE; k++) fprintf(chain->intWaveChainFile[ic],"%.12g ", model_x->extParams[k]);
        for(j=1;j<=model_x->signal->size; j++)  // print out first few sine-gaussians
        {
          i = model_x->signal->index[j];

          //print all parameters
          for(k=0; k<=4; k++) fprintf(chain->intWaveChainFile[ic],"%.12g ", model_x->signal->intParams[i][k]);
        }
        fprintf(chain->intWaveChainFile[ic],"\n");
      }

      //print intrinsic parameters of glitch model
      if(data->glitchFlag)
      {
        for(ifo=0; ifo<NI; ifo++)
        {
          fprintf(chain->intGlitchChainFile[ifo][ic],"%i ",model_x->glitch[ifo]->size);
          for(j=1;j<=model_x->glitch[ifo]->size; j++)  // print out first few sine-gaussians
          {
            i = model_x->glitch[ifo]->index[j];

            //print all parameters
            for(k=0; k<=4; k++) fprintf(chain->intGlitchChainFile[ifo][ic],"%.12g ", model_x->glitch[ifo]->intParams[i][k]);
          }
          fprintf(chain->intGlitchChainFile[ifo][ic],"\n");
        }
      }

      double gmst = XLALGreenwichMeanSiderealTime(&data->epoch);
      //remap gmst back to [0:2pi]
      double intpart;
      double decpart;

      if(data->signalFlag)
      {
        gmst /= LAL_TWOPI;
        intpart = (int)( gmst );
        decpart = gmst - (double)intpart;
        gmst = decpart*LAL_TWOPI;

        fprintf(chain->intChainFile[ic],"%e %e %e ",gmst-model_x->extParams[0], model_x->extParams[1], model_x->extParams[5]);
      }
      fprintf(chain->intChainFile[ic],"\n");
      fflush(chain->intChainFile[ic]);
    }
  }

  // re-do residual calculation occasionally to counter any round-off error from delta updates
  for(i=0; i< N; i++)
  {
    model_x->signal->templates[i] = 0.0;
    for(ifo=0; ifo<NI; ifo++) model_x->glitch[ifo]->templates[i] = 0.0;
  }

  if(data->signalFlag)
  {
    for(j=1; j<=model_x->signal->size; j++)
    {
      i = model_x->signal->index[j];
      SineGaussian(model_x->signal->templates, model_x->signal->intParams[i], N, 1, data->Tobs);
    }
  }

  if(data->glitchFlag)
  {
    for(ifo=0; ifo<NI; ifo++)
    {
      for(j=1; j<=model_x->glitch[ifo]->size; j++)
      {
        i = model_x->glitch[ifo]->index[j];
        SineGaussian(model_x->glitch[ifo]->templates, model_x->glitch[ifo]->intParams[i], N, 1, data->Tobs);
      }
    }
  }

  //recompute likelihood
  for(ifo=0; ifo<NI; ifo++)
  {
    for(i=0; i<data->N; i++)
    {
      data->r[ifo][i] = data->s[ifo][i];
      if(data->glitchFlag) data->r[ifo][i] -= model_x->glitch[ifo]->templates[i];
      if(data->signalFlag) data->r[ifo][i] -= model_x->response[ifo][i];
    }
  }

  model_x->logL = 0.0;
  if(!data->constantLogLFlag)
  {
    for(ifo=0; ifo<NI; ifo++)
    {
      model_x->detLogL[ifo] = loglike(data->imin, data->imax, data->r[ifo], model_x->Sn[ifo], data->Snf[ifo]);
      model_x->logL += model_x->detLogL[ifo];
    }
  }

  free_model(model_y, data, prior);
  free(model_y);

}

void EvolveExtrinsicParameters(struct Data *data, struct Prior *prior, struct Model **model, struct Chain *chain, long *seed, int ic)
{
	int i, j, ifo, c, mc;
  int dim;
  int test;
  double logLx, logLy, logH;
  double alpha;
  double **glitch;
  double *geo;
  double *Sn;
  double *SnGeox;
  double *SnGeoy;
  double *paramsx, *paramsy;
  double **intParams;
  double logpy,logpx;

  //Unpack structures and use convenient (and familiar) names
  int NI = data->NI;

  /* CHAIN */
  int M = chain->cycle;
  //FILE *chainFile = chain->extChainFile[ic];

  /* MODEL */
  // extrinsic parameters
  struct Model *model_x = model[chain->index[ic]];
  struct Wavelet *smodel = model_x->signal;
  Sn      = model_x->Sn;
  geo     = smodel->templates;
  paramsx = model_x->extParams;
  SnGeox  = data->SnGeo;
  paramsy = dvector(0,NE);
  SnGeoy  = dvector(0,data->N/2);

  dim = smodel->size;
  intParams = dmatrix(1,dim,0,4);

  glitch = malloc(NI*sizeof(double*));
  for(ifo=0; ifo<NI; ifo++) glitch[ifo] = model_x->glitch[ifo]->templates;

  //Initialize parameter vectors & likelihood for extrinsic parameter MCMC
  for(i=0; i<NE; i++)  paramsy[i] = paramsx[i];

  //Find minimum and maximum frequencies of signal model and only compute logL over that range
  double fmin = data->fmax;
  double fmax = data->fmin;
  double fi,ff;
  for(i=1; i<=smodel->size; i++)
  {
    SineGaussianBandwidth(smodel->intParams[smodel->index[i]],&fi,&ff);
    if(fi<fmin) fmin=fi;
    if(ff>fmax) fmax=ff;

    //Make sure frequency is in range
    if(fmin < data->fmin)
      fmin = data->fmin;
    if(fmax > data->fmax)
      fmax = data->fmax;

  }

  logLx = data->extrinsic_likelihood(model_x->projection, paramsx, Sn, geo, glitch, data, fmin, fmax);

  double dLogL = model_x->logL - logLx;

  //Compute Fisher Matrix for current sky location
  struct FisherMatrix *fisher = model_x->fisher;

  double draw;

  logpx=0.0;
  if(data->amplitudePriorFlag)
  {
    //model_x->projection has be updated by data->extrinsic_likelihood, so it stores F+ and Fx for params_y.
    Shf_Geocenter(data, model_x->projection, SnGeox, paramsx);
    for(i=1; i<=dim; i++) logpx += (data->signal_amplitude_prior(smodel->intParams[smodel->index[i]],SnGeox, 1.0, data->Tobs, prior->SNRpeak));
  }

  for(mc=0; mc < M; mc++)
	{
    test=0;
    for(i=1;i<=dim;i++)for(j=0; j<5; j++) intParams[i][j] = smodel->intParams[smodel->index[i]][j];
    for(i=0; i<data->N/2; i++) SnGeoy[i]=SnGeox[i];

    if(ran2(seed)< 0.5)    //0.5 || chain->temperature[ic]>25.0)
    {
      draw = ran2(seed);
      /* Uniform Draw */
      if(draw<0.1 || NI==1)extrinsic_uniform_proposal(seed,paramsy);

      /* Psi/Phi draw */
      //else if(draw < 0.35)psi_phi_proposal(paramsx,seed,paramsy,data);

      /* Sky Ring Draw */
      else sky_ring_proposal(paramsx,seed,paramsy,data);
    }
    else
    {
      /* Fisher Matrix Proposal */
      //propose in one eigendirection at a time
      c = (int)floor(ran2(seed) * NE);
      extrinsic_fisher_proposal(fisher, paramsx, seed, c, paramsy);
    }

    //Rescale intrinsic amplitudes with extrinsic dA
    for(i=1; i<=dim; i++)
    {
      intParams[i][3] *= paramsy[4];
      test += checkrange(intParams[i],prior->range);
    }

    if(test==0)logLy = data->extrinsic_likelihood(model_x->projection, paramsy, Sn, geo, glitch, data, fmin, fmax);
    else       logLy = -1.0e60;


    logH  = (logLy - logLx)*chain->beta;

    //amplitude prior
    logpy=0.0;
    if(data->amplitudePriorFlag && test==0)
    {
      //model_x->projection has be updated by data->extrinsic_likelihood, so it stores F+ and Fx for params_y.
      Shf_Geocenter(data, model_x->projection, SnGeoy, paramsy);
      for(i=1; i<=dim; i++) logpy += (data->signal_amplitude_prior(intParams[i],SnGeoy, 1.0,data->Tobs, prior->SNRpeak));
      logH += logpy - logpx;
    }

    alpha = log(ran2(seed));

    if(logH >= alpha)
    {
      chain->xacc++;
      logLx = logLy;
      for(i=0; i<data->N; i++) geo[i] *= paramsy[4]/paramsx[4];
      for(i=0; i< NE; i++) paramsx[i] = paramsy[i];
      for(i=0; i<data->N/2; i++) SnGeox[i]=SnGeoy[i];
      for(i=1; i<=dim; i++) smodel->intParams[smodel->index[i]][3]=intParams[i][3];
      paramsx[4]=paramsy[4]=1.0;
      logpx = logpy;
    }


    if(mc%10 == 0 && (ic==0 || data->verboseFlag) )
    {
      fprintf(chain->extChainFile[ic],"%d %f ", mc/10, dLogL+logLx-data->logLc);
      for(i=0; i< NE; i++) fprintf(chain->extChainFile[ic],"%e ", paramsx[i]);
      fprintf(chain->extChainFile[ic],"%g \n",logpx);
      fflush(chain->extChainFile[ic]);
    }
    chain->xcount++;
  }


  //Now update the full model with current extrinsic parameters
  logLx = 0.0;
  computeProjectionCoeffs(data, model_x->projection, model_x->extParams);
  waveformProject(data, model_x->projection, model_x->extParams, model_x->response, model_x->signal->templates, data->fmin, data->fmax);
  Shf_Geocenter(data, model_x->projection, data->SnGeo, model_x->extParams);

  for(ifo=0; ifo<NI; ifo++)
  {
    for(i=0; i<data->N; i++)
    {
      data->r[ifo][i] = data->s[ifo][i];
      if(data->signalFlag) data->r[ifo][i] -= model_x->response[ifo][i];
      if(data->glitchFlag) data->r[ifo][i] -= model_x->glitch[ifo]->templates[i];
    }
    if(data->constantLogLFlag) model_x->detLogL[ifo] = 0.0;
    else model_x->detLogL[ifo] = loglike(data->imin, data->imax, data->r[ifo], model_x->Sn[ifo], data->Snf[ifo]);
    logLx += model_x->detLogL[ifo];
  }
  model_x->logL = logLx;

  free(glitch);
  free_dvector(paramsy,0,NE-1);
  free_dvector(SnGeoy,0,data->N/2);
  free_dmatrix(intParams,1,dim,0,4);

}

void fisher_information_matrix(struct FisherMatrix *fisher, double *params, double *Sn, double *geo, double **glitch, struct Data *data)
{
	int i, j, k;
  int NI,N;

	double *paramsP, *paramsM;
	double logLP, logLM;
	double **hP, **hM, **hPP, **hMM;
  double ***hdiv;
	double *epsilon;
	double **s;
	double stab;
  double Tobs;
	int NV;

  struct Network *projectionP = malloc(sizeof(struct Network));
  struct Network *projectionM = malloc(sizeof(struct Network));

  NI   = data->NI;
	N    = data->N;
  Tobs = data->Tobs;

  initialize_network(projectionP,N, NI);
  initialize_network(projectionM,N, NI);

  stab = 10.0;

	// Plus and minus parameters:
	paramsP = dvector(0,NE);
	paramsM = dvector(0,NE);

	NV = NE;


	// Plus and minus templates for each detector:
  s     = dmatrix(0,NI-1,0,N-1);
	hP		= dmatrix(0,NI-1,0,N-1);
	hM		= dmatrix(0,NI-1,0,N-1);
	hPP		= dmatrix(0,NI-1,0,N-1);
	hMM		= dmatrix(0,NI-1,0,N-1);

  hdiv = d3tensor(0,NE-1,0,NI-1,0,N-1);

	epsilon = dvector(0,NE-1);

  // Compute SNR of signal model
  computeProjectionCoeffs(data, projectionP, params);
  waveformProject(data, projectionP, params, hP, geo, data->fmin, data->fmax);

  //store data
  for(i=0;i<NI;i++)
  {
    for(j=0;j<N;j++)
    {
      s[i][j]=data->s[i][j];
      data->s[i][j]=hP[i][j];
    }
  }

  double range;
	for(i = 0; i < NE; i++)
	{
    epsilon[i] = 1.0e-3;//fabs(params[i]);
    switch(i)
    {
      case 0:
        range = TPI;
        break;
      case 1:
        range = 2.0;
        break;
      case 2:
        range = TPI;
        break;
      case 3:
        range = 0.99*2;
        break;
      case 4:
        range = 10.0;
        break;
      case 5:
        range = 2.0;
        break;
      case 6:
        range = TPI;
        break;
    }

    epsilon[i] = 1.0e-5*range;

    for(j = 0; j < NE; j++)
    {
      paramsP[j] = params[j];
      paramsM[j] = params[j];
    }

    paramsP[i] += epsilon[i];
    paramsM[i] -= epsilon[i];

    computeProjectionCoeffs(data, projectionP, paramsP);
    computeProjectionCoeffs(data, projectionM, paramsM);

    logLP = EvaluateExtrinsicMarkovianLogLikelihood(projectionP, paramsP, Sn, geo, glitch, data, data->fmin, data->fmax);
    logLM = EvaluateExtrinsicMarkovianLogLikelihood(projectionM, paramsM, Sn, geo, glitch, data, data->fmin, data->fmax);

    epsilon[i] = 0.1*1.0/sqrt(-(logLM+logLP)/(epsilon[i]*epsilon[i]));
    if(epsilon[i]!=epsilon[i])epsilon[i]=1.0e-5;
	}

  //restore data
  for(i=0;i<NI;i++)  for(j=0;j<N;j++) data->s[i][j]=s[i][j];

	/* assumes all the parameters are log or angles */
	for(i=0; i<NE; i++)
	{
    for(j=0; j<NE; j++)
    {
      paramsP[j] = params[j];
      paramsM[j] = params[j];
    }


    paramsP[i] += epsilon[i];
    paramsM[i] -= epsilon[i];

    computeProjectionCoeffs(data, projectionP, paramsP);
    computeProjectionCoeffs(data, projectionM, paramsM);

    waveformProject(data, projectionP, paramsP, hP, geo, data->fmin, data->fmax);
    waveformProject(data, projectionM, paramsM, hM, geo, data->fmin, data->fmax);

    paramsP[i] += epsilon[i];
    paramsM[i] -= epsilon[i];

    computeProjectionCoeffs(data, projectionP, paramsP);
    computeProjectionCoeffs(data, projectionM, paramsM);

    waveformProject(data, projectionP, paramsP, hPP, geo, data->fmin, data->fmax);
    waveformProject(data, projectionM, paramsM, hMM, geo, data->fmin, data->fmax);


    //Central differencing
    for(k=0; k<NI; k++)
    {
      for(j=0; j<N; j++)
      {
        hdiv[i][k][j] = (hMM[k][j] + 8.0*hP[k][j] - 8.0*hM[k][j] - hPP[k][j])/(12.0*epsilon[i]);
      }
    }
	}

  for(i=0; i<NE; i++) for(j=0; j<NE; j++) fisher->matrix[i][j] = network_nwip(data->imin,data->imax,hdiv[i], hdiv[j], data->Snf, Sn, NI);

  // stabilizers
  for(i=0; i<NE; i++)  fisher->matrix[i][i] += stab;


	free_dvector(paramsP,0,NE-1);
	free_dvector(paramsM,0,NE-1);

  free_d3tensor(hdiv,0,NE-1,0,NI-1,0,N-1);

  free_dmatrix(s,0,NI-1,0,N-1);
	free_dmatrix(hP,0,NI-1,0,N-1);
	free_dmatrix(hM,0,NI-1,0,N-1);
  free_dmatrix(hPP,0,NI-1,0,N-1);
  free_dmatrix(hMM,0,NI-1,0,N-1);

  free_dvector(epsilon,0,NE-1);

  free_network(projectionM,N,NI);
  free_network(projectionP,N,NI);
  free(projectionP);
  free(projectionM);

}

void extrinsic_fisher_update(struct Data *data, struct Model *model)
{
	int i, j;

  double *params = model->extParams;
  double *geo    = model->signal->templates;
  double *Sn     = model->Sn;

  double **glitch = NULL;
  glitch = malloc(data->NI*sizeof(double*));
  for(i=0; i<data->NI; i++) glitch[i]=model->glitch[i]->templates;

  struct FisherMatrix *fisher = model->fisher;

	//calculate elements of Fisher
  fisher_information_matrix(fisher, params, Sn, geo, glitch, data);

  // Find eigenvectors and eigenvalues
  gsl_matrix *GSLfisher = gsl_matrix_alloc(NE,NE);
  gsl_matrix *GSLcovari = gsl_matrix_alloc(NE,NE);
  gsl_matrix *GSLevectr = gsl_matrix_alloc(NE,NE);
  gsl_vector *GSLevalue = gsl_vector_alloc(NE);

  for(i=0; i<NE; i++)for(j=0; j<NE; j++) gsl_matrix_set(GSLfisher,i,j,fisher->matrix[i][j]);

  /*
   // Print inverted matrix
   double val;
   printf("Fisher Matrix =\n");
   for(i = 0; i < NE; i++) {
   for(j = 0; j < NE; j++) {
   if(i>=j)val=gsl_matrix_get (GSLfisher, i, j);
   else    val=gsl_matrix_get (GSLfisher, j, i);
   if(val>=0)printf("+");
   printf ("%.1e ",val);
   }
   printf("\n");
   }
   printf ("--------------------\n");
   */

  // sort and put them into evec
  gsl_eigen_symmv_workspace * workspace = gsl_eigen_symmv_alloc (NE);
  gsl_permutation * permutation = gsl_permutation_alloc(NE);
  gsl_eigen_symmv (GSLfisher, GSLevalue, GSLevectr, workspace);
  gsl_eigen_symmv_sort (GSLevalue, GSLevectr, GSL_EIGEN_SORT_ABS_ASC);
  gsl_linalg_LU_decomp(GSLfisher, permutation, &i);
  gsl_linalg_LU_invert(GSLfisher, permutation, GSLcovari);

  //unpack arrays from gsl inversion
  for(i=0; i<NE; i++)
  {
    fisher->evalue[i] = gsl_vector_get(GSLevalue,i);
    for(j=0; j<NE; j++) fisher->evector[i][j] = gsl_matrix_get(GSLevectr,i,j);
  }

  for(i=0;i<NE-1;i++)for(j=i+1;j<NE;j++) gsl_matrix_set(GSLcovari,j,i, gsl_matrix_get(GSLcovari,i,j) );

  /*
   // Print covariance matrix
   double norm;
   printf("Covariance Matrix =\n");
   for(i = 0; i < NE; i++) {
   for(j = 0; j < NE; j++) {
   val=gsl_matrix_get (GSLcovari, i, j);
   if(val>=0)printf("+");
   printf ("%.1e ",val);
   }
   printf("\n");
   }
   printf ("--------------------\n");

   // Print correlation matrix
   printf("Correlation Matrix =\n");
   printf(" alpha   sindelta    psi      ecc     scale     dt      phi0  \n");
   for(i = 0; i < NE; i++) {
   for(j = 0; j < NE; j++) {
   val=gsl_matrix_get (GSLcovari, i, j) / sqrt(gsl_matrix_get (GSLcovari, i, i)) / sqrt(gsl_matrix_get (GSLcovari, j, j));
   if(val>=0)printf("+");
   if(val==val)printf ("%.1e ",val);
   else printf("XXXXXXXX ");
   }
   printf("\n");
   }
   printf ("--------------------\n");

   // print out eigenvalues and eigenvectors
   for (i = 0; i < NE; i++) {
   double eval_i = gsl_vector_get (GSLevalue,i);
   gsl_vector_view evec_i = gsl_matrix_column (GSLevectr, i);
   if(eval_i>=0.0)printf ("eigenvalue = +%.1e: ", eval_i);
   else printf ("eigenvalue = %.1e: ", eval_i);
   printf (" { ");
   for (j = 0; j < NE; j++)
   {
   val=gsl_matrix_get(GSLevectr,i,j);
   if(val>=0.0) printf("+");
   printf ("%.1e ",val);
   }
   printf ("}\n");
   }
   */

	//cap minimum size eigenvalues
	for(i=0; i<NE; i++) if(fisher->evalue[i] < 100.) fisher->evalue[i] = 100.0;

  free(glitch);

  gsl_vector_free (GSLevalue);
  gsl_matrix_free (GSLfisher);
  gsl_matrix_free (GSLcovari);
  gsl_matrix_free (GSLevectr);
  gsl_eigen_symmv_free (workspace);
  gsl_permutation_free (permutation);

}

void extrinsic_fisher_proposal(struct FisherMatrix *fisher, double *params, long *seed, int c, double *y)
{
	int i, j;

	double Amps[NE];
	double jump[NE];

  // not used since we jump along eigendirections
  double sqD = 1.;//sqrt((double)(NE));

	//draw the eigen-jump amplitudes from N[0,1] scaled by evalue & dimension
	for(i=0; i<NE; i++)
	{
		if(i!=c) Amps[i] = 0.0;	      //set all amplitudes to 0.0 except for the one chosen at random.
    if(i==c) Amps[i] = 1.0/sqrt(fisher->evalue[i]);

    //Amps[i] = 1.0/sqrt(fisher->evalue[i]);

		jump[i]	= 0.0;
	}

	//decompose eigenjumps into parameter directions
	for(i=0; i<NE; i++) for (j=0; j<NE; j++) jump[j] += Amps[i]*fisher->evector[j][i]*sqD;

	//jump from current position
  for(i=0; i<NE; i++) y[i] = params[i] + gasdev2(seed)*jump[i];
}

void extrinsic_uniform_proposal(long *seed, double *y)
{
  y[0] = ran2(seed)*TPI;
  y[1] = -1.0 + 2.0*ran2(seed);
  y[2] = ran2(seed)*LAL_PI/2;
  y[3] = -0.99 + 1.98*ran2(seed);
  y[4] = 0. + 100.*ran2(seed);
  y[5] = -0.01 + 0.02*ran2(seed);
  y[6] = ran2(seed)*TPI;
}

void intrinsic_fisher_update(double *params, double *dparams, double *Snf, double Snx, double Tobs)
{
  int i;
  double SNR;
  i = (int)(params[1]*Tobs);

  //SNR defined with Sn(f) but Snf array holdes <n_i^2>
  SNR = params[3]*sqrt((params[2]/(2.0*RT2PI*params[1]))/(Snx*Snf[i]*2.0/Tobs));

  if(SNR < 5.0) SNR = 5.0;  // this caps the size of the proposed jumps
  dparams[0] = 1.0/(LAL_TWOPI*params[1]*SNR);
  dparams[1] = 2.0*params[1]/(params[2]*SNR);
  dparams[2] = 2.0*params[2]/(sqrt(3.0)*SNR);
  dparams[3] = params[3]/SNR;//1.0/SNR;//amp/SNR;
  dparams[4] = 1.0/SNR;
}

void intrinsic_fisher_proposal(double **range, double *paramsx, double *paramsy, double *Snf, double Snx, double *logpx, double *logpy, double scale, long *seed, double Tobs, int *test)
{
  int k;
  double x,y;
  double qyx = *logpy;
  double qxy = *logpx;
  double *dparamsx = dvector(0,5);
  double *dparamsy = dvector(0,5);

  // this is a crude diagonal Fisher matrix for individual sine-gaussians
  intrinsic_fisher_update(paramsx,dparamsx,Snf,Snx,Tobs);

  // [0] t0 [1] f0 [2] Q [3] Amp [4] phi

  //perturb current location by N[0,scale*sigma]
  for(k=0; k<= 4; k++)
  {
    paramsy[k] = paramsx[k] + scale*dparamsx[k]*gasdev2(seed);
  }

  if(checkrange(paramsy, range))
  {
    free_dvector(dparamsx,0,5);
    free_dvector(dparamsy,0,5);

    *test=1;
    return;
  }

  //compute probabiltiy of such a proposal (could just product the draws from gasdev2() above
  y = 1.0;
  for(k=0; k<= 4; k++)
  {
    x = (paramsx[k]-paramsy[k])/(scale*dparamsx[k]);
    y *= scale*dparamsx[k];
    qyx +=  -x*x/2.0;
  }
  qyx -= log(y);

  //need fisher at new location to compute reverse probability
  intrinsic_fisher_update(paramsy,dparamsy,Snf,Snx,Tobs);

  //compute probabiltiy of reverse proposal
  y = 1.0;
  for(k=0; k<= 4; k++)
  {
    x = (paramsx[k]-paramsy[k])/(scale*dparamsy[k]);
    y *= scale*dparamsy[k];
    qxy +=  -x*x/2.0;
  }
  qxy -= log(y);

  // map phase onto [0,2pi]
  while(paramsy[4] > TPI) paramsy[4] -=TPI;
  while(paramsy[4] < 0.0) paramsy[4] +=TPI;

  //update incoming pointers and return
  *logpy += qyx;
  *logpx += qxy;

  free_dvector(dparamsx,0,5);
  free_dvector(dparamsy,0,5);
}

void adapt_temperature_ladder(struct Data *data, struct Model **model, struct Chain *chain, int NC)
{
  int i,ic,ifo;
  double x;

  for(ic=0; ic<NC-1; ic++)
  {
    //only compare after several pt proposals (gives the chains a chance to settle in)
    if(chain->ptprop[ic]%500==0)
    {
      x = (double)chain->ptacc[ic]/(double)chain->ptprop[ic];
      chain->dT[ic] = chain->temperature[ic+1]/chain->temperature[ic];
      //If swaps are too rare...
      if(x<0.4)
      {
        chain->dT[ic] = sqrt(chain->dT[ic]);//0.05;
        fprintf(stdout, "---> chain %i (%.1f) has acceptance rate %.2g\n",ic,chain->temperature[ic],x);
        fprintf(stdout, "     decreasing dT to %.3g\n",chain->dT[ic]);

        if(chain->dT[ic] < 1.1)
        {
          chain->dT[ic] = 1.1;

          if( data->runPhase == 0 )//|| burnFlag == 0)//chain->ptacc[ic] == 0)
          {
            //              if(x<0.01)
            //              {
            fprintf(stdout, "     Warning:  Chain %i and %i have lost contact\n",ic,ic+1);
            fprintf(stdout, "               Now cloning chains w/ T>T[%i]\n",ic);
            //chain->dT[ic] = chain->tempStep;

            for(i=ic; i<NC-1; i++)
            {
              //model[chain->index[i+1]]->logL = model[chain->index[ic]]->logL;
              copy_int_model(model[chain->index[ic]], model[chain->index[i+1]], data->N, data->NI, -1);
              for(ifo=0; ifo<data->NI; ifo++)copy_int_model(model[chain->index[ic]], model[chain->index[i+1]], data->N, data->NI, ifo);
              copy_ext_model(model[chain->index[ic]], model[chain->index[i+1]], data->N, data->NI);
            }
          }
        }
      }

      //If swaps are too frequent...
      else if(x>0.6 && chain->dT[ic]<3.0)
      {
        chain->dT[ic] = chain->dT[ic]*chain->dT[ic];//0.05;
        fprintf(stdout, "---> chain %i (%.3f) has acceptance rate %.2g\n",ic,chain->temperature[ic],x);
        fprintf(stdout, "     increasing dT to %.3g\n",chain->dT[ic]);
      }
    }//end if(ptprop>1000) condition

    if(chain->ptprop[ic]==500)
    {
      chain->ptprop[ic]=1;
      chain->ptacc[ic] =0;
    }

  }//end loop over ic

}//end adapt function

void reset_params(struct Wavelet *wave_x, struct Wavelet *wave_y, double **range)
{
  int i,j,k;

  int sx   = wave_x->size;
  int smax = wave_x->smax;

  int *larrayx = wave_x->index;
  int *larrayy = wave_y->index;

  double **paramsx = wave_x->intParams;
  double **paramsy = wave_y->intParams;



  for(j=1; j<= smax; j++) for(k=0; k <= 1; k++) paramsy[j][k] = paramsx[j][k];
  for(j=1; j<= smax; j++) for(k=0; k <= 1; k++) paramsx[j][k] = range[k][0];
  for(j=1; j<= sx; j++)
  {
    i = larrayx[j];
    for(k=0; k <= 1; k++) paramsx[i][k] = paramsy[i][k];
  }
  for(j=1; j<= smax; j++) for(k=0; k <= 1; k++) paramsy[j][k] = range[k][0];

  for(j=1; j<= sx; j++)
  {
    larrayy[j] = larrayx[j];
    i = larrayx[j];
    for(k=0; k <= 4; k++) paramsy[i][k] = paramsx[i][k];
  }
}

int checkrange(double *p, double **r)
{
  //TODO:  Check physical parameters (t,phase,amp have int. and ext. components)
  int k;

  for(k=0; k<=3; k++) if(p[k] < r[k][0] || p[k] > r[k][1])
  {
    //printf("parameter %i out of range (%g):  [%g,%g]\n",k,p[k],r[k][0],r[k][1]);
    return 1;
  }
  return 0;
}

int extrinsic_checkrange(double *p)
{
  //TODO:  Check physical parameters (t,phase,amp have int. and ext. components)
  if(p[0] < 0.0)  p[0] += TPI;
  if(p[0] >= TPI) p[0] -= TPI;

//  while(p[2] < 0.0) p[2] += LAL_PI/2.;
//  while(p[2] >=LAL_PI/2) p[2] -= LAL_PI/2.;
  while(p[2] < 0.0)
   p[2] += 0.5*LAL_PI;
   while(p[2] >= 0.5*LAL_PI)
   p[2] -= 0.5*LAL_PI;

  if(p[6] < 0.0) p[6] += TPI;
  if(p[6] >=TPI) p[6] -= TPI;

  if(p[1] < -1.0 || p[1] >= 1.0) return 1;

  if(p[3] < -0.99 || p[3] >= 0.99) return 1;

  //Each wavelet's amplitude is checked against intrinsic prior during extrinsic updates
  //if(p[4] < 0  || p[4] >= 100.0) return 1;

  if(p[5] < -0.1 || p[5] >=0.1) return 1;//{printf("***********5 %g,%g\n",p[5],Tobs);return 1;}//return 1;

  return 0;
}


void average_log_likelihood_via_direct_downsampling(int M, double **logLchain, struct Chain *chain)
{
  int i,j,ic;

  int lag;
  int count;

  int NC      = chain->NC;
  int nPoints = M/2/chain->cycle;
  int imax    = nPoints/2;

  double cumACF;
  double ACL[NC];
  double ACF;
  double ACLmin;

  double norm;
  double mean;
  double *temp = malloc(sizeof(double)*nPoints);
  double *lagged = malloc(sizeof(double)*nPoints);

  ACLmin = (double)nPoints/100.;

  // Compute autocorrelation length of each logL chain
  for(ic=0; ic<NC; ic++)
  {

    //temporarily store chain for each temperature
    for(i=0; i<nPoints; i++) temp[i] = logLchain[ic][i];

    //de-trend logL chain
    mean = gsl_stats_mean(temp, 1, nPoints);
    for (i=0; i<nPoints; i++) temp[i] -= mean;
    norm=0.0;
    for(i=0; i<nPoints; i++)norm+=temp[i]*temp[i];


    lag=1;
    ACL[ic] = 1.0;
    ACF     = 1.0;
    cumACF  = 1.0;

    while (cumACF >= ACL[ic])
    {
      ACF=0.0;
      for(j=0; j<=nPoints-lag-1; j++) ACF += temp[j]*temp[j+lag]/norm;
      cumACF += 2.0 * ACF;
      lag++;
      ACL[ic] = (REAL8)lag/(REAL8)5;
      if (lag > imax)
      {
        ACL[ic]=nPoints;
        break;
      }
    }

    /*
     We want 100 effective samples unless the ACL is longer than Npoints/100
     where we will use the full variance of logL as var(<logL>) -- dramatically
     reducing that samples weight in the integral
     For this case we set ACL[ic] = -1 to flag it as bad chain
     */
    if(ACL[ic]<=ACLmin) ACL[ic]=ACLmin;
    else                ACL[ic]=-1;
  }

  //convert sumLogL to average logL
  for(ic=0; ic<NC; ic++)
  {
    count=0;
    chain->avgLogLikelihood[ic]=0.0;
    chain->varLogLikelihood[ic]=0.0;
    for(i=0; i<nPoints; i++)
    {
      if(i%(int)ACL[ic]==0 || ACL[ic]<0.0)
      {
        chain->avgLogLikelihood[ic] += logLchain[ic][i];
        chain->varLogLikelihood[ic] += logLchain[ic][i]*logLchain[ic][i];
        count++;
      }
    }

    chain->avgLogLikelihood[ic] /= (double)count;
    chain->varLogLikelihood[ic] /= (double)count;

    //Var[logL] = <logL^2> - <logL>^2
    chain->varLogLikelihood[ic] = chain->varLogLikelihood[ic] - chain->avgLogLikelihood[ic]*chain->avgLogLikelihood[ic];

    //Var[<logL>] = Var[logL]/N **unless N is small (chain is bad) in which case we'll just use Var[logL]
    if(ACL[ic]>0.0) chain->varLogLikelihood[ic] /= (double)count;
  }

  free(temp);
  free(lagged);
}


void average_log_likelihood_via_ACL_downsampling(int M, double **logLchain, struct Chain *chain)
{
  int i,j,ic;

  int lag;
  int count;

  int NC      = chain->NC;
  int nPoints = M/2/chain->cycle;
  int imax    = nPoints/2;

  double cumACF;
  double ACL[NC];
  double ACF;

  double norm;
  double mean;
  double *temp = malloc(sizeof(double)*nPoints);
  double *lagged = malloc(sizeof(double)*nPoints);

  // Compute autocorrelation length of each logL chain
  for(ic=0; ic<NC; ic++)
  {

    //temporarily store chain for each temperature
    for(i=0; i<nPoints; i++) temp[i] = logLchain[ic][i];

    //de-trend logL chain
    mean = gsl_stats_mean(temp, 1, nPoints);
    for (i=0; i<nPoints; i++) temp[i] -= mean;
    norm=0.0;
    for(i=0; i<nPoints; i++)norm+=temp[i]*temp[i];


    lag=1;
    ACL[ic] = 1.0;
    ACF     = 1.0;
    cumACF  = 1.0;

    while (cumACF >= ACL[ic])
    {
      ACF=0.0;
      for(j=0; j<=nPoints-lag-1; j++) ACF += temp[j]*temp[j+lag]/norm;
      cumACF += 2.0 * ACF;
      lag++;
      ACL[ic] = (REAL8)lag/(REAL8)5;
      if (lag > imax)
      {
        ACL[ic]=nPoints;
        break;
      }
    }
    if(ACL[ic]<1.)ACL[ic]=1.;


    printf("Total number of post burn-in samples for chain %i:  %i\n",ic,nPoints);
    printf("                                  ACL of chain %i:  %i\n",ic,(int)ACL[ic]);
    printf("         Effective number of samples for chain %i:  %i\n",ic,(int)nPoints/(int)ACL[ic]);

    if(nPoints/(int)ACL[ic] < 100)
    {
      printf("**WARNING**:  CHAIN %i CAN NOT BE TRUSTED -- SETTING Neff = 1\n",ic);
      ACL[ic] = nPoints;
    }

  }

  //convert sumLogL to average logL
  double beta[NC];
  for(ic=0; ic<NC; ic++)
  {
    count=0;
    chain->avgLogLikelihood[ic]=0.0;
    chain->varLogLikelihood[ic]=0.0;
    for(i=0; i<nPoints; i++)
    {
      if(i%(int)ACL[ic]==0 || ACL[ic]==nPoints)
      {
        chain->avgLogLikelihood[ic] += logLchain[ic][i];
        chain->varLogLikelihood[ic] += logLchain[ic][i]*logLchain[ic][i];
        count++;
      }
    }

    beta[ic] = 1./chain->temperature[ic];
    chain->avgLogLikelihood[ic] /= (double)count;
    chain->varLogLikelihood[ic] /= (double)count;

    //Var[logL] = <logL^2> - <logL>^2
    chain->varLogLikelihood[ic] = chain->varLogLikelihood[ic] - chain->avgLogLikelihood[ic]*chain->avgLogLikelihood[ic];

    if(ACL[ic]<nPoints)chain->varLogLikelihood[ic] /= (double)count;
  }

  free(temp);
  free(lagged);
}



void average_log_likelihood_via_recursion_downsampling(int M, double **logLchain, struct Chain *chain)
{
  int i,ic;

  int lag;
  int count;

  int NC      = chain->NC;
  int nPoints = M/2/chain->cycle;
  int imax    = nPoints/2;

  double s;
  double cumACF;
  double ACL[NC];
  double ACF;

  double mean;
  double *temp = malloc(sizeof(double)*nPoints);
  double *lagged = malloc(sizeof(double)*nPoints);

  double variance;
  double beta[NC];
  double norm;
  for(ic=0; ic<NC; ic++)
  {
    norm=0.0;
    for(i=0; i<nPoints; i++)norm+=temp[i]*temp[i];
    imax=nPoints/2;
    cumACF=1.0;
    ACL[ic]=1.0;
    ACF=1.0;
    s=1.0;
    lag=1;

    while (cumACF >= s)
    {
      for(i=0; i<nPoints-lag; i++)
      {
        temp[i]   = logLchain[ic][i];
        lagged[i] = logLchain[ic][i+lag];
      }
      if(lag==1)
      {
        mean = gsl_stats_mean(temp, 1, nPoints-lag);
        variance = gsl_stats_variance(temp, 1, nPoints-lag);
      }
      for (i=0; i<nPoints-lag; i++)
      {
        temp[i]   -= mean;
        lagged[i] -= mean;
      }

      ACF = gsl_stats_correlation(temp, 1, lagged, 1, nPoints-lag);
      cumACF += 2.0 * ACF;
      lag++;
      s = (REAL8)lag/(REAL8)5;
      if (lag > imax)
      {
        ACL[ic]=nPoints;
        break;
      }
      ACL[ic]=s;
    }

    printf("Total number of post burn-in samples for chain %i:  %i\n",ic,nPoints);
    printf("Effective sample size of chain %i:                  %g\n",ic,1+cumACF);
  }


  /******************************************************************************/
  /*                                                                            */
  /*  Compute (log) evidence via Thermodynamic Integration                      */
  /*                                                                            */
  /******************************************************************************/

  //convert sumLogL to average logL
  //double beta[NC];
  for(ic=0; ic<NC; ic++)
  {
    count=0;
    chain->avgLogLikelihood[ic]=0.0;
    chain->varLogLikelihood[ic]=0.0;
    for(i=0; i<nPoints; i++)
    {
      if(i%(int)ACL[ic]==0)
      {
        chain->avgLogLikelihood[ic] += logLchain[ic][i];
        chain->varLogLikelihood[ic] += logLchain[ic][i]*logLchain[ic][i];
        count++;
      }
    }

    printf("number of samples in average = %i\n",count);
    beta[ic] = 1./chain->temperature[ic];
    chain->avgLogLikelihood[ic] /= (double)count;
    chain->varLogLikelihood[ic] /= (double)count;

    //Var[logL] = <logL^2> - <logL>^2
    chain->varLogLikelihood[ic] = chain->varLogLikelihood[ic] - chain->avgLogLikelihood[ic]*chain->avgLogLikelihood[ic];

    /*
     We use the variance of the logL chains to get the variance in our
     estimate for the mean

     so Var[<logL>] = Var[logL] / N
     */
    chain->varLogLikelihood[ic] /= (double)count;//((double)M/chain->cycle/4.);//pow( (double)M/chain->cycle/4., 2.0);
    printf("OLD VARIANCE[%i] = %g\n",ic,chain->varLogLikelihood[ic]);

    free(temp);
    free(lagged);

  }
}

void average_log_likelihood_via_ACF(int M, double **logLchain, struct Chain *chain)
{
  int i,j,ic;
  int NC = chain->NC;
  int lag;
  int imax;
  int count=0;
  double ACL[NC];
  double ACF;
  double s;
  double mean;
  int nPoints = M/2/chain->cycle;//M/4/chain->cycle;

  double *temp = malloc(sizeof(double)*nPoints);
  double norm;

  //convert sumLogL to average logL
  double beta[NC];

  double correction;// = 0.0;
  for(ic=0; ic<NC; ic++)
  {
    correction = 0.0;
    count=0;
    chain->avgLogLikelihood[ic]=0.0;
    chain->varLogLikelihood[ic]=0.0;
    for(i=0; i<nPoints; i++)
    {
      chain->avgLogLikelihood[ic] += logLchain[ic][i];
      chain->varLogLikelihood[ic] += logLchain[ic][i]*logLchain[ic][i];
      count++;
    }


    for(i=0; i<nPoints; i++) temp[i] = logLchain[ic][i];

    mean = gsl_stats_mean(temp, 1, nPoints);
    for (i=0; i<nPoints; i++) temp[i] -= mean;
    norm=0.0;
    for(i=0; i<nPoints; i++)norm+=temp[i]*temp[i];
    imax=nPoints/2;
    lag=1;
    ACL[ic]=1.0;
    ACF=1.0;
    s=1.0;
    for(lag=1; lag<nPoints; lag++)
    {
      ACF=0.0;
      for(j=0; j<=nPoints-lag-1; j++) ACF += temp[j]*temp[j+lag]/norm;

      correction += 2.0*(1.0 - ((double)lag/(double)nPoints))*ACF;
    }

    printf("number of samples in average = %i\n",count);
    beta[ic] = 1./chain->temperature[ic];
    chain->avgLogLikelihood[ic] /= (double)count;
    chain->varLogLikelihood[ic] /= (double)count;

    //Var[logL] = <logL^2> - <logL>^2
    chain->varLogLikelihood[ic] = chain->varLogLikelihood[ic] - chain->avgLogLikelihood[ic]*chain->avgLogLikelihood[ic];

    /*
     We use the variance of the logL chains to get the variance in our
     estimate for the mean

     so Var[<logL>] = Var[logL] / N
     */
    chain->varLogLikelihood[ic] /= (double)count;//((double)M/chain->cycle/4.);//pow( (double)M/chain->cycle/4., 2.0);
    chain->varLogLikelihood[ic] *= (1.0 + correction);
    printf("NEW VARIANCE[%i] = %g\n",ic,chain->varLogLikelihood[ic]);

  }

  free(temp);
}

static void splint(double *xa, double *ya, double *y2a, int n, double x, double *y)
{
  //	void nrerror(char error_text[]);
	int klo,khi,k;
	double h,b,a;

	klo=1;
	khi=n;
	while (khi-klo > 1) {
		k=(khi+klo) >> 1;
		if (xa[k] > x) khi=k;
		else klo=k;
		//printf("%d %f\n", k, xa[k]);
	}
	h=xa[khi]-xa[klo];
	//printf("%f %d %f %d %f\n", x, klo, xa[klo], khi, xa[khi]);
  //	if (h == 0.0) nrerror("Bad xa input to routine splint");
	a=(xa[khi]-x)/h;
	b=(x-xa[klo])/h;
	*y=a*ya[klo]+b*ya[khi]+((a*a*a-a)*y2a[klo]+(b*b*b-b)*y2a[khi])*(h*h)/6.0;
}

#define NRANSI
static void spline(double *x, double *y, int n, double yp1, double ypn, double *y2)
{
	int i,k;
	double p,qn,sig,un;
	double *u;

	u=dvector(1,n-1);
	if (yp1 > 0.99e30)
    y2[1]=u[1]=0.0;
	else {
		y2[1] = -0.5;
		u[1]=(3.0/(x[2]-x[1]))*((y[2]-y[1])/(x[2]-x[1])-yp1);
	}
	for (i=2;i<=n-1;i++) {
		sig=(x[i]-x[i-1])/(x[i+1]-x[i-1]);
		p=sig*y2[i-1]+2.0;
		y2[i]=(sig-1.0)/p;
		u[i]=(y[i+1]-y[i])/(x[i+1]-x[i]) - (y[i]-y[i-1])/(x[i]-x[i-1]);
		/*if(i==n-1)printf("%i y[i+1] = %.12g y[i] = %.12g x[i+1] = %.12g x[i] = %.12g u[i-1] = %.12g u[i] = %.12g sig = %.12g \n",i, y[i+1], y[i], x[i+1], x[i], u[i-1], u[i], sig);*/
		u[i]=(6.0*u[i]/(x[i+1]-x[i-1])-sig*u[i-1])/p;
		/*if(i==n-1) printf("x[i+1]-x[i-1] = %.12g \n", x[i+1]-x[i-1]);*/
	}
	if (ypn > 0.99e30)	qn=un=0.0;

	else {
		qn=0.5;
		un=(3.0/(x[n]-x[n-1]))*(ypn-(y[n]-y[n-1])/(x[n]-x[n-1]));
  }
	y2[n]=(un-qn*u[n-1])/(qn*y2[n-1]+1.0);
	//printf("qn = %.12g  y2[n-1] = %.12g  u[n-1] = %.12g\n", qn, y2[n-1], u[n-1]);

	for (k=n-1;k>=1;k--) y2[k]=y2[k]*y2[k+1]+u[k];

	free_dvector(u,1,n-1);
}
#undef NRANSI

void ThermodynamicIntegration(double *temperature, double *avgLogLikelihood, double *varLogLikelihood, int NC, double *logEvidence, double *varLogEvidence)
{

  int N, Nx, Ny, ND;
  int mc, i, j, ii, test, scnt;
  long seed;
  int acc;
  int fixedD;
  double logLx, logLy;
  double logpx, logpy;
  double *ref, *sprd, *datax, *datay;
  int *activex, *activey;
  double *counts;

  double alpha, H, av;
  double model;

  double max, min;

  double x, y, y3, y4, x3, x4, *sigma;
  double q, secd;

  double trap;
  double *sdatay, *sderiv, *sgm;
  double *spoints, *sdatax, *tdata, *tpoints;

  double ep;
  double avr=0, var;


  N = 10000000;  // number of MCMC steps

	ep = 0.0001;

	ND = NC;


  datax = dvector(0,ND-1);
  datay = dvector(0,ND-1);
  sigma = dvector(0,ND-1);
  counts = dvector(1,ND);


  seed = -946673524;

  // Read in  Data
  for(j=0; j< ND; j++)
  {
    i = ND-j-1;
    datax[i] = log(1./temperature[j]);
    datay[i] = avgLogLikelihood[j];
    sigma[i] = sqrt(varLogLikelihood[j]);
  }


	max = datax[ND-1];
	min = datax[0];

	ref = dvector(1,ND);
  sprd = dvector(1,ND);
	spoints = dvector(1,ND);
	tdata = dvector(1,ND);
	tpoints = dvector(1,ND);
  sdatax = dvector(1,ND);
  sdatay = dvector(1,ND);
  sderiv = dvector(1,ND);
	sgm = dvector(1,ND);

	activex = ivector(1,ND);
	activey = ivector(1,ND);
	for(i=1; i<= ND; i++)
	{
		activex[i] = 1;
		activey[i] = 1;
	}

	// copy over the input data
  for(i=1; i<= ND; i++)
	{
		sdatax[i]= datay[i-1];
		spoints[i] = datax[i-1];
		sderiv[i] = 0.0;
		ref[i] = datay[i-1];
		sprd[i] = 100.0+0.5*fabs(ref[i]);
		//printf("%d %f %f\n", i, ref[i], sprd[i]);
	}

	for(i=1; i< ND; i++)
	{
		sgm[i] = 10.0+ fabs((sdatax[i+1]-sdatax[i])/pow((spoints[i+1]-spoints[i]),2.0));
	}
	sgm[ND] = sgm[ND-1];

	spline(spoints, sdatax, ND, 1.0e31, 1.0e31, sderiv);

  Nx = Ny = ND;

	// set this flag to 1 if you want to do a fixed dimension run
  fixedD = 0;

  for(i=1; i<= ND; i++) counts[i] = 0.0;


  // compute the initial log likelihood
  av = 0.0;
  for(j=0; j< ND; j++)
  {
		splint(spoints, sdatax, sderiv, Nx, datax[j], &model);
    y = (datay[j]-model)/sigma[j];
    av -= y*y;
  }
  logLx = av/2.0;


	logpx = 0.0;
	for(i=2; i< ND; i++)
	{
		x = 0.5*(datax[i]+datax[i-1]);
		splint(spoints, sdatax, sderiv, Nx, x, &y);
		x3 = x - ep;
		splint(spoints, sdatax, sderiv, Nx, x3, &y3);
		x4 = x + ep;
		splint(spoints, sdatax, sderiv, Nx, x4, &y4);
		secd = (y3+y4-2.0*y)/(ep*ep);
		logpx -= (secd*secd)/(2.0*sgm[i]*sgm[i]);
	}

  acc = 1;
	av = 0.0;
	var = 0.0;
	scnt = 0;

  // start the RJMCMC
  for(mc=0; mc< N; mc++)
  {

		test = 0;

    for(i=1; i<= ND; i++)
    {
      sdatay[i] = sdatax[i];
      activey[i] = activex[i];
    }

    alpha = ran2(&seed);

    q = 0.5;
    if(fixedD == 1) q = 10.0;

    if(alpha > q)   // propose a dimension change
    {

      alpha = ran2(&seed);

      if(alpha < 0.5)
      {
        Ny = Nx + 1;
      }
      else
      {
        Ny = Nx - 1;
      }

      if(Ny < Nx) // propose a kill
      {
        if(Ny > 1 && Ny <= ND)
		    {

          do
          {
            i = 1 + (int)(ran2(&seed)*(double)(ND)); // pick one to kill
          } while(activex[i] == 0);  // can't kill it if already dead
          activey[i] = 0;
		    }
        else
		    {
          test = 1;
        }
      }
      else
      {
        if(Ny >= 1 && Ny < ND)
		    {

          do
          {
            i = 1 + (int)(ran2(&seed)*(double)(ND)); // pick one to add
          } while(activex[i] == 1);  // can't add if already in use
          activey[i] = 1;

          sdatay[i] = ref[i]+sprd[i]*(1.0-2.0*ran2(&seed));

		    }
        else
		    {
          test = 1;
        }


      }


    }
    else     // within dimension update
    {

      Ny = Nx;

      alpha = ran2(&seed);

      for(ii=1; ii<= ND; ii++)
      {
        // variety of jump sizes
        if(alpha > 0.8)
        {
          sdatay[ii] += sprd[ii]*1.0e-1*gasdev2(&seed);
        }
        else if (alpha > 0.5)
        {
          sdatay[ii] += sprd[ii]*1.0e-2*gasdev2(&seed);
        }
        else if (alpha > 0.3)
        {
          sdatay[ii] += sprd[ii]*1.0e-3*gasdev2(&seed);
        }
        else
        {
          sdatay[ii] += sprd[ii]*1.0e-4*gasdev2(&seed);
        }

      }
		}

    // check that proposed values are within the prior range
    if(Ny >= 1 && Ny <= ND)
    {
			for(ii=1; ii<= ND; ii++) // check the active points are in range
			{
        if(activey[ii] == 1)
			  {
				  if(sdatay[ii] > ref[ii]+sprd[ii]) test = 1;
				  if(sdatay[ii] < ref[ii]-sprd[ii]) test = 1;
			  }
			}
    }
    else
    {
      test = 1;
    }


    logLy = 0.0;

    if(test == 0)
    {

      i = 0;
      for(ii=1; ii<= ND; ii++)
      {
        if(activey[ii] == 1)  // only use active points
        {
          i++;
          tpoints[i] = spoints[ii];
          tdata[i] = sdatay[ii];
        }
      }


      spline(tpoints, tdata, Ny, 1.0e31, 1.0e31, sderiv);
      av = 0.0;
      for(j=0; j< ND; j++)
      {
        splint(tpoints, tdata, sderiv, Ny, datax[j], &model);
        y = (datay[j]-model)/sigma[j];
        av -= y*y;
      }
      logLy = av/2.0;

      logpy = 0.0;
      for(i=2; i< ND; i++)
      {
        x = 0.5*(datax[i]+datax[i-1]);
        splint(tpoints, tdata, sderiv, Ny, x, &y);
        x3 = x - ep;
        splint(tpoints, tdata, sderiv, Ny, x3, &y3);
        x4 = x + ep;
        splint(tpoints, tdata, sderiv, Ny, x4, &y4);
        secd = (y3+y4-2.0*y)/(ep*ep);
        logpy -= (secd*secd)/(2.0*sgm[i]*sgm[i]);
      }
    }

    H = logLy-logLx +logpy - logpx;
    alpha = log(ran2(&seed));

    if((H > alpha) && (test==0))
    {
      acc++;
      logLx = logLy;
			logpx = logpy;
      Nx = Ny;
      for(i=1; i<= ND; i++)
      {
        sdatax[i] = sdatay[i];
        activex[i] = activey[i];
      }

    }

    // This gives us the data for Bayes factors. Don't start collecting it until after burn-in
    if(mc > N/2) counts[Nx] += 1.0;

    if(mc%1000 == 0)
    {

      i = 0;
      for(ii=1; ii<= ND; ii++)
      {
        if(activex[ii] == 1)  // only use active points
        {
          i++;
          tpoints[i] = spoints[ii];
          tdata[i] = sdatax[ii];
        }
      }


      spline(tpoints, tdata, Nx, 1.0e31, 1.0e31, sderiv);

      trap = 0.0;
      x3 = min;
      splint(tpoints, tdata, sderiv, Nx, x3, &y3);
      for(i=1; i<= 1000; i++)
      {
        x = x3;
        y = y3;
        x3 = min + (max-min)*(double)(i)/1000.0;
        splint(tpoints, tdata, sderiv, Nx, x3, &y3);
        trap += 0.5*(x3-x)*(exp(x3)*y3+exp(x)*y);
      }

      scnt++;
      avr += trap;
      var += trap*trap;


    }



  }

	x = avr/(double)(scnt);
	y = var/(double)(scnt);

	printf("\n Average Evidence %f\n", x);
	printf(" Sigma Evidence %f\n\n", sqrt(y-x*x));

	*logEvidence = x;
  *varLogEvidence = y-x*x;

  free_dvector(datax,0,ND-1);
	free_dvector(datay,0,ND-1);
	free_dvector(sigma,0,ND-1);
	free_dvector(counts,1,ND);

 	free_dvector(ref,1,ND);
  free_dvector(sprd,1,ND);
	free_dvector(spoints,1,ND);
  free_dvector(tdata,1,ND);
	free_dvector(tpoints,1,ND);
  free_dvector(sdatax,1,ND);
	free_dvector(sdatay,1,ND);
  free_dvector(sderiv,1,ND);
	free_dvector(sgm,1,ND);

	free_ivector(activex,1,ND);
	free_ivector(activey,1,ND);



}


void bootcorrposdef(double **data, int NC, int N, double **cmat, double *mu, long *seed)
{
	double av, cc, vc;
	double x, y, shrink;
	int i, j, k, cmin, icm, cmax, icx;
	int ii, jj, kk, nn;
	int *cyc;
	int cycles;
	int so, sc, cp;
	int b, bs, bsteps;
	int *cnts;
	double **fmean;
	double *var, *mx;

	// This code uses the "threshold bootstrap" method to compute the covariance matrix
	// See http://www.sciencedirect.com/science/article/pii/S0377221700002095 and
	/* @inproceedings{Kim:1993:TBN:256563.256697,
   author = {Kim, Y. B. and Willemain, T. R. and Haddock, J. and Runger, G. C.},
   title = {The Threshold Bootstrap: A New Approach to Simulation Output Analysis},
   booktitle = {Proceedings of the 25th Conference on Winter Simulation},
   series = {WSC '93},
   year = {1993},
   isbn = {0-7803-1381-X},
   location = {Los Angeles, California, USA},
   pages = {498--502},
   numpages = {5},
   url = {http://doi.acm.org/10.1145/256563.256697},
   doi = {10.1145/256563.256697},
   acmid = {256697},
   publisher = {ACM},
   address = {New York, NY, USA},
   } */

	// The "shrinkage" method described here is uses to ensure the covariance matrices are
  // postive definite and well balanced https://www.stat.wisc.edu/courses/st992-newton/smmb/files/expression/shrinkcov2005.pdf

	bsteps = 10000;

	var = dvector(0,NC);
	cnts = ivector(0,NC);

	for(ii=0; ii< NC; ii++)
	{
		mu[ii] = 0.0;
		var[ii] = 0.0;
		for(i=0; i< N; i++)
		{
			mu[ii] += data[ii][i];
			var[ii] += data[ii][i]*data[ii][i];
		}
		mu[ii] /= (double)(N);
		var[ii] /= (double)(N);
		var[ii] -= mu[ii]*mu[ii];
		var[ii] = sqrt(var[ii]);
	}

	b = 8;
	// if b > 1, then what I'm calling cycles are actually chunks, made up of b cycles

  cmin=0;
	do
	{
    kk = 0;
    nn = 0;
    // cycle lengths for individual chains
    cmin = 100000;
    cmax = 0;
    for(ii=0; ii< NC; ii++)
    {
      cycles = 0;
      so = -1;
      if((data[ii][0]-mu[ii]) > 0.0) so = 1;
      cp = 0;

      for(i=0; i< N; i++)
      {
        sc = -1;
        if((data[ii][i]-mu[ii]) > 0.0) sc = 1;
        if(sc != so) cp++;   // count the change points
        so = sc;
        if(cp == 2*b)
        {
          cycles++;
          cp = 0;
        }
      }

      cnts[ii] = cycles;
      if(cycles < cmin)
      {
        cmin = cycles;
        icm = ii;
      }
      if(cycles > cmax)
      {
        cmax = cycles;
        icx = ii;
      }

      kk += cycles;
      nn += cycles*cycles;

    }

    cc = ((double)(kk)/(double)(NC));
    vc = sqrt((double)(nn)/(double)(NC)-(double)(kk*kk)/(double)(NC*NC));

    // looking for between 100 and 1000 chunks
    // This assumes that we have a decent amount of data to begin with

    if(cmin <  100) b /= 2;
    if(cmin > 1000) b *= 2;

    if(b < 2)
    {
      b = 2;
      cmin = 500;
    }


	} while(cmin < 100 || cmin > 1000);

	// record the change point for each cycle
	// if cyc = i, then one cycle ends at i and the next starts at i+1
	cyc = ivector(0,N/2);

	cyc[0] = -1;



  av = 0.0;

  for(i=0; i< N; i++)
  {
    av += data[icm][i];    // correlation length determined by most corelated
  }
  av /= (double)(N);


  cycles = 0;
  so = -1;
  if((data[icm][0]-av) > 0.0) so = 1;

  cp = 0;

  for(i=0; i< N; i++)
  {
    sc = -1;
    if((data[icm][i]-av) > 0.0) sc = 1;
    if(sc != so) cp++;   // count the change points
    so = sc;
    if(cp == 2*b)
    {
      cycles++;
      cyc[cycles] = i-1;
      cp = 0;
    }
  }

	// create bootstrap data sets

  fmean = dmatrix(0,NC,0,bsteps);  // averages for each bootstrap at each temperature

  for(bs=0; bs< bsteps; bs++)
  {
		for(ii=0; ii< NC; ii++) fmean[ii][bs] = 0.0;
  }

  for(bs=0; bs< bsteps; bs++)
  {

    k = -1;
    do
    {
      i = (double)(cycles)*ran2(seed)+1;  // pick a chunk
      for(j=(cyc[i-1]+1); j<= cyc[i]; j++)
      {
        k++;
        if(k < N)
        {
          for(ii=0; ii< NC; ii++) fmean[ii][bs] += data[ii][j];
        }
      }

    }while(k < N);

    for(ii=0; ii< NC; ii++) fmean[ii][bs] /= (double)(N);

  }

	// means and variances of the bootstraps
	mx = dvector(0,NC);

  for(ii=0; ii< NC; ii++)
	{
		mx[ii] = 0.0;
		var[ii] = 0.0;
		for(bs=0; bs< bsteps; bs++)
		{
			mx[ii] += fmean[ii][bs];
			var[ii] += fmean[ii][bs]*fmean[ii][bs];
		}
		mx[ii] /= (double)(bsteps);
		var[ii] /= (double)(bsteps);
		var[ii] -= mx[ii]*mx[ii];
		var[ii] = sqrt(var[ii]);
		//standardize the data
		for(bs=0; bs< bsteps; bs++)
		{
			fmean[ii][bs] = (fmean[ii][bs]-mx[ii])/var[ii];
		}
	}

	double **rmat, **rvar;

	rmat = dmatrix(0,NC,0,NC);
	rvar = dmatrix(0,NC,0,NC);

  rvar[0][0] = 0.0;

	// next we compute the correlation r_ij and its variance
	for(ii=0; ii< NC; ii++)
	{
		for(jj=0; jj< NC; jj++)
		{
			rmat[ii][jj] = 0.0;
			for(bs=0; bs< bsteps; bs++)	rmat[ii][jj] += fmean[ii][bs]*fmean[jj][bs];
			rmat[ii][jj] /= (double)(bsteps);
		}
	}

	/*for(ii=0; ii< 10; ii++)
   {
   for(jj=0; jj< 10; jj++)
   {
   printf("%f ", rmat[ii][jj]);
   }
   printf("\n");
   }*/


	for(ii=0; ii< NC; ii++)
	{
		for(jj=0; jj< NC; jj++)
		{
			rvar[ii][jj] = 0.0;
			for(bs=0; bs< bsteps; bs++)	rvar[ii][jj] += pow((fmean[ii][bs]*fmean[jj][bs]-rmat[ii][jj]),2.0);
			rvar[ii][jj] *= (double)(bsteps)/pow((double)(bsteps-1),3.0);
		}
	}

	x = 0.0;
	y = 0.0;
	for(ii=0; ii< NC; ii++)
	{
		for(jj=0; jj< NC; jj++)
		{
			if(ii != jj)
			{
				x += rvar[ii][jj];
				y += rmat[ii][jj]*rmat[ii][jj];
			}
		}
	}

	shrink = 1.0-x/y;
	if(shrink < 0.0) shrink = 0.0;
	if(shrink > 1.0) shrink = 1.0;


	for(ii=0; ii< NC; ii++)
	{
		for(jj=0; jj< NC; jj++)
		{
			if(ii == jj)
			{
				cmat[ii][jj] = var[ii]*var[jj];
      }
			else
			{
				cmat[ii][jj] = shrink*var[ii]*var[jj]*rmat[ii][jj];
			}
		}
	}

	free_dmatrix(rmat,0,NC,0,NC);
	free_dmatrix(rvar,0,NC,0,NC);
	free_ivector(cyc,0,N);
	free_ivector(cnts,0,NC);
	free_dmatrix(fmean,0,NC,0,bsteps);
	free_dvector(var,0,NC);
	free_dvector(mx,0,NC);


}

void BootstrapThermodynamicIntegration(double *temperature, double **loglikelihood, int M, int NC, char modelname[], double *logEvidence, double *varLogEvidence)
{


  int N, Nx, Ny, ND, NSP;
  int mc, i, j, ii, test, scnt;
  long seed;
  int acc;
  int fixedD;
  double logLx, logLy;
  double logpx, logpy=0.0;
  double *ref, *sprd, *datax, *datay;
  int *activex, *activey;
  int NL;
  double *counts;
  double alpha, H, av;
  double model;
  double max, min;
  double x, y, y3, y4, x3, x4, dx, *sigma;
  double q;
  double trap,fird,secd;
  double *sdatay, *sderiv, *sgm, *srm;
  double *spoints, *sdatax, *tdata, *tpoints;
  double ep;
  double avr=0.0, var;
  double *mdl;
  double **cov,**chains;

  char filename[100];
  FILE *outfile;

  N = 1000000;  // number of MCMC steps

	// sets the degree of smoothing for the spline
	double smooth = 0.1;

	ep = 0.0001;


	ND = NC;
	NL = M;

  // maximum number of spline points - we initially overcover
	NSP = 2*ND-1;

	datax = dvector(0,ND-1);
	chains = dmatrix(0,ND-1,0,NL-1);
	datay = dvector(0,ND-1);
  cov = dmatrix(0,ND-1,0,ND-1);
  counts = dvector(1,NSP);
	sigma = dvector(0,ND-1);
	mdl = dvector(0,ND-1);

  seed = -673524;

  FILE *temp = fopen("likelihood.dat","w");

  fprintf(temp,"0 ");
  for(j=0; j< ND; j++)
  {
    datax[ND-1-j] = log(1./temperature[j]);
    fprintf(temp,"%lg ",1./temperature[j]);
  }fprintf(temp,"\n");

  for(j=0; j< NL; j++)
  {
    fprintf(temp,"%i ",j);
    for(i=0; i< ND; i++)
    {
      chains[ND-1-i][j] = loglikelihood[i][j];
      fprintf(temp,"%lg ",loglikelihood[i][j]);
    }
    fprintf(temp,"\n");
  }
  fclose(temp);

  printf("\n");
  printf("Starting bootstrap estimate of correlation matrix\n");
  bootcorrposdef(chains, ND, NL, cov, datay, &seed);
  printf("\n");


  /******************************************/
  /*                                        */
  /*  Linear algebra gymnastics with the    */
  /*  correlation matrix                    */
  /*                                        */
  /******************************************/

  /* allocate memory for GSL linear algebra */

	gsl_matrix *matrix  = gsl_matrix_alloc (ND, ND);
	gsl_matrix *inverse = gsl_matrix_alloc (ND, ND);

  gsl_vector *eval = gsl_vector_alloc (ND);
  gsl_matrix *evec = gsl_matrix_alloc (ND, ND);

  gsl_permutation *perm = gsl_permutation_alloc(ND);

  gsl_eigen_symmv_workspace *w = gsl_eigen_symmv_alloc (ND);


  // get eigenvalues and eigenvectors of correlation matrix (matrix is destroyed)
  for (i = 0; i < ND; i++)
  {
    sigma[i] = sqrt(cov[i][i]);
    for (j = 0; j < ND; j++)
    {
      gsl_matrix_set(matrix,i,j,cov[i][j]);
    }
  }

  // the eigenstuff calculator
  gsl_eigen_symmv(matrix, eval, evec, w);

  /*
   if the correlation matrix is up to snuff
   calculate its inverse
   */
  for(i=0; i<ND; i++)
  {
    if(gsl_vector_get(eval,i) < 0.0)
    {
      KILL("Correlation matrix not positive definite\n");
    }
    else
    {
      for(j=0; j<ND; j++) gsl_matrix_set(matrix,i,j,cov[i][j]);
		}
  }

	// Make LU decomposition of matrix
	gsl_linalg_LU_decomp (matrix, perm, &i);

	// Invert the matrix
	gsl_linalg_LU_invert (matrix, perm, inverse);

  double val=0;
  for(i=0; i<ND; i++)
  {
    for(j=0; j<ND; j++)
    {
      val=gsl_matrix_get(inverse,i,j);
    }
  }

  // print out data with error bars

  sprintf(filename,"%s_bootstrap_evidence.dat",modelname);
	outfile = fopen(filename,"w");
	for(j=0; j< ND; j++)
  {
		fprintf(outfile,"%f %f %f\n", datax[j], datay[j], sigma[j]);
  }
	fclose(outfile);


	// compute simple trapezoid integrand and error
	trap = 0.0;
	for(j=1; j< ND; j++)
  {
		trap += 0.5*(datax[j]-datax[j-1])*(exp(datax[j])*datay[j]+exp(datax[j-1])*datay[j-1]);
	}

	double b1, b0;
	double s1, s0;

	x = 0.0;
	for(j=1; j< (ND-1); j++)
	{


		if(j==1)
		{
			b0 = 0.0;
		}
		else
		{
			b0 = (datax[j-1]-datax[j-2]);
		}
		b1 = (datax[j]-datax[j-1]);
		s0 = sigma[j-1]*exp(datax[j-1]);
		s1 = sigma[j]*exp(datax[j]);
		s0 = s0*s0;
		s1 = s1*s1;

    x += 0.25*((s1+s0)*b1*b1+2.0*s0*b1*b0);


  }

	printf("Trapezoid integrand %f and error %f\n", trap, sqrt(x));



	max = datax[ND-1];
	min = datax[0];

  /* Allocate memory for spline fit*/
	ref = dvector(1,NSP);
  sprd = dvector(1,NSP);
	spoints = dvector(1,NSP);
	tdata = dvector(1,NSP);
	tpoints = dvector(1,NSP);
  sdatax = dvector(1,NSP);
  sdatay = dvector(1,NSP);
  sderiv = dvector(1,NSP);
	sgm = dvector(1,NSP);
	srm = dvector(1,NSP);

	activex = ivector(1,NSP);
	activey = ivector(1,NSP);
	for(i=1; i<= NSP; i++)
	{
		activex[i] = 1;
		activey[i] = 1;
	}

	// copy over the input data
  for(i=1; i<= ND; i++)
	{
		sdatax[2*i-1]= datay[i-1];
		spoints[2*i-1] = datax[i-1];
		sderiv[2*i-1] = 0.0;
		ref[2*i-1] = datay[i-1];
		sprd[2*i-1] = 2.0*sigma[i-1];
		if(2*i <= NSP)
		{
      sdatax[2*i]= 0.5*(datay[i-1]+datay[i]);
      spoints[2*i] = 0.5*(datax[i-1]+datax[i]);
      sderiv[2*i] = 0.0;
      ref[2*i] = sdatax[2*i];
      sprd[2*i] = sigma[i-1]+sigma[i];
		}
	}

	spline(spoints, sdatax, NSP, 1.0e31, 1.0e31, sderiv);

  Nx = Ny = NSP;

	// set this flag to 1 if you want to do a fixed dimension run
  fixedD = 0;

  for(i=1; i<= NSP; i++) counts[i] = 0.0;


  // compute the initial log likelihood


  for(j=0; j< ND; j++)
  {
		splint(spoints, sdatax, sderiv, Nx, datax[j], &model);
		mdl[j] = model;
  }

	trap = 0.0;
	x3 = min;
	splint(spoints, sdatax, sderiv, Nx, x3, &y3);
	for(i=1; i<= 1000; i++)
	{
		x = x3;
		y = y3;
		x3 = min + (max-min)*(double)(i)/1000.0;
		splint(spoints, sdatax, sderiv, Nx, x3, &y3);
		trap += 0.5*(x3-x)*(exp(x3)*y3+exp(x)*y);
	}
	printf("Spline Evidence = %f\n", trap);

  /* Initialize Likelhood */
  av = 0.0;
  for (i = 0; i < ND; i++)
  {
		for (j = 0; j < ND; j++)
		{
			av += gsl_matrix_get(inverse,i,j)*(datay[i]-mdl[i])*(datay[j]-mdl[j]);
		}
  }
  logLx = -av/2.0;

  /* Initialize Prior */
  logpx = 0.0;
	for(i=2; i< NSP; i++)
	{
		x = spoints[i];
		dx = spoints[i]-spoints[i-1];
		splint(spoints, sdatax, sderiv, Nx, x, &y);
		x3 = x - ep;
		splint(spoints, sdatax, sderiv, Nx, x3, &y3);
		x4 = x + ep;
		splint(spoints, sdatax, sderiv, Nx, x4, &y4);
		secd = (y3+y4-2.0*y)/(ep*ep);
		fird = (y4-y3)/(2.0*ep);
		logpx -= smooth*(secd*secd*dx*dx)/(fird*fird);
	}

  acc = 1;
	av = 0.0;
	var = 0.0;
	scnt = 0;

  // start the RJMCMC
  for(mc=0; mc< N; mc++)
  {

		test = 0;

    for(i=1; i<= NSP; i++)
    {
      sdatay[i] = sdatax[i];
      activey[i] = activex[i];
    }

    alpha = ran2(&seed);

    q = 0.5;
    if(fixedD == 1) q = 10.0;

    if(alpha > q)   // propose a dimension change
    {

      alpha = ran2(&seed);

      if(alpha < 0.5)
      {
        Ny = Nx + 1;
      }
      else
      {
        Ny = Nx - 1;
      }

      if(Ny < Nx) // propose a kill
      {
        if(Ny > 1 && Ny <= NSP)
		    {

          do
          {
            i = 1 + (int)(ran2(&seed)*(double)(NSP)); // pick one to kill
          } while(activex[i] == 0);  // can't kill it if already dead
          activey[i] = 0;
		    }
        else
		    {
          test = 1;
        }
      }
      else
      {
        if(Ny >= 1 && Ny < NSP)
		    {

          do
          {
            i = 1 + (int)(ran2(&seed)*(double)(NSP)); // pick one to add
          } while(activex[i] == 1);  // can't add if already in use
          activey[i] = 1;

          sdatay[i] = ref[i]+10.0*sprd[i]*(1.0-2.0*ran2(&seed));

		    }
        else
		    {
          test = 1;
        }


      }


    }
    else     // within dimension update
    {

      Ny = Nx;

      alpha = ran2(&seed);

      for(ii=1; ii<= NSP; ii++)
      {
        // variety of jump sizes
        if(alpha > 0.8)
        {
          sdatay[ii] += sprd[ii]*gasdev2(&seed);
        }
        else if (alpha > 0.5)
        {
          sdatay[ii] += sprd[ii]*1.0e-1*gasdev2(&seed);
        }
        else if (alpha > 0.2)
        {
          sdatay[ii] += sprd[ii]*1.0e-2*gasdev2(&seed);
        }
        else
        {
          sdatay[ii] += sprd[ii]*1.0e-3*gasdev2(&seed);
        }

      }
		}

    // check that proposed values are within the prior range
    if(Ny >= 1 && Ny <= NSP)
    {
			for(ii=1; ii<= NSP; ii++) // check the active points are in range
			{
        if(activey[ii] == 1)
			  {
				  if(sdatay[ii] > ref[ii]+10.0*sprd[ii]) test = 1;
				  if(sdatay[ii] < ref[ii]-10.0*sprd[ii]) test = 1;
			  }
			}
    }
    else
    {
      test = 1;
    }


    logLy = 0.0;

    if(test == 0)
    {
      i = 0;
      for(ii=1; ii<= NSP; ii++)
      {
        if(activey[ii] == 1)  // only use active points
        {
          i++;
          tpoints[i] = spoints[ii];
          tdata[i] = sdatay[ii];
        }
      }


      spline(tpoints, tdata, Ny, 1.0e31, 1.0e31, sderiv);
      for(j=0; j< ND; j++)
      {
        splint(tpoints, tdata, sderiv, Ny, datax[j], &model);
        mdl[j] = model;
      }
      av = 0.0;
      for (i = 0; i < ND; i++)
      {
        for (j = 0; j < ND; j++)
        {
          av += gsl_matrix_get(inverse,i,j)*(datay[i]-mdl[i])*(datay[j]-mdl[j]);

        }
      }
      logLy = -av/2.0;

      logpy = 0.0;
      for(i=2; i< NSP; i++)
      {
        x = spoints[i];
        dx = spoints[i]-spoints[i-1];
        splint(tpoints, tdata, sderiv, Ny, x, &y);
        x3 = x - ep;
        splint(tpoints, tdata, sderiv, Ny, x3, &y3);
        x4 = x + ep;
        splint(tpoints, tdata, sderiv, Ny, x4, &y4);
        secd = (y3+y4-2.0*y)/(ep*ep);
        fird = (y4-y3)/(2.0*ep);
        logpy -= smooth*(secd*secd*dx*dx)/(fird*fird);
      }
    }

    H = logLy-logLx +logpy - logpx;
    // H = logLy-logLx;  // without the smoothing prior
    alpha = log(ran2(&seed));

    if((H > alpha) && (test==0))
    {
      acc++;
      logLx = logLy;
			logpx = logpy;
      Nx = Ny;
      for(i=1; i<= NSP; i++)
      {
        sdatax[i] = sdatay[i];
        activex[i] = activey[i];
      }

    }

    // This gives us the data for Bayes factors. Don't start collecting it until after burn-in
    if(mc > N/2) counts[Nx] += 1.0;

    if(mc%1000 == 0)
    {

      i = 0;
      for(ii=1; ii<= NSP; ii++)
      {
        if(activex[ii] == 1)  // only use active points
        {
          i++;
          tpoints[i] = spoints[ii];
          tdata[i] = sdatax[ii];
        }
      }


      spline(tpoints, tdata, Nx, 1.0e31, 1.0e31, sderiv);

      trap = 0.0;
      x3 = min;
      splint(tpoints, tdata, sderiv, Nx, x3, &y3);
      for(i=1; i<= 1000; i++)
      {
        x = x3;
        y = y3;
        x3 = min + (max-min)*(double)(i)/1000.0;
        splint(tpoints, tdata, sderiv, Nx, x3, &y3);
        trap += 0.5*(x3-x)*(exp(x3)*y3+exp(x)*y);
      }

      fprintf(outfile, "%d %f %d %f\n", mc/1000, logLx, Nx, trap);

      scnt++;
      avr += trap;
      var += trap*trap;
    }
  } //End RJMCMC

  i = 0;
	for(ii=1; ii<= NSP; ii++)
	{
		if(activex[ii] == 1)  // only use active points
		{
			i++;
			tpoints[i] = spoints[ii];
			tdata[i] = sdatax[ii];
		}
	}
	spline(tpoints, tdata, Nx, 1.0e31, 1.0e31, sderiv);

  sprintf(filename,"%s_bootstrap_fit.dat",modelname);
	outfile = fopen(filename,"w");
	for(i=1; i<= 1000; i++)
	{
		x3 = min + (max-min)*(double)(i)/1000.0;
		splint(tpoints, tdata, sderiv, Nx, x3, &y3);
		fprintf(outfile,"%f %f\n", x3, y3);
	}
	fclose(outfile);

  x = avr/(double)(scnt);
	y = var/(double)(scnt);

	printf("\n Average Evidence %f\n", x);
	printf(" Sigma Evidence %f\n\n", sqrt(y-x*x));

	*logEvidence = x;
  *varLogEvidence = y-x*x;


  gsl_matrix_free(matrix);
	gsl_matrix_free(inverse);

  gsl_vector_free(eval);
  gsl_matrix_free(evec);

  gsl_permutation_free(perm);

  gsl_eigen_symmv_free(w);

	free_dmatrix(chains,0,ND-1,0,NL-1);
  free_dmatrix(cov,0,ND-1,0,ND-1);
	free_dvector(mdl,0,ND-1);

  free_dvector(datax,0,ND-1);
	free_dvector(datay,0,ND-1);
	free_dvector(sigma,0,ND-1);
	free_dvector(counts,1,NSP);


  /* Free memory for spline fit*/
 	free_dvector(ref,1,NSP);
  free_dvector(sprd,1,NSP);
	free_dvector(spoints,1,NSP);
  free_dvector(tdata,1,NSP);
	free_dvector(tpoints,1,NSP);
  free_dvector(sdatax,1,NSP);
	free_dvector(sdatay,1,NSP);
  free_dvector(sderiv,1,NSP);
	free_dvector(sgm,1,NSP);
  free_dvector(srm,1,NSP);

	free_ivector(activex,1,NSP);
	free_ivector(activey,1,NSP);
}

/* ********************************************************************************** */
/*                                                                                    */
/*                            Wavelet basis functions                                 */
/*                                                                                    */
/* ********************************************************************************** */
static void recursive_phase_evolution(double dre, double dim, double *cosPhase, double *sinPhase)
{
  /* Update re and im for the next iteration. */
  double cosphi = *cosPhase;
  double sinphi = *sinPhase;

  double newRe = cosphi + cosphi*dre - sinphi*dim;
  double newIm = sinphi + cosphi*dim + sinphi*dre;

  *cosPhase = newRe;
  *sinPhase = newIm;

}
void SineGaussian(double *hs, double *sigpar, int N, int flag, double Tobs)
{
  double f0, t0, Q, sf, sx, Amp;
  double fmax, fmin;//, fac;
  double phi, f;
  double tau;
  double re,im;

  int i,istart,istop,imin,imax;

  t0  = sigpar[0];
  f0  = sigpar[1];
  Q   = sigpar[2];
  Amp = sigpar[3];//exp(sigpar[3]);
  phi = sigpar[4];

  SineGaussianBandwidth(sigpar,&fmin,&fmax);

  tau = Q/(TPI*f0);

  //Make sure frequency is in range
  if(fmin < 0.0)
    fmin=0.0;
  if(fmax > (double)(N/2-1)/Tobs)
    fmax = (double)(N/2-1)/Tobs;

  imin = (int)(fmin*Tobs);
  imax = (int)(fmax*Tobs);

  //fac = sqrt(Tobs);

  if(flag==0)
  {
    istart = 0;
    istop  = N/2;
  }
  else
  {
    istart = imin;
    istop  = imax;
  }

  /* Use recursion relationship for cos(Phase) & sin(Phase) */

  //initial values of exp(iPhase)
  double phase = TPI*fmin*t0;
  double cosPhase_m  = cos(phase-phi);
  double sinPhase_m  = sin(phase-phi);
  double cosPhase_p  = cos(phase+phi);
  double sinPhase_p  = sin(phase+phi);

  //incremental values of exp(iPhase)
  double dim = -sin(TPI*t0/Tobs);
  double dre = -2.0*sin(0.5*(TPI*t0/Tobs))*sin(0.5*(TPI*t0/Tobs));

  double amplitude = (Amp)*RTPI/2.0*tau;//(Amp/fac)*RTPI/2.0*tau;
  double pi2tau2   = PI*PI*tau*tau;
  double Q2        = Q*Q/f0;

  for(i = istart; i < istop; i++)
  {
    f = (double)(i)/Tobs;

    if(flag==0) hs[2*i] = hs[2*i+1] = 0.0;

    if(i >= imin && i <= imax)
    {
      sf = amplitude*exp(-pi2tau2*(f-f0)*(f-f0));
      sx = exp(-Q2*f);
      re = sf*(cosPhase_m+sx*cosPhase_p);//cos(phi-TPI*f*(t0-Tobs));
      im = sf*(sinPhase_m+sx*sinPhase_p);//sin(phi-TPI*f*(t0-Tobs));

      switch(flag)
      {
        case 1: // Add new wavelet to hs[]
          hs[2*i]   += re;
          hs[2*i+1] += im;
          break;
        case -1: // Remove new wavelet from hs[]
          hs[2*i]   -= re;
          hs[2*i+1] -= im;
          break;
        case 0:  // Replace hs[] with new wavelet
          hs[2*i]    = re;
          hs[2*i+1]  = im;
          break;
        default:
          KILL("Unsupported SineGaussian flag\n");
          break;
      }

      /* Now update re and im for the next iteration. */
      recursive_phase_evolution(dre, dim, &cosPhase_m, &sinPhase_m);
      recursive_phase_evolution(dre, dim, &cosPhase_p, &sinPhase_p);
    }

  }
}

void SineGaussianBandwidth(double *sigpar, double *fmin, double *fmax)
{

  double f0 = sigpar[1];
  double Q  = sigpar[2];

  double tau = Q/(TPI*f0);

  *fmax = f0 + 3.0/(tau);  // no point evaluating waveform past this time (many efolds down)
  *fmin = f0 - 3.0/(tau);  // no point evaluating waveform before this time (many efolds down)

}

double network_snr(int imin, int imax, double **h, double **psd, double *eta, int NI )
{
  return sqrt( network_nwip(imin,imax,h,h,psd,eta,NI) );
}

double detector_snr(int imin, int imax, double *g, double *psd, double eta)
{
  return sqrt( fourier_nwip(imin,imax,g,g,psd)/eta );
}

double wavlet_proposal_density(int ifo, double f, double t, struct Data *data)
{
  int i, j;
  double dt, df;

  dt = data->Tobs/(double)(data->tsize);
  df = 0.5/dt;

  i = (int)(f/df);
  j = (int)(t/dt);

  /*
   ifo coming from EvolveIntrinsicParameters
   ranges from [-1:NI-1], w/ -1 corresponding
   to geocenter.

   TFdensity ranges from [0:NI] w/ 0 corresponding
   to geocenter
   */
  //return(log(data->TFdensity[ifo+1][i][j]));
  return(-log((double)(data->N)));
}

static void normalization(double **params, double **range, double *norm, int *larray, int cnt)
{
  int i, j;
  double tau;
  double tmin, tmax, sigma_t_thin, sigma_t_fat;
  double fmin, fmax, sigma_f_thin, sigma_f_fat;

  for(i=1; i<= cnt; i++)
  {

    j = larray[i];

    tau = params[j][2]/(LAL_TWOPI*params[j][1]);

    sigma_t_fat  = SPRD*tau;
    sigma_t_thin = HLW*tau;

    sigma_f_fat  = SPRD/LAL_PI/tau;
    sigma_f_thin = HLW/LAL_PI/tau;

    tmin = (range[0][0]-params[j][0]);
    tmax = (range[0][1]-params[j][0]);
    fmin = (range[1][0]-params[j][1]);
    fmax = (range[1][1]-params[j][1]);

    norm[j] = 1./(gaussian_norm(tmin,tmax,sigma_t_fat)*gaussian_norm(fmin,fmax,sigma_f_fat) - gaussian_norm(tmin,tmax,sigma_t_thin)*gaussian_norm(fmin,fmax,sigma_f_thin));
  }
  
}



double wavelet_proximity_density(double f, double t, double **params, int *larray, int cnt, struct Prior *prior)
{
  int jj, kk;
  double tau;
  double sigwt, signt, sigwf, signf;
  double sigwt2, signt2, sigwf2, signf2;
  double df, dt, df2, dt2;
  double pden;

  double **range = prior->range;
  double TFV = prior->TFV;

  // distribution we are drawing from is hollowed out 2-d Gaussian


  // we give equal weight to each wavelet (equally likely to draw from any of them)
  // so have to sum up densities for each wavelet and divide by the number of wavelets

  double *norm = NULL;


  // if no other active wavelets, the draw would have been from the full TF volume
  if(cnt == 0)
  {
    pden = 1.0/TFV;
  }
  else
  {

    //get normalization for prior
    norm = malloc( (prior->smax+1)*sizeof(double)  );
    normalization(params, range, norm, larray, cnt);


    pden = 0.0;

    // loop over the the existing wavelets
    for(jj=1; jj<= cnt; jj++)
    {

      kk = larray[jj];

      df = fabs(f-params[kk][1]);
      dt = fabs(t-params[kk][0]);

      tau = params[kk][2]/(TPI*params[kk][1]);

      // printf("%f %f %f %f %f %f\n", params[kk][1], params[kk][0], dt, df, 6.0*tau, 2.0/tau);

      sigwt = SPRD*tau;
      sigwf = SPRD/(PI*tau);

      signt = HLW*tau;
      signf = HLW/(PI*tau);


      if(dt < 4.0*sigwt && df < 4.0*sigwf)   // don't bother adding in contribution wavelets that are too far away
      {

        sigwt2 = sigwt*sigwt;
        signt2 = signt*signt;

        sigwf2 = sigwf*sigwf;
        signf2 = signf*signf;

        df2 = df*df;
        dt2 = dt*dt;

        pden += norm[kk]*(exp(-df2/(2.0*sigwf2)-dt2/(2.0*sigwt2))-exp(-df2/(2.0*signf2)-dt2/(2.0*signt2)));

      }

    }

    pden /= (double)(cnt);

    pden *= (1.0-UNIFORM);

    pden += UNIFORM/TFV;   // the uniform fraction

    free(norm);

  }


//  printf("wavelet_proximity_density returning %g\n",pden);

  return pden;
}

void wavelet_proximity(double *paramsy, double **params, int *larray, double **range, int cnt, long *seed)
{
  int i, jj, kk;
  double alpha;
  double tau;
  double sigwt, signt, sigwf, signf;
  double sigwt2, signt2, sigwf2, signf2;
  double df, dt, df2, dt2;
  double rat;


  alpha = ran2(seed);

  if(alpha < UNIFORM || cnt == 0)  // 100% uniform if no other pixels lit
  {
    for(i=0; i<= 1; i++)
    {
      paramsy[i] = range[i][0] + (range[i][1]-range[i][0])*ran2(seed);
    }
  }
  else
  {

    // select one of the existing pixels and propose near it
    jj = (int)(ran2(seed)*(double)(cnt))+1;

    kk = larray[jj];

    tau = params[kk][2]/(TPI*params[kk][1]);

    // scaling have been chosen to give distances of about 1 to 5 ish

    sigwt = SPRD*tau;
    signt = HLW*tau;
    sigwt2 = sigwt*sigwt;
    signt2 = signt*signt;

    sigwf = SPRD/(PI*tau);
    signf = HLW/(PI*tau);
    sigwf2 = sigwf*sigwf;
    signf2 = signf*signf;

    // distribution we are drawing from is hollowed out Gaussian
    // 1/(sqrt(2 Pi)(sigmaW-sigmaN))*(exp(-x^2/2 sigmaW^2) - exp(-x^2/2 sigmaN^2))

    // rejection sample to find t2, f2
    do
    {
      df = sigwf*gasdev2(seed);
      df2 = df*df;

      dt = sigwt*gasdev2(seed);
      dt2 = dt*dt;

      rat = 1.0-exp(df2*(1.0/(2.0*sigwf2) - 1.0/(2.0*signf2)))*exp(dt2*(1.0/(2.0*sigwt2) - 1.0/(2.0*signt2)));

      alpha = ran2(seed);

      paramsy[1] = params[kk][1]+df;
      paramsy[0] = params[kk][0]+dt;

      if
        (paramsy[0]<range[0][0] || paramsy[0]>range[0][1]) rat = -1.0;
      else if
        (paramsy[1]<range[1][0] || paramsy[1]>range[1][1]) rat = -1.0;

    } while(alpha > rat);
  }
  return;
}

void wavelet_sample(double *params, double **range, double **prop, int tsize, double Tobs, double pmax, long *seed)
{
  int i, j;
  double f, t, alpha;
  double df, dt;

  // Note: it is possible to speed the rejection sampling up by separating out the high density pixels
  // and the uniform piece.

  dt = Tobs/(double)tsize;
  df = 0.5/dt;

  // k = 0;

  do
  {
    f = range[1][0]+(range[1][1]-range[1][0])*ran2(seed);
    t = range[0][0]+(range[0][1]-range[0][0])*ran2(seed);

    i = (int)(f/df);
    j = (int)(t/dt);

    alpha = 2.*pmax*ran2(seed);
    //printf("%d %d %f %f %f %f\n", i, j, f, t, alpha, prop[i][j]);
    // k++;
  } while(alpha > prop[i][j]);

  // if(k > 100) printf("%d\n", k);

  params[0] = t;
  params[1] = f;

  return;
}

static double proximity_prior_density(double t, double f, double **params, int *larray, double *norm, int cnt, int exc, double TFV)
{
  int jj, kk, cx;
  double tau;
  double sigwt, signt, sigwf, signf;
  double sigwt2, signt2, sigwf2, signf2;
  double df, dt, df2, dt2;
  double pden, uf;

  // To maintain the same contrast between regions away from the cluster and in the cluster, the
  // uniform fraction has to go down as we shrink the TF volume. Denisty ratio is TFV*alpha*(1-uf)/uf

  uf = 1.0/(3776.0/TFV+1.0);

  // if no other active wavelets, then prior is uniform in TF volume
  if(cnt <= 1)
  {
    pden = 1.0/TFV;
  }
  else
  {

    pden = 0.0;
    cx = 0;

    // loop over the the existing wavelets
    for(jj=1; jj<= cnt; jj++)
    {

      if(jj != exc)  // the prior does not include the density from the wavelet itself
      {

        cx++;

        kk = larray[jj];

        df = fabs(f-params[kk][1]);
        dt = fabs(t-params[kk][0]);

        tau = params[kk][2]/(TPI*params[kk][1]);

        sigwt = SPRD*tau;
        sigwf = SPRD/(PI*tau);


        if(dt < 6.0*sigwt && df < 6.0*sigwf)   // don't bother adding in contribution from wavelets that are too far away
        {

          signt = HLW*tau;
          sigwt2 = sigwt*sigwt;
          signt2 = signt*signt;

          signf = HLW/(PI*tau);
          sigwf2 = sigwf*sigwf;
          signf2 = signf*signf;

          df2 = df*df;
          dt2 = dt*dt;

          pden += norm[kk]*(exp(-df2/sigwf2-dt2/sigwt2) -exp(-df2/signf2-dt2/signt2));
        }

      }

    }

    pden /= (double)(cx);

    pden *= (1.0-uf);

    pden += uf/TFV;   // the uniform fraction

  }

  return pden;
}

double wavelet_proximity_prior(struct Wavelet *wave, struct Prior *prior)
{
  int j;
  double logp = 0.0;

  // unpack wave structure
  int cnt = wave->size;
  int *larray  = wave->index;
  double **params = wave->intParams;
  double TFV = prior->TFV;

  double *norm = malloc((prior->smax+1)*sizeof(double));


  //get normalization for prior
  normalization(params, prior->range, norm, larray, cnt);

  for(j=1; j<= cnt; j++) logp += log(proximity_prior_density(params[larray[j]][0], params[larray[j]][1], params, larray, norm, cnt, j, TFV));

  //correct for bias caused by clustering prior
  logp -= prior->bias[cnt];

  free(norm);

  return logp;
}


void draw_wavelet_params(int ifo, double *params, struct Data *data, double **range, long *seed)
{
  int i;

  // Draw the rest from uniform distribution
  for(i=0; i<=4; i++)
  {
    params[i] = range[i][0] + (range[i][1]-range[i][0])*ran2(seed);
  }

  return;
}

void draw_time_frequency(int ifo, int ii, struct Wavelet *wx, struct Wavelet *wy, struct Data *data, double **range, long *seed, double fr)
{
  if(ran2(seed)<fr)
    wavelet_sample(wy->intParams[ii],range,wx->TFdensity,data->tsize, data->Tobs, wx->TFdensityMax,seed);
  else
    wavelet_proximity(wy->intParams[ii],wx->intParams,wy->index,range,wx->size, seed);
}

double TFprop(double f, double t, double **prop, int tsize, int fsize, double Tobs)
{
  int i, j;
  double dt, df, alpha;

  dt = Tobs/(double)tsize;
  df = 0.5/dt;

  i = (int)(f/df);
  j = (int)(t/dt);

  alpha = -1.0e60;

  if((i > -1) && (j > -1) && (i < fsize) && (j < tsize)) alpha = prop[i][j];

  return(alpha);
}


double glitch_amplitude_prior(double *params, double *Snf, double Sa, double Tobs, double SNRpeak)
{
  int i;
  double SNR, alpha;

  //TODO:  HACK!!! no PSD scale parameters in glitch_amplitude_prior()
  Sa = 1.0;


  // x/a^2 exp(-x/a) prior on SNR. Peaks at x = a. Good choice is a=5

  i = (int)(params[1]*Tobs);

  //SNR defined with Sn(f) but Snf array holdes <n_i^2>
  SNR = params[3]*sqrt((params[2]/(2.0*RT2PI*params[1]))/(Sa*Snf[i]*2.0/Tobs));
  alpha = log(SNR/(SNRpeak*SNRpeak))+(-SNR/SNRpeak);

  return(alpha);

}

double signal_amplitude_prior(double *params, double *Snf, double Sa, double Tobs, double SNRpeak)
{
  int i;
  double SNR, alpha;
  double SP3, SP5, S5;

  //TODO:  HACK!!! no PSD scale parameters in signal_amplitude_prior()
  Sa = 1.0;

  SP3 = SNRpeak*SNRpeak*SNRpeak;
	SP5 = SP3*SNRpeak*SNRpeak;

  // x/a^2 exp(-x/a) prior on SNR. Peaks at x = a. Good choice is a=5

  i = (int)(params[1]*Tobs);

  //SNR defined with Sn(f) but Snf array holdes <n_i^2>
  SNR = params[3]*sqrt((params[2]/(2.0*RT2PI*params[1]))/(Sa*Snf[i]*2.0/Tobs));

  S5 = SNR*SNR*SNR*SNR*SNR;

  alpha = log(APN*SNR*SP3/(4.0*SP5+S5));

  return(alpha);

}


void draw_glitch_amplitude(double *params, double *Snf, double Sa, long *seed, double Tobs, double SNRpeak)
{
  int i, k;
  double SNR, den, alpha;
  double max;

  //TODO:  HACK!!! no PSD scale parameters in draw_glitch_amplitude()
  Sa = 1.0;

  // x/a^2 exp(-x/a) prior on SNR. Peaks at x = a. Good choice is a=5

  max = 1.0/(SNRpeak*E);

  // rejection sample. Envelope function on runs out to ten times the peak
  // don't even bother putting in this minute correction to the normalization
  // (it is a 5e-4 correction).

  k = 0;
  do
  {

    SNR = 10.0*SNRpeak*ran2(seed);

    den = SNR/(SNRpeak*SNRpeak)*exp(-SNR/SNRpeak);

    den /= max;

    alpha = ran2(seed);

    k++;

  } while(alpha > den);

  if(k > 100) printf("%d\n", k);

  i = (int)(params[1]*Tobs);

  //SNR defined with Sn(f) but Snf array holdes <n_i^2>
  params[3] = SNR/sqrt((params[2]/(2.0*RT2PI*params[1]))/(Sa*Snf[i]*2.0/Tobs));

}

void draw_signal_amplitude(double *params, double *Snf, double Sa, long *seed, double Tobs, double SNRpeak)
{
  int i, k;
  double SNR, den, alpha;
  double max;

  double SP3, SP5, S5;

	SP3 = SNRpeak*SNRpeak*SNRpeak;
	SP5 = SP3*SNRpeak*SNRpeak;

  //TODO:  HACK!!! no PSD scale parameters in draw_signal_amplitude()
  Sa = 1.0;

  // x/a^2 exp(-x/a) prior on SNR. Peaks at x = a. Good choice is a=5

  max = APN/(SNRpeak*5.0);

  // rejection sample. Envelope function on runs out to ten times the peak
  // don't even bother putting in this minute correction to the normalization
  // (it is a 5e-4 correction).

  k = 0;
  do
  {

    SNR = 20.0*SNRpeak*ran2(seed);

    S5 = SNR*SNR*SNR*SNR*SNR;

		den = APN*SNR*SP3/(4.0*SP5+S5);

    den /= max;

    alpha = ran2(seed);

    k++;

  } while(alpha > den);

  //if(k > 100) printf("%d\n", k);

  i = (int)(params[1]*Tobs);

  //SNR defined with Sn(f) but Snf array holdes <n_i^2>
  params[3] = SNR/sqrt((params[2]/(2.0*RT2PI*params[1]))/(Sa*Snf[i]*2.0/Tobs));


  return;
}


/* ********************************************************************************** */
/*                                                                                    */
/*                             Likelihood Functions                                   */
/*                                                                                    */
/* ********************************************************************************** */

void max_array_element(double *max, int *index, double *array, int n)
{
  int i;

  *max = array[0];
  *index = 0;

  for(i = 1; i <= n-1; i++)
  {
    if(array[i] > *max)
	  {
	    *max = array[i];
	    *index = i;
	  }
  }
}

void phase_blind_time_shift(double *corr, double *corrf, double *data1, double *data2, double *psd, int n)
{
  int nb2, i, l, k;

  nb2 = n / 2;

  corr[0] = 0.0;
  corr[1] = 0.0;
  corrf[0] = 0.0;
  corrf[1] = 0.0;

  for (i=1; i < nb2; i++)
  {
    l=2*i;
    k=l+1;

    corr[l]	= ( data1[l]*data2[l] + data1[k]*data2[k]) / psd[i];
    corr[k]	= -( data1[k]*data2[l] - data1[l]*data2[k]) / psd[i];
    corrf[l] = ( data1[l]*data2[k] - data1[k]*data2[l]) / psd[i];
    corrf[k] = -( data1[k]*data2[k] + data1[l]*data2[l]) / psd[i];
  }

  drealft(corr-1, n, -1);
  drealft(corrf-1, n, -1);

}

void computeProjectionCoeffs(struct Data *data, struct Network *projection, double *extParams)
{
  int i,re,im;

  // get needed constants from data structure
  int N  = data->N;
  int NI = data->NI;
  double Tobs = data->Tobs;
  // network time-delays and antennae patterns
  int ifo;

  double df,freq;
  double toff;
  double dtimes[4];
  double Fplus, Fcross;

  // common extrinsic parameters
  //[0] alpha, [1] sin(delta) [2] psi [3] ellipticity [4] scale [5] dt [6] phi0

  double alpha    = extParams[0];
  double sindelta = extParams[1];
  double psi      = extParams[2];
  double dt       = extParams[5];
  double phi0     = extParams[6];

  double cosPhase;
  double sinPhase;

  double dim,dre;

  for(ifo = 0; ifo<NI; ifo++)
  {
    //Get time offsets for detectors
    dtimes[ifo] = XLALTimeDelayFromEarthCenter(data->detector[ifo]->location, alpha, asin(sindelta), &data->epoch);

    //Calculate antenna pattern once for each detector
    XLALComputeDetAMResponse(&Fplus, &Fcross, (const REAL4(*)[3])data->detector[ifo]->response, alpha, asin(sindelta), psi, XLALGreenwichMeanSiderealTime(&data->epoch));

    //initialize times
    toff	=  dt + dtimes[ifo];					//offset time for detector ifo

    //store the timedelay to each IFO for use elsewhere
    projection->deltaT[ifo] = toff;

    //calculate frequency domain waveforms:
    freq  = 0.0;							// initialize frequency
    df    = 1./Tobs;

    //Use recursion relationship for cos(Phase) & sin(Phase)
    cosPhase = cos(LAL_TWOPI*freq*toff+phi0);
    sinPhase = sin(LAL_TWOPI*freq*toff+phi0);

    /* Incremental values, using cos(theta) - 1 = -2*sin(theta/2)^2 */
    dim = -sin(LAL_TWOPI*toff/Tobs);
    dre = -2.0*sin(0.5*(LAL_TWOPI*toff/Tobs))*sin(0.5*(LAL_TWOPI*toff/Tobs));


    for(i=0; i<N/2; i++)
    {
      re = 2*i;
      im = re+1;

      projection->expPhase[ifo][re] = cosPhase;
      projection->expPhase[ifo][im] = sinPhase;

      recursive_phase_evolution(dre, dim, &cosPhase, &sinPhase);
    }

    projection->Fplus[ifo]  = Fplus;
    projection->Fcross[ifo] = Fcross;

  }// end loop over detectors

}

void waveformProject(struct Data *data, struct Network *projection, double *extParams, double **response, double *geo, double fmin, double fmax)
{
  double hpi, hpr, hci, hcr;
  double Fplus, Fcross;

  double df, freq;
  double sn, cn;

  int i, id;

  double ecc      = extParams[3];
  double scale    = extParams[4];

  int n = data->N;
  double Tobs = data->Tobs;
  int NI = data->NI;
  fmin = data->fmin;
  fmax = data->fmax;

  df = 1./(Tobs);

  for(id = 0; id<NI; id++)
  {

    Fplus  = projection->Fplus[id];
    Fcross = projection->Fcross[id];

    freq  = 0.0;							// initialize frequency

    for(i=0; i<n/2; i++)
    {
      if((freq< fmin) || (freq > fmax))		    // frequency below LIGO band or above termination point
      {
        response[id][2*i] = 0.0;
        response[id][2*i+1] = 0.0;
      }
      else								// frequency detectable and present
      {
        cn = projection->expPhase[id][2*i];
        sn = projection->expPhase[id][2*i+1];

        hpr = scale*(geo[2*i]*cn - geo[2*i+1]*sn);
        hpi = scale*(geo[2*i]*sn + geo[2*i+1]*cn);

        hcr = -ecc*hpi;
        hci =  ecc*hpr;

        response[id][2*i]   = (Fplus*hpr + Fcross*hcr);
        response[id][2*i+1] = (Fplus*hpi + Fcross*hci);

      }
      freq += df;
    }
  }   // end loop over detectors
}

void compute_reconstructed_plus_and_cross(struct Data *data, struct Network *projection, double *extParams, double **hrecPlus, double **hrecCross, double *geo, double fmin, double fmax)
{
  double hpi, hpr, hci, hcr;

  double df, freq;
  double sn, cn;

  int i, id;

  double ecc      = extParams[3];
  double scale    = extParams[4];

  int n = data->N;
  double Tobs = data->Tobs;
  int NI = data->NI;
  fmin = data->fmin;
  fmax = data->fmax;

  df = 1./(Tobs);

  for(id = 0; id<NI; id++)
  {

    freq  = 0.0;              // initialize frequency

    for(i=0; i<n/2; i++)
    {
      if((freq< fmin) || (freq > fmax))        // frequency below LIGO band or above termination point
      {
        hrecPlus[id][2*i] = 0.0;
        hrecPlus[id][2*i+1] = 0.0;
        hrecCross[id][2*i] = 0.0;
        hrecCross[id][2*i+1] = 0.0;
      }
      else                // frequency detectable and present
      {
        cn = projection->expPhase[id][2*i];
        sn = projection->expPhase[id][2*i+1];

        hpr = scale*(geo[2*i]*cn - geo[2*i+1]*sn);
        hpi = scale*(geo[2*i]*sn + geo[2*i+1]*cn);

        hcr = -ecc*hpi;
        hci = ecc*hpr;

        hrecPlus[id][2*i]   = hpr;
        hrecPlus[id][2*i+1] = hpi;
        hrecCross[id][2*i]   = hcr;
        hrecCross[id][2*i+1] = hci;

      }
      freq += df;
    }
  }   // end loop over detectors
}


void waveformTrivialProject(double **response, double *geo, int N, int NI)
{
  int i,ifo;
  for(ifo=0; ifo<NI; ifo++)  for(i=0; i<N; i++) response[ifo][i]=geo[i];
}

double loglike(int imin, int imax, double *r, double Sn, double *Snf)
{
  /*
   logL = -(r|r)/2 - Nlog(Sn)
   */
  return -0.5*(fourier_nwip(imin, imax, r, r, Snf)/Sn) - (double)(imax-imin)*log(Sn);
}

double loglike_maxNET(struct Data *data, struct Prior *prior, double **sNET, double **hNET, double *SnNET, double t0, double *dt, double *dphi, double *dA)
{
  int N = data->N;
  int NI = data->NI;
  int imin = data->imin;
  int imax = data->imax;
  double **SnfNET = data->Snf;
  double Tobs = data->Tobs;

  int tmin = (int)floor(prior->range[0][0]/(Tobs/(double)N));
  int tmax = (int)floor(prior->range[0][1]/(Tobs/(double)N));

	int i;
  int ifo;

	double logL;
  double norm, delt, pshift;


	//Chi^2 = (s-h|s-h)/Sn
  double *h = NULL;
  double *s = NULL;
  double *Snf = NULL;
  double Sn;

  double rhoNET = 0.0;
  double SSNET  = 0.0;
  double HSNET  = 0.0;

  logL = 0.0;
  for(ifo=0; ifo<NI; ifo++)
  {
    h   = hNET[ifo];
    s   = sNET[ifo];
    Snf = SnfNET[ifo];
    Sn  = SnNET[ifo];

    rhoNET += fourier_nwip(imin,imax,h,h,Snf)/Sn;
    SSNET  += fourier_nwip(imin,imax,s,s,Snf)/Sn;

  }


  //Create PSD scaled by Sn parameter for Sum_Extreme
  double **Snf_scaled = dmatrix(0,NI-1,0,N/2-1);
  for(ifo=0; ifo<NI; ifo++) for(i=0; i<N/2; i++) Snf_scaled[ifo][i] = SnfNET[ifo][i]*SnNET[ifo];

  /*
   HS = Single_Extreme(s, h, Snf, N, &delt, &pshift);
   HS *= 4.0/Sn;
   */

  HSNET = Sum_Extreme(sNET, hNET, Snf_scaled, N, &delt, &pshift, Tobs, NI, tmin, tmax, t0);
  free_dmatrix(Snf_scaled,0,NI-1,0,N/2-1);

  norm = HSNET/rhoNET;

  logL = -0.5*(SSNET - HSNET*HSNET/rhoNET);

  //Amplitude
  //params[3] *= norm;
  *dA = norm;

  //Time
  //params[0] += delt;
  *dt = -delt;

  //Phase
  //params[4] -= pshift;
  //*dphi = -pshift;
  *dphi = pshift;

  /*
   if(params[4] < 0.0) params[4] += TPI;
   if(params[4] > TPI) params[4] -= TPI;

   if(params[0] < 0.0) params[0] += Tobs;
   if(params[0] > Tobs) params[0] -= Tobs;
   */
  for(ifo=0; ifo<NI; ifo++) logL -= (double)(imax-imin)*log(SnNET[ifo]);

  if(logL!=logL)
  {
    printf("encountered nan in loglike_maxNet:\n");
    for(ifo=0; ifo<NI; ifo++)printf("     SnNET[%i]=%g\n",ifo,SnNET[ifo]);
    printf("     HSNET = %g\n",HSNET);
    printf("     SSNET = %g\n",SSNET);
    printf("     rhoNET = %g\n",rhoNET);
    FILE *dump = fopen("loglikedump.dat","w");

    for(i=imin; i<imax; i++)
    {
      fprintf(dump,"%i ",i);
      for(ifo=0; ifo<NI; ifo++)
      {
        h   = hNET[ifo];
        s   = sNET[ifo];
        Snf = SnfNET[ifo];
        Sn  = SnNET[ifo];

        fprintf(dump,"%.12g %.12g %.12g %.12g %.12g ",h[i*2],h[i*2+1],s[i*2],s[i*2+1],Snf[i]);
      }
      fprintf(dump,"\n");
    }
    fclose(dump);
    abort();
  }
	return logL;
}

double Sum_Extreme(double **a, double **b, double **Sn, int n, double *delt, double *pshift, double Tobs, int NI, int imin, int imax, double t0)
{
  double max;
  int i, j;
  int index;
  double **AC, **AF;
  double *corr;
  double *corrcos, *corrsin;

  double srate = (Tobs/(double)n);
  int newtime = 0;
  int oldtime = (int)floor(t0/srate);

  // this version simultaneously maximizes across the Network

  AC=dmatrix(0,NI-1,0,n-1);
  AF=dmatrix(0,NI-1,0,n-1);
  corr = dvector(0,n-1);
  corrcos = dvector(0,n-1);
  corrsin = dvector(0,n-1);

  for(i = 0; i < n; i++)
  {
    corr[i] = 0.0;
    corrcos[i] = 0.0;
    corrsin[i] = 0.0;
  }

  for(i = 0; i < NI; i++)  phase_blind_time_shift(AC[i], AF[i], a[i], b[i], Sn[i], n);

  for(j = 0; j < NI; j++)
  {
    for(i = 0; i < n; i++)
    {
      corrcos[i] += AC[j][i];
      corrsin[i] += AF[j][i];
    }
  }
  for(i = 0; i < n; i++) corr[i] = sqrt(corrcos[i]*corrcos[i]+corrsin[i]*corrsin[i]);

  while(newtime<imin || newtime >= imax)
  {
    //TODO: Might be worth sorting corr[] in Sum_Extreme()
    max_array_element(&max, &index, corr, n);

    if(index < (n/2)-1)
      *delt = ((double) index)/((double) n)*Tobs;
    else if(index >= n/2)
      *delt = ((double) (index - n))/((double) n)*Tobs;

    newtime = oldtime + (*delt)/srate;
    corr[index]=0.0;
  }
  max = 2.0*(max);
  *pshift = atan2(corrsin[index],corrcos[index]);

  free_dmatrix(AC, 0,NI-1,0,n-1);
  free_dmatrix(AF, 0,NI-1,0,n-1);
  free_dvector(corr, 0,n-1);
  free_dvector(corrcos, 0,n-1);
  free_dvector(corrsin, 0,n-1);

  return(max);
}

double EvaluateSearchLogLikelihood(int typ, int ii, int det, struct Model *mx, struct Model *my, struct Prior *prior, struct Chain *chain, struct Data *data)
{
  int i, ifo;
  double logLy = 0.0;
  double dt,dphi,dA;
  double fmin,fmax;

  //Unpack model_ & data_parameter structures
  int N  = data->N;
  int NI = data->NI;
  int imin = data->imin;
  int imax = data->imax;

  double Tobs  = data->Tobs;
  double **r   = data->r;
  double **s   = data->s;
  double **Snf = data->Snf;

  //work-space vectors
  double *h   = data->h;
  double **rh = data->rh;

  for(i=0; i<N; i++)
  {
    h[i]=0.0;
    for(ifo=0; ifo<NI; ifo++) rh[ifo][i]=0.0;
  }
  //pointer to model being varied (glitch or signal)
  struct Wavelet *wave_x = NULL;
  struct Wavelet *wave_y = NULL;
  if(det == -1)
  {
    wave_x = mx->signal;
    wave_y = my->signal;
  }
  else
  {
    wave_x = mx->glitch[det];
    wave_y = my->glitch[det];
  }

  //current model
  double **rx = mx->response;

  double *tx      = wave_x->templates;
  double *paramsx = wave_x->intParams[ii];

  //proposed model
  double *Sny = my->Sn;
  double **ry = my->response;

  int sy          = wave_y->size;
  double *ty      = wave_y->templates;
  double *paramsy = wave_y->intParams[ii];


  //short-cut pointer to glitch models
  double **gx = NULL;
  double **gy = NULL;
  gx = malloc(NI*sizeof(double*));
  gy = malloc(NI*sizeof(double*));

  for(ifo=0; ifo<NI; ifo++)
  {
    gx[ifo] = mx->glitch[ifo]->templates;
    gy[ifo] = my->glitch[ifo]->templates;
  }

  // Rejection sample on prior range
  if(wave_y->size>0 )
  {
    if(checkrange(paramsy,prior->range) )
    {
      free(gx);
      free(gy);
      return -1.0e60;
    }
  }

  // copy over current multi-template & current residual
  for(i=0; i< N; i++)
  {
    ty[i] = tx[i];
    for(ifo=0; ifo<NI; ifo++)
    {
      ry[ifo][i] = gy[ifo][i] = 0.0;
      r[ifo][i] = s[ifo][i];
      if(mx->signal->size>0)
      {
        ry[ifo][i] = rx[ifo][i];
        r[ifo][i] -= ry[ifo][i];
      }
      if(mx->glitch[ifo]->size>0)
      {
        gy[ifo][i] = gx[ifo][i];
        r[ifo][i] -= gy[ifo][i];
      }
    }
  }
  switch(typ)
  {
      // updating a single wavelet
    case 1:

      // we are in the noise-only model - can't use the maximized likelihood
      if(sy == 0)
      {
        for(ifo=0; ifo<NI; ifo++)
        {
          my->detLogL[ifo] = loglike(imin, imax, r[ifo], Sny[ifo], Snf[ifo]);
          logLy += my->detLogL[ifo];//loglike(imin, imax, r[ifo], Sny[ifo], Snf[ifo]);
        }

        free(gx);
        free(gy);

        return logLy;
      }
      else
      {
        // remove old template
        SineGaussian(ty, paramsx, N, -1, Tobs);
        if(det==-1)
        {
          SineGaussianBandwidth(paramsx,&fmin,&fmax);

          //Make sure frequency is in range
          if(fmin < data->fmin)
            fmin = data->fmin;
          if(fmax > data->fmax)
            fmax = data->fmax;

          waveformProject(data, my->projection, my->extParams, ry, ty, fmin, fmax);
        }
        // form up residual using all other wavelets
        for(ifo=0; ifo<NI; ifo++)
        {
          for(i=0; i< N; i++)
          {
            r[ifo][i] = s[ifo][i];//-ry[ifo][i]-gy[ifo][i];
            if(data->signalFlag) r[ifo][i] -= ry[ifo][i];
            if(data->glitchFlag) r[ifo][i] -= gy[ifo][i];
          }
        }
        // generate the new template
        SineGaussian(h, paramsy, N, 0, Tobs);
        if(det==-1)
        {
          SineGaussianBandwidth(paramsy,&fmin,&fmax);

          //Make sure frequency is in range
          if(fmin < data->fmin)
            fmin = 0.0;
          if(fmax > data->fmax)
            fmax = data->fmax;

          waveformProject(data, my->projection, my->extParams, rh, h, fmin, fmax);
        }
        else for(i=0; i<N; i++) rh[det][i]=h[i];
      }
      break;

      // adding a new wavelet
    case 2:

      // generate the new template
      SineGaussian(h, paramsy, N, 0, Tobs);

      if(det==-1)
      {
        /* Delayed Rejection for extrinsic parameters??
         if(chain->burn)
         {
         long seed = -(int)time(NULL);
         struct Model *temp = malloc(sizeof(struct Model));
         initialize_model(temp, data, prior, &seed);

         copy_model(my, temp, N);

         if(sy==1)extrinsic_uniform_proposal(&seed,temp->extParams);

         //add new template to model
         for(i=0; i<N; i++) temp->signal->templates[i] += h[i];

         EvolveExtrinsicParameters(data, prior, temp, chain, &seed, 100, 0);

         for(i=0; i<NE; i++) my->extParams[i] = temp->extParams[i];

         for(ifo=0; ifo<NI; ifo++)
         {
         my->projection->Fplus[ifo]  = temp->projection->Fplus[ifo];
         my->projection->Fcross[ifo] = temp->projection->Fcross[ifo];
         for(i=0; i<N; i++)
         {
         my->projection->expPhase[ifo][i] = temp->projection->expPhase[ifo][i];
         }
         }

         free_model(temp, data, prior);
         free(temp);

         }*/
        SineGaussianBandwidth(paramsy,&fmin,&fmax);
        //Make sure frequency is in range
        if(fmin < data->fmin)
          fmin = 0.0;
        if(fmax > data->fmax)
          fmax = data->fmax;

        waveformProject(data, my->projection, my->extParams, rh, h, fmin, fmax);
      }
      else for(i=0; i<N; i++) rh[det][i]=h[i];

      break;

      //remove existing wavelet
    case 3:
      logLy = EvaluateMarkovianLogLikelihood(typ, ii, det, mx, my, prior, chain, data);
      break;

    default:
      KILL("Unsupported case for SearchLogLikelihood()\n");
      break;
  }

  //If likelihood for new wavelet needs to be evaluated (either by birth or move)
  if(typ!=3)
  {
    // note the the residual here is the signal minus all the other SGs
    logLy = loglike_maxNET(data, prior, r, rh, Sny, paramsy[0], &dt, &dphi, &dA);

    //Amplitude
    //paramsy[3] *= dA;//+= log(dA);

    //Time
    paramsy[0] += dt;

    //Phase
    paramsy[4] += dphi;

//    printf("%g :",2.-paramsy[0]);
//    FILE *prop=fopen("prop.dat","a");
//    for(i=0;i<5;i++)fprintf(prop," %g ",paramsy[i]);fprintf(prop,"\n");
//    fclose(prop);
    //abort();

    if(paramsy[4] < 0.0) paramsy[4] += TPI;
    if(paramsy[4] > TPI) paramsy[4] -= TPI;

    if(paramsy[0] < 0.0) paramsy[0] += Tobs;
    if(paramsy[0] > Tobs) paramsy[0] -= Tobs;

    // check that we haven't mapped out of range
    if(wave_y->size>0)
    {
      if(checkrange(paramsy,prior->range))
      {
        free(gx);
        free(gy);
        return -1.0e60;
      }
    }
    // now include the newly proposed SG in the proposed combined template
    // note that we can't just use h since the parameters get updated by the maximization procedure
    SineGaussian(ty, paramsy, N, 1, Tobs);
    if(det==-1)
    {
      SineGaussianBandwidth(paramsy,&fmin,&fmax);

      //Make sure frequency is in range
      if(fmin < data->fmin)
        fmin = 0.0;
      if(fmax > data->fmax)
        fmax = data->fmax;

      waveformProject(data, my->projection, my->extParams, ry, ty, fmin, fmax);
    }
  }

  free(gx);
  free(gy);

  return logLy;
}

//double EvaluateConstantLogLikelihood(int typ, int ii, int det, struct Model *mx, struct Model *my, double **range, struct Data *data)
double EvaluateConstantLogLikelihood(int typ, int ii, int det, struct Model *mx, struct Model *my, struct Prior *prior, struct Chain *chain, struct Data *data)
{
  // Rejection sample on prior range
  struct Wavelet *wave = NULL;

  if(det==-1) wave = my->signal;
  else        wave = my->glitch[det];

  /*
   type 3 is a death move -- params[ii] is out of the picture
   and doesn't need to be checked
   */
  if(typ!=3 && wave->size > 0)
  {
    if( checkrange(wave->intParams[ii],prior->range)) return -1.0e60;
  }

  return 0.0;
}

//double EvaluateMarkovianLogLikelihood(int typ, int ii, int det, struct Model *mx, struct Model *my, double **range, struct Data *data)
double EvaluateMarkovianLogLikelihood(int typ, int ii, int det, struct Model *mx, struct Model *my, struct Prior *prior, struct Chain *chain, struct Data *data)
{
  int i,j,k,ifo,jj;
  double logLy = 0.0;
  double logLseg = 0.0;
  double fmin=data->fmin;
  double fmax=data->fmax;
  double fi,ff;

  //Unpack model_ & data_parameter structures
  int N  = data->N;
  int NI = data->NI;
  int imin = data->imin;
  int imax = data->imax;
  int jmin = imin;
  int jmax = imax;

  double **r   = data->r;
  double **s   = data->s;
  double **Snf = data->Snf;
  double Tobs  = data->Tobs;

  //pointer to model being varied (glitch or signal)
  struct Wavelet *wave_x = NULL;
  struct Wavelet *wave_y = NULL;
  if(det == -1)
  {
    wave_x = mx->signal;
    wave_y = my->signal;
  }
  else
  {
    wave_x = mx->glitch[det];
    wave_y = my->glitch[det];
  }

  //current model
  double **rx = mx->response;

  double *tx      = wave_x->templates;
  double *paramsx = wave_x->intParams[ii];

  //proposed model
  double *Sny = my->Sn;
  double **ry = my->response;

  int sy          = wave_y->size;
  double *ty      = wave_y->templates;
  double *paramsy = wave_y->intParams[ii];

  //short-cut pointer to glitch models
  double **gx = NULL;
  double **gy = NULL;
  gx = malloc(NI*sizeof(double*));
  gy = malloc(NI*sizeof(double*));

  for(ifo=0; ifo<NI; ifo++)
  {
    gx[ifo] = mx->glitch[ifo]->templates;
    gy[ifo] = my->glitch[ifo]->templates;

    my->detLogL[ifo] = mx->detLogL[ifo];

  }

  // Rejection sample on prior range
  /*
   type 3 is a death move -- params[ii] is out of the picture
   and doesn't need to be checked
   */
  if( (data->signalFlag || data->glitchFlag) && typ!=3 && wave_y->size>0 )
  {
    if(checkrange(paramsy,prior->range))
    {
      free(gx);
      free(gy);
      return -1.0e60;
    }
  }

  // copy over current multi-template
  for(i=0; i< N; i++)
  {
    ty[i] = tx[i];
    for(ifo=0; ifo<NI; ifo++)
    {
      r[ifo][i]  = s[ifo][i];
      if(mx->glitch[ifo]->size>0)
      {
        gy[ifo][i] = gx[ifo][i];
        r[ifo][i] -= gx[ifo][i];
      }
      if(mx->signal->size>0)
      {
        ry[ifo][i] = rx[ifo][i];
        r[ifo][i] -= rx[ifo][i];
      }
    }
  }

  switch(typ)
  {
      // Fixed dimension MCMC move
    case 1:
      if(sy > 0)
      {

        // subtract the current wavelet
        SineGaussian(ty, paramsx, N, -1, Tobs);
        //printf("1:paramsx={%g,%g,%g,%g,%g}\n",paramsx[0],paramsx[1],paramsx[2],paramsx[3],paramsx[4]);
        SineGaussianBandwidth(paramsx, &fmin,&fmax);

        // add the new wavelet
        SineGaussian(ty, paramsy, N, 1, Tobs);

        SineGaussianBandwidth(paramsy, &fi,&ff);
        //Make sure frequency is in range
        if(fi < fmin)
          fmin = fi;
        if(ff > fmax)
          fmax = ff;

        //Make sure frequency is in range
        if(fmin < data->fmin)
          fmin = data->fmin;
        if(fmax > data->fmax)
          fmax = data->fmax;

      }
      break;

      // RJMCMC birth move
    case 2:

      SineGaussian(ty, paramsy, N, 1, Tobs);

      SineGaussianBandwidth(paramsy, &fmin,&fmax);
      //Make sure frequency is in range
      if(fmin < data->fmin)
        fmin = data->fmin;
      if(fmax > data->fmax)
        fmax = data->fmax;

      break;

      // RJMCMC death move
    case 3:

      // subtract the current wavelet
      SineGaussian(ty, paramsx, N, -1, Tobs);

      SineGaussianBandwidth(paramsx, &fmin,&fmax);
      //Make sure frequency is in range
      if(fmin < data->fmin)
        fmin = data->fmin;
      if(fmax > data->fmax)
        fmax = data->fmax;

      break;

      //RJMCMC birth-death move
    case 4:

      //Some helper pointers
      if(det==-1) ifo = (int)floor(ran2(&chain->seed)*NI);

      struct Wavelet *wBirth = NULL;
      struct Wavelet *wDeath = NULL;

      if(det==-1)
      {
        wDeath = my->signal;
        wBirth = my->glitch[ifo];
      }
      else
      {
        wDeath = my->glitch[det];
        wBirth = my->signal;
      }

      // subtract the current wavelet from the signal model
      SineGaussian(wDeath->templates, wDeath->intParams[ii], N, -1, Tobs);
      SineGaussianBandwidth(wDeath->intParams[ii], &fmin, &fmax);
      //Make sure frequency is in range
      if(fmin < data->fmin)
        fmin = data->fmin;
      if(fmax > data->fmax)
        fmax = data->fmax;

      waveformProject(data,my->projection, my->extParams, ry, wDeath->templates, fmin, fmax);

      // where in glitch model do I put the new parameters?
      // find a label that isn't in use
      jj = 0;
      do
      {
        jj++;
        k = 0;
        for(j=0; j< wBirth->size; j++) if(jj == wBirth->index[j]) k = 1;
      } while(k == 1);

      wBirth->index[wBirth->size] = jj;

      wBirth->size++;

      // shift time of signal
      for(i=0; i<5; i++) wBirth->intParams[jj][i] = wDeath->intParams[ii][i];
      if(det==-1)
      {
        wBirth->intParams[jj][0] += my->projection->deltaT[ifo];
        wBirth->intParams[jj][3] *= sqrt(my->projection->Fplus[ifo]*my->projection->Fplus[ifo]+my->projection->Fcross[ifo]*my->projection->Fcross[ifo]);
        wBirth->intParams[jj][4] = ran2(&chain->seed)*TPI;
      }
      else wBirth->intParams[jj][0] -= my->projection->deltaT[ifo];

      // add wavelet
      SineGaussian(wBirth->templates, wBirth->intParams[jj], N, 1, Tobs);

      break;

    default:
      KILL("Unsupported MCMC move type\n");
      break;
  }
  //Make sure template array is zero if we have no wavelets in model
  if(sy == 0) for(i=0; i< N; i++) ty[i] = 0.0;

  // form up the resdial s-h
  if(det==-1 && data->signalFlag)
    waveformProject(data,my->projection, my->extParams, ry, ty, fmin, fmax);

  //try just computing deltaLogL for birth-death moves in glitch model
  if(typ==1 && data->intrinsic_likelihood == EvaluateMarkovianLogLikelihood)
  {
    logLy = 0.0;

    /*
     Compute logLy if we only change Sn:
     logLx = -(r|r)/2/Snx - NlogSnx
     logLy = -(r|r)/2/Sny - NlogSny
     -(r|r)/2 = (logLx + NlogSnx)*Snx
     so
     logLy = (logLx + NlogSnx)*(Snx/Sny) - NlogSny
     */
    for(ifo=0; ifo<NI; ifo++)
    {
      my->detLogL[ifo] = (mx->detLogL[ifo] + (imax-imin)*log(mx->Sn[ifo]))*(mx->Sn[ifo]/Sny[ifo]) -(imax-imin)*log(Sny[ifo]);
      logLy += my->detLogL[ifo];
    }

    if(sy > 0)
    {
      /*
       Now adjust logLy for changes in the glitch/signal parameters
       Keep the cost down by only computing deltaLogL in bins
       where modified wavelets have support
      */
      double fminx,fmaxx;
      double fminy,fmaxy;

      // subtract the current wavelet
      SineGaussianBandwidth(paramsx,&fminx,&fmaxx);

      // add the new wavelet
      SineGaussianBandwidth(paramsy,&fminy,&fmaxy);

      // make sure bandwidth spans both wavelets
      fmin = fminx;
      if(fminy<fmin) fmin = fminy;

      fmax = fmaxx;
      if(fmaxy>fmax) fmax = fmaxy;

      jmin = (int)floor(fmin*Tobs);
      jmax = (int)floor(fmax*Tobs);

      if(jmin<imin) jmin = imin;
      if(jmax>imax) jmax = imax;

      if(det==-1)
      {
        for(ifo=0; ifo<NI; ifo++)
        {
          //remove contribution between imin and imax of old model
          logLseg = loglike(jmin, jmax, r[ifo], Sny[ifo], Snf[ifo]);
          my->detLogL[ifo] -= logLseg;
          logLy -= logLseg;

          //add contribution between imin and imax of new model
          for(i=jmin; i<jmax; i++)
          {
            j = 2*i;
            k = j+1;
            r[ifo][j] = s[ifo][j];
            r[ifo][k] = s[ifo][k];
            if(data->glitchFlag)
            {
              r[ifo][j] -= gy[ifo][j];
              r[ifo][k] -= gy[ifo][k];
            }
            if(data->signalFlag)
            {
              r[ifo][j] -= ry[ifo][j];
              r[ifo][k] -= ry[ifo][k];
            }
          }
          logLseg = loglike(jmin, jmax, r[ifo], Sny[ifo], Snf[ifo]);
          my->detLogL[ifo] += logLseg;
          logLy += logLseg;
        }
      }
      else
      {
        //remove contribution between imin and imax of old model
        logLseg = loglike(jmin, jmax, r[det], Sny[det], Snf[det]);
        my->detLogL[det] -= logLseg;
        logLy -= logLseg;

        //add contribution between imin and imax of new model
        for(i=jmin; i<jmax; i++)
        {
          j = 2*i;
          k = j+1;
          r[det][j] = s[det][j] - ry[det][j] - gy[det][j];
          r[det][k] = s[det][k] - ry[det][k] - gy[det][k];
        }
        logLseg = loglike(jmin, jmax, r[det], Sny[det], Snf[det]);
        my->detLogL[det] += logLseg;
        logLy += logLseg;
      }
    }
  }
  else if((typ==2 || typ==3) && data->intrinsic_likelihood == EvaluateMarkovianLogLikelihood)
  {
    if(typ==2)
      SineGaussianBandwidth(paramsy,&fmin,&fmax);
    else
      SineGaussianBandwidth(paramsx,&fmin,&fmax);

    for(ifo=0; ifo<NI; ifo++) my->detLogL[ifo] = mx->detLogL[ifo];
    logLy = mx->logL;
    jmin = (int)floor(fmin*Tobs);
    jmax = (int)floor(fmax*Tobs);

    if(jmin<imin) jmin = imin;
    if(jmax>imax) jmax = imax;

    if(det==-1)
    {
      for(ifo=0; ifo<NI; ifo++)
      {
        my->detLogL[ifo] = mx->detLogL[ifo];

        //remove contribution between imin and imax of old model
        logLseg = loglike(jmin, jmax, r[ifo], mx->Sn[ifo], Snf[ifo]);
        my->detLogL[ifo] -= logLseg;
        logLy -= logLseg;

        //add contribution between imin and imax of new model
        for(i=jmin; i<jmax; i++)
        {
          j = 2*i;
          k = j+1;
          r[ifo][j] = s[ifo][j];// - ry[ifo][j] - gy[ifo][j];
          r[ifo][k] = s[ifo][k];// - ry[ifo][k] - gy[ifo][k];
          if(data->glitchFlag)
          {
            r[ifo][j] -= gy[ifo][j];
            r[ifo][k] -= gy[ifo][k];
          }
          if(data->signalFlag)
          {
            r[ifo][j] -= ry[ifo][j];
            r[ifo][k] -= ry[ifo][k];
          }
        }
        logLseg = loglike(jmin, jmax, r[ifo], Sny[ifo], Snf[ifo]);
        my->detLogL[ifo] += logLseg;
        logLy += logLseg;
      }
    }
    else
    {
      my->detLogL[det] = mx->detLogL[det];

      //remove contribution between imin and imax of old model
      logLseg = loglike(jmin, jmax, r[det], mx->Sn[det], Snf[det]);
      my->detLogL[det] -= logLseg;
      logLy -=  logLseg;

      //add contribution between imin and imax of new model
      for(i=jmin; i<jmax; i++)
      {
        j = 2*i;
        k = j+1;
        r[det][j] = s[det][j] - ry[det][j] - gy[det][j];
        r[det][k] = s[det][k] - ry[det][k] - gy[det][k];
      }
      logLseg = loglike(jmin, jmax, r[det], Sny[det], Snf[det]);
      my->detLogL[det] += logLseg;
      logLy += logLseg;
    }
  }
  else
  {
    logLy=0.0;
    for(ifo=0; ifo<NI; ifo++)
    {
      for(i=0; i< N; i++)
      {
        r[ifo][i] = s[ifo][i];
        //if(data->signalFlag || data->glitchFlag) r[ifo][i] -= ry[ifo][i] + gy[ifo][i];
        if(data->glitchFlag)
        {
          r[ifo][i] -= gy[ifo][i];
        }
        if(data->signalFlag)
        {
          r[ifo][i] -= ry[ifo][i];
        }
      }
      my->detLogL[ifo] = loglike(imin, imax, r[ifo], Sny[ifo], Snf[ifo]);
      logLy += my->detLogL[ifo];
    }
  }

  free(gx);
  free(gy);

  return logLy;

}

double EvaluateExtrinsicConstantLogLikelihood(struct Network *projection, double *params, double *Sn, double *geo, double **g, struct Data *data, double fmin, double fmax)
{
  if(extrinsic_checkrange(params)) return -1.0e60;

  else return 0.0;
}

double EvaluateExtrinsicMarkovianLogLikelihood(struct Network *projection, double *params, double *Sn, double *geo, double **g, struct Data *data, double fmin, double fmax)
{
	int i, n;
  int NI,N;
  int ifo;
  int imin,imax;

	//double sum = 0.0;
	double logL = 0.0;
	double **h, **d, **r, **psd;

  if(extrinsic_checkrange(params)) return -1.0e60;

  NI   = data->NI;
	N = data->N;
  imin = (int)floor(fmin*data->Tobs);//data->imin;
  imax = (int)floor(fmax*data->Tobs);//data->imax;

  d   = data->s;
	h   = dmatrix(0,NI-1,0,N-1); //template
	r   = dmatrix(0,NI-1,0,N-1); //residual
  psd = data->Snf;


	//initialize template and residual to 0's
	for(i=0; i<NI; i++) for(n=0; n<N; n++) h[i][n] = 0.0;

  //compute instrument response h[] to geocenter waveform geo[]
  computeProjectionCoeffs(data, projection, params);
  waveformProject(data, projection, params, h, geo, fmin, fmax);


	//Form up residual
	for(ifo=0; ifo<NI; ifo++)
  {
    for(n=0; n<N; n++)
    {
      r[ifo][n] = d[ifo][n];
      if(data->signalFlag) r[ifo][n] -= h[ifo][n];
      if(data->glitchFlag) r[ifo][n] -= g[ifo][n];
    }
    //sum	+= fourier_nwip(r[ifo], r[ifo], data->Snf[ifo], N)/Sn[ifo];
    logL += loglike(imin, imax, r[ifo], Sn[ifo], data->Snf[ifo]);
  }
  //logL = -0.5*sum;

	free_dmatrix(h,0,NI-1,0,N-1);
	free_dmatrix(r,0,NI-1,0,N-1);

	return logL;
}

double EvaluateExtrinsicSearchLogLikelihood(struct Network *projection, double *params, double *Sn, double *geo, double **g, struct Data *data, double fmin, double fmax)
{
	int i, n;
  int NI,N;
  int ifo;
  int imin,imax;

  double Tobs;
	double logL = 0.0;
  double dt=0,dphi=0,dA=1;
  double **h, **d, **r, **psd;

  if(extrinsic_checkrange(params)) return -1.0e60;

  NI   = data->NI;
	N = data->N;
  Tobs = data->Tobs;
  imin = (int)floor(fmin*data->Tobs);//data->imin;
  imax = (int)floor(fmax*data->Tobs);//data->imax;

  d   = data->s;
	h   = dmatrix(0,NI-1,0,N-1); //template
	r   = dmatrix(0,NI-1,0,N-1); //residual
  psd = data->Snf;


	//initialize template and residual to 0's
	for(i=0; i<NI; i++) for(n=0; n<N; n++) h[i][n] = 0.0;

  //compute instrument response h[] to geocenter waveform geo[]
  computeProjectionCoeffs(data, projection, params);
  waveformProject(data, projection, params, h, geo, data->fmin, data->fmax);

	//Form up residual
	for(ifo=0; ifo<NI; ifo++)
  {
    for(n=0; n<N; n++)
    {
      r[ifo][n] = d[ifo][n];
      if(data->signalFlag) r[ifo][n] -= h[ifo][n];
      if(data->glitchFlag) r[ifo][n] -= g[ifo][n];
    }
  }


  //logL = loglike_maxNET(data, prior, r, h, Sn, &dt, &dphi, &dA);

  //Amplitude[4]
  /*
   model->extParams[0] = 1.2;                   //right ascension
   model->extParams[1] = -0.3;                  //sin(declination)
   model->extParams[2] = 4.1;                   //polarization angle
   model->extParams[3] = -2.0*0.2/(1.0+0.2*0.2);//ellipticity
   model->extParams[4] = 1.0;                   //amplitude scale
   model->extParams[5] = 0.0;                   //geocenter time shift
   model->extParams[6] = 0.0;                   //geocenter phase shift
   */

  //Amplitude
  params[1] *= dA;

  //Time
  params[5] += dt;

  //Phase
  params[6] += dphi;

  if(params[6] < 0.0) params[6] += TPI;
  if(params[6] > TPI) params[6] -= TPI;

  if(params[5] < 0.0) params[5] += Tobs;
  if(params[5] > Tobs) params[5] -= Tobs;

  // check that we haven't mapped out of range
  if(extrinsic_checkrange(params)) return -1.0e60;

  //Now update the full model with current extrinsic parameters
  computeProjectionCoeffs(data, projection, params);
  waveformProject(data, projection, params, h, geo, fmin, fmax);
	//Form up residual
	for(ifo=0; ifo<NI; ifo++)
  {
    for(n=0; n<N; n++)
    {
      r[ifo][n] = d[ifo][n];
      if(data->signalFlag) r[ifo][n] -= h[ifo][n];
      if(data->glitchFlag) r[ifo][n] -= g[ifo][n];
    }
  }
  logL = 0.0;
  for(ifo=0; ifo<NI; ifo++) logL += loglike(imin, imax, r[ifo], Sn[ifo], data->Snf[ifo]);

	free_dmatrix(h,0,NI-1,0,N-1);
	free_dmatrix(r,0,NI-1,0,N-1);

	return logL;
}


/* ********************************************************************************** */
/*                                                                                    */
/*                                    Math tools                                      */
/*                                                                                    */
/* ********************************************************************************** */



double fourier_nwip(int imin, int imax, double *a, double *b, double *Sn)
{
	int i, j, k;
	double arg, product;
	double ReA, ReB, ImA, ImB;

	arg = 0.0;
	for(i=imin; i<imax; i++)
  {
		j = i * 2;
		k = j + 1;
		ReA = a[j]; ImA = a[k];
		ReB = b[j]; ImB = b[k];
		product = ReA*ReB + ImA*ImB;
		arg += product/(Sn[i]);
  }

	return(2.0*arg);
}

double network_nwip(int imin, int imax, double **a, double **b, double **Sn, double *eta, int NI)
{
	int i;
	double sum=0.0;

  for(i=0; i<NI; i++) sum += fourier_nwip(imin,imax,a[i],b[i],Sn[i])/eta[i];

	return(sum);
}

double gaussian_norm(double min, double max, double sigma)
{
  double sq2  = 1.4142135623731; //sqrt(2)
  double sqp2 = 1.2533141373155; //sqrt(pi/2)
  return sqp2*sigma*(erfc(min/sq2/sigma) - erfc(max/sq2/sigma) );
}

double gasdev2(long *idum)
{
	double ran2(long *idum);
	static int iset=0;
	static double gset;
	double fac, rsq, v1, v2;

	if(*idum < 0) iset = 0;
	if(iset == 0){
		do{
			v1 = 2.0 * ran2(idum)-1.0;
			v2 = 2.0 * ran2(idum)-1.0;
			rsq = v1*v1+v2*v2;
		} while (rsq >= 1.0 || rsq == 0.0);
		fac=sqrt(-2.0*log(rsq)/rsq);
		gset = v1 * fac;
		iset = 1;
		return(v2*fac);
	} else {
		iset = 0;
		return (gset);
	}
}

double ran2(long *idum)
{
  int j;
  long k;
  static long idum2=123456789;
  static long iy=0;
  static long iv[NTAB];
  double temp;

  if (*idum <= 0) {
    if (-(*idum) < 1) *idum=1;
    else *idum = -(*idum);
    idum2=(*idum);
    for (j=NTAB+7;j>=0;j--) {
      k=(*idum)/IQ1;
      *idum=IA1*(*idum-k*IQ1)-k*IR1;
      if (*idum < 0) *idum += IM1;
      if (j < NTAB) iv[j] = *idum;
    }
    iy=iv[0];
  }
  k=(*idum)/IQ1;
  *idum=IA1*(*idum-k*IQ1)-k*IR1;
  if (*idum < 0) *idum += IM1;
  k=idum2/IQ2;
  idum2=IA2*(idum2-k*IQ2)-k*IR2;
  if (idum2 < 0) idum2 += IM2;
  j=iy/NDIV;
  iy=iv[j]-idum2;
  iv[j] = *idum;
  if (iy < 1) iy += IMM1;
  if ((temp=AM*iy) > RNMX) return RNMX;
  else return temp;
}

/* ********************************************************************************** */
/*                                                                                    */
/*                             Instrument noise routines                              */
/*                                                                                    */
/* ********************************************************************************** */


void Shf_Geocenter(struct Data *data, struct Network *projection, double *SnGeo, double *params)
{
  int ifo,i;
  int NI = data->NI;

  int imin = data->imin;
  int imax = data->imax;

  double AntennaPattern;
  double ecc = params[3];

  for(i=imin; i<imax; i++) SnGeo[i] = 0.0;

  for(ifo=0; ifo<NI; ifo++)
  {
    AntennaPattern = projection->Fplus[ifo]*projection->Fplus[ifo] + ecc*ecc*projection->Fcross[ifo]*projection->Fcross[ifo];
    for(i=imin; i<imax; i++) SnGeo[i] += AntennaPattern/(data->Snf[ifo][i]);
  }
  for(i=0; i<imin; i++)SnGeo[i]=1.0;
  for(i=imax; i<data->N/2; i++)SnGeo[i]=1.0;
  for(i=imin; i<imax; i++)
  {
    SnGeo[i] = 1./SnGeo[i];
    //TODO:  Hack in Shf_Geocenter() to tray and keep signal model amplitude from going crazy
    if(SnGeo[i]<0.1*data->Snf[0][i] || SnGeo[i]>10.*data->Snf[0][i]) SnGeo[i] = data->Snf[0][i];
  }
}


/* ********************************************************************************** */
/*																					                                          */
/*                                 Fourier Routines                                   */
/*																					                                          */
/* ********************************************************************************** */

void dfour1(double data[], unsigned long nn, int isign)
{
  unsigned long n,mmax,m,j,istep,i;
  double wtemp,wr,wpr,wpi,wi,theta;
  double tempr,tempi, swap;

  n=nn << 1;
  j=1;
  for (i=1;i<n;i+=2) {
    if (j > i) {
      SWAP(data[j],data[i]);
      SWAP(data[j+1],data[i+1]);
    }
    m=n >> 1;
    while (m >= 2 && j > m) {
      j -= m;
      m >>= 1;
    }
    j += m;
  }
  mmax=2;
  while (n > mmax) {
    istep=mmax << 1;
    theta=isign*(6.28318530717959/mmax);
    wtemp=sin(0.5*theta);
    wpr = -2.0*wtemp*wtemp;
    wpi=sin(theta);
    wr=1.0;
    wi=0.0;
    for (m=1;m<mmax;m+=2) {
      for (i=m;i<=n;i+=istep) {
        j=i+mmax;
        tempr=wr*data[j]-wi*data[j+1];
        tempi=wr*data[j+1]+wi*data[j];
        data[j]=data[i]-tempr;
        data[j+1]=data[i+1]-tempi;
        data[i] += tempr;
        data[i+1] += tempi;
      }
      wr=(wtemp=wr)*wpr-wi*wpi+wr;
      wi=wi*wpr+wtemp*wpi+wi;
    }
    mmax=istep;
  }
}

void drealft(double data[], unsigned long n, int isign)
{
  void dfour1(double data[], unsigned long nn, int isign);
  unsigned long i,i1,i2,i3,i4,np3;
  double c1=0.5,c2,h1r,h1i,h2r,h2i;
  double wr,wi,wpr,wpi,wtemp,theta;

  theta=3.141592653589793/(double) (n>>1);
  if (isign == 1) {
    c2 = -0.5;
    dfour1(data,n>>1,1);
  } else {
    c2=0.5;
    theta = -theta;
  }
  wtemp=sin(0.5*theta);
  wpr = -2.0*wtemp*wtemp;
  wpi=sin(theta);
  wr=1.0+wpr;
  wi=wpi;
  np3=n+3;
  for (i=2;i<=(n>>2);i++) {
    i4=1+(i3=np3-(i2=1+(i1=i+i-1)));
    h1r=c1*(data[i1]+data[i3]);
    h1i=c1*(data[i2]-data[i4]);
    h2r = -c2*(data[i2]+data[i4]);
    h2i=c2*(data[i1]-data[i3]);
    data[i1]=h1r+wr*h2r-wi*h2i;
    data[i2]=h1i+wr*h2i+wi*h2r;
    data[i3]=h1r-wr*h2r+wi*h2i;
    data[i4] = -h1i+wr*h2i+wi*h2r;
    wr=(wtemp=wr)*wpr-wi*wpi+wr;
    wi=wi*wpr+wtemp*wpi+wi;
  }
  if (isign == 1) {
    data[1] = (h1r=data[1])+data[2];
    data[2] = h1r-data[2];
  } else {
    data[1]=c1*((h1r=data[1])+data[2]);
    data[2]=c1*(h1r-data[2]);
    dfour1(data,n>>1,-1);
  }
}

void Fourier_Scaling(double *ft, int n, double Tobs)
{
  double fix;
  int i;

  /* convert Numerical Recipes FFT output into a strain spectral density */

  fix = sqrt(Tobs)/((double)n);
  for(i = 0; i < n; i++) ft[i] *= fix;
}

/* ********************************************************************************** */
/*                                                                                    */
/*                        Data handling and injection routines                        */
/*                                                                                    */
/* ********************************************************************************** */

REAL8TimeSeries *readTseries(CHAR *cachefile, CHAR *channel, LIGOTimeGPS start, REAL8 length)
{
	LALStatus status;
	memset(&status,0,sizeof(status));
	LALCache *cache = NULL;
	LALFrStream *stream = NULL;
	REAL8TimeSeries *out = NULL;

	cache  = XLALCacheImport( cachefile );
  int err;
  err = *XLALGetErrnoPtr();
	if(cache==NULL) {fprintf(stderr,"ERROR: Unable to import cache file \"%s\",\n       XLALError: \"%s\".\n",cachefile, XLALErrorString(err)); exit(-1);}
	stream = XLALFrStreamCacheOpen( cache );
	if(stream==NULL) {fprintf(stderr,"ERROR: Unable to open stream from frame cache file\n"); exit(-1);}
	out = XLALFrStreamInputREAL8TimeSeries( stream, channel, &start, length , 0 );
	if(out==NULL) fprintf(stderr,"ERROR: unable to read channel %s from %s at time %i\nCheck the specified data duration is not too long\n",channel,cachefile,start.gpsSeconds);
	XLALDestroyCache(cache);
	LALFrClose(&status,&stream);
	return out;
}

void InjectFromMDC(ProcessParamsTable *commandLine, LALInferenceIFOData *IFOdata, double *SNR)
{

  /* Read time domain WF present in an mdc frame file, FFT it and inject into the frequency domain stream */

  char mdcname[]="GW";
  char **mdc_caches=NULL;
  char **mdc_channels=NULL;
  ProcessParamsTable * ppt=commandLine;

  UINT4 nIFO=0;
  int i=0;
  UINT4 j=0;
  LALInferenceIFOData *data=IFOdata;
  REAL8 prefactor =1.0;
  ppt=LALInferenceGetProcParamVal(commandLine,"--MDC-prefactor");
  if (ppt){

    prefactor=atof(ppt->value);
    fprintf(stdout,"Using prefactor=%f to scale the MDC injection\n",prefactor);
  }
  REAL8 tmp=0.0;
  REAL8 net_snr=0.0;
  while (data) {nIFO++; data=data->next;}
  UINT4 Nmdc=0,Nchannel=0;

  ppt=LALInferenceGetProcParamVal(commandLine,"--MDC-cache");
  if (!ppt){

    fprintf(stderr,"You must provide the path of an MDC lal cache file for each IFO, using --MDC-cache [ XXX, YYY, ZZZ]\n");
    exit(1);

  }
  ppt=LALInferenceGetProcParamVal(commandLine,"--inj");
  if (ppt){

    fprintf(stderr,"You cannot use both injfile (--inj) and MDCs (--MDC-cache) Exiting... \n");
    exit(1);

  }

  ppt=LALInferenceGetProcParamVal(commandLine,"--MDC-cache");
  LALInferenceParseCharacterOptionString(ppt->value,&mdc_caches,&Nmdc);

  if (Nmdc!= nIFO){
    fprintf(stderr, "You have to provide an MDC cache file for each IFO\n");
    exit(1);
  }
  else printf("got %i ifos and %i MDC cache files\n",nIFO,Nmdc);


  ppt=LALInferenceGetProcParamVal(commandLine,"--MDC-channel");
  if (ppt){

    LALInferenceParseCharacterOptionString(ppt->value,&mdc_channels,&Nchannel);
    if (Nchannel!=Nmdc){
      fprintf(stderr,"You must provide a channel name for eache mdc frame, using --MDC-channel [X, Y, Z] . Exiting...\n");
      exit(1);
    }
  }
  else{
    fprintf(stdout,"WARNING: You did not provide the name(s) of channel(s) to use with the injection mdc. Using the default which may not be what you want!\n");
    mdc_channels=  malloc((nIFO+1)*sizeof(char*));
    data=IFOdata;
    i=0;
    while (data){
      mdc_channels[i] =  malloc(512*sizeof(char));
      if(!strcmp(data->name,"H1")) {
        sprintf(mdc_channels[i],"H1:%s-H",mdcname);}
      else if(!strcmp(data->name,"L1")) {
        sprintf(mdc_channels[i],"L1:%s-H",mdcname); }
      else if(!strcmp(data->name,"V1")) {
        sprintf(mdc_channels[i],"V1:%s-16K",mdcname);}
      data=data->next;
      i++;

    }
  }

  LIGOTimeGPS epoch=IFOdata->timeData->epoch;
  REAL8 deltaT=IFOdata->timeData->deltaT ;
  int seglen=IFOdata->timeData->data->length;
  REAL8 SampleRate=4096.0,SegmentLength=0.0;
  if(LALInferenceGetProcParamVal(commandLine,"--srate")) SampleRate=atof(LALInferenceGetProcParamVal(commandLine,"--srate")->value);
  SegmentLength=(REAL8) seglen/SampleRate;

  REAL8TimeSeries * timeData=NULL;
  REAL8TimeSeries * windTimeData=(REAL8TimeSeries *)XLALCreateREAL8TimeSeries("WindMDCdata",&epoch,0.0,deltaT,&lalDimensionlessUnit,(size_t)seglen);
  COMPLEX16FrequencySeries* injF=(COMPLEX16FrequencySeries *)XLALCreateCOMPLEX16FrequencySeries("injF",&IFOdata->timeData->epoch,0.0,IFOdata->freqData->deltaF,&lalDimensionlessUnit,  IFOdata->freqData->data->length);

  if(!injF) {
    XLALPrintError("Unable to allocate memory for injection buffer\n");
    XLAL_ERROR_VOID(XLAL_EFUNC);
  }

  REAL4 WinNorm = sqrt(IFOdata->window->sumofsquares/IFOdata->window->data->length);

  data=IFOdata;
  i=0;
  UINT4 lower = (UINT4)ceil(data->fLow / injF->deltaF);
  UINT4 upper = (UINT4)floor(data->fHigh /injF-> deltaF);

  /* Inject into FD data stream and calculate optimal SNR */
  while(data){

    char foutname[50]="";
    sprintf(foutname,"MDC_freq_%s_%d",data->name,epoch.gpsSeconds);
    FILE * fout = fopen(foutname,"w");

    char outname[50]="";
    sprintf(outname,"MDC_time_%s_%d",data->name,epoch.gpsSeconds);
    FILE * out = fopen(outname,"w");

    tmp=0.0;

    /* Read MDC frame */
    timeData=readTseries(mdc_caches[i],mdc_channels[i],epoch,SegmentLength);
    /* downsample */
    XLALResampleREAL8TimeSeries(timeData,1.0/SampleRate);
    /* window timeData and store it in windTimeData */
    XLALDDVectorMultiply(windTimeData->data,timeData->data,IFOdata->window->data);

    for(j=0;j< timeData->data->length;j++)
      fprintf(out,"%lf %10.10e %10.10e %10.10e \n",epoch.gpsSeconds + j*deltaT,data->timeData->data->data[j],data->timeData->data->data[j]+timeData->data->data[j],timeData->data->data[j]);
    fclose(out);

    /* set the whole seq to 0 */
    for(j=0;j<injF->data->length;j++) injF->data->data[j]*=0.0;//+1j*0.0;

    /* FFT */
    XLALREAL8TimeFreqFFT(injF,windTimeData,IFOdata->timeToFreqFFTPlan);


    for(j=lower;j<upper;j++){
      fprintf(fout,"%lf %10.10e %10.10e %10.10e\n", j*injF->deltaF,creal(injF->data->data[j])/WinNorm,cimag(injF->data->data[j])/WinNorm,data->oneSidedNoisePowerSpectrum->data->data[j]);
      //injF ->data->data[j]/=sqrt(data->window->sumofsquares / data->window->data->length);
      windTimeData->data->data[j] /= sqrt(data->window->sumofsquares / data->window->data->length);

      /* Add data in freq stream */
      data->freqData->data->data[j]+=crect(prefactor *creal(injF->data->data[j])/WinNorm,prefactor *cimag(injF->data->data[j])/WinNorm);
      tmp+= prefactor*prefactor*(creal(injF ->data->data[j])*creal(injF ->data->data[j])/WinNorm/WinNorm+cimag(injF ->data->data[j])*cimag(injF ->data->data[j])/WinNorm/WinNorm)/data->oneSidedNoisePowerSpectrum->data->data[j];
    }

    tmp*=2.*injF->deltaF;
    printf("Injected SNR %.3f in IFO %s from MDC \n",sqrt(2*tmp),data->name);
    data->SNR=sqrt(2*tmp);
    net_snr+=2*tmp;
    fclose(fout);
    SNR[i] = sqrt(2*tmp);

    i++;
    data=data->next;
  }
  printf("Injected network SNR %.3f from MDC\n",sqrt(net_snr));
  SNR[i]=sqrt(net_snr);
}


void getChannels(ProcessParamsTable *commandLine, char **channel)
{
  ProcessParamsTable *ppt = NULL;
  char ifo[10];
  char arg[20];

  ProcessParamsTable *argument=commandLine;

  int nifo=0;

  fprintf(stdout," ============= Channels ==============\n");
  while(argument)
  {
    if(!strcmp(argument->param,"--ifo") || !strcmp(argument->param,"--IFO"))
    {
      //Get name of one of the interferferometers (channel name is preceeded by IFO name)
      ppt = LALInferenceGetProcParamVal(commandLine, "--ifo");
      sprintf(ifo,"%s",argument->value);
      fprintf(stdout,"Detector %s: ",ifo);

      //Get channel name of interferometers
      sprintf(arg,"--%s-channel",ifo);
      ppt = LALInferenceGetProcParamVal(commandLine,arg);
      channel[nifo] = (char*)malloc(strlen(ppt->value) + 1 * sizeof(char));
      strcpy(channel[nifo], ppt->value);

      //This removes the ifo name from the channel string
      memmove(channel[nifo], channel[nifo]+3, strlen(channel[nifo]));
      fprintf(stdout,"channel %s\n",channel[nifo]);
      nifo++;
    }
    argument=argument->next;
  }
}

int outputFrameFile(ProcessParamsTable *commandLine, struct Data *data, struct Model *model)
{
  int N = data->N;

  LALFrameH *frame;       /* output frame           */
  char filename[256];     /* output file name       */

  REAL8TimeSeries *ht;    /* input strain data      */
  REAL8TimeSeries *rt;    /* output strain data     */

  char cache_name_1[256]; /* input frame cache name */
  LALCache *cache;        /* input frame cache      */
  LALFrStream *stream;    /* input stream           */

  /*** Get GMST for the trigger time ***/
  LALStatus status;
  char *chartmp=NULL;
  LIGOTimeGPS GPStrig, gps_start, gps_end;

  memset(&status,0,sizeof(status));

  LALStringToGPS(&status,&GPStrig,LALInferenceGetProcParamVal(commandLine,"--trigtime")->value,&chartmp);
  int Tobs = atoi(LALInferenceGetProcParamVal(commandLine,"--seglen")->value);

  /* define time segment */
  gps_start.gpsSeconds     = GPStrig.gpsSeconds+2-Tobs;
  gps_start.gpsNanoSeconds = 0;
  gps_end.gpsSeconds     = GPStrig.gpsSeconds+2;
  gps_end.gpsNanoSeconds = 0;

  cache_name_1[0]='\0';
  filename[0]='\0';
  ht=NULL;
  rt=NULL;

  int i,n;
  char **ifos=NULL;
  ProcessParamsTable *this=commandLine;
  char tmp[128];

  /* Construct a list of IFOs */
  int NI=0;
  for(this=commandLine;this;this=this->next)
  {
    if(!strcmp(this->param,"--ifo")||!strcmp(this->param,"--IFO"))
    {
      NI++;
      ifos=XLALRealloc(ifos,NI*sizeof(char *));
      ifos[NI-1]=XLALStringDuplicate(this->value);
    }
  }

  char **caches=XLALCalloc(NI,sizeof(char *));
  char **channels=XLALCalloc(NI,sizeof(char *));

  double *r = malloc(N*sizeof(double));

  double fix = (sqrt((double)Tobs))*sqrt(0.5*(double)N)/8.0;

  /* For each IFO, fetch the other options if available */
  for(i=0;i<NI;i++)
  {
    for(n=0; n<N/2; n++)
    {
      r[2*n]    =  (data->s[i][2*n]  - model->glitch[i]->templates[2*n]  );
      r[2*n+1]  = -(data->s[i][2*n+1]- model->glitch[i]->templates[2*n+1]);
    }
    drealft(r-1,N,-1);

    for(n=0; n<N; n++) r[n] /= fix;

    /* Cache */
    sprintf(tmp,"--%s-cache",ifos[i]);
    this=LALInferenceGetProcParamVal(commandLine,tmp);
    caches[i]=XLALStringDuplicate(this->value);

    /* Channel */
    sprintf(tmp,"--%s-channel",ifos[i]);
    this=LALInferenceGetProcParamVal(commandLine,tmp);
    channels[i]=XLALStringDuplicate(this?this->value:"Unknown channel");

    cache  = XLALCacheImport(caches[i]);
    stream = XLALFrStreamCacheOpen(cache);

    sprintf(filename,"%s_CLEAN",channels[i]);


    /* read raw data */
    ht = XLALFrStreamReadREAL8TimeSeries(stream, channels[i], &gps_start, XLALGPSDiff(&gps_end, &gps_start), 0);

    /* units */
    ht->sampleUnits = lalStrainUnit;

    rt = XLALCreateREAL8TimeSeries(filename, &gps_start, 0, (double)N/data->Tobs, &ht->sampleUnits, data->N);

    /* fill residual data vector */
    for(n=0;n<data->N; n++) rt->data->data[n]=r[n];

    /* save time series into a frame */
    frame = XLALFrameNew(&gps_start, XLALGPSDiff(&gps_end, &gps_start), "BayesWaveResidual", 0, 1, 0);
    XLALFrameAddREAL8TimeSeriesProcData(frame, rt);

    sprintf(filename,"%c-%s_BAYESWAVE_RESIDUAL-%d-%d.gwf", ifos[i][0],ifos[i],gps_start.gpsSeconds,(int)Tobs);
    printf("writing %s\n",filename);
    XLALFrameWrite(frame, filename);

    /* cleaning */
    XLALDestroyREAL8TimeSeries(ht);
    XLALDestroyREAL8TimeSeries(rt);
    XLALFrameFree(frame);

  }
  /* close streams */
  XLALFrStreamClose(stream);

  free(r);

  return 0;
}

void parse_command_line(struct Data *data, struct Chain *chain, struct Prior *prior, ProcessParamsTable *commandLine)
{
  ProcessParamsTable *ppt = NULL;

  ppt = LALInferenceGetProcParamVal(commandLine, "--Twin");
  if(ppt) data->Twin = atof(ppt->value);
  else data->Twin = 1.0;

  ppt = LALInferenceGetProcParamVal(commandLine, "--NCmin");
  if(ppt) chain->NCmin = atoi(ppt->value);
  else chain->NCmin = 25;

  ppt = LALInferenceGetProcParamVal(commandLine, "--NCmax");
  if(ppt) chain->NCmax = atoi(ppt->value);
  else chain->NCmax = 25;

  ppt = LALInferenceGetProcParamVal(commandLine, "--Niter");
  if(ppt) chain->count = atoi(ppt->value);
  else chain->count = 2000000;

  ppt = LALInferenceGetProcParamVal(commandLine, "--Ncycle");
  if(ppt) chain->cycle = atoi(ppt->value);
  else chain->cycle = 100;

  ppt = LALInferenceGetProcParamVal(commandLine, "--Nburnin");
  if(ppt) chain->burn = atoi(ppt->value);
  else chain->burn = 100000;

  ppt = LALInferenceGetProcParamVal(commandLine, "--chainSeed");
  if(ppt) chain->seed = atoi(ppt->value);
  else chain->seed = 1234;

  ppt = LALInferenceGetProcParamVal(commandLine, "--Dmin");
  if(ppt) data->Dmin = atoi(ppt->value);
  else data->Dmin = 1;

  ppt = LALInferenceGetProcParamVal(commandLine, "--Dmax");
  if(ppt) data->Dmax = atoi(ppt->value);
  else data->Dmax = 20;

  ppt = LALInferenceGetProcParamVal(commandLine, "--Qmax");
  if(ppt) data->Qmax = atof(ppt->value);
  else data->Qmax = 40.0;

  ppt = LALInferenceGetProcParamVal(commandLine, "--tempMin");
  if(ppt) chain->tempMin = (double)atof(ppt->value);
  else chain->tempMin = 1.0;

  ppt = LALInferenceGetProcParamVal(commandLine, "--tempSpacing");
  if(ppt) chain->tempStep = (double)atof(ppt->value);
  else chain->tempStep = 1.5;

  ppt = LALInferenceGetProcParamVal(commandLine, "--adaptTemperature");
  if(ppt) data->adaptTemperatureFlag = 1;
  else data->adaptTemperatureFlag = 0;

  ppt = LALInferenceGetProcParamVal(commandLine, "--noClean");
  if(ppt) data->skipCleanFlag = 1;
  else data->skipCleanFlag = 0;

  ppt = LALInferenceGetProcParamVal(commandLine, "--cleanOnly");
  if(ppt) data->cleanOnlyFlag = 1;
  else data->cleanOnlyFlag = 0;

  ppt = LALInferenceGetProcParamVal(commandLine, "--noiseOnly");
  if(ppt) data->noiseOnlyFlag = 1;
  else data->noiseOnlyFlag = 0;

  ppt = LALInferenceGetProcParamVal(commandLine, "--glitchOnly");
  if(ppt) data->glitchOnlyFlag = 1;
  else data->glitchOnlyFlag = 0;

  ppt = LALInferenceGetProcParamVal(commandLine, "--signalOnly");
  if(ppt) data->signalOnlyFlag = 1;
  else data->signalOnlyFlag = 0;

  ppt = LALInferenceGetProcParamVal(commandLine, "--fullOnly");
  if(ppt) data->fullOnlyFlag = 1;
  else data->fullOnlyFlag = 0;

  ppt = LALInferenceGetProcParamVal(commandLine, "--clusterPrior");
  if(ppt) data->clusterPriorFlag = 1;
  else data->clusterPriorFlag = 0;

  ppt = LALInferenceGetProcParamVal(commandLine, "--noOrientationProposal");
  if(ppt) data->orientationProposalFlag = 0;
  else data->orientationProposalFlag = 1;

  ppt = LALInferenceGetProcParamVal(commandLine, "--noAmplitudePrior");
  if(ppt) data->amplitudePriorFlag = 0;
  else data->amplitudePriorFlag = 1;

  ppt = LALInferenceGetProcParamVal(commandLine, "--sameAmplitudePrior");
  if(ppt)data->signalAmplitudePriorFlag = 0;
  else data->signalAmplitudePriorFlag = 1;

  ppt = LALInferenceGetProcParamVal(commandLine, "--ampPriorPeak");
  if(ppt) prior->SNRpeak = (double)atof(ppt->value);
  else prior->SNRpeak=5.0;

  ppt = LALInferenceGetProcParamVal(commandLine, "--dimensionDecayRate");
  if(ppt) prior->Dtau = (double)atof(ppt->value);
  else prior->Dtau=1000.0;

  ppt = LALInferenceGetProcParamVal(commandLine, "--noClusterProposal");
  if(ppt) data->clusterProposalFlag = 0;
  else data->clusterProposalFlag = 1;

  ppt = LALInferenceGetProcParamVal(commandLine, "--runName");
  if(ppt) sprintf(data->runName,"%s_",ppt->value);
  else data->runName[0]=0;

  ppt = LALInferenceGetProcParamVal(commandLine, "--outputDir");
  if(ppt) sprintf(data->outputDirectory,"%s/",ppt->value);
  else sprintf(data->outputDirectory,"./");

  ppt = LALInferenceGetProcParamVal(commandLine, "--noPSDfit");
  if(ppt) data->psdFitFlag = 0;
  else data->psdFitFlag = 1;

  ppt = LALInferenceGetProcParamVal(commandLine, "--bayesLine");
  if(ppt) data->bayesLineFlag = 1;
  else data->bayesLineFlag = 0;

  ppt = LALInferenceGetProcParamVal(commandLine, "--0noise");
  if(ppt) data->noiseSimFlag = 0;
  else data->noiseSimFlag = 1;

  ppt = LALInferenceGetProcParamVal(commandLine, "--zeroLogL");
  if(ppt) data->constantLogLFlag = 1;
  else data->constantLogLFlag = 0;

  ppt = LALInferenceGetProcParamVal(commandLine, "--gnuplot");
  if(ppt) data->gnuplotFlag = 1;
  else data->gnuplotFlag = 0;

  ppt = LALInferenceGetProcParamVal(commandLine, "--verbose");
  if(ppt) data->verboseFlag = 1;
  else data->verboseFlag = 0;

  ppt = LALInferenceGetProcParamVal(commandLine, "--bootstrapEvidence");
  if(ppt) data->bootstrapEvidenceFlag = 1;
  else data->bootstrapEvidenceFlag = 0;

  /*** Cluster prior ***/

  // what we mean by close
  ppt = LALInferenceGetProcParamVal(commandLine, "--clusterAlpha");
  if(ppt) prior->alpha = (double)atof(ppt->value);
  else prior->alpha = 2.0;

  // cluster prior cpen^-K
  ppt = LALInferenceGetProcParamVal(commandLine, "--clusterBeta");
  if(ppt) prior->beta = (double)atof(ppt->value);
  else prior->beta = 4.0;

  /*
   Crediting the Occam penalty for each wavelet if it is in a cluster
   Penalty is D for D degrees of freedom.
   So 5 if we want to cover all the parameters in a wavelet.
   */
  ppt = LALInferenceGetProcParamVal(commandLine, "--clusterGamma");
  if(ppt) prior->gamma = (double)atof(ppt->value);
  else prior->gamma = 0.0;

  /*
   path to files containing normalization for cluster prior
   */
  ppt = LALInferenceGetProcParamVal(commandLine, "--clusterPath");
  if(ppt) sprintf(prior->path,"%s",ppt->value);
  else prior->path[0]=0;


  /*** Get GMST for the trigger time ***/
  LALStatus status;
  char *chartmp=NULL;
  LIGOTimeGPS GPStrig;

  memset(&status,0,sizeof(status));

  ppt = LALInferenceGetProcParamVal(commandLine,"--trigtime");
  LALStringToGPS(&status,&GPStrig,ppt->value,&chartmp);
  double gmst=(double)XLALGreenwichMeanSiderealTime(&GPStrig);

  //remap gmst back to [0:2pi]
  int intpart;
  double decpart;
  gmst /= LAL_TWOPI;
  intpart = (int)( gmst );
  decpart = gmst - (double)intpart;
  gmst = decpart*LAL_TWOPI;
  data->gmst = gmst;

  /*** Get channel names for each interferometer ***/
  data->channels = malloc(3 * sizeof(char*));
  getChannels(commandLine, data->channels);

  //store command line in data structure for printing later on
  data->commandLine = commandLine;

  //bayesLine takes precedent over PSD fitting
  //if(data->bayesLineFlag) data->psdFitFlag = 0;

}

/* ********************************************************************************** */
/*                                                                                    */
/*                           Memory (de)allocation routines                           */
/*                                                                                    */
/* ********************************************************************************** */


void initialize_fisher(struct FisherMatrix *fisher)
{
  int i;
  fisher->sigma	  = dvector(0,NE-1);
  fisher->evalue	= dvector(0,NE-1);
  fisher->matrix	= dmatrix(0,NE-1,0,NE-1);
  fisher->evector	= dmatrix(0,NE-1,0,NE-1);
  fisher->aamps	  = dvector(0,NE-1);
  fisher->count   = dmatrix(0,NE-1,0,NE-1);

  for(i=0;i<NE;i++)
  {
    fisher->aamps[i] = 1.0;
    fisher->count[0][i] = 1;
    fisher->count[1][i] = 1;
  }
}

void free_fisher(struct FisherMatrix *fisher)
{
  free_dvector(fisher->sigma,0,NE-1);
  free_dvector(fisher->evalue,0,NE-1);
  free_dmatrix(fisher->matrix,0,NE-1,0,NE-1);
  free_dmatrix(fisher->evector,0,NE-1,0,NE-1);
  free_dvector(fisher->aamps,0,NE-1);
  free_dmatrix(fisher->count,0,NE-1,0,NE-1);
}


void initialize_chain(struct Chain *chain, int flag)
{
  chain->mod0     = 0;     //Counter keeping track of noise model
  chain->mod1     = 0;     //Counter keeping track of signal model
  chain->mod2     = 0;     //Counter keeping track of noise model
  chain->mod3     = 0;     //Counter keeping track of signal model
  chain->mcount   = 1;     //# of MCMC proposals
  chain->scount   = 1;     //# of RJMCMC proposals
  chain->xcount   = 1;     //# of extrinsic proposals
  chain->macc     = 0;     //# of MCMC successes
  chain->sacc     = 0;     //# of RJMCMC successes
  chain->xacc     = 0;     //# of extrinsic successes
  chain->burnFlag = flag;  //Flag for burn-in (1 = yes)
}

void free_chain(struct Chain *chain, int NI)
{
  free(chain->ptacc);
  free(chain->ptprop);
  free(chain->index);

  free(chain->dT);
  free(chain->temperature);
  free(chain->avgLogLikelihood);
  free(chain->varLogLikelihood);

  free(chain->extChainFile);
  free(chain->intChainFile);

  free(chain->intWaveChainFile);
  int ifo;
  for(ifo=0; ifo<NI; ifo++) free(chain->intGlitchChainFile[ifo]);
  free(chain->intGlitchChainFile);
}

void allocate_chain(struct Chain *chain, int NI)
{
  chain->ptacc        = malloc(chain->NC*sizeof(int));
  chain->ptprop       = malloc(chain->NC*sizeof(int));
  chain->index        = malloc(chain->NC*sizeof(int));

  chain->dT               = malloc(chain->NC*sizeof(double));
  chain->temperature      = malloc(chain->NC*sizeof(double));
  chain->avgLogLikelihood = malloc(chain->NC*sizeof(double));
  chain->varLogLikelihood = malloc(chain->NC*sizeof(double));
  //chain->logLstep = malloc(chain->NC*sizeof(double));

  chain->extChainFile = malloc(chain->NC*sizeof(FILE*));
  chain->intChainFile = malloc(chain->NC*sizeof(FILE*));

  chain->intWaveChainFile = malloc(chain->NC*sizeof(FILE*));

  int ifo;
  chain->intGlitchChainFile = malloc(NI*sizeof(FILE**));
  for(ifo=0; ifo<NI; ifo++) chain->intGlitchChainFile[ifo] = malloc(chain->NC*sizeof(FILE*));
}

void reset_chain(struct Chain *chain)
{
  chain->sacc   = 0;
  chain->scount = 1;
  chain->macc   = 0;
  chain->mcount = 1;
  chain->xacc   = 0;
  chain->xcount = 1;
}

void initialize_priors(struct Data *data, struct Prior *prior, int omax)
{
  int i, n;

  /* Model dimension parameters */
  // number of sineGaussians we print to file
  prior->omax = omax;

  // maximum number of sineGaussians
  prior->gmin = ivector(0,data->NI-1);
  prior->gmax = ivector(0,data->NI-1);

  prior->smin = 0;
  prior->smax = data->Dmax;
  for(i=0; i<data->NI; i++)
  {
    prior->gmin[i] = 0;
    prior->gmax[i] = data->Dmax;
  }

  if(data->signalFlag || data->glitchFlag) prior->smin = 1;


  /* PSD paramters */
  prior->Snmin = 0.5;
  prior->Snmax = 2.0;


  /* Wavelet parameters */
  prior->range = dmatrix(0,4,0,1);

  double Amin=0.0, Amax=1.0;
  //Set Amplitude priors based on the data channel being used
  //S6 strain
  if(!strcmp(data->channels[0],"LDAS-STRAIN"))
  {
    Amin = 5.e-24;
    Amax = 1.e-15;
  }

  //VSR2 strain
  else if (!strcmp(data->channels[0],"h_16384Hz"))
  {
    Amin = 5.e-24;
    Amax = 1.e-15;
  }

  //S6 LIGO DARM Error
  else if (!strcmp(data->channels[0],"LSC-DARM_ERR"))
  {
    Amin = 5.0e-9;
    Amax = 1.0e-3;
  }

  //Use S6 strain as default
  else
  {
    Amin = 5.e-24;
    Amax = 1.e-15;
  }

  //double dt   = data->Tobs/(double)data->tsize;
  double Qmin = 1.0;
  double Qmax = data->Qmax;
  double tmin = 0;//dt*(int)((data->Tobs/16.)/dt);
  double tmax = data->Tobs;//-tmin;
  double fmin = data->fmin;
  double fmax = data->fmax;

  /*
  if(data->amplitudePriorFlag)
  {
    Amin=0.0;
    Amax=1.0;
  }
   */

  prior->range[0][0] = tmin;
  prior->range[0][1] = tmax;
  prior->range[1][0] = fmin;
  prior->range[1][1] = fmax;
  prior->range[2][0] = Qmin;
  prior->range[2][1] = Qmax;
  prior->range[3][0] = Amin;
  prior->range[3][1] = Amax;
  prior->range[4][0] = 0.0;
  prior->range[4][1] = TPI;

  prior->TFV    = (tmax-tmin)*(fmax-fmin);
  prior->logTFV = log(prior->TFV);


  /*
   cluster prior
   */
  prior->bias = dvector(0,prior->smax);

  if(data->clusterPriorFlag)
  {
    FILE *normFile;
    char filename[1000];

    //sprintf(filename,"%s/cluster_norm_flow%i_seglen%i_srate%i_Qmax%i.dat",prior->path,(int)fmin,(int)data->Tobs,(int)fmax*2,(int)Qmax);
    sprintf(filename,"%s/cluster_norm_flow%i_twin%i_srate%i_Qmax%i.dat",prior->path,(int)fmin,(int)data->Twin,(int)fmax*2,(int)Qmax);
    normFile = fopen(filename,"r");

    for(i=0; i<prior->smax; i++)
    {
      fscanf(normFile,"%i %lg",&n,&prior->bias[i]);
      prior->bias[i] = log(prior->bias[i]);
    }

    fclose(normFile);
  }
}

void reset_priors(struct Data *data, struct Prior *prior)
{
  int i;

  prior->smin = 0;
  prior->smax = data->Dmax;
  for(i=0; i<data->NI; i++)
  {
    prior->gmin[i] = 0;
    prior->gmax[i] = data->Dmax;
  }

  if(data->signalFlag || data->glitchFlag) prior->smin = data->Dmin;

  /* PSD paramters */
  prior->Snmin = 0.5;
  prior->Snmax = 2.0;

  double Amin=0.0, Amax=1.0;
  //Set Amplitude priors based on the data channel being used
  //S6 strain
  if(!strcmp(data->channels[0],"LDAS-STRAIN"))
  {
    Amin = 5.e-24;
    Amax = 1.e-15;
  }

  //VSR2 strain
  else if (!strcmp(data->channels[0],"h_16384Hz"))
  {
    Amin = 5.e-24;
    Amax = 1.e-15;
  }

  //S6 LIGO DARM Error
  else if (!strcmp(data->channels[0],"LSC-DARM_ERR"))
  {
    Amin = 5.0e-9;
    Amax = 1.0e-3;
  }

  //Use S6 strain as default
  else
  {
    Amin = 5.e-24;
    Amax = 1.e-15;
  }

  double Qmin = 1.0;
  double Qmax = data->Qmax;
  double tmin = 0;//dt*(int)((data->Tobs/16.)/dt);
  double tmax = data->Tobs;//-tmin;
  double fmin = data->fmin;
  double fmax = data->fmax;

  //restrict prior range for wavelet time to 1 second around trigger
  if(data->runPhase==1)
  {
    tmin = data->Tobs-2.0-data->Twin/2.0;
    tmax = data->Tobs-2.0+data->Twin/2.0;
  }
  /*
  if(data->amplitudePriorFlag)
  {
    Amin=0.0;
    Amax=1.0;
  }
*/
  prior->range[0][0] = tmin;
  prior->range[0][1] = tmax;
  prior->range[1][0] = fmin;
  prior->range[1][1] = fmax;
  prior->range[2][0] = Qmin;
  prior->range[2][1] = Qmax;
  prior->range[3][0] = Amin;
  prior->range[3][1] = Amax;
  prior->range[4][0] = 0.0;
  prior->range[4][1] = TPI;

  prior->TFV    = (tmax-tmin)*(fmax-fmin);
  prior->logTFV = log(prior->TFV);

}

void initialize_data(struct Data *data, double **s, int N, int tsize, double **psd, double Tobs, int NI, double fmin, double fmax)
{
  int i,ifo;

  data->N     = N;
  data->NI    = NI;
  data->tsize = tsize;
  data->fsize = N/tsize;
  data->Tobs  = Tobs;
  data->fmin  = fmin;
  data->fmax  = fmax;
  data->imin  = (int)floor(fmin*Tobs);
  data->imax  = (int)floor(fmax*Tobs);
  data->dt    = Tobs/(double)(tsize);
  data->df    = 0.5/data->dt;

  // allocate data vectors
  data->h  = dvector(0,N-1); // workspace for individual templates
  data->s  = dmatrix(0,NI-1,0,N-1); // data
  data->r  = dmatrix(0,NI-1,0,N-1); // residual
  data->rh = dmatrix(0,NI-1,0,N-1);
  data->SnS= dmatrix(0,NI-1,0,N/2-1);
  data->Snf= dmatrix(0,NI-1,0,N/2-1);


  // initialize data vectors
  for(i=0; i<N; i++)
  {
    data->h[i] = 0.0;

    for(ifo=0; ifo<NI; ifo++)
    {
      data->s[ifo][i]  = s[ifo][i];
      data->r[ifo][i]  = 0.0;
      data->rh[ifo][i] = 0.0;
    }
  }

  data->SnGeo = dvector(0,N/2-1);

  data->logLc = 0.0;


  //store sum of 1/psd in 1/SnGeo
  for(i=0; i<N/2; i++)
  {
    data->SnGeo[i] = 0.0;
    for(ifo=0; ifo<NI; ifo++)
    {
      data->Snf[ifo][i] = psd[ifo][i];
      data->SnS[ifo][i] = psd[ifo][i];
      data->SnGeo[i] += 1./psd[ifo][i];
    }
    data->SnGeo[i] = 1./data->SnGeo[i];
  }
}

void initialize_model(struct Model *model, struct Data *data, struct Prior *prior, long *seed)
{
  //store psd parameters
  model->Sn = dvector(0,data->NI-1);

  //extrinsic parameters
  model->extParams = dvector(0,NE);

  //full instrument response model (projected geocenter + glitches)
  model->response = dmatrix(0,data->NI-1,0,data->N-1);

  //likelhood is a pointer to a function which returns logL
  model->logL = 0.0;

  //detLogL stores the likelihood in each detector (for quick computation of full logL)
  model->detLogL = dvector(0,data->NI-1);

  /*
   model->extParams[0] = 98.5*M_PI/180.;//40.87*M_PI/180.;//1.2;                   //right ascension
   model->extParams[1] = sin(38.0*M_PI/180.);//sin(52.0*M_PI/180.);//-0.3;                  //sin(declination)
   model->extParams[2] = 0.9;                   //polarization angle
   model->extParams[3] = -2.0*0.2/(1.0+0.2*0.2);//ellipticity
   model->extParams[4] = 1.0;                   //amplitude scale
   model->extParams[5] = 0.0;                   //geocenter time shift
   model->extParams[6] = 0.0;                   //geocenter phase shift
   */
  //draw extrinsic parameters from prior
  extrinsic_uniform_proposal(seed,model->extParams);

  //set "delta" parameters to not modify signal model
  model->extParams[4] = 1.0;  //amplitude scale
  model->extParams[5] = 0.0;  //geocenter time shift
  model->extParams[6] = 0.0;  //geocenter phase shift


  //allocate memory for signal wavelet structure
  model->signal = malloc(sizeof(struct Wavelet));

  //Setup wavelet structure holding signal model
  if(data->signalFlag)
    initialize_wavelet(model->signal,data->N, prior->smax, 1, data->fsize, data->tsize);
  else
    initialize_wavelet(model->signal,data->N, prior->smax, 0, data->fsize, data->tsize);

  //allocate memory for structure holding projection coefficients
  model->projection = malloc(sizeof(struct Network));

  //set up projection coefficients
  initialize_network(model->projection, data->N, data->NI);

  //Allocate and initialize glitch structures
  int i;
  model->glitch = malloc(data->NI * sizeof(struct Wavelet *));
  for(i=0; i<data->NI; i++) model->glitch[i] = malloc(sizeof(struct Wavelet));
  for(i=0; i<data->NI; i++)
  {
    if(data->glitchFlag)
      initialize_wavelet(model->glitch[i],data->N, prior->smax, 1, data->fsize, data->tsize);
    else
      initialize_wavelet(model->glitch[i],data->N, prior->smax, 0, data->fsize, data->tsize);
  }

  //allocate memory for structure holding extrinsic fisher matrix
  model->fisher = malloc(sizeof(struct FisherMatrix));
  initialize_fisher(model->fisher);

}

void reset_model(struct Data *data, struct Chain *chain, struct Prior *prior, struct Model **model)
{
  int i,j,ic,ifo;
  int NC= chain->NC;
  int NI= data->NI;

  struct Wavelet *signal;
  struct Wavelet **glitch;

  for(ic=0; ic<NC; ic++)
  {
    //reset model
    signal = model[ic]->signal;
    glitch = model[ic]->glitch;
    model[ic]->size=0;
    signal->size=0;
    for(ifo=0; ifo<NI; ifo++) glitch[ifo]->size=0;

    signal->smax = prior->smax;
    for(j=1; j<=signal->smax; j++) signal->index[j]=j;
    for(i=0; i<NI; i++)
    {
      glitch[i]->smax = prior->smax;
      for(j=1; j<=glitch[i]->smax; j++) glitch[i]->index[j]=j;
    }

    for(ifo=0; ifo<NI; ifo++)
    {
      if(data->glitchFlag)
      {
        glitch[ifo]->size = 1;
        model[ic]->size++;
      }
    }

    if(data->signalFlag)
    {
      signal->size = 1;
      model[ic]->size++;
    }

    model[ic]->logL=data->logLc;
  }
}

void reset_likelihood(struct Data *data)
{
  if(data->constantLogLFlag)
  {
    // logL=const to test detailed balance
    data->intrinsic_likelihood = EvaluateConstantLogLikelihood;
    data->extrinsic_likelihood = EvaluateExtrinsicConstantLogLikelihood;
  }
  else
  {
    // Start with search-phase log likelihood
    data->intrinsic_likelihood = EvaluateMarkovianLogLikelihood;//EvaluateSearchLogLikelihood;
    //data->intrinsic_likelihood = EvaluateMarkovianLogLikelihood;

    // Jump right into characterization likelihood if noise-only run
    if(!data->signalFlag && !data->glitchFlag) data->intrinsic_likelihood = EvaluateMarkovianLogLikelihood;

    //data->extrinsic_likelihood = EvaluateExtrinsicSearchLogLikelihood;
    data->extrinsic_likelihood = EvaluateExtrinsicMarkovianLogLikelihood;
  }
}

void initialize_network(struct Network *projection, int N, int NI)
{
  projection->expPhase = dmatrix(0,NI-1,0,N-1);
  projection->deltaT   = dvector(0,NI-1);
  projection->Fplus    = dvector(0,NI-1);
  projection->Fcross   = dvector(0,NI-1);
}

void free_network(struct Network *projection, int N, int NI)
{
  free_dmatrix(projection->expPhase,0,NI-1,0,N-1);
  free_dvector(projection->deltaT,0,NI-1);
  free_dvector(projection->Fplus,0,NI-1);
  free_dvector(projection->Fcross,0,NI-1);
}

void copy_int_model(struct Model *origin, struct Model *copy, int N, int NI, int det)
{
  int i,ifo;

  copy->logL=origin->logL;

  copy->size=origin->size;

  if(det==-1)
  {
    copy_wavelet(origin->signal, copy->signal, N);
    for(ifo=0; ifo<NI; ifo++) for(i=0; i<N; i++) copy->response[ifo][i] = origin->response[ifo][i];
  }
  else
  {
    for(ifo=0; ifo<NI; ifo++)copy_wavelet(origin->glitch[ifo],copy->glitch[ifo], N);
  }

  for(ifo=0; ifo<NI; ifo++)
  {
    copy->detLogL[ifo] = origin->detLogL[ifo];

    copy->Sn[ifo] = origin->Sn[ifo]; //noise PSD scale
  }
}

void copy_ext_model(struct Model *origin, struct Model *copy, int N, int NI)
{
  int i,ifo;
  for(i=0; i<NE; i++) copy->extParams[i] = origin->extParams[i];

  for(ifo=0; ifo<NI; ifo++)
  {
    copy->projection->Fplus[ifo]  = origin->projection->Fplus[ifo];
    copy->projection->Fcross[ifo] = origin->projection->Fcross[ifo];
    copy->projection->deltaT[ifo] = origin->projection->deltaT[ifo];
    for(i=0; i<N; i++) copy->projection->expPhase[ifo][i] = origin->projection->expPhase[ifo][i];
  }
}

void free_model(struct Model *model, struct Data *data, struct Prior *prior)
{
  free_dvector(model->Sn,0,data->NI-1);
  free_dvector(model->detLogL,0,data->NI-1);
  free_dvector(model->extParams,0,NE);
  free_dmatrix(model->response,0,data->NI-1,0,data->N-1);

  free_wavelet(model->signal,data->N, prior->smax, data->fsize, data->tsize);
  free(model->signal);

  free_network(model->projection, data->N, data->NI);
  free(model->projection);

  free_fisher(model->fisher);
  free(model->fisher);

  int i;
  for(i=0; i<data->NI; i++)
  {
    free_wavelet(model->glitch[i],data->N,prior->smax, data->fsize, data->tsize);
    free(model->glitch[i]);
  }
  free(model->glitch);
}

void initialize_wavelet(struct Wavelet *wave, int N, int smax, int size, int fsize, int tsize)
{
  int j;

  // initial number of wavelets
  wave->size = size;

  // maximum number of basis functions
  /*
   redundent to make Wavelet more self-contained
   should consider doing the same with N, and for
   other structures as well...
   //TODO: give all structures a copy of N??
   */
  wave->smax = smax;

  //store wavelet parameters
  wave->intParams = dmatrix(1,smax,0,4);

  // this keeps track of who is who
  wave->index = ivector(1,smax);

  // initialize reference order
  for(j=1; j<= smax; j++) wave->index[j] = j;

  // Fourier domain linear combination of wavelets
  wave->templates = dvector(0,N-1);
  for(j=0; j<N; j++) wave->templates[j] = 0.0;

  // clustering prior
  wave->Ncluster = 0;
  wave->cosy = 0.0;
  wave->logp = 0.0;

  //density proposal
  wave->TFhistory = dmatrix(0,fsize-1,0,tsize-1);
  wave->TFdensity = dmatrix(0,fsize-1,0,tsize-1);

}

void copy_wavelet(struct Wavelet *origin, struct Wavelet *copy, int N)
{
  int i,j;
  for(i=1; i<=origin->smax; i++)
  {
    copy->index[i] = origin->index[i];
    for(j=0; j<5; j++) copy->intParams[i][j] = origin->intParams[i][j];
  }
  for(j=0; j<N; j++) copy->templates[j] = origin->templates[j];

  copy->size = origin->size;

  // clustering prior
  copy->Ncluster = origin->Ncluster;
  copy->cosy     = origin->cosy;
  copy->logp     = origin->logp;

}

void free_wavelet(struct Wavelet *wave, int N, int smax, int fsize, int tsize)
{
  free_dmatrix(wave->intParams,1,smax,0,4);
  free_ivector(wave->index,1,smax);
  free_dvector(wave->templates,0,N-1);
  free_dmatrix(wave->TFhistory,0,fsize-1,0,tsize-1);
  free_dmatrix(wave->TFdensity,0,fsize-1,0,tsize-1);
}

void initialize_TFproposal(struct Data *data, struct Prior *prior, struct Wavelet *wave)
{
  //Initialize TFdensity to uniform over allowed TF volume
  int i,j;
  int count = 0;
  double t,f;
  double dt = data->Tobs/(double)data->tsize;
  double df = data->fmax/(double)data->fsize;


  /*
  prior->range[0][0] = tmin;
  prior->range[0][1] = tmax;
  prior->range[1][0] = fmin;
  prior->range[1][1] = fmax;
   */

  double tmin = prior->range[0][0];//dt*(int)((data->Tobs/16.)/dt);
  double tmax = prior->range[0][1];//data->Tobs-tmin;
  double fmin = prior->range[1][0];//data->fmin;
  double fmax = prior->range[1][1];//data->fmax;


  for(i=0; i<data->fsize; i++)
  {
    f = i*df;
    for(j=0; j<data->tsize; j++)
    {
      t = j*dt;
      if(f>=fmin && f<fmax && t>=tmin && t<tmax)
      {
        wave->TFdensity[i][j] = 1.0;
        count++;
      }
      else wave->TFdensity[i][j] = 0.0;
    }
  }

  wave->TFdensityMax = 1./(double)count/0.5;
  for(i=0; i<data->fsize; i++)
  {
    for(j=0; j<data->tsize;j++)
    {
      wave->TFhistory[i][j]=wave->TFdensity[i][j];
      wave->TFdensity[i][j]/=(double)count/0.5;
    }
  }
}

int checkfile(char *name)
{
  FILE *fp=fopen(name,"r");
  if(fp)
  {
    fclose(fp);
    return 1;
  }
  else return 0;
}


int *ivector(long nl, long nh)
/* allocate an int vector with subscript range v[nl..nh] */
{
  int *v;

  v=(int *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(int)));
  if (!v) fprintf(stderr,"allocation failure in ivector()");
  return v-nl+NR_END;
}

void free_ivector(int *v, long nl, long nh)
/* free an int vector allocated with ivector() */
{
  free((FREE_ARG) (v+nl-NR_END));
}

double *dvector(long nl, long nh)
/* allocate a double vector with subscript range v[nl..nh] */
{
  double *v=0;

  v=(double *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(double)));
  if (!v) fprintf(stderr,"allocation failure in dvector()");
  return v-nl+NR_END;
}

void free_dvector(double *v, long nl, long nh)
/* free a double vector allocated with dvector() */
{
  free((FREE_ARG) (v+nl-NR_END));
}

unsigned long *lvector(long nl, long nh)
/* allocate an unsigned long vector with subscript range v[nl..nh] */
{
  unsigned long *v;

  v=(unsigned long *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(long)));
  if (!v) fprintf(stderr,"allocation failure in lvector()");
  return v-nl+NR_END;
}

void free_lvector(unsigned long *v, long nl, long nh)
/* free an unsigned long vector allocated with lvector() */
{
  free((FREE_ARG) (v+nl-NR_END));
}

int **imatrix(long nrl, long nrh, long ncl, long nch)
/* allocate a int matrix with subscript range m[nrl..nrh][ncl..nch] */
{
  long i, nrow=nrh-nrl+1,ncol=nch-ncl+1;
  int **m;

  /* allocate pointers to rows */
  m=(int **) malloc((size_t)((nrow+NR_END)*sizeof(int*)));
  if (!m) fprintf(stderr, "allocation failure 1 in matrix()");
  m += NR_END;
  m -= nrl;


  /* allocate rows and set pointers to them */
  m[nrl]=(int *) malloc((size_t)((nrow*ncol+NR_END)*sizeof(int)));
  if (!m) fprintf(stderr, "allocation failure 2 in matrix()");
  m[nrl] += NR_END;
  m[nrl] -= ncl;

  for(i=nrl+1;i<=nrh;i++) m[i]=m[i-1]+ncol;

  /* return pointer to array of pointers to rows */
  return m;
}

void free_imatrix(int **m, long nrl, long nrh, long ncl, long nch)
/* free an int matrix allocated by imatrix() */
{
  free((FREE_ARG) (m[nrl]+ncl-NR_END));
  free((FREE_ARG) (m+nrl-NR_END));
}

double **dmatrix(long nrl, long nrh, long ncl, long nch)
/* allocate a double matrix with subscript range m[nrl..nrh][ncl..nch] */
{
  long i, nrow=nrh-nrl+1,ncol=nch-ncl+1;
  double **m;

  /* allocate pointers to rows */
  m=(double **) malloc((size_t)((nrow+NR_END)*sizeof(double*)));
  if (!m) fprintf(stderr, "allocation failure 1 in matrix()");
  m += NR_END;
  m -= nrl;

  /* allocate rows and set pointers to them */
  m[nrl]=(double *) malloc((size_t)((nrow*ncol+NR_END)*sizeof(double)));
  if (!m[nrl]) fprintf(stderr,"allocation failure 2 in matrix()");
  m[nrl] += NR_END;
  m[nrl] -= ncl;

  for(i=nrl+1;i<=nrh;i++) m[i]=m[i-1]+ncol;

  /* return pointer to array of pointers to rows */
  return m;
}

void free_dmatrix(double **m, long nrl, long nrh, long ncl, long nch)
/* free a double matrix allocated by dmatrix() */
{
  free((FREE_ARG) (m[nrl]+ncl-NR_END));
  free((FREE_ARG) (m+nrl-NR_END));
}

double ***d3tensor(long nrl, long nrh, long ncl, long nch, long ndl, long ndh)
/* allocate a double 3tensor with range t[nrl..nrh][ncl..nch][ndl..ndh] */
{
  long i,j,nrow=nrh-nrl+1,ncol=nch-ncl+1,ndep=ndh-ndl+1;
  double ***t;

  /* allocate pointers to pointers to rows */
  t=(double ***) malloc((size_t)((nrow+NR_END)*sizeof(double**)));
  if (!t) fprintf(stderr,"allocation failure 1 in d3tensor()");
  t += NR_END;
  t -= nrl;

  /* allocate pointers to rows and set pointers to them */
  t[nrl]=(double **) malloc((size_t)((nrow*ncol+NR_END)*sizeof(double*)));
  if (!t[nrl]) fprintf(stderr,"allocation failure 2 in d3tensor()");
  t[nrl] += NR_END;
  t[nrl] -= ncl;

  /* allocate rows and set pointers to them */
  t[nrl][ncl]=(double *) malloc((size_t)((nrow*ncol*ndep+NR_END)*sizeof(double)));
  if (!t[nrl][ncl]) fprintf(stderr,"allocation failure 3 in f3tensor()");
  t[nrl][ncl] += NR_END;
  t[nrl][ncl] -= ndl;

  for(j=ncl+1;j<=nch;j++) t[nrl][j]=t[nrl][j-1]+ndep;
  for(i=nrl+1;i<=nrh;i++) {
    t[i]=t[i-1]+ncol;
    t[i][ncl]=t[i-1][ncl]+ncol*ndep;
    for(j=ncl+1;j<=nch;j++) t[i][j]=t[i][j-1]+ndep;
  }

  /* return pointer to array of pointers to rows */
  return t;
}

void free_d3tensor(double ***t, long nrl, long nrh, long ncl, long nch, long ndl, long ndh)
/* free a float d3tensor allocated by d3tensor() */
{
  free((FREE_ARG) (t[nrl][ncl]+ndl-NR_END));
  free((FREE_ARG) (t[nrl]+ncl-NR_END));
  free((FREE_ARG) (t+nrl-NR_END));
}

int ***i3tensor(long nrl, long nrh, long ncl, long nch, long ndl, long ndh)
/* allocate a double 3tensor with range t[nrl..nrh][ncl..nch][ndl..ndh] */
{
  long i,j,nrow=nrh-nrl+1,ncol=nch-ncl+1,ndep=ndh-ndl+1;
  int ***t;

  /* allocate pointers to pointers to rows */
  t=(int ***) malloc((size_t)((nrow+NR_END)*sizeof(int**)));
  if (!t) fprintf(stderr,"allocation failure 1 in d3tensor()");
  t += NR_END;
  t -= nrl;

  /* allocate pointers to rows and set pointers to them */
  t[nrl]=(int **) malloc((size_t)((nrow*ncol+NR_END)*sizeof(int*)));
  if (!t[nrl]) fprintf(stderr,"allocation failure 2 in d3tensor()");
  t[nrl] += NR_END;
  t[nrl] -= ncl;

  /* allocate rows and set pointers to them */
  t[nrl][ncl]=(int *) malloc((size_t)((nrow*ncol*ndep+NR_END)*sizeof(int)));
  if (!t[nrl][ncl]) fprintf(stderr,"allocation failure 3 in f3tensor()");
  t[nrl][ncl] += NR_END;
  t[nrl][ncl] -= ndl;

  for(j=ncl+1;j<=nch;j++) t[nrl][j]=t[nrl][j-1]+ndep;
  for(i=nrl+1;i<=nrh;i++) {
    t[i]=t[i-1]+ncol;
    t[i][ncl]=t[i-1][ncl]+ncol*ndep;
    for(j=ncl+1;j<=nch;j++) t[i][j]=t[i][j-1]+ndep;
  }

  /* return pointer to array of pointers to rows */
  return t;
}

void free_i3tensor(int ***t, long nrl, long nrh, long ncl, long nch, long ndl, long ndh)
/* free an int d3tensor allocated by i3tensor() */
{
  free((FREE_ARG) (t[nrl][ncl]+ndl-NR_END));
  free((FREE_ARG) (t[nrl]+ncl-NR_END));
  free((FREE_ARG) (t+nrl-NR_END));
}

void KILL(char* Message)
{
  fprintf(stderr,"\a\n");
  fprintf(stderr,"%s",Message);
  fprintf(stderr,"Terminating the program.\n\n\n");
  exit(1);

  return;
}


/* ********************************************************************************** */
/*                                                                                    */
/*                The Sausage Factory:  Custom proposal distributions                 */
/*                                                                                    */
/* ********************************************************************************** */

static void flip(double *ra, double *dec, double *psi, long *seed)
{
  double sindec = sin(*dec);

  /* randomly flip ra & dec */
  if(ran2(seed)<0.5)
    *ra = LAL_TWOPI - (*ra);
  if(ran2(seed)<0.5)
    *dec = asin(-1.0*(sindec));

  /* always flip psi */
  *psi = 0.5*LAL_PI - (*psi);
}


void sky_ring_proposal(double *x, long *seed, double *y, struct Data *data)
{
  int i,j,l;
  int NI = data->NI;

  double n[3];
  double k[3];

  double newRA, newDec, newTime;//, newPsi, newPhi, newEcc; //newScale;

  double gmst = XLALGreenwichMeanSiderealTime(&data->epoch);
  //remap gmst back to [0:2pi]
  double intpart;
  double decpart;
  gmst /= LAL_TWOPI;
  intpart = (int)( gmst );
  decpart = gmst - (double)intpart;
  gmst = decpart*LAL_TWOPI;

  double ra       = x[0];
  double sindelta = x[1];
  double dec      = asin(sindelta);
  double time     = x[5];


  //first copy parameters to new array
  for(i=0; i<7; i++) y[i]=x[i];

  /*
   line-of-sight vector
   */
  k[0] = cos(gmst-ra)*cos(dec);
  k[1] =-sin(gmst-ra)*cos(dec);
  k[2] = sin(dec);

  double IFO1[3];
  double IFO2[3];
  double IFOX[3];


  /*
   Store location for each detector
   */
  double **IFO = dmatrix(0,NI-1,0,2);

  int ifo;
  for(ifo=0; ifo<NI; ifo++)
  {
    memcpy(IFOX, data->detector[ifo]->location, 3*sizeof(double));
    for(i=0; i<3; i++) IFO[ifo][i] = IFOX[i];
  }

  /*
   Randomly select two detectors from the network
   -this assumes there are no co-located detectors
   */
  i=j=0;
  while(i==j)
  {
    i=(int)floor(ran2(seed)*(double)NI);
    j=(int)floor(ran2(seed)*(double)NI);
  }

  for(l=0; l<3; l++)
  {
    IFO1[l] = IFO[i][l];
    IFO2[l] = IFO[j][l];
  }

  /*
   detector axis
   */
  double normalize;

  normalize=0.0;
  for(i=0; i<3; i++)
  {
    n[i]  = IFO1[i]-IFO2[i];
    normalize += n[i]*n[i];
  }
  normalize = 1./sqrt(normalize);
  for(i=0; i<3; i++) n[i] *= normalize;

  /*
   rotation angle
   */
  double omega    = LAL_TWOPI*ran2(seed);
  double cosomega = cos(omega);
  double sinomega = sin(omega);
  double c1momega = 1.0 - cosomega;


  /*
   rotate k' = Rk
   */
  double kp[3];
  kp[0] = (c1momega*n[0]*n[0] + cosomega)     *k[0] + (c1momega*n[0]*n[1] - sinomega*n[2])*k[1] + (c1momega*n[0]*n[2] + sinomega*n[1])*k[2];
  kp[1] = (c1momega*n[0]*n[1] + sinomega*n[2])*k[0] + (c1momega*n[1]*n[1] + cosomega)     *k[1] + (c1momega*n[1]*n[2] - sinomega*n[0])*k[2];
  kp[2] = (c1momega*n[0]*n[2] - sinomega*n[1])*k[0] + (c1momega*n[1]*n[2] + sinomega*n[0])*k[1] + (c1momega*n[2]*n[2] + cosomega)     *k[2];

  /*
   convert k' back to ra' and dec'
   */
  newDec   = asin(kp[2]);
  newRA    = atan2(kp[1],kp[0]) + gmst;

  if(newRA <0.0)
    newRA+=LAL_TWOPI;
  else if (newRA >= LAL_TWOPI)
    newRA -= LAL_TWOPI;

  /*
   compute new geocenter time using
   fixed arrival time at IFO1 (arbitrary)
   */
  double tx = 0.0; //old time shift = k * n
  double ty = 0.0; //new time shift = k' * n
  for(i=0; i<3; i++)
  {
    tx += -IFO1[i]*k[i] /CLIGHT;
    ty += -IFO1[i]*kp[i]/CLIGHT;
  }
  newTime = tx + time - ty;

  /* This proposal is cheating
   double Fx,Fy;
   double Fp,Fc;
   Fx=0.0;
   Fy=0.0;
   for(i=0; i<NI; i++)
   {
   XLALComputeDetAMResponse(&Fp, &Fc, (const REAL4(*)[3])data->detector[i]->response, ra, dec, psi, XLALGreenwichMeanSiderealTime(&data->epoch));
   Fx += Fp*Fp+Fc*Fc;

   XLALComputeDetAMResponse(&Fp, &Fc, (const REAL4(*)[3])data->detector[i]->response, newRA, newDec, newPsi, XLALGreenwichMeanSiderealTime(&data->epoch));
   Fy += Fp*Fp+Fc*Fc;
   }
   newScale = scale/sqrt(Fy/Fx);
   */

  /* Uniform draws on psi, phi, ecc */
  if(data->orientationProposalFlag)
  {
    if(ran2(seed)<0.5)
      amoeba_orientation_proposal(x, seed, y, data);
    else
      uniform_orientation_proposal(x, seed, y, data);

    /* try flipping sky position */
    if(ran2(seed)<0.5)
    {
      double newPsi;
      flip(&newRA, &newDec, &newPsi,seed);
      y[2] = newPsi;
    }
  }
  else
    uniform_orientation_proposal(x, seed, y, data);

  y[0] = newRA;
  y[1] = sin(newDec);
  y[5] = newTime;


  free_dmatrix(IFO,0,NI-1,0,2);

}

void uniform_orientation_proposal(double *x, long *seed, double *y, struct Data *data)
{
  double psi      = x[2];
  double phi      = x[6];

  double newPsi,newPhi,newEcc;
	double alpha,beta;
	double draw;

	//big jump in either alpha (beta) or beta (alpha)
  if(ran2(seed)<0.5)
  {
    alpha = psi + phi;
    beta  = psi - phi;

    //alpha =>   0:3pi
    //beta  => -2pi:pi

    draw=ran2(seed);
    if(draw < 0.5) alpha = ran2(seed)*3.0*LAL_PI;
    else           beta  = -LAL_TWOPI+ran2(seed)*3.0*LAL_PI;

    //transform back to psi,phi space
    newPsi =  (alpha + beta)*0.5;
    newPhi =  (alpha - beta)*0.5;

    //map back in range
    if(newPhi>LAL_TWOPI) newPhi-=LAL_TWOPI;
    else if(newPhi<0)    newPhi+=LAL_TWOPI;

    if(newPsi>LAL_TWOPI) newPsi-=LAL_TWOPI;
    else if(newPsi<0)    newPsi+=LAL_TWOPI;
  }
  //uniform draw on psi and phi
  else
  {
    newPsi = ran2(seed)*LAL_TWOPI;
    newPhi = ran2(seed)*LAL_TWOPI;
  }

  //uniform draw on ecc
  newEcc = -1. + 2.*ran2(seed);

  y[2] = newPsi;
  y[3] = newEcc;
  y[6] = newPhi;

}

void burst2bayes_sky(double phi, double theta, struct Data *data, double *ra, double *dec)
{
  double sid = XLALGreenwichMeanSiderealTime(&data->epoch);
  double quotiant = fmod(sid,LAL_TWOPI);

  (*ra) = phi + quotiant;

  while((*ra)>LAL_TWOPI) (*ra) -= LAL_TWOPI;

  (*dec) = (LAL_PI/2.0 - theta);
}

void amoeba_orientation_proposal(double *x, long *seed, double *y, struct Data *data)
{
	double tl,t1,t2;
	double logH;
	double *phase, *dp;
	double *Fcross, *Fplus;
	int i;
  int NI = data->NI;

	double dphi, *pass, *pv, *yv, **pm;
	int nfunk;

	pass = dvector(0,NI+2);
	pv = dvector(1,2);
	pm = dmatrix(1,3,1,2);
	yv = dvector(1,3);
	phase = dvector(1,NI);
	Fplus = dvector(1,NI);
	Fcross = dvector(1,NI);
	dp = dvector(1,NI);

	// sets the tolerance, probably should be in a header file
	tl = 0.2;
	tl = tl*tl;
  t1 = 0.001;
	t2 = 0.00001;

	// randomly pick extrinsic parameters to simulate current state of chain
  //	ra = ran2(seed)*TPI;
  //	sindelta = -1.0+2.0*ran2(seed);
  //	psi = ran2(seed)*PI;
  //	ecc = -1.0+2.0*ran2(seed);
  double ra       = x[0];
  double sindelta = x[1];
  double psi = x[2];
  double ecc = x[3];

  double dec = asin(sindelta);
  for(i=0; i<NI; i++)
    XLALComputeDetAMResponse(&Fplus[i+1], &Fcross[i+1], (const REAL4(*)[3])data->detector[i]->response, ra, dec, psi, XLALGreenwichMeanSiderealTime(&data->epoch));

	for(i=1; i<= NI; i++) phase[i] = atan2(ecc*Fcross[i],Fplus[i]);


	// randomly pick new sky location to simulate sky proposal
  //	alpha = ran2(seed)*TPI;
  //	sindelta = -1.0+2.0*ran2(seed);
  double newRa       = y[0];
  double newSindelta = y[1];
  double newDec = asin(newSindelta);
  double newPsi      = y[2];
  double newEcc      = y[3];
  double newPhi      = y[6];

	// need to pass these into the minimization routine
	pass[0] = newRa;
	pass[1] = newSindelta;
	//pass[1] = newDec;
	pass[2] = tl;
	for(i=1; i<= NI; i++) pass[i+2] = phase[i];
	logH = 1.0e40;


  int quad,draw;

  //randomly select 4 quadrants to evolve amoeba
  for(draw=0; draw<4; draw++)
  {
    quad = 1+(int)floor(ran2(seed)*4);

    // set the three legs of the amoeba (we are going to try 4 quadrants of the eccentricty, polarization angle plane if needed)
    switch(quad)
    {
      default:
      case 1:
        pm[1][1] = PI/16.0;
        pm[1][2] = 0.2;
        pm[2][1] = PI/8.0;
        pm[2][2] = 0.8;
        pm[3][1] = 3.0*PI/16.0;
        pm[3][2] = 0.2;
        break;

      case 2:
        pm[1][1] = PI/16.0;
        pm[1][2] = -0.8;
        pm[2][1] = PI/8.0;
        pm[2][2] = -0.2;
        pm[3][1] = 3.0*PI/16.0;
        pm[3][2] = -0.8;
        break;

      case 3:
        pm[1][1] = 5.0*PI/16.0;
        pm[1][2] = 0.2;
        pm[2][1] = 3.0*PI/8.0;
        pm[2][2] = 0.8;
        pm[3][1] = 7.0*PI/16.0;
        pm[3][2] = 0.2;
        break;

      case 4:
        pm[1][1] = 5.0*PI/16.0;
        pm[1][2] = -0.8;
        pm[2][1] = 3.0*PI/8.0;
        pm[2][2] = -0.2;
        pm[3][1] = 7.0*PI/16.0;
        pm[3][2] = -0.8;
        break;
    }

    // evaluate amoeba at each footpoint
    yv[1] = mfunc(pm[1],pass,data);
    yv[2] = mfunc(pm[2],pass,data);
    yv[3] = mfunc(pm[3],pass,data);

    // evolve the amoeba
    amoeba(pm, yv, pass, 2, t2, t1, &nfunk, data);

    //	printf("%d\n", nfunk);
    //	printf("%f %f %f %f %f %f %f %f %f\n", pm[1][1], pm[1][2], yv[1], pm[2][1], pm[2][2], yv[2], pm[3][1], pm[3][2], yv[3]);

    // find the best foot point
    for(i=1; i <= 3; i++)
    {
      //printf("yv=%g ",yv[i]);
      if(yv[i] < logH)
      {
        logH = yv[i];
        newPsi = pm[i][1];
        newEcc = pm[i][2];
      }
    }
  }

  /*
   if logH is still above tolerance t1 kill the jump
   Rejection sampling only on ecc (all others are angles)
  */
  if(logH>t1)
  {
    //printf("failed jump:  logH=%g, t1=%g\n",logH,t1);
    newEcc = 2.;
  }
  for(i=0; i<NI; i++)
    XLALComputeDetAMResponse(&Fplus[i+1], &Fcross[i+1], (const REAL4(*)[3])data->detector[i]->response, newRa, newDec, newPsi, XLALGreenwichMeanSiderealTime(&data->epoch));
  
	dphi = 0.0;
	for(i=1; i<= NI; i++)
	{
		dp[i] = atan2(newEcc*Fcross[i],Fplus[i])-phase[i];
		while(dp[i] < -PI) dp[i] += TPI;
		while(dp[i] > PI) dp[i] -= TPI;
		dphi += dp[i];
	}
	dphi /= (double)(NI);
  
	// Note: -dphi is the amount we have to shift all the wavelet phases by to compensate for the polarization phase shift
  newPhi += dphi;

	free_dvector(pass,0,NI+2);
	free_dvector(pv,1,2);
	free_dvector(yv,1,3);
	free_dvector(phase,1,NI);
	free_dvector(Fplus,1,NI);
	free_dvector(Fcross,1,NI);
	free_dvector(dp,1,NI);
	free_dmatrix(pm,1,3,1,2);
  
  y[2] = newPsi;
  y[3] = newEcc;
  y[6] = newPhi;


}

void amoeba(double **p, double *y, double *pass, int ndim, double ftol, double fmin, int *nfunk, struct Data *data)
{
	int i,ihi,ilo,inhi,j,mpts=ndim+1;
	double rtol,sum,swap,ysave,ytry,*psum;
	int NMAX = 200;
  
	psum=dvector(1,ndim);
	*nfunk=0;
	GET_PSUM
	for (;;) {
		ilo=1;
		ihi = y[1]>y[2] ? (inhi=2,1) : (inhi=1,2);
		for (i=1;i<=mpts;i++) {
			if (y[i] <= y[ilo]) ilo=i;
			if (y[i] > y[ihi]) {
				inhi=ihi;
				ihi=i;
			} else if (y[i] > y[inhi] && i != ihi) inhi=i;
		}
		rtol=2.0*fabs(y[ihi]-y[ilo])/(fabs(y[ihi])+fabs(y[ilo])+TTINY);
		if (rtol < ftol || y[ihi] < fmin) {
			SWAP(y[1],y[ilo])
			for (i=1;i<=ndim;i++) SWAP(p[1][i],p[ilo][i])
        break;
		}
		if (*nfunk >= NMAX) return;
		*nfunk += 2;
		ytry=amotry(p,y,psum,pass,ndim,ihi,-1.0,data);
		if (ytry <= y[ilo])
      ytry=amotry(p,y,psum,pass,ndim,ihi,2.0,data);
		else if (ytry >= y[inhi]) {
			ysave=y[ihi];
			ytry=amotry(p,y,psum,pass,ndim,ihi,0.5,data);
			if (ytry >= ysave) {
				for (i=1;i<=mpts;i++) {
					if (i != ilo) {
						for (j=1;j<=ndim;j++)
              p[i][j]=psum[j]=0.5*(p[i][j]+p[ilo][j]);
						y[i]=mfunc(psum,pass,data);
					}
				}
				*nfunk += ndim;
				GET_PSUM
			}
		} else --(*nfunk);
	}
	free_dvector(psum,1,ndim);
}

double amotry(double **p, double *y, double *psum, double *pass, int ndim, int ihi, float fac, struct Data *data)
{
	int j;
	double fac1,fac2,ytry,*ptry;
  
	ptry=dvector(1,ndim);
	fac1=(1.0-fac)/ndim;
	fac2=fac1-fac;
	for (j=1;j<=ndim;j++) ptry[j]=psum[j]*fac1-p[ihi][j]*fac2;
	ytry=mfunc(ptry,pass,data);
	if (ytry < y[ihi]) {
		y[ihi]=ytry;
		for (j=1;j<=ndim;j++) {
			psum[j] += ptry[j]-p[ihi][j];
			p[ihi][j]=ptry[j];
		}
	}
	free_dvector(ptry,1,ndim);
	return ytry;
}

double mfunc(double *psum, double *pass, struct Data *data)
{
	double ecc, psi, alpha, sindelta;
	double *phase;
	double *dp;
	double *del;
	double x, tl;
	double *Fplus, *Fcross;
	int i, j, k, KN;
  int NI = data->NI;
  
	KN = (NI*(NI-1))/2;
  
	Fplus = dvector(0,NI-1);
	Fcross = dvector(0,NI-1);
	dp = dvector(1,NI);
	del = dvector(1,KN);
	phase = dvector(1,NI);
  
	x = 0.0;
  
  
	if((fabs(psum[2]) > 1.0) || psum[1] < 0.0 || psum[1] > PI/2.0)
	{
		x = 1.0e40;
	}
	else
	{
    alpha = pass[0];
    sindelta = pass[1];
    tl = pass[2];
    for(i=1; i<= NI; i++) phase[i] = pass[i+2];
    psi = psum[1];
    ecc = psum[2];
    
    //    for(i=1; i<= NI; i++)
    //    {
    //      if(i==1) F_LHO(psi, alpha, sindelta, &Fplus[i], &Fcross[i]);
    //      if(i==2) F_LLO(psi, alpha, sindelta, &Fplus[i], &Fcross[i]);
    //      if(i==3) F_VIRGO(psi, alpha, sindelta, &Fplus[i], &Fcross[i]);
    //    }
    
    for(i=0; i<NI; i++)
      XLALComputeDetAMResponse(&Fplus[i], &Fcross[i], (const REAL4(*)[3])data->detector[i]->response, alpha, asin(sindelta), psi, XLALGreenwichMeanSiderealTime(&data->epoch));
    
    for(i=1; i<= NI; i++) {
      //printf("dp[%i]=%g, F+(%g,%g,%g)=%g, Fx=%g, phase=%g\n",i,dp[i],alpha, asin(sindelta), psi,Fplus[i-1],Fcross[i-1],phase[i]);
    dp[i] = atan2(ecc*Fcross[i-1],Fplus[i-1])-phase[i];

    }
		k = 0;
    for(i=1; i<= NI; i++)
    {
      for(j=i+1; j<= NI; j++)
      {
        k++;
        del[k] = fabs(dp[i]-dp[j]);
        //printf("%f ", del[k]);
        while(fabs(del[k] -TPI) < del[k]) del[k] = fabs(del[k] -TPI);
        // printf("%f\n", del[k]);
        x += del[k]*del[k]/(2.0*tl);
      }
    }
    
		
	}
	
	free_dvector(Fplus,0,NI-1);
	free_dvector(Fcross,0,NI-1);
	free_dvector(dp,1,NI);
	free_dvector(del,1,KN);
	free_dvector(phase,1,NI);
  
  
	return x;
  
}





